#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>

#include "xig.h"

/* Path to lock file */
static gchar *lock_path = NULL;

/* Display number being served */
static int display_number = 0;

/* X server */
static XigServer *xserver = NULL;

static void
cleanup ()
{
    if (lock_path)
        unlink (lock_path);
    if (xserver)
        g_object_unref (xserver);
}

static void
quit (int status)
{
    cleanup ();
    exit (status);
}

static void
indicate_ready ()
{
    void *handler;  
    handler = signal (SIGUSR1, SIG_IGN);
    if (handler == SIG_IGN)
        kill (getppid (), SIGUSR1);
    signal (SIGUSR1, handler);
}

static void
signal_cb (int signum)
{
    if (signum == SIGHUP)
        indicate_ready ();
    else
        quit (EXIT_SUCCESS);
}

static void
client_connected_cb (XigServer *server, XigRemoteClient *client)
{
    xig_codec_logger_new (XIG_CODEC (client));
}

static void
client_disconnected_cb (XigServer *server, XigRemoteClient *client)
{
    if (xig_server_get_clients (server) == NULL)
        indicate_ready ();
}

int
main (int argc, char **argv)
{
    int i;
    char *pid_string;
    GMainLoop *loop;
    int lock_file;
    XigScreen *screen;
    XigVisual *visual;
    GError *error = NULL;

    signal (SIGINT, signal_cb);
    signal (SIGTERM, signal_cb);
    signal (SIGHUP, signal_cb);
  
    g_type_init ();
  
    for (i = 1; i < argc; i++)
    {
        char *arg = argv[i];

        if (arg[0] == ':')
        {
            display_number = atoi (arg + 1);
        }
        else
        {
            g_printerr ("Unrecognized option: %s\n"
                        "Use: %s [:<display>]\n",
                        arg, argv[0]);
            return EXIT_FAILURE;
        }
    }

    xserver = xig_server_new ("Xig", display_number);
    g_signal_connect (xserver, "client-connected", G_CALLBACK (client_connected_cb), NULL);
    g_signal_connect (xserver, "client-disconnected", G_CALLBACK (client_disconnected_cb), NULL);
    xig_server_add_pixmap_format (xserver, 1, 1, 32);
    xig_server_add_pixmap_format (xserver, 4, 8, 32);
    xig_server_add_pixmap_format (xserver, 8, 8, 32);
    xig_server_add_pixmap_format (xserver, 15, 16, 32);
    xig_server_add_pixmap_format (xserver, 16, 16, 32);
    xig_server_add_pixmap_format (xserver, 24, 32, 32);
    xig_server_add_pixmap_format (xserver, 32, 32, 32);
    screen = xig_server_add_screen (xserver, 0x00FFFFFF, 0x00000000, 1024, 768, 1024, 768);
    visual = xig_screen_add_visual (screen, 24, XIG_VISUAL_CLASS_TrueColor, 8, 1, 0x00FF0000, 0x0000FF00, 0x000000FF);
    xig_screen_add_root (screen, visual);

    loop = g_main_loop_new (NULL, FALSE);

    lock_path = g_strdup_printf ("/tmp/.X%d-lock", display_number);
    lock_file = open (lock_path, O_CREAT | O_EXCL | O_WRONLY, 0444);
    if (lock_file < 0)
    {
        g_printerr ("Fatal server error:\n"
                    "Server is already active for display %d\n"
                    "	If this server is no longer running, remove %s\n"
                    "	and start again.\n", display_number, lock_path);
        g_free (lock_path);
        lock_path = NULL;
        quit (EXIT_FAILURE);
    }
    pid_string = g_strdup_printf ("%10ld", (long) getpid ());
    if (write (lock_file, pid_string, strlen (pid_string)) < 0)
    {
        g_warning ("Error writing PID file: %s", strerror (errno));
        quit (EXIT_FAILURE);
    }
    g_free (pid_string);

    if (!xig_server_start (xserver, &error))
    {
        g_warning ("Failed to start Xig server: %s", error->message);
        quit (EXIT_FAILURE);
    }

    /* Indicate ready if parent process has requested it */
    indicate_ready ();

    g_main_loop_run (loop);

    return EXIT_SUCCESS;
}
