/*
 * Copyright (C) 2011 Canonical Ltd.
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation. See http://www.gnu.org/copyleft/lgpl.html the
 * full text of the license.
 */

#include <string.h>

#include "xig/xig-client-codec.h"
#include "xig-common.h" // FIXME: Rename to xig-codec.h

G_DEFINE_TYPE (XigClientCodec, xig_client_codec, xig_codec_get_type ());

struct XigClientCodecPrivate
{
    /* TRUE if client has connected */
    gboolean connected;

    /* Byte order client is using */
    ByteOrder byte_order;

    /* TRUE if client can send big requests */
    gboolean BIGREQ_enabled;
};

XigClientCodec *
xig_client_codec_new (void)
{
    return g_object_new (xig_client_codec_get_type (), NULL);
}

void
xig_client_codec_set_BIGREQ_enabled (XigClientCodec *codec, gboolean BIGREQ_enabled)
{
    g_return_if_fail (codec != NULL);
    codec->priv->BIGREQ_enabled = BIGREQ_enabled;
}

static void
decode_success (XigClientCodec *codec, const guint8 *buffer, gsize buffer_length)
{
    XigSuccess reply;
    gsize offset = 0;
    guint16 vendor_length;
    int i;
 
    read_padding (1, &offset);
    reply.protocol_major_version = read_card16 (buffer, buffer_length, codec->priv->byte_order, &offset);
    reply.protocol_minor_version = read_card16 (buffer, buffer_length, codec->priv->byte_order, &offset);
    /*n =*/ read_card16 (buffer, buffer_length, codec->priv->byte_order, &offset);
    reply.release_number = read_card32 (buffer, buffer_length, codec->priv->byte_order, &offset);
    reply.resource_id_base = read_card32 (buffer, buffer_length, codec->priv->byte_order, &offset);
    reply.resource_id_mask = read_card32 (buffer, buffer_length, codec->priv->byte_order, &offset);
    reply.motion_buffer_size = read_card32 (buffer, buffer_length, codec->priv->byte_order, &offset);
    vendor_length = read_card16 (buffer, buffer_length, codec->priv->byte_order, &offset);
    reply.maximum_request_length = read_card16 (buffer, buffer_length, codec->priv->byte_order, &offset);
    reply.screens_length = read_card8 (buffer, buffer_length, &offset);
    reply.pixmap_formats_length = read_card8 (buffer, buffer_length, &offset);
    reply.image_byte_order = read_card8 (buffer, buffer_length, &offset);
    reply.bitmap_format_bit_order = read_card8 (buffer, buffer_length, &offset);
    reply.bitmap_format_scanline_unit = read_card8 (buffer, buffer_length, &offset);
    reply.bitmap_format_scanline_pad = read_card8 (buffer, buffer_length, &offset);
    reply.min_keycode = read_card8 (buffer, buffer_length, &offset);
    reply.max_keycode = read_card8 (buffer, buffer_length, &offset);
    read_padding (4, &offset);
    reply.vendor = read_padded_string (buffer, buffer_length, vendor_length, &offset);

    reply.pixmap_formats = calloc (reply.pixmap_formats_length, sizeof (XigFORMAT));
    for (i = 0; i < reply.pixmap_formats_length; i++)
    {
         reply.pixmap_formats[i].depth = read_card8 (buffer, buffer_length, &offset);
         reply.pixmap_formats[i].bits_per_pixel = read_card8 (buffer, buffer_length, &offset);
         reply.pixmap_formats[i].scanline_pad = read_card8 (buffer, buffer_length, &offset);
         read_padding (5, &offset);
    }

    reply.screens = calloc (reply.screens_length, sizeof (XigSCREEN));  
    for (i = 0; i < reply.screens_length; i++)
    {
        XigSCREEN *screen = &reply.screens[i];
        int j;

        screen->root = read_card32 (buffer, buffer_length, codec->priv->byte_order, &offset);
        screen->default_colormap = read_card32 (buffer, buffer_length, codec->priv->byte_order, &offset);
        screen->white_pixel = read_card32 (buffer, buffer_length, codec->priv->byte_order, &offset);
        screen->black_pixel = read_card32 (buffer, buffer_length, codec->priv->byte_order, &offset);
        screen->current_input_masks = read_card32 (buffer, buffer_length, codec->priv->byte_order, &offset);
        screen->width_in_pixels = read_card16 (buffer, buffer_length, codec->priv->byte_order, &offset);
        screen->height_in_pixels = read_card16 (buffer, buffer_length, codec->priv->byte_order, &offset);
        screen->width_in_millimeters = read_card16 (buffer, buffer_length, codec->priv->byte_order, &offset);
        screen->height_in_millimeters = read_card16 (buffer, buffer_length, codec->priv->byte_order, &offset);
        screen->min_installed_maps = read_card16 (buffer, buffer_length, codec->priv->byte_order, &offset);
        screen->max_installed_maps = read_card16 (buffer, buffer_length, codec->priv->byte_order, &offset);
        screen->root_visual = read_card32 (buffer, buffer_length, codec->priv->byte_order, &offset);
        screen->backing_stores = read_card8 (buffer, buffer_length, &offset);
        screen->save_unders = read_card8 (buffer, buffer_length, &offset) != 0;
        screen->root_visual = read_card8 (buffer, buffer_length, &offset);
        screen->allowed_depths_length = read_card8 (buffer, buffer_length, &offset);

        screen->allowed_depths = calloc (screen->allowed_depths_length, sizeof (XigDEPTH));
        for (j = 0; j < screen->allowed_depths_length; j++)
        {
            XigDEPTH *depth = &screen->allowed_depths[j];
            int k;

            depth->depth = read_card8 (buffer, buffer_length, &offset);
            read_padding (1, &offset);
            depth->visuals_length = read_card16 (buffer, buffer_length, codec->priv->byte_order, &offset);
            read_padding (4, &offset);
            depth->visuals = calloc (depth->visuals_length, sizeof (XigVISUALTYPE));
            for (k = 0; k < depth->visuals_length; k++)
            {
                XigVISUALTYPE *visual = &depth->visuals[k];

                visual->visual_id = read_card32 (buffer, buffer_length, codec->priv->byte_order, &offset);
                visual->class = read_card8 (buffer, buffer_length, &offset);
                visual->bits_per_rgb_value = read_card8 (buffer, buffer_length, &offset);
                visual->colormap_entries = read_card16 (buffer, buffer_length, codec->priv->byte_order, &offset);
                visual->red_mask = read_card32 (buffer, buffer_length, codec->priv->byte_order, &offset);
                visual->green_mask = read_card32 (buffer, buffer_length, codec->priv->byte_order, &offset);
                visual->blue_mask = read_card32 (buffer, buffer_length, codec->priv->byte_order, &offset);
                read_padding (4, &offset);
            }
        }
    }

    xig_codec_feed_success (XIG_CODEC (codec), &reply);

    g_free (reply.pixmap_formats);
    for (i = 0; i < reply.screens_length; i++)
    {
        int j;
        for (j = 0; j < reply.screens[i].allowed_depths_length; j++)
            g_free (reply.screens[i].allowed_depths[j].visuals);
        g_free (reply.screens[i].allowed_depths);
    }
    g_free (reply.screens);
}

static void
decode_get_window_attributes_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGetWindowAttributesReply reply;

    reply.sequence_number = 0; // FIXME
    reply.backing_store = data;
    reply.visual = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.class = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.bit_gravity = read_card8 (buffer, buffer_length, offset);
    reply.win_gravity = read_card8 (buffer, buffer_length, offset);
    reply.backing_planes = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.backing_pixel = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.save_under = read_card8 (buffer, buffer_length, offset) != 0;
    reply.map_is_installed = read_card8 (buffer, buffer_length, offset) != 0;
    reply.map_state = read_card8 (buffer, buffer_length, offset);
    reply.override_redirect = read_card8 (buffer, buffer_length, offset) != 0;
    reply.colormap = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.all_event_masks = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.your_event_mask = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.do_not_propagate_mask = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (2, offset);

    xig_codec_feed_get_window_attributes_reply (XIG_CODEC (codec), &reply);
}

static void
decode_get_geometry_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGetGeometryReply reply;

    reply.sequence_number = 0; // FIXME
    reply.depth = data;
    reply.root = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.x = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.y = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.width = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.height = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.border_width = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (2, offset);

    xig_codec_feed_get_geometry_reply (XIG_CODEC (codec), &reply);
}

static void
decode_query_tree_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigQueryTreeReply reply;
    int i;

    reply.sequence_number = 0; // FIXME
    reply.root = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.parent = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.children_length = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (14, offset);
    reply.children = calloc (reply.children_length, sizeof (guint32));
    for (i = 0; i < reply.children_length; i++)
        reply.children[i] = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_query_tree_reply (XIG_CODEC (codec), &reply);

    g_free (reply.children);
}

static void
decode_intern_atom_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigInternAtomReply reply;

    reply.sequence_number = 0; // FIXME
    reply.atom = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_intern_atom_reply (XIG_CODEC (codec), &reply);
}

static void
decode_get_atom_name_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGetAtomNameReply reply;
    gsize name_length;

    reply.sequence_number = 0; // FIXME
    name_length = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (22, offset);
    reply.name = read_padded_string (buffer, buffer_length, name_length, offset);

    xig_codec_feed_get_atom_name_reply (XIG_CODEC (codec), &reply);

    g_free (reply.name);
}

static void
decode_get_property_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGetPropertyReply reply;
    guint8 *value8;
    guint16 *value16;
    guint32 *value32;
    int i;

    reply.sequence_number = 0; // FIXME
    reply.value_format = data;
    reply.type = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.bytes_after = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.value_length = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (12, offset);
    switch (reply.value_format)
    {
    case 8:
        value8 = reply.value = calloc (reply.value_length, sizeof (guint8));
        for (i = 0; i < reply.value_length; i++)
            value8[i] = read_card8 (buffer, buffer_length, offset);
        break;
    case 16:
        value16 = reply.value = calloc (reply.value_length, sizeof (guint16));
        for (i = 0; i < reply.value_length; i++)
            value16[i] = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
        break;
    case 32:
        value32 = reply.value = calloc (reply.value_length, sizeof (guint32));
        for (i = 0; i < reply.value_length; i++)
            value32[i] = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
        break;
    default:
        reply.value = NULL;
        break;
    }

    xig_codec_feed_get_property_reply (XIG_CODEC (codec), &reply);

    g_free (reply.value);
}

static void
decode_list_properties_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigListPropertiesReply reply;
    int i;

    reply.sequence_number = 0; // FIXME
    reply.atoms_length = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (22, offset);
    reply.atoms = calloc (reply.atoms_length, sizeof (guint32));
    for (i = 0; i < reply.atoms_length; i++)
        reply.atoms[i] = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_list_properties_reply (XIG_CODEC (codec), &reply);

    g_free (reply.atoms);
}

static void
decode_get_selection_owner_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGetSelectionOwnerReply reply;

    reply.sequence_number = 0; // FIXME
    reply.owner = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_get_selection_owner_reply (XIG_CODEC (codec), &reply);
}

static void
decode_grab_pointer_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGrabPointerReply reply;

    reply.sequence_number = 0; // FIXME
    reply.status = data;

    xig_codec_feed_grab_pointer_reply (XIG_CODEC (codec), &reply);
}

static void
decode_grab_keyboard_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGrabKeyboardReply reply;

    reply.sequence_number = 0; // FIXME
    reply.status = data;

    xig_codec_feed_grab_keyboard_reply (XIG_CODEC (codec), &reply);
}

static void
decode_query_pointer_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigQueryPointerReply reply;

    reply.sequence_number = 0; // FIXME
    reply.same_screen = data != 0;
    reply.root = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.child = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.root_x = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.root_y = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.win_x = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.win_y = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.mask = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (2, offset);

    xig_codec_feed_query_pointer_reply (XIG_CODEC (codec), &reply);
}

static void
decode_TIMECOORD (XigClientCodec *codec, const guint8 *buffer, gsize buffer_length, gsize *offset, XigTIMECOORD *value)
{
    value->time = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    value->x = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    value->y = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
}

static void
decode_get_motion_events_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGetMotionEventsReply reply;
    int i;

    reply.sequence_number = 0; // FIXME
    reply.events_length = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (20, offset);
    reply.events = calloc (reply.events_length, sizeof (XigTIMECOORD));
    for (i = 0; i < reply.events_length; i++)
        decode_TIMECOORD (codec, buffer, buffer_length, offset, &reply.events[i]);

    xig_codec_feed_get_motion_events_reply (XIG_CODEC (codec), &reply);

    g_free (reply.events);
}

static void
decode_translate_coordinates_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigTranslateCoordinatesReply reply;

    reply.sequence_number = 0; // FIXME
    reply.same_screen = data != 0;
    reply.child = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.dst_x = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.dst_y = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_translate_coordinates_reply (XIG_CODEC (codec), &reply);
}

static void
decode_get_input_focus_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGetInputFocusReply reply;

    reply.sequence_number = 0; // FIXME
    reply.revert_to = data;
    reply.focus = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_get_input_focus_reply (XIG_CODEC (codec), &reply);
}

static void
decode_query_keymap_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigQueryKeymapReply reply;
    int i;

    for (i = 0; i < 32; i++)
        reply.keys[i] = read_card8 (buffer, buffer_length, offset);

    xig_codec_feed_query_keymap_reply (XIG_CODEC (codec), &reply);
}

static void
decode_FONTPROP (XigClientCodec *codec, const guint8 *buffer, gsize buffer_length, gsize *offset, XigFONTPROP *value)
{
    value->name = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    value->value = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
}

static void
decode_CHARINFO (XigClientCodec *codec, const guint8 *buffer, gsize buffer_length, gsize *offset, XigCHARINFO *value)
{
    value->left_side_bearing = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    value->right_side_bearing = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    value->character_width = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    value->ascent = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    value->descent = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    value->attributes = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
}

static void
decode_query_font_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigQueryFontReply reply;
    int i;

    reply.sequence_number = 0; // FIXME
    decode_CHARINFO (codec, buffer, buffer_length, offset, &reply.min_bounds);
    read_padding (4, offset);
    decode_CHARINFO (codec, buffer, buffer_length, offset, &reply.max_bounds);
    read_padding (4, offset);
    reply.min_char_or_byte2 = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.max_char_or_byte2 = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.default_char = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.properties_length = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.draw_direction = read_card8 (buffer, buffer_length, offset);
    reply.min_byte1 = read_card8 (buffer, buffer_length, offset);
    reply.max_byte1 = read_card8 (buffer, buffer_length, offset);
    reply.all_chars_exist = read_card8 (buffer, buffer_length, offset) != 0;
    reply.font_ascent = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.font_descent = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.char_infos_length = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.properties = calloc (reply.properties_length, sizeof (XigFONTPROP));
    for (i = 0; i < reply.properties_length; i++)
        decode_FONTPROP (codec, buffer, buffer_length, offset, &reply.properties[i]);
    reply.char_infos = calloc (reply.char_infos_length, sizeof (XigCHARINFO));
    for (i = 0; i < reply.char_infos_length; i++)
        decode_CHARINFO (codec, buffer, buffer_length, offset, &reply.char_infos[i]);

    xig_codec_feed_query_font_reply (XIG_CODEC (codec), &reply);

    g_free (reply.properties);
    g_free (reply.char_infos);
}

static void
decode_query_text_extents_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigQueryTextExtentsReply reply;

    reply.sequence_number = 0; // FIXME
    reply.draw_direction = data;
    reply.font_ascent = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.font_descent = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.overall_ascent = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.overall_descent = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.overall_width = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.overall_left = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.overall_right = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_query_text_extents_reply (XIG_CODEC (codec), &reply);
}

static void
decode_list_fonts_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigListFontsReply reply;
    int i;
    gsize names_length;

    reply.sequence_number = 0; // FIXME
    names_length = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (22, offset);
    reply.names = calloc (names_length + 1, sizeof (gchar *));
    for (i = 0; i < names_length; i++)
        reply.names[i] = read_string (buffer, buffer_length, read_card8 (buffer, buffer_length, offset), offset);
    reply.names[i] = NULL;

    xig_codec_feed_list_fonts_reply (XIG_CODEC (codec), &reply);

    g_strfreev (reply.names);
}

static void
decode_list_fonts_with_info_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigListFontsWithInfoReply reply;
    int i;
    gsize name_length;

    reply.sequence_number = 0; // FIXME
    name_length = data;
    decode_CHARINFO (codec, buffer, buffer_length, offset, &reply.min_bounds);
    read_padding (4, offset);
    decode_CHARINFO (codec, buffer, buffer_length, offset, &reply.max_bounds);
    read_padding (4, offset);
    reply.min_char_or_byte2 = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.max_char_or_byte2 = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.default_char = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.properties_length = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.draw_direction = read_card8 (buffer, buffer_length, offset);
    reply.min_byte1 = read_card8 (buffer, buffer_length, offset);
    reply.max_byte1 = read_card8 (buffer, buffer_length, offset);
    reply.all_chars_exist = read_card8 (buffer, buffer_length, offset) != 0;
    reply.font_ascent = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.font_descent = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.replies_hint = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.properties = calloc (reply.properties_length, sizeof (XigFONTPROP));
    for (i = 0; i < reply.properties_length; i++)
        decode_FONTPROP (codec, buffer, buffer_length, offset, &reply.properties[i]);
    reply.name = read_padded_string (buffer, buffer_length, name_length, offset);

    xig_codec_feed_list_fonts_with_info_reply (XIG_CODEC (codec), &reply);

    g_free (reply.properties);
    g_free (reply.name);
}

static void
decode_get_font_path_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGetFontPathReply reply;
    int i;
    gsize path_length;

    reply.sequence_number = 0; // FIXME
    path_length = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (22, offset);
    reply.path = calloc (path_length + 1, sizeof (gchar *));
    for (i = 0; i < path_length; i++)
        reply.path[i] = read_string (buffer, buffer_length, read_card8 (buffer, buffer_length, offset), offset);
    reply.path[i] = NULL;

    xig_codec_feed_get_font_path_reply (XIG_CODEC (codec), &reply);

    g_free (reply.path);
}

static void
decode_get_image_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGetImageReply reply;
    int i;

    reply.sequence_number = 0; // FIXME
    reply.depth = data;
    reply.visual = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (20, offset);
    reply.data = calloc (reply.data_length, sizeof (guint8));
    for (i = 0; i < reply.data_length; i++)
        reply.data[i] = read_card8 (buffer, buffer_length, offset);

    xig_codec_feed_get_image_reply (XIG_CODEC (codec), &reply);

    g_free (reply.data);
}

static void
decode_list_installed_colormaps_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigListInstalledColormapsReply reply;
    int i;

    reply.sequence_number = 0; // FIXME
    reply.cmaps_length = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (22, offset);
    reply.cmaps = calloc (reply.cmaps_length, sizeof (guint32));
    for (i = 0; i < reply.cmaps_length; i++)
        reply.cmaps[i] = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_list_installed_colormaps_reply (XIG_CODEC (codec), &reply);

    g_free (reply.cmaps);
}

static void
decode_alloc_color_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigAllocColorReply reply;

    reply.sequence_number = 0; // FIXME
    reply.red = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.green = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.blue = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (2, offset);
    reply.pixel = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_alloc_color_reply (XIG_CODEC (codec), &reply);
}

static void
decode_alloc_named_color_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigAllocNamedColorReply reply;

    reply.sequence_number = 0; // FIXME
    reply.pixel = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.exact_red = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.exact_green = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.exact_blue = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.visual_red = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.visual_green = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.visual_blue = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_alloc_named_color_reply (XIG_CODEC (codec), &reply);
}

static void
decode_alloc_color_cells_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigAllocColorCellsReply reply;
    int i;

    reply.sequence_number = 0; // FIXME
    reply.pixels_length = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.masks_length = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (20, offset);
    reply.pixels = calloc (reply.pixels_length, sizeof (guint32));
    for (i = 0; i < reply.pixels_length; i++)
        reply.pixels[i] = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.masks = calloc (reply.masks_length, sizeof (guint32));
    for (i = 0; i < reply.masks_length; i++)
        reply.masks[i] = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_alloc_color_cells_reply (XIG_CODEC (codec), &reply);

    g_free (reply.pixels);
    g_free (reply.masks);
}

static void
decode_alloc_color_planes_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigAllocColorPlanesReply reply;
    int i;

    reply.sequence_number = 0; // FIXME
    reply.pixels_length = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (2, offset);
    reply.red_mask = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.green_mask = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.blue_mask = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (8, offset);
    reply.pixels = calloc (reply.pixels_length, sizeof (guint32));
    for (i = 0; i < reply.pixels_length; i++)
        reply.pixels[i] = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_alloc_color_planes_reply (XIG_CODEC (codec), &reply);

    g_free (reply.pixels);
}

static void
decode_RGB (XigClientCodec *codec, const guint8 *buffer, gsize buffer_length, gsize *offset, XigRGB *value)
{
    value->red = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    value->green = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    value->blue = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (2, offset);
}

static void
decode_query_colors_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigQueryColorsReply reply;
    int i;

    reply.sequence_number = 0; // FIXME
    reply.colors_length = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (22, offset);
    reply.colors = calloc (reply.colors_length, sizeof (XigRGB));
    for (i = 0; i < reply.colors_length; i++)
        decode_RGB (codec, buffer, buffer_length, offset, &reply.colors[i]);

    xig_codec_feed_query_colors_reply (XIG_CODEC (codec), &reply);

    g_free (reply.colors);
}

static void
decode_lookup_color_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigLookupColorReply reply;

    reply.sequence_number = 0; // FIXME
    reply.exact_red = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.exact_green = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.exact_blue = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.visual_red = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.visual_green = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.visual_blue = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_lookup_color_reply (XIG_CODEC (codec), &reply);
}

static void
decode_query_best_size_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigQueryBestSizeReply reply;

    reply.sequence_number = 0; // FIXME
    reply.width = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.height = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_query_best_size_reply (XIG_CODEC (codec), &reply);
}

static void
decode_query_extension_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigQueryExtensionReply reply;

    reply.sequence_number = 0; // FIXME
    reply.present = read_card8 (buffer, buffer_length, offset) != 0;
    reply.major_opcode = read_card8 (buffer, buffer_length, offset);
    reply.first_event = read_card8 (buffer, buffer_length, offset);
    reply.first_error = read_card8 (buffer, buffer_length, offset);

    xig_codec_feed_query_extension_reply (XIG_CODEC (codec), &reply);
}

static void
decode_list_extensions_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigListExtensionsReply reply;
    int i;
    gsize names_length;

    reply.sequence_number = 0; // FIXME
    names_length = data;
    read_padding (24, offset);
    reply.names = calloc (names_length + 1, sizeof (gchar *));
    for (i = 0; i < names_length; i++)
        reply.names[i] = read_string (buffer, buffer_length, read_card8 (buffer, buffer_length, offset), offset);
    reply.names[i] = NULL;

    xig_codec_feed_list_extensions_reply (XIG_CODEC (codec), &reply);

    g_strfreev (reply.names);
}

static void
decode_get_keyboard_mapping_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGetKeyboardMappingReply reply;
    int i;

    reply.sequence_number = 0; // FIXME
    reply.keysyms_per_keycode = data;
    read_padding (24, offset);
    reply.keysyms = calloc (reply.keysyms_length, sizeof (guint32));
    for (i = 0; i < reply.keysyms_length; i++)
        reply.keysyms[i] = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_get_keyboard_mapping_reply (XIG_CODEC (codec), &reply);

    g_free (reply.keysyms);
}

static void
decode_get_keyboard_control_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGetKeyboardControlReply reply;
    int i;

    reply.sequence_number = 0; // FIXME
    reply.global_auto_repeat = data;
    reply.led_mask = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.key_click_percent = read_card8 (buffer, buffer_length, offset);
    reply.bell_percent = read_card8 (buffer, buffer_length, offset);
    reply.bell_pitch = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.bell_duration = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (2, offset);
    for (i = 0; i < 32; i++)
        reply.auto_repeats[i] = read_card8 (buffer, buffer_length, offset);

    xig_codec_feed_get_keyboard_control_reply (XIG_CODEC (codec), &reply);
}

static void
decode_get_pointer_control_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGetPointerControlReply reply;

    reply.sequence_number = 0; // FIXME
    reply.acceleration_numerator = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.acceleration_denominator = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.threshold = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (18, offset);

    xig_codec_feed_get_pointer_control_reply (XIG_CODEC (codec), &reply);
}

static void
decode_get_screen_saver_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGetScreenSaverReply reply;

    reply.sequence_number = 0; // FIXME
    reply.timeout = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.interval = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    reply.prefer_blanking = read_card8 (buffer, buffer_length, offset);
    reply.allow_exposures = read_card8 (buffer, buffer_length, offset);
    read_padding (18, offset);

    xig_codec_feed_get_screen_saver_reply (XIG_CODEC (codec), &reply);
}

static void
decode_HOST (XigClientCodec *codec, const guint8 *buffer, gsize buffer_length, gsize *offset, XigHOST *value)
{
    int i;
    value->family = read_card8 (buffer, buffer_length, offset);
    read_padding (1, offset);
    value->address_length = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    value->address = calloc (value->address_length, sizeof (guint8));
    for (i = 0; i < value->address_length; i++)
        value->address[i] = read_card8 (buffer, buffer_length, offset);
    read_padding (pad (value->address_length), offset);
}

static void
decode_list_hosts_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigListHostsReply reply;
    int i;

    reply.sequence_number = 0; // FIXME
    reply.mode = data;
    reply.hosts_length = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (22, offset);
    reply.hosts = calloc (reply.hosts_length, sizeof (XigHOST));
    for (i = 0; i < reply.hosts_length; i++)
        decode_HOST (codec, buffer, buffer_length, offset, &reply.hosts[i]);

    xig_codec_feed_list_hosts_reply (XIG_CODEC (codec), &reply);

    for (i = 0; i < reply.hosts_length; i++)
        g_free (reply.hosts[i].address);
    g_free (reply.hosts);
}

static void
decode_set_pointer_mapping_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigSetPointerMappingReply reply;

    reply.sequence_number = 0; // FIXME
    reply.status = data;

    xig_codec_feed_set_pointer_mapping_reply (XIG_CODEC (codec), &reply);
}

static void
decode_get_pointer_mapping_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGetPointerMappingReply reply;
    int i;

    reply.sequence_number = 0; // FIXME
    reply.map_length = data;
    read_padding (24, offset);
    reply.map = calloc (reply.map_length, sizeof (guint8));
    for (i = 0; i < reply.map_length; i++)
        reply.map[i] = read_card8 (buffer, buffer_length, offset);

    xig_codec_feed_get_pointer_mapping_reply (XIG_CODEC (codec), &reply);

    g_free (reply.map);
}

static void
decode_set_modifier_mapping_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigSetModifierMappingReply reply;

    reply.sequence_number = 0; // FIXME
    reply.status = data;

    xig_codec_feed_set_modifier_mapping_reply (XIG_CODEC (codec), &reply);
}

static void
decode_get_modifier_mapping_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGetModifierMappingReply reply;
    int i;

    reply.sequence_number = 0; // FIXME
    reply.keycodes_per_modifier = data;
    read_padding (24, offset);
    reply.keycodes = calloc (reply.keycodes_length, sizeof (guint8));
    for (i = 0; i < reply.keycodes_length; i++)
        reply.keycodes[i] = read_card8 (buffer, buffer_length, offset);

    xig_codec_feed_get_modifier_mapping_reply (XIG_CODEC (codec), &reply);

    g_free (reply.keycodes);
}

static void
decode_BIGREQ_enable_reply (XigClientCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigBIGREQEnableReply reply;

    reply.sequence_number = 0; // FIXME
    reply.maximum_request_length = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_BIGREQ_enable_reply (XIG_CODEC (codec), &reply);
}

static void
decode_reply (XigClientCodec *codec, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    guint8 opcode = 0; // FIXME
    guint8 data = 0; // FIXME

    switch (opcode)
    {
    case XIG_REQUEST_GetWindowAttributes:
        decode_get_window_attributes_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_GetGeometry:
        decode_get_geometry_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_QueryTree:
        decode_query_tree_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_InternAtom:
        decode_intern_atom_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_GetAtomName:
        decode_get_atom_name_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_GetProperty:
        decode_get_property_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_ListProperties:
        decode_list_properties_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_GetSelectionOwner:
        decode_get_selection_owner_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_GrabPointer:
        decode_grab_pointer_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_GrabKeyboard:
        decode_grab_keyboard_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_QueryPointer:
        decode_query_pointer_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_GetMotionEvents:
        decode_get_motion_events_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_TranslateCoordinates:
        decode_translate_coordinates_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_GetInputFocus:
        decode_get_input_focus_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_QueryKeymap:
        decode_query_keymap_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_QueryFont:
        decode_query_font_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_QueryTextExtents:
        decode_query_text_extents_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_ListFonts:
        decode_list_fonts_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_ListFontsWithInfo:
        decode_list_fonts_with_info_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_GetFontPath:
        decode_get_font_path_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_GetImage:
        decode_get_image_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_ListInstalledColormaps:
        decode_list_installed_colormaps_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_AllocColor:
        decode_alloc_color_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_AllocNamedColor:
        decode_alloc_named_color_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_AllocColorCells:
        decode_alloc_color_cells_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_AllocColorPlanes:
        decode_alloc_color_planes_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_QueryColors:
        decode_query_colors_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_LookupColor:
        decode_lookup_color_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_QueryBestSize:
        decode_query_best_size_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_QueryExtension:
        decode_query_extension_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_ListExtensions:
        decode_list_extensions_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_GetKeyboardMapping:
        decode_get_keyboard_mapping_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_GetKeyboardControl:
        decode_get_keyboard_control_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_GetPointerControl:
        decode_get_pointer_control_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_GetScreenSaver:
        decode_get_screen_saver_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_ListHosts:
        decode_list_hosts_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_SetPointerMapping:
        decode_set_pointer_mapping_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_GetPointerMapping:
        decode_get_pointer_mapping_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_SetModifierMapping:
        decode_set_modifier_mapping_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_GetModifierMapping:
        decode_get_modifier_mapping_reply (codec, data, buffer, buffer_length, offset);
        break;
    case XIG_REQUEST_BIGREQ:
        decode_BIGREQ_enable_reply (codec, data, buffer, buffer_length, offset);
        break;
    default:
        break;
    }
}

static gboolean
decode_response (XigClientCodec *codec, guint8 opcode, guint8 value, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    guint8 response_code;

    response_code = read_card8 (buffer, buffer_length, offset);

    if (!codec->priv->connected)
    {
        switch (response_code)
        {
        case Success:
            decode_success (codec, buffer + *offset, buffer_length - *offset);
            break;
        }

        return TRUE;
    }

    switch (response_code)
    {
    case XIG_RESPONSE_Error:
        break;
    case XIG_RESPONSE_Reply:
        decode_reply (codec, buffer, buffer_length, offset);
        break;
    case XIG_RESPONSE_KeyPress:
    case XIG_RESPONSE_KeyRelease:
    case XIG_RESPONSE_ButtonPress:
    case XIG_RESPONSE_ButtonRelease:
    case XIG_RESPONSE_MotionNotify:
    case XIG_RESPONSE_EnterNotify:
    case XIG_RESPONSE_LeaveNotify:
    case XIG_RESPONSE_FocusIn:
    case XIG_RESPONSE_FocusOut:
    case XIG_RESPONSE_KeymapNotify:
    case XIG_RESPONSE_Expose:
    case XIG_RESPONSE_GraphicsExposure:
    case XIG_RESPONSE_NoExposure:
    case XIG_RESPONSE_VisibilityNotify:
    case XIG_RESPONSE_CreateNotify:
    case XIG_RESPONSE_DestroyNotify:
    case XIG_RESPONSE_UnmapNotify:
    case XIG_RESPONSE_MapNotify:
    case XIG_RESPONSE_MapRequest:
    case XIG_RESPONSE_ReparentNotify:
    case XIG_RESPONSE_ConfigureNotify:
    case XIG_RESPONSE_ConfigureRequest:
    case XIG_RESPONSE_GravityNotify:
    case XIG_RESPONSE_ResizeRequest:
    case XIG_RESPONSE_CirculateNotify:
    case XIG_RESPONSE_CirculateRequest:
    case XIG_RESPONSE_PropertyNotify:
    case XIG_RESPONSE_SelectionClear:
    case XIG_RESPONSE_SelectionRequest:
    case XIG_RESPONSE_SelectionNotify:
    case XIG_RESPONSE_ColormapNotify:
    case XIG_RESPONSE_ClientMessage:
    case XIG_RESPONSE_MappingNotify:
        break;
    }

    return TRUE;
}

static void
xig_client_codec_real_feed_data (XigCodec *codec, const guint8 *data, gsize data_length)
{
    XigClientCodec *client_codec = XIG_CLIENT_CODEC (codec);
    int opcode;
    gsize offset = 0;

    if (!client_codec->priv->connected)
    {
        decode_success (client_codec, data, data_length);
        return;
    }

    while (offset < data_length)
    {
        guint8 value;
        /*guint16 sequence_number;*/
        guint32 reply_length;
        //guint8 opcode = 0; // FIXME

        value = read_card8 (data, data_length, &offset);
        /*sequence_number = */read_card16 (data, data_length, client_codec->priv->byte_order, &offset);
        reply_length = read_card32 (data, data_length, client_codec->priv->byte_order, &offset);

        decode_response (client_codec, opcode, value, data, offset + reply_length, &offset);

        //offset = start_offset + length;
    }
}

static void
xig_client_codec_real_connect (XigCodec *codec, XigConnect *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->byte_order, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->protocol_major_version, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->protocol_minor_version, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, strlen (message->authorization_protocol_name), &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->authorization_protocol_data_length, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 2, &n_written);
    write_padded_string (buffer, MAXIMUM_REQUEST_LENGTH, message->authorization_protocol_name, &n_written);
    write_string8 (buffer, MAXIMUM_REQUEST_LENGTH, message->authorization_protocol_data, message->authorization_protocol_data_length, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, pad (message->authorization_protocol_data_length), &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_create_window (XigCodec *codec, XigCreateWindow *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_CreateWindow, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->depth, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->wid, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->parent, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->x, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->y, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->width, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->height, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->border_width, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->class, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->visual, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_change_window_attributes (XigCodec *codec, XigChangeWindowAttributes *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_ChangeWindowAttributes, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->window, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_get_window_attributes (XigCodec *codec, XigGetWindowAttributes *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_GetWindowAttributes, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->window, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_destroy_window (XigCodec *codec, XigDestroyWindow *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_DestroyWindow, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->window, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_destroy_subwindows (XigCodec *codec, XigDestroySubwindows *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_DestroySubwindows, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->window, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_change_save_set (XigCodec *codec, XigChangeSaveSet *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_ChangeSaveSet, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->mode, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->window, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_reparent_window (XigCodec *codec, XigReparentWindow *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_ReparentWindow, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->window, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->parent, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->x, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->y, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_map_window (XigCodec *codec, XigMapWindow *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_MapWindow, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->window, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_map_subwindows (XigCodec *codec, XigMapSubwindows *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_MapSubwindows, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->window, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_unmap_window (XigCodec *codec, XigUnmapWindow *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_UnmapWindow, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->window, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_unmap_subwindows (XigCodec *codec, XigUnmapSubwindows *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_UnmapSubwindows, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->window, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_configure_window (XigCodec *codec, XigConfigureWindow *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_ConfigureWindow, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->window, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->value_mask, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 2, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_circulate_window (XigCodec *codec, XigCirculateWindow *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_CirculateWindow, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->direction, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->window, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_get_geometry (XigCodec *codec, XigGetGeometry *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_GetGeometry, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->drawable, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_query_tree (XigCodec *codec, XigQueryTree *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_QueryTree, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->window, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_intern_atom (XigCodec *codec, XigInternAtom *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_InternAtom, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->only_if_exists ? 1 : 0, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, strlen (message->name), &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 2, &n_written);
    write_padded_string (buffer, MAXIMUM_REQUEST_LENGTH, message->name, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_get_atom_name (XigCodec *codec, XigGetAtomName *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_GetAtomName, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->atom, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_change_property (XigCodec *codec, XigChangeProperty *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;
    guint8 *data8;
    guint16 *data16;
    guint32 *data32;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_ChangeProperty, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->mode, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->window, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->property, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->type, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->data_format, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 3, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->data_length, &n_written);
    switch (message->data_format)
    {
    case 8:
        data8 = message->data;
        for (i = 0; i < message->data_length; i++)
            write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, data8[i], &n_written);
        break;
    case 16:
        data16 = message->data;
        for (i = 0; i < message->data_length; i++)
            write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, data16[i], &n_written);
        break;
    case 32:
        data32 = message->data;
        for (i = 0; i < message->data_length; i++)
            write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, data32[i], &n_written);
        break;
    }

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_delete_property (XigCodec *codec, XigDeleteProperty *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_DeleteProperty, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->window, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->property, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_get_property (XigCodec *codec, XigGetProperty *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_GetProperty, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->delete ? 1 : 0, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->window, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->property, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->type, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->long_offset, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->long_length, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_list_properties (XigCodec *codec, XigListProperties *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_ListProperties, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->window, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_set_selection_owner (XigCodec *codec, XigSetSelectionOwner *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_SetSelectionOwner, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->owner, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->selection, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->time, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_get_selection_owner (XigCodec *codec, XigGetSelectionOwner *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_GetSelectionOwner, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->selection, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_convert_selection (XigCodec *codec, XigConvertSelection *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_ConvertSelection, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->requestor, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->selection, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->target, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->property, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->time, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_send_event (XigCodec *codec, XigSendEvent *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_SendEvent, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->propagate ? 1 : 0, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->destination, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->event_mask, &n_written);
    //write_padded_string (buffer, MAXIMUM_REQUEST_LENGTH, message->event, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_grab_pointer (XigCodec *codec, XigGrabPointer *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_GrabPointer, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->owner_events ? 1 : 0, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->grab_window, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->event_mask, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->pointer_mode, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->keyboard_mode, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->confine_to, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->cursor, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->time, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_ungrab_pointer (XigCodec *codec, XigUngrabPointer *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_UngrabPointer, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->time, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_grab_button (XigCodec *codec, XigGrabButton *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_GrabButton, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->owner_events ? 1 : 0, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->grab_window, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->event_mask, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->pointer_mode, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->keyboard_mode, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->confine_to, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->cursor, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->button, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->modifiers, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_ungrab_button (XigCodec *codec, XigUngrabButton *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_UngrabButton, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->button, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->grab_window, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->modifiers, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 2, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_change_active_pointer_grab (XigCodec *codec, XigChangeActivePointerGrab *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_ChangeActivePointerGrab, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->cursor, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->time, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->event_mask, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 2, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_grab_keyboard (XigCodec *codec, XigGrabKeyboard *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_GrabKeyboard, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->owner_events ? 1 : 0, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->grab_window, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->time, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->pointer_mode, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->keyboard_mode, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 2, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_ungrab_keyboard (XigCodec *codec, XigUngrabKeyboard *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_UngrabKeyboard, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->time, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_grab_key (XigCodec *codec, XigGrabKey *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_GrabKey, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->owner_events ? 1 : 0, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->grab_window, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->modifiers, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->key, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->pointer_mode, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->keyboard_mode, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 3, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_ungrab_key (XigCodec *codec, XigUngrabKey *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_UngrabKey, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->key, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->grab_window, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->modifiers, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 2, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_allow_events (XigCodec *codec, XigAllowEvents *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_AllowEvents, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->mode, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->time, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_grab_server (XigCodec *codec, XigGrabServer *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_GrabServer, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_ungrab_server (XigCodec *codec, XigUngrabServer *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_UngrabServer, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_query_pointer (XigCodec *codec, XigQueryPointer *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_QueryPointer, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->window, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_get_motion_events (XigCodec *codec, XigGetMotionEvents *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_GetMotionEvents, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->window, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->start, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->stop, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_translate_coordinates (XigCodec *codec, XigTranslateCoordinates *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_TranslateCoordinates, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->src_window, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->dst_window, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->src_x, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->src_y, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_warp_pointer (XigCodec *codec, XigWarpPointer *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_WarpPointer, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->src_window, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->dst_window, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->src_x, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->src_y, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->src_width, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->src_height, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->dst_x, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->dst_y, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_set_input_focus (XigCodec *codec, XigSetInputFocus *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_SetInputFocus, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->revert_to, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->focus, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->time, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_get_input_focus (XigCodec *codec, XigGetInputFocus *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_GetInputFocus, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_query_keymap (XigCodec *codec, XigQueryKeymap *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_QueryKeymap, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_open_font (XigCodec *codec, XigOpenFont *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_OpenFont, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->fid, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, strlen (message->name), &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 2, &n_written);
    write_padded_string (buffer, MAXIMUM_REQUEST_LENGTH, message->name, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_close_font (XigCodec *codec, XigCloseFont *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_CloseFont, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->font, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_query_font (XigCodec *codec, XigQueryFont *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_QueryFont, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->font, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_query_text_extents (XigCodec *codec, XigQueryTextExtents *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_QueryTextExtents, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->font, &n_written);
    for (i = 0; i < message->string_length; i++)
        write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->string[i], &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_list_fonts (XigCodec *codec, XigListFonts *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_ListFonts, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->max_names, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, strlen (message->pattern), &n_written);
    write_padded_string (buffer, MAXIMUM_REQUEST_LENGTH, message->pattern, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_list_fonts_with_info (XigCodec *codec, XigListFontsWithInfo *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_ListFontsWithInfo, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->max_names, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, strlen (message->pattern), &n_written);
    write_padded_string (buffer, MAXIMUM_REQUEST_LENGTH, message->pattern, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_set_font_path (XigCodec *codec, XigSetFontPath *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;
    gchar **s;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_SetFontPath, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, g_strv_length (message->path), &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 2, &n_written);
    for (s = message->path; *s; s++)
    {
        write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, strlen (*s), &n_written);
        write_string (buffer, MAXIMUM_REQUEST_LENGTH, *s, &n_written);
    }

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_get_font_path (XigCodec *codec, XigGetFontPath *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_GetFontPath, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_create_pixmap (XigCodec *codec, XigCreatePixmap *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_CreatePixmap, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->depth, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->pid, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->drawable, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->width, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->height, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_free_pixmap (XigCodec *codec, XigFreePixmap *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_FreePixmap, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->pixmap, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_create_gc (XigCodec *codec, XigCreateGC *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_CreateGC, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->cid, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->drawable, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_change_gc (XigCodec *codec, XigChangeGC *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_ChangeGC, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->gc, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_copy_gc (XigCodec *codec, XigCopyGC *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_CopyGC, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->src_gc, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->dst_gc, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->value_mask, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_set_dashes (XigCodec *codec, XigSetDashes *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_SetDashes, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->gc, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->dash_offset, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->dashes_length, &n_written);
    for (i = 0; i < message->dashes_length; i++)
        write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->dashes[i], &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_set_clip_rectangles (XigCodec *codec, XigSetClipRectangles *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_SetClipRectangles, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->ordering, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->gc, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->clip_x_origin, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->clip_y_origin, &n_written);
    for (i = 0; i < message->rectangles_length; i++)
        ;//write_RECTANGLE (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->rectangles[i], &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_free_gc (XigCodec *codec, XigFreeGC *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_FreeGC, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->gc, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_clear_area (XigCodec *codec, XigClearArea *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_ClearArea, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->exposures ? 1 : 0, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->window, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->x, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->y, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->width, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->height, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_copy_area (XigCodec *codec, XigCopyArea *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_CopyArea, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->src_drawable, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->dst_drawable, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->gc, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->src_x, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->src_y, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->dst_x, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->dst_y, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->width, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->height, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_copy_plane (XigCodec *codec, XigCopyPlane *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_CopyPlane, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->src_drawable, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->dst_drawable, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->gc, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->src_x, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->src_y, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->dst_x, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->dst_y, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->width, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->height, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->bit_plane, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_poly_point (XigCodec *codec, XigPolyPoint *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_PolyPoint, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->coordinate_mode, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->drawable, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->gc, &n_written);
    for (i = 0; i < message->points_length; i++)
        ;//write_POINT (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->points[i], &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_poly_line (XigCodec *codec, XigPolyLine *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_PolyLine, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->coordinate_mode, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->drawable, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->gc, &n_written);
    for (i = 0; i < message->points_length; i++)
        ;//write_POINT (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->points[i], &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_poly_segment (XigCodec *codec, XigPolySegment *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_PolySegment, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->drawable, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->gc, &n_written);
    for (i = 0; i < message->segments_length; i++)
        ;//write_SEGMENT (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->segments[i], &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_poly_rectangle (XigCodec *codec, XigPolyRectangle *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_PolyRectangle, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->drawable, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->gc, &n_written);
    for (i = 0; i < message->rectangles_length; i++)
        ;//write_RECTANGLE (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->rectangles[i], &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_poly_arc (XigCodec *codec, XigPolyArc *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_PolyArc, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->drawable, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->gc, &n_written);
    for (i = 0; i < message->arcs_length; i++)
        ;//write_ARC (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->arcs[i], &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_fill_poly (XigCodec *codec, XigFillPoly *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_FillPoly, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->drawable, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->gc, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->shape, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->coordinate_mode, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 2, &n_written);
    for (i = 0; i < message->points_length; i++)
        ;//write_POINT (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->points[i], &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_poly_fill_rectangle (XigCodec *codec, XigPolyFillRectangle *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_PolyFillRectangle, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->drawable, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->gc, &n_written);
    for (i = 0; i < message->rectangles_length; i++)
        ;//write_RECTANGLE (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->rectangles[i], &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_poly_fill_arc (XigCodec *codec, XigPolyFillArc *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_PolyFillArc, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->drawable, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->gc, &n_written);
    for (i = 0; i < message->arcs_length; i++)
        ;//write_ARC (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->arcs[i], &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_put_image (XigCodec *codec, XigPutImage *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_PutImage, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->format, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->drawable, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->gc, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->width, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->height, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->dst_x, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->dst_y, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->left_pad, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->depth, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 2, &n_written);
    for (i = 0; i < message->data_length; i++)
        write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->data[i], &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_get_image (XigCodec *codec, XigGetImage *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_GetImage, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->format, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->drawable, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->x, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->y, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->width, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->height, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->plane_mask, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_poly_text8 (XigCodec *codec, XigPolyText8 *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_PolyText8, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->drawable, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->gc, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->x, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->y, &n_written);
    for (i = 0; i < message->items_length; i++)
        write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->items[i], &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_poly_text16 (XigCodec *codec, XigPolyText16 *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_PolyText16, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->drawable, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->gc, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->x, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->y, &n_written);
    for (i = 0; i < message->items_length; i++)
        write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->items[i], &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_image_text8 (XigCodec *codec, XigImageText8 *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_ImageText8, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, strlen (message->string), &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->drawable, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->gc, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->x, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->y, &n_written);
    write_padded_string (buffer, MAXIMUM_REQUEST_LENGTH, message->string, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_image_text16 (XigCodec *codec, XigImageText16 *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_ImageText16, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->string_length, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->drawable, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->gc, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->x, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->y, &n_written);
    for (i = 0; i < message->string_length; i++)
        write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->string[i], &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_create_colormap (XigCodec *codec, XigCreateColormap *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_CreateColormap, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->alloc, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->mid, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->window, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->visual, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_free_colormap (XigCodec *codec, XigFreeColormap *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_FreeColormap, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->cmap, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_copy_colormap_and_free (XigCodec *codec, XigCopyColormapAndFree *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_CopyColormapAndFree, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->mid, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->src_cmap, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_install_colormap (XigCodec *codec, XigInstallColormap *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_InstallColormap, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->cmap, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_uninstall_colormap (XigCodec *codec, XigUninstallColormap *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_UninstallColormap, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->cmap, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_list_installed_colormaps (XigCodec *codec, XigListInstalledColormaps *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_ListInstalledColormaps, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->window, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_alloc_color (XigCodec *codec, XigAllocColor *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_AllocColor, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->cmap, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->red, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->green, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->blue, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 2, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_alloc_named_color (XigCodec *codec, XigAllocNamedColor *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_AllocNamedColor, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->cmap, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, strlen (message->name), &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 2, &n_written);
    write_padded_string (buffer, MAXIMUM_REQUEST_LENGTH, message->name, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_alloc_color_cells (XigCodec *codec, XigAllocColorCells *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_AllocColorCells, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->contiguous ? 1 : 0, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->cmap, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->colors, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->planes, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_alloc_color_planes (XigCodec *codec, XigAllocColorPlanes *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_AllocColorPlanes, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->contiguous ? 1 : 0, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->cmap, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->colors, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->reds, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->greens, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->blues, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_free_colors (XigCodec *codec, XigFreeColors *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_FreeColors, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->cmap, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->plane_mask, &n_written);
    for (i = 0; i < message->pixels_length; i++)
        write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->pixels[i], &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_store_colors (XigCodec *codec, XigStoreColors *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_StoreColors, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->cmap, &n_written);
    for (i = 0; i < message->items_length; i++)
        ;//write_COLORITEM (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->items[i], &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_store_named_color (XigCodec *codec, XigStoreNamedColor *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;
    guint8 flags = 0x00;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_StoreNamedColor, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    if (message->do_red)
        flags |= 0x01;
    if (message->do_green)
        flags |= 0x02;
    if (message->do_blue)
        flags |= 0x04;
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, flags, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->cmap, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->pixel, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, strlen (message->name), &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 2, &n_written);
    write_padded_string (buffer, MAXIMUM_REQUEST_LENGTH, message->name, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_query_colors (XigCodec *codec, XigQueryColors *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_QueryColors, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->cmap, &n_written);
    for (i = 0; i < message->pixels_length; i++)
        write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->pixels[i], &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_lookup_color (XigCodec *codec, XigLookupColor *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_LookupColor, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->cmap, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, strlen (message->name), &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 2, &n_written);
    write_padded_string (buffer, MAXIMUM_REQUEST_LENGTH, message->name, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_create_cursor (XigCodec *codec, XigCreateCursor *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_CreateCursor, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->cid, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->source, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->mask, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->fore_red, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->fore_green, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->fore_blue, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->back_red, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->back_green, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->back_blue, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->x, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->y, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_create_glyph_cursor (XigCodec *codec, XigCreateGlyphCursor *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_CreateGlyphCursor, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->cid, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->source_font, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->mask_font, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->source_char, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->mask_char, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->fore_red, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->fore_green, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->fore_blue, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->back_red, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->back_green, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->back_blue, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_free_cursor (XigCodec *codec, XigFreeCursor *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_FreeCursor, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->cursor, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_recolor_cursor (XigCodec *codec, XigRecolorCursor *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_RecolorCursor, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->cursor, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->fore_red, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->fore_green, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->fore_blue, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->back_red, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->back_green, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->back_blue, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_query_best_size (XigCodec *codec, XigQueryBestSize *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_QueryBestSize, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->class, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->drawable, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->width, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->height, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_query_extension (XigCodec *codec, XigQueryExtension *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_QueryExtension, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, strlen (message->name), &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 2, &n_written);
    write_padded_string (buffer, MAXIMUM_REQUEST_LENGTH, message->name, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_list_extensions (XigCodec *codec, XigListExtensions *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_ListExtensions, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_change_keyboard_mapping (XigCodec *codec, XigChangeKeyboardMapping *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_ChangeKeyboardMapping, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->keycode_count, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->first_keycode, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->keysyms_per_keycode, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 2, &n_written);
    for (i = 0; i < message->keysyms_length; i++)
        write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->keysyms[i], &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_get_keyboard_mapping (XigCodec *codec, XigGetKeyboardMapping *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_GetKeyboardMapping, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->first_keycode, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->count, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_change_keyboard_control (XigCodec *codec, XigChangeKeyboardControl *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_ChangeKeyboardControl, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_get_keyboard_control (XigCodec *codec, XigGetKeyboardControl *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_GetKeyboardControl, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_bell (XigCodec *codec, XigBell *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_Bell, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->percent, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_change_pointer_control (XigCodec *codec, XigChangePointerControl *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_ChangePointerControl, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->acceleration_numerator, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->acceleration_denominator, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->threshold, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->do_acceleration ? 1 : 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->do_threshold ? 1 : 0, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_get_pointer_control (XigCodec *codec, XigGetPointerControl *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_GetPointerControl, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_set_screen_saver (XigCodec *codec, XigSetScreenSaver *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_SetScreenSaver, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->timeout, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->interval, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->prefer_blanking, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->allow_exposures, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_get_screen_saver (XigCodec *codec, XigGetScreenSaver *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_GetScreenSaver, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_change_hosts (XigCodec *codec, XigChangeHosts *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_ChangeHosts, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->mode, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->family, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->address_length, &n_written);
    for (i = 0; i < message->address_length; i++)
        write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->address[i], &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_list_hosts (XigCodec *codec, XigListHosts *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_ListHosts, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_set_access_control (XigCodec *codec, XigSetAccessControl *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_SetAccessControl, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->mode, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_set_close_down_mode (XigCodec *codec, XigSetCloseDownMode *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_SetCloseDownMode, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->mode, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_kill_client (XigCodec *codec, XigKillClient *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_KillClient, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->resource, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_rotate_properties (XigCodec *codec, XigRotateProperties *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_RotateProperties, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->window, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->properties_length, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->delta, &n_written);
    for (i = 0; i < message->properties_length; i++)
        write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, message->properties[i], &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_force_screen_saver (XigCodec *codec, XigForceScreenSaver *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_ForceScreenSaver, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->mode, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_set_pointer_mapping (XigCodec *codec, XigSetPointerMapping *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_SetPointerMapping, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->map_length, &n_written);
    for (i = 0; i < message->map_length; i++)
        write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->map[i], &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_get_pointer_mapping (XigCodec *codec, XigGetPointerMapping *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_GetPointerMapping, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_set_modifier_mapping (XigCodec *codec, XigSetModifierMapping *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_SetModifierMapping, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->keycodes_per_modifier, &n_written);
    for (i = 0; i < message->keycodes_length; i++)
        write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->keycodes[i], &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_get_modifier_mapping (XigCodec *codec, XigGetModifierMapping *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_GetModifierMapping, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_no_operation (XigCodec *codec, XigNoOperation *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_NoOperation, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_real_BIGREQ_enable (XigCodec *codec, XigBIGREQEnable *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_REQUEST_BIGREQ, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, 0, &n_written); // FIXME
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_CLIENT_CODEC (codec)->priv->byte_order, 0, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_client_codec_init (XigClientCodec *codec)
{
    codec->priv = G_TYPE_INSTANCE_GET_PRIVATE (codec, xig_client_codec_get_type (), XigClientCodecPrivate);
}

static void
xig_client_codec_class_init (XigClientCodecClass *klass)
{
    XigCodecClass *codec_class = XIG_CODEC_CLASS (klass);

    g_type_class_add_private (klass, sizeof (XigClientCodecPrivate));

    codec_class->feed_data = xig_client_codec_real_feed_data;
    codec_class->connect = xig_client_codec_real_connect;
    codec_class->create_window = xig_client_codec_real_create_window;
    codec_class->change_window_attributes = xig_client_codec_real_change_window_attributes;
    codec_class->get_window_attributes = xig_client_codec_real_get_window_attributes;
    codec_class->destroy_window = xig_client_codec_real_destroy_window;
    codec_class->destroy_subwindows = xig_client_codec_real_destroy_subwindows;
    codec_class->change_save_set = xig_client_codec_real_change_save_set;
    codec_class->reparent_window = xig_client_codec_real_reparent_window;
    codec_class->map_window = xig_client_codec_real_map_window;
    codec_class->map_subwindows = xig_client_codec_real_map_subwindows;
    codec_class->unmap_window = xig_client_codec_real_unmap_window;
    codec_class->unmap_subwindows = xig_client_codec_real_unmap_subwindows;
    codec_class->configure_window = xig_client_codec_real_configure_window;
    codec_class->circulate_window = xig_client_codec_real_circulate_window;
    codec_class->get_geometry = xig_client_codec_real_get_geometry;
    codec_class->query_tree = xig_client_codec_real_query_tree;
    codec_class->intern_atom = xig_client_codec_real_intern_atom;
    codec_class->get_atom_name = xig_client_codec_real_get_atom_name;
    codec_class->change_property = xig_client_codec_real_change_property;
    codec_class->delete_property = xig_client_codec_real_delete_property;
    codec_class->get_property = xig_client_codec_real_get_property;
    codec_class->list_properties = xig_client_codec_real_list_properties;
    codec_class->set_selection_owner = xig_client_codec_real_set_selection_owner;
    codec_class->get_selection_owner = xig_client_codec_real_get_selection_owner;
    codec_class->convert_selection = xig_client_codec_real_convert_selection;
    codec_class->send_event = xig_client_codec_real_send_event;
    codec_class->grab_pointer = xig_client_codec_real_grab_pointer;
    codec_class->ungrab_pointer = xig_client_codec_real_ungrab_pointer;
    codec_class->grab_button = xig_client_codec_real_grab_button;
    codec_class->ungrab_button = xig_client_codec_real_ungrab_button;
    codec_class->change_active_pointer_grab = xig_client_codec_real_change_active_pointer_grab;
    codec_class->grab_keyboard = xig_client_codec_real_grab_keyboard;
    codec_class->ungrab_keyboard = xig_client_codec_real_ungrab_keyboard;
    codec_class->grab_key = xig_client_codec_real_grab_key;
    codec_class->ungrab_key = xig_client_codec_real_ungrab_key;
    codec_class->allow_events = xig_client_codec_real_allow_events;
    codec_class->grab_server = xig_client_codec_real_grab_server;
    codec_class->ungrab_server = xig_client_codec_real_ungrab_server;
    codec_class->query_pointer = xig_client_codec_real_query_pointer;
    codec_class->get_motion_events = xig_client_codec_real_get_motion_events;
    codec_class->translate_coordinates = xig_client_codec_real_translate_coordinates;
    codec_class->warp_pointer = xig_client_codec_real_warp_pointer;
    codec_class->set_input_focus = xig_client_codec_real_set_input_focus;
    codec_class->get_input_focus = xig_client_codec_real_get_input_focus;
    codec_class->query_keymap = xig_client_codec_real_query_keymap;
    codec_class->open_font = xig_client_codec_real_open_font;
    codec_class->close_font = xig_client_codec_real_close_font;
    codec_class->query_font = xig_client_codec_real_query_font;
    codec_class->query_text_extents = xig_client_codec_real_query_text_extents;
    codec_class->list_fonts = xig_client_codec_real_list_fonts;
    codec_class->list_fonts_with_info = xig_client_codec_real_list_fonts_with_info;
    codec_class->set_font_path = xig_client_codec_real_set_font_path;
    codec_class->get_font_path = xig_client_codec_real_get_font_path;
    codec_class->create_pixmap = xig_client_codec_real_create_pixmap;
    codec_class->free_pixmap = xig_client_codec_real_free_pixmap;
    codec_class->create_gc = xig_client_codec_real_create_gc;
    codec_class->change_gc = xig_client_codec_real_change_gc;
    codec_class->copy_gc = xig_client_codec_real_copy_gc;
    codec_class->set_dashes = xig_client_codec_real_set_dashes;
    codec_class->set_clip_rectangles = xig_client_codec_real_set_clip_rectangles;
    codec_class->free_gc = xig_client_codec_real_free_gc;
    codec_class->clear_area = xig_client_codec_real_clear_area;
    codec_class->copy_area = xig_client_codec_real_copy_area;
    codec_class->copy_plane = xig_client_codec_real_copy_plane;
    codec_class->poly_point = xig_client_codec_real_poly_point;
    codec_class->poly_line = xig_client_codec_real_poly_line;
    codec_class->poly_segment = xig_client_codec_real_poly_segment;
    codec_class->poly_rectangle = xig_client_codec_real_poly_rectangle;
    codec_class->poly_arc = xig_client_codec_real_poly_arc;
    codec_class->fill_poly = xig_client_codec_real_fill_poly;
    codec_class->poly_fill_rectangle = xig_client_codec_real_poly_fill_rectangle;
    codec_class->poly_fill_arc = xig_client_codec_real_poly_fill_arc;
    codec_class->put_image = xig_client_codec_real_put_image;
    codec_class->get_image = xig_client_codec_real_get_image;
    codec_class->poly_text8 = xig_client_codec_real_poly_text8;
    codec_class->poly_text16 = xig_client_codec_real_poly_text16;
    codec_class->image_text8 = xig_client_codec_real_image_text8;
    codec_class->image_text16 = xig_client_codec_real_image_text16;
    codec_class->create_colormap = xig_client_codec_real_create_colormap;
    codec_class->free_colormap = xig_client_codec_real_free_colormap;
    codec_class->copy_colormap_and_free = xig_client_codec_real_copy_colormap_and_free;
    codec_class->install_colormap = xig_client_codec_real_install_colormap;
    codec_class->uninstall_colormap = xig_client_codec_real_uninstall_colormap;
    codec_class->list_installed_colormaps = xig_client_codec_real_list_installed_colormaps;
    codec_class->alloc_color = xig_client_codec_real_alloc_color;
    codec_class->alloc_named_color = xig_client_codec_real_alloc_named_color;
    codec_class->alloc_color_cells = xig_client_codec_real_alloc_color_cells;
    codec_class->alloc_color_planes = xig_client_codec_real_alloc_color_planes;
    codec_class->free_colors = xig_client_codec_real_free_colors;
    codec_class->store_colors = xig_client_codec_real_store_colors;
    codec_class->store_named_color = xig_client_codec_real_store_named_color;
    codec_class->query_colors = xig_client_codec_real_query_colors;
    codec_class->lookup_color = xig_client_codec_real_lookup_color;
    codec_class->create_cursor = xig_client_codec_real_create_cursor;
    codec_class->create_glyph_cursor = xig_client_codec_real_create_glyph_cursor;
    codec_class->free_cursor = xig_client_codec_real_free_cursor;
    codec_class->recolor_cursor = xig_client_codec_real_recolor_cursor;
    codec_class->query_best_size = xig_client_codec_real_query_best_size;
    codec_class->query_extension = xig_client_codec_real_query_extension;
    codec_class->list_extensions = xig_client_codec_real_list_extensions;
    codec_class->change_keyboard_mapping = xig_client_codec_real_change_keyboard_mapping;
    codec_class->get_keyboard_mapping = xig_client_codec_real_get_keyboard_mapping;
    codec_class->change_keyboard_control = xig_client_codec_real_change_keyboard_control;
    codec_class->get_keyboard_control = xig_client_codec_real_get_keyboard_control;
    codec_class->bell = xig_client_codec_real_bell;
    codec_class->change_pointer_control = xig_client_codec_real_change_pointer_control;
    codec_class->get_pointer_control = xig_client_codec_real_get_pointer_control;
    codec_class->set_screen_saver = xig_client_codec_real_set_screen_saver;
    codec_class->get_screen_saver = xig_client_codec_real_get_screen_saver;
    codec_class->change_hosts = xig_client_codec_real_change_hosts;
    codec_class->list_hosts = xig_client_codec_real_list_hosts;
    codec_class->set_access_control = xig_client_codec_real_set_access_control;
    codec_class->set_close_down_mode = xig_client_codec_real_set_close_down_mode;
    codec_class->kill_client = xig_client_codec_real_kill_client;
    codec_class->rotate_properties = xig_client_codec_real_rotate_properties;
    codec_class->force_screen_saver = xig_client_codec_real_force_screen_saver;
    codec_class->set_pointer_mapping = xig_client_codec_real_set_pointer_mapping;
    codec_class->get_pointer_mapping = xig_client_codec_real_get_pointer_mapping;
    codec_class->set_modifier_mapping = xig_client_codec_real_set_modifier_mapping;
    codec_class->get_modifier_mapping = xig_client_codec_real_get_modifier_mapping;
    codec_class->no_operation = xig_client_codec_real_no_operation;
    codec_class->BIGREQ_enable = xig_client_codec_real_BIGREQ_enable;
}
