/*
 * Copyright (C) 2011 Canonical Ltd.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation. See http://www.gnu.org/copyleft/lgpl.html the
 * full text of the license.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <gio/gio.h>
#include <gio/gunixsocketaddress.h>

#include "xig/xig-client.h"
#include "xig/xig-version.h"
#include "xig-common.h"

G_DEFINE_TYPE (XigClient, xig_client, xig_client_codec_get_type ());

struct XigClientPrivate
{
    /* Hostname to connect to */
    gchar *hostname;
  
    /* Display number to connect to */
    guint16 display_number;

    /* Screen to use */
    guint8 screen;

    /* Authorization protocol to use */
    gchar *authorization_protocol_name;
    guint8 *authorization_protocol_data;
    guint16 authorization_protocol_data_length;

    /* Socket to server */
    GSocket *socket;  
    GIOChannel *channel;

    /* Byte order being used */
    ByteOrder byte_order;

    /* TRUE if client has connected */
    gboolean connected;

    /* Server properties */
    guint16 protocol_major_version;
    guint16 protocol_minor_version;
    guint32 release_number;
    guint32 motion_buffer_size;
    guint32 maximum_request_length;
    guint8 image_byte_order;
    guint8 bitmap_format_bit_order;
    guint8 bitmap_format_scanline_unit;
    guint8 bitmap_format_scanline_pad;
    guint8 min_keycode;
    guint8 max_keycode;
    gchar *vendor;

    /* Allocated resource IDs */
    guint32 resource_id_base;
    guint32 resource_id_mask;
};

enum
{
    DISCONNECTED,
    LAST_SIGNAL
};
static guint signals[LAST_SIGNAL] = { 0 };

XigClient *
xig_client_new (void)
{
    return g_object_new (xig_client_get_type (), NULL);
}

XigClient *
xig_client_new_from_environment (GError **error)
{
    XigClient *client;
    const gchar *display, *c;
    gchar *authority;

    client = xig_client_new ();

    display = g_getenv ("DISPLAY");
    if (!display)
        return NULL;

    c = strchr (display, ':');
    if (!c)
        return NULL;

    client->priv->display_number = atoi (c + 1);

    authority = g_strdup (g_getenv ("XAUTHORITY"));
    if (!authority)
        authority = g_build_filename (g_get_home_dir (), ".Xauthority", NULL);
    g_free (authority);

    return client;
}

const gchar *
xig_client_get_vendor (XigClient *client)
{
    return client->priv->vendor;
}

GInetAddress *
xig_client_get_server_address (XigClient *client)
{
    GSocketAddress *socket_address;
    GError *error = NULL;

    socket_address = g_socket_get_remote_address (client->priv->socket, &error);
    if (error)
        g_warning ("Error getting remote socket address");
    g_clear_error (&error);
    if (!socket_address)
        return NULL;

    if (G_IS_INET_SOCKET_ADDRESS (socket_address))
        return g_inet_socket_address_get_address (G_INET_SOCKET_ADDRESS (socket_address));
    else
        return NULL;
}

void
xig_client_disconnect (XigClient *client)
{
    g_io_channel_shutdown (client->priv->channel, TRUE, NULL);
}

static gboolean
socket_data_cb (GIOChannel *channel, GIOCondition condition, gpointer data)
{
    XigClient *client = data;
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gssize n_read;
    GError *error = NULL;

    n_read = g_socket_receive (client->priv->socket, (gchar *) buffer, MAXIMUM_REQUEST_LENGTH, NULL, &error);
    if (error)
        g_warning ("Error reading from socket: %s", error->message);
    g_clear_error (&error);
    if (n_read < 0)
        return FALSE;
    if (n_read == 0)
    {
        g_signal_emit (client, signals[DISCONNECTED], 0);
        return FALSE;
    }

    xig_codec_feed_data (XIG_CODEC (client), buffer, n_read);

    return TRUE;
}

void
xig_client_connect (XigClient *client, GCancellable *cancellable, GAsyncReadyCallback callback, gpointer user_data)
{
}

gboolean
xig_client_connect_finish (XigClient *client, GAsyncResult *res, GError **error)
{
    return FALSE;
}

gboolean
xig_client_connect_sync (XigClient *client, GError **error)
{
    gchar *socket_path;
    GSocketAddress *socket_address;
    gboolean result;
    XigConnect request;

    client->priv->socket = g_socket_new (G_SOCKET_FAMILY_UNIX, G_SOCKET_TYPE_STREAM, G_SOCKET_PROTOCOL_DEFAULT, error);
    if (client->priv->socket == NULL)
        return FALSE;

    socket_path = g_strdup_printf ("/tmp/.X11-unix/X%d", client->priv->display_number);
    socket_address = g_unix_socket_address_new (socket_path);
    g_free (socket_path);
  
    result = g_socket_connect (client->priv->socket, socket_address, NULL, error);
    g_object_unref (socket_address);

    if (!result)
        return FALSE;

    client->priv->channel = g_io_channel_unix_new (g_socket_get_fd (client->priv->socket));
    g_io_add_watch (client->priv->channel, G_IO_IN, socket_data_cb, client);
  
    /* Send connection request */
    request.byte_order = client->priv->byte_order == BYTE_ORDER_MSB ? 'B' : 'l';
    request.protocol_major_version = XIG_PROTOCOL_MAJOR_VERSION;
    request.protocol_minor_version = XIG_PROTOCOL_MINOR_VERSION;
    request.authorization_protocol_name = client->priv->authorization_protocol_name;
    request.authorization_protocol_data = NULL;
    request.authorization_protocol_data_length = 0;
    xig_codec_feed_connect (XIG_CODEC (client), &request);

    /* Wait for response */
    if (!socket_data_cb (client->priv->channel, G_IO_IN, client))
        return FALSE;

    return TRUE;
}

static void
xig_client_real_write (XigCodec *codec, const guint8 *data, gsize data_length)
{
    XigClient *client = XIG_CLIENT (codec);
    GError *error = NULL;

    gssize n_sent;

    n_sent = g_socket_send (client->priv->socket, (gchar *) data, data_length, NULL, &error);
    if (error)
        g_warning ("Error writing to socket: %s", error->message);
    g_clear_error (&error);
    if (n_sent != data_length)
        ;//return FALSE;
}

static void
xig_client_real_success (XigCodec *codec, XigSuccess *message)
{
    XigClient *client = XIG_CLIENT (codec);

    client->priv->protocol_major_version = message->protocol_major_version;
    client->priv->protocol_minor_version = message->protocol_minor_version;
    client->priv->release_number = message->release_number;
    client->priv->resource_id_base = message->resource_id_base;
    client->priv->resource_id_mask = message->resource_id_mask;
    client->priv->motion_buffer_size = message->motion_buffer_size;
    client->priv->maximum_request_length = message->maximum_request_length;
    client->priv->image_byte_order = message->image_byte_order;
    client->priv->bitmap_format_bit_order = message->bitmap_format_bit_order;
    client->priv->bitmap_format_scanline_unit = message->bitmap_format_scanline_unit;
    client->priv->bitmap_format_scanline_pad = message->bitmap_format_scanline_pad;
    client->priv->min_keycode = message->min_keycode;
    client->priv->max_keycode = message->max_keycode;
    client->priv->vendor = g_strdup (message->vendor);
}  

static void
xig_client_init (XigClient *client)
{
    client->priv = G_TYPE_INSTANCE_GET_PRIVATE (client, xig_client_get_type (), XigClientPrivate);
    client->priv->byte_order = BYTE_ORDER_LSB;
    client->priv->authorization_protocol_name = g_strdup ("");
}

static void
xig_client_finalize (GObject *object)
{
    XigClient *client = XIG_CLIENT (object);

    g_free (client->priv->hostname);
    g_free (client->priv->authorization_protocol_name);
    if (client->priv->authorization_protocol_data)
        g_free (client->priv->authorization_protocol_data);
    g_free (client->priv->vendor);

    G_OBJECT_CLASS (xig_client_parent_class)->finalize (object);
}

static void
xig_client_class_init (XigClientClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);
    XigCodecClass *codec_class = XIG_CODEC_CLASS (klass);

    codec_class->write = xig_client_real_write;
    codec_class->success = xig_client_real_success;
    object_class->finalize = xig_client_finalize;

    g_type_class_add_private (klass, sizeof (XigClientPrivate));

    signals[DISCONNECTED] =
        g_signal_new ("disconnected",
                      G_TYPE_FROM_CLASS (klass),
                      G_SIGNAL_RUN_LAST,
                      G_STRUCT_OFFSET (XigClientClass, disconnected),
                      NULL, NULL,
                      g_cclosure_marshal_VOID__VOID,
                      G_TYPE_NONE, 0);
}
