/*
 * Copyright (C) 2011 Canonical Ltd.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation. See http://www.gnu.org/copyleft/lgpl.html the
 * full text of the license.
 */

#include "xig/xig-codec-logger.h"
#include "xig/xig-events.h"
#include "xig-common.h"

G_DEFINE_TYPE (XigCodecLogger, xig_codec_logger, G_TYPE_OBJECT);

struct XigCodecLoggerPrivate
{
    /* Codec being logged */
    XigCodec *codec;
};

static void
connect_cb (XigCodec *codec, XigConnect *message, XigCodecLogger *logger)
{
    GString *text;

    text = g_string_new ("Connect");
    g_string_append_printf (text, " byte-order=%d", message->byte_order);
    g_string_append_printf (text, " protocol-major-version=%d", message->protocol_major_version);
    g_string_append_printf (text, " protocol-minor-version=%d", message->protocol_minor_version);
    g_string_append_printf (text, " authorization-protocol-name=%s", message->authorization_protocol_name);
    g_string_append (text, " authorization-protocol-data=");
    if (message->authorization_protocol_data_length > 0)
    {
        int i;
        g_string_append (text, "0x");
        for (i = 0; i < message->authorization_protocol_data_length; i++)
            g_string_append_printf (text, "%02X", message->authorization_protocol_data[i]);
    }

    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text->str);
    g_string_free (text, TRUE);
}

static void
failed_cb (XigCodec *codec, XigFailed *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("Failed protocol-major-version=%d protocol-minor-version=%d reason=%s", message->protocol_major_version, message->protocol_minor_version, message->reason);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
authenticate_cb (XigCodec *codec, XigAuthenticate *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("Authenticate reason=%s", message->reason);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
log_pixmap_format (GString *text, XigFORMAT *value)
{
    g_string_append_printf (text, "depth=%d", value->depth);
    g_string_append_printf (text, " bits-per-pixel=%d", value->bits_per_pixel);
    g_string_append_printf (text, " scanline-pad=%d", value->scanline_pad);
}

static void
log_screen (GString *text, XigSCREEN *value)
{
    int i;

    g_string_append_printf (text, "root=0x%08X", value->root);
    g_string_append_printf (text, " default-colormap=0x%08X", value->default_colormap);
    g_string_append_printf (text, " white-pixel=0x%08X", value->white_pixel);
    g_string_append_printf (text, " black-pixel=0x%08X", value->black_pixel);
    g_string_append_printf (text, " current-input-masks=0x%08X", value->current_input_masks);
    g_string_append_printf (text, " width-in-pixels=%d", value->width_in_pixels);
    g_string_append_printf (text, " height-in-pixels=%d", value->height_in_pixels);
    g_string_append_printf (text, " width-in-millimeters=%d", value->width_in_millimeters);
    g_string_append_printf (text, " height-in-millimeters=%d", value->height_in_millimeters);
    g_string_append_printf (text, " min-installed-maps=%d", value->min_installed_maps);
    g_string_append_printf (text, " max-installed-maps=%d", value->max_installed_maps);
    g_string_append_printf (text, " root-visual=0x%08X", value->root_visual);
    g_string_append_printf (text, " backing-stores=%d", value->backing_stores);
    g_string_append_printf (text, " save-unders=%s", value->save_unders ? "true" : "false");
    g_string_append_printf (text, " root-depth=%d", value->root_depth);
    g_string_append (text, " allowed-depths=");
    for (i = 0; i < value->allowed_depths_length; i++)
    {
        int j;

        if (i != 0)
            g_string_append (text, ",");
        g_string_append (text, "{");
        g_string_append_printf (text, "depth=%d", value->allowed_depths[i].depth);
        g_string_append (text, " visuals=");
        for (j = 0; j < value->allowed_depths[i].visuals_length; j++)
        {
            XigVISUALTYPE *visual = &value->allowed_depths[i].visuals[i];

            if (i != 0)
                g_string_append (text, ",");
            g_string_append (text, "{");
            g_string_append_printf (text, "visual-id=%d", visual->visual_id);
            g_string_append_printf (text, " class=%d", visual->class);
            g_string_append_printf (text, " bits-per-rgb-value=%d", visual->bits_per_rgb_value);
            g_string_append_printf (text, " colormap-entries=%d", visual->colormap_entries);
            g_string_append_printf (text, " red-mask=0x%08X", visual->red_mask);
            g_string_append_printf (text, " green-mask=0x%08X", visual->green_mask);
            g_string_append_printf (text, " blue-mask=0x%08X", visual->blue_mask);
            g_string_append (text, "}");
        }
        g_string_append (text, "}");
    }
}

static void
success_cb (XigCodec *codec, XigSuccess *message, XigCodecLogger *logger)
{
    GString *text;
    int i;

    text = g_string_new ("Success");
    g_string_append_printf (text, " protocol-major-version=%d", message->protocol_major_version);
    g_string_append_printf (text, " protocol-minor-version=%d", message->protocol_minor_version);
    g_string_append_printf (text, " release-number=%d", message->release_number);
    g_string_append_printf (text, " resource_id_base=0x%08X", message->resource_id_base);
    g_string_append_printf (text, " resource_id_mask=0x%08X", message->resource_id_mask);
    g_string_append_printf (text, " motion-buffer-size=%d", message->motion_buffer_size);
    g_string_append_printf (text, " maximum-request-length=%d", message->maximum_request_length);
    g_string_append_printf (text, " image-byte-order=%d", message->image_byte_order);
    g_string_append_printf (text, " bitmap-format-bit-order=%d", message->bitmap_format_bit_order);
    g_string_append_printf (text, " bitmap-format-scanline-unit=%d", message->bitmap_format_scanline_unit);
    g_string_append_printf (text, " bitmap-format-scanline-pad=%d", message->bitmap_format_scanline_pad);
    g_string_append_printf (text, " min-keycode=%d", message->min_keycode);
    g_string_append_printf (text, " max-keycode=%d", message->max_keycode);
    g_string_append_printf (text, " vendor=%s", message->vendor);
    g_string_append (text, " pixmap-formats=");
    for (i = 0; i < message->pixmap_formats_length; i++)
    {
        if (i != 0)
            g_string_append (text, ",");
        g_string_append (text, "{");
        log_pixmap_format (text, &message->pixmap_formats[i]);
        g_string_append (text, "}");
    }
    g_string_append (text, " screens=");
    for (i = 0; i < message->screens_length; i++)
    {
        if (i != 0)
            g_string_append (text, ",");
        g_string_append (text, "{");
        log_screen (text, &message->screens[i]);
        g_string_append (text, "}");
    }

    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text->str);
    g_string_free (text, TRUE);
}

static void
error_cb (XigCodec *codec, XigError *message, XigCodecLogger *logger)
{
    gchar *text;
    if (message->type == Value)
        text = g_strdup_printf ("Error type=Value major_opcode=%d minor_opcode=%d value=0x%08X", message->major_opcode, message->minor_opcode, message->data);
    else if (message->type == Window)
        text = g_strdup_printf ("Error type=Window major_opcode=%d minor_opcode=%d id=0x%08X", message->major_opcode, message->minor_opcode, message->data);
    else if (message->type == Pixmap)
        text = g_strdup_printf ("Error type=Pixmap major_opcode=%d minor_opcode=%d id=0x%08X", message->major_opcode, message->minor_opcode, message->data);
    else if (message->type == Atom)
        text = g_strdup_printf ("Error type=Atom major_opcode=%d minor_opcode=%d id=0x%08X", message->major_opcode, message->minor_opcode, message->data);
    else if (message->type == Drawable)
        text = g_strdup_printf ("Error type=Drawable major_opcode=%d minor_opcode=%d id=0x%08X", message->major_opcode, message->minor_opcode, message->data);
    else if (message->type == IDChoice)
        text = g_strdup_printf ("Error type=IDChoice major_opcode=%d minor_opcode=%d id=0x%08X", message->major_opcode, message->minor_opcode, message->data);
    else if (message->type == Implementation)
        text = g_strdup_printf ("Error type=Implementation major_opcode=%d minor_opcode=%d", message->major_opcode, message->minor_opcode);
    else
        text = g_strdup_printf ("Error type=%d major_opcode=%d minor_opcode=%d", message->type, message->major_opcode, message->minor_opcode);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
create_notify_cb (XigCodec *codec, XigCreateNotify *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("CreateNotify window=0x%08X child=0x%08X x=%d y=%d, width=%u, height=%u",
                            message->parent, message->window,
                            message->x, message->y,
                            message->width, message->height);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text); 
}

static void
destroy_notify_cb (XigCodec *codec, XigDestroyNotify *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("DestroyNotify event=0x%08X window=0x%08X", message->event, message->window);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
unmap_notify_cb (XigCodec *codec, XigUnmapNotify *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("UnmapNotify event=0x%08X window=0x%08X", message->event, message->window);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
map_notify_cb (XigCodec *codec, XigMapNotify *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("MapNotify event=0x%08X window=0x%08X", message->event, message->window);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
map_request_cb (XigCodec *codec, XigMapRequest *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("MapRequest parent=0x%08X window=0x%08X", message->parent, message->window);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
reparent_notify_cb (XigCodec *codec, XigReparentNotify *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("ReparentNotify event=0x%08X window=0x%08X", message->event, message->window);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
configure_notify_cb (XigCodec *codec, XigConfigureNotify *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("ConfigureNotify event=0x%08X window=0x%08X x=%d y=%d, width=%u, height=%u border-width=%u",
                            message->event, message->window,
                            message->x, message->y,
                            message->width, message->height, message->border_width);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
configure_request_cb (XigCodec *codec, XigConfigureRequest *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("ConfigureRequest parent=0x%08X window=0x%08X", message->parent, message->window);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
circulate_notify_cb (XigCodec *codec, XigCirculateNotify *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("CirculateNotify event=0x%08X window=0x%08X", message->event, message->window);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
circulate_request_cb (XigCodec *codec, XigCirculateRequest *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("CirculateRequest parent=0x%08X window=0x%08X", message->parent, message->window);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
property_notify_cb (XigCodec *codec, XigPropertyNotify *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("PropertyNotify window=0x%08X property=%d", message->window, message->atom);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
create_window_cb (XigCodec *codec, XigCreateWindow *message, XigCodecLogger *logger)
{
    GString *text;

    text = g_string_new ("CreateWindow");
    g_string_append_printf (text, " wid=0x%08X parent=0x%08X x=%d y=%d width=%u height=%u",
                            message->wid, message->parent,
                            message->x, message->y,
                            message->width, message->height);
    if ((message->value_mask & XIG_WINDOW_VALUE_MASK_background_pixmap) != 0)
        g_string_append_printf (text, " background-pixmap=0x%08X", message->background_pixmap);
    if ((message->value_mask & XIG_WINDOW_VALUE_MASK_background_pixel) != 0)
        g_string_append_printf (text, " background-pixel=0x%08X", message->background_pixel);
    if ((message->value_mask & XIG_WINDOW_VALUE_MASK_border_pixmap) != 0)
        g_string_append_printf (text, " border-pixmap=0x%08X", message->border_pixmap);
    if ((message->value_mask & XIG_WINDOW_VALUE_MASK_border_pixel) != 0)
        g_string_append_printf (text, " border-pixel=0x%08X", message->border_pixel);
    if ((message->value_mask & XIG_WINDOW_VALUE_MASK_bit_gravity) != 0)
        g_string_append_printf (text, " bit-gravity=%d", message->bit_gravity);
    if ((message->value_mask & XIG_WINDOW_VALUE_MASK_win_gravity) != 0)
        g_string_append_printf (text, " win-gravity=%d", message->win_gravity);
    if ((message->value_mask & XIG_WINDOW_VALUE_MASK_backing_store) != 0)
        g_string_append_printf (text, " backing-store=%d", message->backing_store);
    if ((message->value_mask & XIG_WINDOW_VALUE_MASK_backing_planes) != 0)
        g_string_append_printf (text, " backing-planes=%d", message->backing_planes);
    if ((message->value_mask & XIG_WINDOW_VALUE_MASK_backing_pixel) != 0)
        g_string_append_printf (text, " backing-pixel=0x%08X", message->backing_pixel);
    if ((message->value_mask & XIG_WINDOW_VALUE_MASK_override_redirect) != 0)
        g_string_append_printf (text, " override-redirect=%s", message->override_redirect ? "true" : "false");
    if ((message->value_mask & XIG_WINDOW_VALUE_MASK_save_under) != 0)
        g_string_append_printf (text, " save-under=%s", message->save_under ? "true" : "false");
    if ((message->value_mask & XIG_WINDOW_VALUE_MASK_event_mask) != 0)
        g_string_append_printf (text, " event-mask=0x%08X", message->event_mask);
    if ((message->value_mask & XIG_WINDOW_VALUE_MASK_do_not_propagate_mask) != 0)
        g_string_append_printf (text, " do-not-propagate-mask=0x%08X", message->do_not_propagate_mask);
    if ((message->value_mask & XIG_WINDOW_VALUE_MASK_colormap) != 0)
        g_string_append_printf (text, " colormap=0x%08X", message->colormap);
    if ((message->value_mask & XIG_WINDOW_VALUE_MASK_cursor) != 0)
        g_string_append_printf (text, " cursor=0x%08X", message->cursor);

    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text->str);
    g_string_free (text, TRUE);
}

static void 
change_window_attributes_cb (XigCodec *codec, XigChangeWindowAttributes *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("ChangeWindowAttributes window=0x%08X value-mask=0x%08X", message->window, message->value_mask);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
get_window_attributes_cb (XigCodec *codec, XigGetWindowAttributes *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("GetWindowAttributes window=0x%08X", message->window);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
get_window_attributes_reply_cb (XigCodec *codec, XigGetWindowAttributesReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("GetWindowAttributes class=%d visual=0x%08X backing-store=%d bit-gravity=%d win-gravity=%d backing-planes=0x%08X backing-pixel=0x%08X save-under=%s override-redirect=%s all-event-masks=0x%08X events=0x%08X",
                            message->class,
                            message->visual,
                            message->backing_store,
                            message->bit_gravity,
                            message->win_gravity,
                            message->backing_planes,
                            message->backing_pixel,
                            message->save_under ? "true" : "false",
                            message->override_redirect ? "true" : "false",
                            message->all_event_masks,
                            message->your_event_mask);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void 
destroy_window_cb (XigCodec *codec, XigDestroyWindow *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("DestroyWindow window=0x%08X", message->window);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
destroy_subwindows_cb (XigCodec *codec, XigDestroySubwindows *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("DestroySubwindows window=0x%08X", message->window);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
change_save_set_cb (XigCodec *codec, XigChangeSaveSet *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("ChangeSaveSet");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
reparent_window_cb (XigCodec *codec, XigReparentWindow *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("ReparentWindow window=0x%08X parent=0x%08X", message->window, message->parent);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
map_window_cb (XigCodec *codec, XigMapWindow *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("MapWindow window=0x%08X", message->window);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
map_subwindows_cb (XigCodec *codec, XigMapSubwindows *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("MapSubwindows window=0x%08X", message->window);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
unmap_window_cb (XigCodec *codec, XigUnmapWindow *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("UnmapWindow window=0x%08X", message->window);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
unmap_subwindows_cb (XigCodec *codec, XigUnmapSubwindows *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("UnmapSubwindows window=0x%08X", message->window);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
configure_window_cb (XigCodec *codec, XigConfigureWindow *message, XigCodecLogger *logger)
{
    GString *text;

    text = g_string_new ("ConfigureWindow");
    g_string_append_printf (text, " window=0x%08X", message->window);
    if ((message->value_mask & XIG_CONFIGURE_WINDOW_VALUE_MASK_x) != 0)
        g_string_append_printf (text, " x=%d", message->x);
    if ((message->value_mask & XIG_CONFIGURE_WINDOW_VALUE_MASK_y) != 0)
        g_string_append_printf (text, " y=%d", message->y);
    if ((message->value_mask & XIG_CONFIGURE_WINDOW_VALUE_MASK_width) != 0)
        g_string_append_printf (text, " width=%u", message->width);
    if ((message->value_mask & XIG_CONFIGURE_WINDOW_VALUE_MASK_height) != 0)
        g_string_append_printf (text, " height=%u", message->height);
    if ((message->value_mask & XIG_CONFIGURE_WINDOW_VALUE_MASK_border_width) != 0)
        g_string_append_printf (text, " border-width=%u", message->border_width);
    if ((message->value_mask & XIG_CONFIGURE_WINDOW_VALUE_MASK_sibling) != 0)
        g_string_append_printf (text, " sibling=0x%08X", message->sibling);
    if ((message->value_mask & XIG_CONFIGURE_WINDOW_VALUE_MASK_stack_mode) != 0)
        g_string_append_printf (text, " stack-mode=%d", message->stack_mode);

    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text->str);
    g_string_free (text, TRUE);
}

static void 
circulate_window_cb (XigCodec *codec, XigCirculateWindow *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("CirculateWindow window=0x%08X", message->window);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
get_geometry_cb (XigCodec *codec, XigGetGeometry *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("GetGeometry drawable=0x%08X", message->drawable);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
get_geometry_reply_cb (XigCodec *codec, XigGetGeometryReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("GetGeometry depth=%d root=0x%08X x=%d y=%d width=%u height=%u, border-width=%u",
                            message->depth,
                            message->root,
                            message->x, message->y,
                            message->width, message->height,
                            message->border_width);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void 
query_tree_cb (XigCodec *codec, XigQueryTree *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("QueryTree window=0x%08X", message->window);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
query_tree_reply_cb (XigCodec *codec, XigQueryTreeReply *message, XigCodecLogger *logger)
{
    GString *text;
    int i;
  
    text = g_string_new ("QueryTree");
    g_string_append_printf (text, " root=0x%08X parent=0x%08X children=", message->root, message->parent);
    if (message->children_length > 0)
        g_string_append_printf (text, "0x%08X", message->children[0]);
    for (i = 1; i < message->children_length; i++)
        g_string_append_printf (text, ",0x%08X", message->children[i]);

    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text->str);
    g_string_free (text, TRUE);
}

static void 
intern_atom_cb (XigCodec *codec, XigInternAtom *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("InternAtom name=%s only-if-exists=%s", message->name, message->only_if_exists ? "true" : "false");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
intern_atom_reply_cb (XigCodec *codec, XigInternAtomReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("InternAtom atom=%d", message->atom);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void 
get_atom_name_cb (XigCodec *codec, XigGetAtomName *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("GetAtomName atom=%d", message->atom);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
get_atom_name_reply_cb (XigCodec *codec, XigGetAtomNameReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("GetAtomName name=%s", message->name);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void 
change_property_cb (XigCodec *codec, XigChangeProperty *message, XigCodecLogger *logger)
{
    GString *text;
    //const gchar *name, *type;
    guint8 *data8;
    guint16 *data16;
    guint32 *data32;
    int i;

    text = g_string_new ("ChangeProperty");
    g_string_append_printf (text, " window=0x%08X", message->window);
    //name = xig_server_get_atom_name (xig_codec_get_server (codec), message->property);
    //if (name)
    //    g_string_append_printf (text, " property=%s", name);
    //else
        g_string_append_printf (text, " property=%d", message->property);
    //g_free (name);
    //type = xig_server_get_atom_name (xig_codec_get_server (codec), message->type);
    //if (type)
    //    g_string_append_printf (text, " type=%s", type);
    //else
        g_string_append_printf (text, " type=%d", message->type);
    //g_free (type);
    g_string_append_printf (text, " mode=%d", message->mode);
    g_string_append (text, " data=");
    switch (message->data_format)
    {
    case 8:
        data8 = message->data;
        for (i = 0; i < message->data_length; i++)
            g_string_append_printf (text, "%02X", data8[i]);
        break;
    case 16:
        data16 = message->data;
        if (message->data_length > 0)
            g_string_append_printf (text, "0x%04X", data16[0]);
        for (i = 1; i < message->data_length; i++)
            g_string_append_printf (text, ",0x%04X", data16[i]);
        break;
    case 32:
        data32 = message->data;
        if (message->data_length > 0)
            g_string_append_printf (text, "0x%08X", data32[0]);
        for (i = 1; i < message->data_length; i++)
            g_string_append_printf (text, ",0x%08X", data32[i]);
        break;
    }

    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text->str);
    g_string_free (text, TRUE);
}

static void 
delete_property_cb (XigCodec *codec, XigDeleteProperty *message, XigCodecLogger *logger)
{
    gchar *text;
    //const gchar *name;

    //name = xig_server_get_atom_name (xig_codec_get_server (codec), message->property);
    //if (name)
    //    text = g_strdup_printf ("DeleteProperty window=0x%08X %s", message->window, name);
    //else
        text = g_strdup_printf ("DeleteProperty window=0x%08X %d", message->window, message->property);

    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
get_property_cb (XigCodec *codec, XigGetProperty *message, XigCodecLogger *logger)
{
    gchar *text;
    //const gchar *name, *type_name = NULL;

    //name = xig_server_get_atom_name (xig_codec_get_server (codec), message->property);
    //if (message->type != 0)
    //    type_name = xig_server_get_atom_name (xig_codec_get_server (codec), message->type);

    //if (name && type_name)
    //    text = g_strdup_printf ("GetProperty window=0x%08X property=%s type=%s delete=%s long-offset=%d long-length=%d", message->window, name, type_name, message->delete ? "true" : "false", message->long_offset, message->long_length);
    //else if (name)
    //    text = g_strdup_printf ("GetProperty window=0x%08X property=%s type=%d delete=%s long-offset=%d long-length=%d", message->window, name, message->type, message->delete ? "true" : "false", message->long_offset, message->long_length);
    //else
        text = g_strdup_printf ("GetProperty window=0x%08X property=%d type=%d delete=%s long-offset=%d long-length=%d", message->window, message->property, message->type, message->delete ? "true" : "false", message->long_offset, message->long_length);

    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
get_property_reply_cb (XigCodec *codec, XigGetPropertyReply *message, XigCodecLogger *logger)
{
    gchar *text;
    //const gchar *type;
      
    //type = xig_server_get_atom_name (xig_codec_get_server (codec), message->type);
    //if (type)
    //    text = g_strdup_printf ("GetProperty type=%s format=%d bytes-after=%d length=%zi", type, message->value_format, message->bytes_after, message->value_length);
    //else
        text = g_strdup_printf ("GetProperty type=%d format=%d bytes-after=%d length=%zi", message->type, message->value_format, message->bytes_after, message->value_length);

    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void 
list_properties_cb (XigCodec *codec, XigListProperties *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("ListProperties window=0x%08X", message->window);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
list_properties_reply_cb (XigCodec *codec, XigListPropertiesReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("ListProperties");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
set_selection_owner_cb (XigCodec *codec, XigSetSelectionOwner *message, XigCodecLogger *logger)
{
    gchar *text;
    //const gchar *name;

    //name = xig_server_get_atom_name (xig_codec_get_server (codec), message->selection);
    //if (name)
    //    text = g_strdup_printf ("SetSelectionOwner owner=0x%08X selection=%s", message->owner, name);
    //else
        text = g_strdup_printf ("SetSelectionOwner owner=0x%08X selection=%d", message->owner, message->selection);

    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
get_selection_owner_cb (XigCodec *codec, XigGetSelectionOwner *message, XigCodecLogger *logger)
{
    gchar *text;
    //const gchar *name;

    //name = xig_server_get_atom_name (xig_codec_get_server (codec), message->selection);
    //if (name)
    //    text = g_strdup_printf ("GetSelectionOwner selection=%s", name);
    //else
        text = g_strdup_printf ("GetSelectionOwner selection=%d", message->selection);

    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
get_selection_owner_reply_cb (XigCodec *codec, XigGetSelectionOwnerReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("GetSelectionOwner owner=0x%08X", message->owner);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
convert_selection_cb (XigCodec *codec, XigConvertSelection *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("ConvertSelection");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
send_event_cb (XigCodec *codec, XigSendEvent *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("SendEvent window=0x%08X", message->destination);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
grab_pointer_cb (XigCodec *codec, XigGrabPointer *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("GrabPointer grab-window=0x%08X", message->grab_window);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
grab_pointer_reply_cb (XigCodec *codec, XigGrabPointerReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("GrabPointer status=%d", message->status);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
ungrab_pointer_cb (XigCodec *codec, XigUngrabPointer *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("UngrabPointer");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
grab_button_cb (XigCodec *codec, XigGrabButton *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("GrabButton button=%d modifiers=0x%04X grab-window=0x%08X", message->button, message->modifiers, message->grab_window);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
ungrab_button_cb (XigCodec *codec, XigUngrabButton *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("UngrabButton button=%d modifiers=0x%04X grab-window=0x%08X", message->button, message->modifiers, message->grab_window);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
change_active_pointer_grab_cb (XigCodec *codec, XigChangeActivePointerGrab *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("ChangeActivePointerGrab cursor=0x%08X time=%d event-mask=0x%04X", message->cursor, message->time, message->event_mask);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
grab_keyboard_cb (XigCodec *codec, XigGrabKeyboard *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("GrabKeyboard owner-events=%s grab-window=0x%08X time=%d pointer-mode=%d keyboard-mode=%d", message->owner_events ? "true" : "false", message->grab_window, message->time, message->pointer_mode, message->keyboard_mode);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
grab_keyboard_reply_cb (XigCodec *codec, XigGrabKeyboardReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("GrabKeyboardReply status=%d", message->status);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
ungrab_keyboard_cb (XigCodec *codec, XigUngrabKeyboard *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("UngrabKeyboard time=%d", message->time);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
grab_key_cb (XigCodec *codec, XigGrabKey *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("GrabKey");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
ungrab_key_cb (XigCodec *codec, XigUngrabKey *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("UngrabKey");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
allow_events_cb (XigCodec *codec, XigAllowEvents *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("AllowEvents");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
grab_server_cb (XigCodec *codec, XigGrabServer *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("GrabServer");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
ungrab_server_cb (XigCodec *codec, XigUngrabServer *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("UngrabServer");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
query_pointer_cb (XigCodec *codec, XigQueryPointer *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("QueryPointer window=0x%08X", message->window);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
query_pointer_reply_cb (XigCodec *codec, XigQueryPointerReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("QueryPointer same-screen=%s root=0x%08X child=0x%08X root-x=%d root-y=%d win-x=%d win-y=%d mask=0x%04X",
                            message->same_screen ? "true" : "false",
                            message->root,
                            message->child,
                            message->root_x,
                            message->root_y,
                            message->win_x,
                            message->win_y,
                            message->mask);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
get_motion_events_cb (XigCodec *codec, XigGetMotionEvents *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("GetMotionEvents window=0x%08X start=%d stop=%d", message->window, message->start, message->stop);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
get_motion_events_reply_cb (XigCodec *codec, XigGetMotionEventsReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("GetMotionEvents events-length=%zi", message->events_length);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
translate_coordinates_cb (XigCodec *codec, XigTranslateCoordinates *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("TranslateCoordinates src-window=0x%08X dst-window=0x%08X src-x=%d src-y=%d", message->src_window, message->dst_window, message->src_x, message->src_y);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
translate_coordinates_reply_cb (XigCodec *codec, XigTranslateCoordinatesReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("TranslateCoordinates child=0x%08X dst-x=%d dst-y=%d", message->child, message->dst_x, message->dst_y);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
warp_pointer_cb (XigCodec *codec, XigWarpPointer *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("WarpPointer src-window=0x%08X dst-window=0x%08X src-x=%d src-y=%d src-width=%d src-height=%d dst-x=%d dst-y=%d", message->src_window, message->dst_window, message->src_x, message->src_y, message->src_width, message->src_height, message->dst_x, message->dst_y);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
set_input_focus_cb (XigCodec *codec, XigSetInputFocus *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("SetInputFocus focus=0x%08X time=%d revert-to=%d", message->focus, message->time, message->revert_to);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
get_input_focus_cb (XigCodec *codec, XigGetInputFocus *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("GetInputFocus");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
get_input_focus_reply_cb (XigCodec *codec, XigGetInputFocusReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("GetInputFocus focus=0x%08X revert-to=%d", message->focus, message->revert_to);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void 
query_keymap_cb (XigCodec *codec, XigQueryKeymap *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("QueryKeymap");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
query_keymap_reply_cb (XigCodec *codec, XigQueryKeymapReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("QueryKeymap keys=0x%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X",
                            message->keys[0],
                            message->keys[1],
                            message->keys[2],
                            message->keys[3],
                            message->keys[4],
                            message->keys[5],
                            message->keys[6],
                            message->keys[7],
                            message->keys[8],
                            message->keys[9],
                            message->keys[10],
                            message->keys[11],
                            message->keys[12],
                            message->keys[13],
                            message->keys[14],
                            message->keys[15],
                            message->keys[16],
                            message->keys[17],
                            message->keys[18],
                            message->keys[19],
                            message->keys[20],
                            message->keys[21],
                            message->keys[22],
                            message->keys[23],
                            message->keys[24],
                            message->keys[25],
                            message->keys[26],
                            message->keys[27],
                            message->keys[28],
                            message->keys[29],
                            message->keys[30],
                            message->keys[31]);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void 
open_font_cb (XigCodec *codec, XigOpenFont *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("OpenFont fid=0x%08X name=%s", message->fid, message->name);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
close_font_cb (XigCodec *codec, XigCloseFont *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("CloseFont font=0x%08X", message->font);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
query_font_cb (XigCodec *codec, XigQueryFont *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("QueryFont font=0x%08X", message->font);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
query_font_reply_cb (XigCodec *codec, XigQueryFontReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("QueryFont");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void 
query_text_extents_cb (XigCodec *codec, XigQueryTextExtents *message, XigCodecLogger *logger)
{
    GString *text;
    int i;

    text = g_string_new ("QueryTextExtents");
    g_string_append_printf (text, " font=0x%08X", message->font);
    g_string_append_printf (text, " string=");
    for (i = 0; i < message->string_length; i++)
        g_string_append_unichar (text, message->string[i]);

    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text->str);
    g_string_free (text, TRUE);
}

static void
query_text_extents_reply_cb (XigCodec *codec, XigQueryTextExtentsReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("QueryTextExtents draw-direction=%d font-ascent=%d font-descent=%d overall-ascent=%d overall-descent=%d overall-width=%d overall-left=%d overall-right=%d", message->draw_direction, message->font_ascent, message->font_descent, message->overall_ascent, message->overall_descent, message->overall_width, message->overall_left, message->overall_right);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void 
list_fonts_cb (XigCodec *codec, XigListFonts *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("ListFonts max-names=%d pattern=%s", message->max_names, message->pattern);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
list_fonts_reply_cb (XigCodec *codec, XigListFontsReply *message, XigCodecLogger *logger)
{
    GString *text;
    int i;
  
    text = g_string_new ("ListFonts names=");
    if (message->names[0])
    {
        g_string_append_printf (text, "%s", message->names[0]);
        for (i = 1; message->names[i]; i++)
            g_string_append_printf (text, ",%s", message->names[i]);
    }
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text->str);
    g_string_free (text, TRUE);
}

static void 
list_fonts_with_info_cb (XigCodec *codec, XigListFontsWithInfo *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("ListFontsWithInfo max-names=%d pattern=%s", message->max_names, message->pattern);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
list_fonts_with_info_reply_cb (XigCodec *codec, XigListFontsWithInfoReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("ListFontsWithInfo ...");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
set_font_path_cb (XigCodec *codec, XigSetFontPath *message, XigCodecLogger *logger)
{
    GString *text;
    int i;
  
    text = g_string_new ("SetFontPath path=");
    if (message->path[0])
    {
        g_string_append_printf (text, "%s", message->path[0]);
        for (i = 1; message->path[i]; i++)
            g_string_append_printf (text, ",%s", message->path[i]);
    }

    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text->str);
    g_string_free (text, TRUE);
}

static void
get_font_path_cb (XigCodec *codec, XigGetFontPath *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("GetFontPath");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
get_font_path_reply_cb (XigCodec *codec, XigGetFontPathReply *message, XigCodecLogger *logger)
{
    GString *text;
    int i;
  
    text = g_string_new ("GetFontPath path=");
    if (message->path[0])
    {
        g_string_append_printf (text, "%s", message->path[0]);
        for (i = 1; message->path[i]; i++)
            g_string_append_printf (text, ",%s", message->path[i]);
    }

    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text->str);
    g_string_free (text, TRUE);
}

static void 
create_pixmap_cb (XigCodec *codec, XigCreatePixmap *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("CreatePixmap pid=0x%08X drawable=0x%08X width=%d height=%d depth=%d", message->pid, message->drawable, message->width, message->height, message->depth);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
free_pixmap_cb (XigCodec *codec, XigFreePixmap *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("FreePixmap pixmap=0x%08X", message->pixmap);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
create_gc_cb (XigCodec *codec, XigCreateGC *message, XigCodecLogger *logger)
{
    GString *text;

    text = g_string_new ("CreateGC");
    g_string_append_printf (text, " cid=0x%08X", message->cid);
    g_string_append_printf (text, " drawable=0x%08X", message->drawable);
    if (message->value_mask & XIG_GC_VALUE_MASK_function)
        g_string_append_printf (text, " function=%d", message->function);
    if (message->value_mask & XIG_GC_VALUE_MASK_plane_mask)
        g_string_append_printf (text, " plane_mask=0x%08X", message->plane_mask);
    if (message->value_mask & XIG_GC_VALUE_MASK_foreground)
        g_string_append_printf (text, " foreground=0x%08X", message->foreground);
    if (message->value_mask & XIG_GC_VALUE_MASK_background)
        g_string_append_printf (text, " background=0x%08X", message->background);
    if (message->value_mask & XIG_GC_VALUE_MASK_line_width)
        g_string_append_printf (text, " line_width=%d", message->line_width);
    if (message->value_mask & XIG_GC_VALUE_MASK_line_style)
        g_string_append_printf (text, " line_style=%d", message->line_style);
    if (message->value_mask & XIG_GC_VALUE_MASK_cap_style)
        g_string_append_printf (text, " cap_style=%d", message->cap_style);
    if (message->value_mask & XIG_GC_VALUE_MASK_join_style)
        g_string_append_printf (text, " join_style=%d", message->join_style);
    if (message->value_mask & XIG_GC_VALUE_MASK_fill_style)
        g_string_append_printf (text, " fill_style=%d", message->fill_style);
    if (message->value_mask & XIG_GC_VALUE_MASK_fill_rule)
        g_string_append_printf (text, " fill_rule=%d", message->fill_rule);
    if (message->value_mask & XIG_GC_VALUE_MASK_tile)
        g_string_append_printf (text, " tile=0x%08X", message->tile);
    if (message->value_mask & XIG_GC_VALUE_MASK_stipple)
        g_string_append_printf (text, " stipple=0x%08X", message->stipple);
    if (message->value_mask & XIG_GC_VALUE_MASK_tile_stipple_x_origin)
        g_string_append_printf (text, " tile_stipple_x_origin=%d", message->tile_stipple_x_origin);
    if (message->value_mask & XIG_GC_VALUE_MASK_tile_stipple_y_origin)
        g_string_append_printf (text, " tile_stipple_y_origin=%d", message->tile_stipple_y_origin);
    if (message->value_mask & XIG_GC_VALUE_MASK_font)
        g_string_append_printf (text, " font=0x%08X", message->font);
    if (message->value_mask & XIG_GC_VALUE_MASK_subwindow_mode)
        g_string_append_printf (text, " subwindow_mode=%d", message->subwindow_mode);
    if (message->value_mask & XIG_GC_VALUE_MASK_graphics_exposures ? "true" : "false")
        g_string_append_printf (text, " graphics_exposures=%s", message->graphics_exposures ? "true" : "false");
    if (message->value_mask & XIG_GC_VALUE_MASK_clip_x_origin)
        g_string_append_printf (text, " clip_x_origin=%d", message->clip_x_origin);
    if (message->value_mask & XIG_GC_VALUE_MASK_clip_y_origin)
        g_string_append_printf (text, " clip_y_origin=%d", message->clip_y_origin);
    if (message->value_mask & XIG_GC_VALUE_MASK_clip_mask)
        g_string_append_printf (text, " clip_mask=0x%08X", message->clip_mask);
    if (message->value_mask & XIG_GC_VALUE_MASK_dash_offset)
        g_string_append_printf (text, " dash_offset=%d", message->dash_offset);
    if (message->value_mask & XIG_GC_VALUE_MASK_dashes)
        g_string_append_printf (text, " dashes=%d", message->dashes);
    if (message->value_mask & XIG_GC_VALUE_MASK_arc_mode)
        g_string_append_printf (text, " arc_mode=%d", message->arc_mode);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text->str);
    g_string_free (text, TRUE);
}

static void 
change_gc_cb (XigCodec *codec, XigChangeGC *message, XigCodecLogger *logger)
{
    GString *text;

    text = g_string_new ("ChangeGC");
    g_string_append_printf (text, " gc=0x%08X", message->gc);
    if (message->value_mask & XIG_GC_VALUE_MASK_function)
        g_string_append_printf (text, " function=%d", message->function);
    if (message->value_mask & XIG_GC_VALUE_MASK_plane_mask)
        g_string_append_printf (text, " plane_mask=0x%08X", message->plane_mask);
    if (message->value_mask & XIG_GC_VALUE_MASK_foreground)
        g_string_append_printf (text, " foreground=0x%08X", message->foreground);
    if (message->value_mask & XIG_GC_VALUE_MASK_background)
        g_string_append_printf (text, " background=0x%08X", message->background);
    if (message->value_mask & XIG_GC_VALUE_MASK_line_width)
        g_string_append_printf (text, " line_width=%d", message->line_width);
    if (message->value_mask & XIG_GC_VALUE_MASK_line_style)
        g_string_append_printf (text, " line_style=%d", message->line_style);
    if (message->value_mask & XIG_GC_VALUE_MASK_cap_style)
        g_string_append_printf (text, " cap_style=%d", message->cap_style);
    if (message->value_mask & XIG_GC_VALUE_MASK_join_style)
        g_string_append_printf (text, " join_style=%d", message->join_style);
    if (message->value_mask & XIG_GC_VALUE_MASK_fill_style)
        g_string_append_printf (text, " fill_style=%d", message->fill_style);
    if (message->value_mask & XIG_GC_VALUE_MASK_fill_rule)
        g_string_append_printf (text, " fill_rule=%d", message->fill_rule);
    if (message->value_mask & XIG_GC_VALUE_MASK_tile)
        g_string_append_printf (text, " tile=0x%08X", message->tile);
    if (message->value_mask & XIG_GC_VALUE_MASK_stipple)
        g_string_append_printf (text, " stipple=0x%08X", message->stipple);
    if (message->value_mask & XIG_GC_VALUE_MASK_tile_stipple_x_origin)
        g_string_append_printf (text, " tile_stipple_x_origin=%d", message->tile_stipple_x_origin);
    if (message->value_mask & XIG_GC_VALUE_MASK_tile_stipple_y_origin)
        g_string_append_printf (text, " tile_stipple_y_origin=%d", message->tile_stipple_y_origin);
    if (message->value_mask & XIG_GC_VALUE_MASK_font)
        g_string_append_printf (text, " font=0x%08X", message->font);
    if (message->value_mask & XIG_GC_VALUE_MASK_subwindow_mode)
        g_string_append_printf (text, " subwindow_mode=%d", message->subwindow_mode);
    if (message->value_mask & XIG_GC_VALUE_MASK_graphics_exposures ? "true" : "false")
        g_string_append_printf (text, " graphics_exposures=%s", message->graphics_exposures ? "true" : "false");
    if (message->value_mask & XIG_GC_VALUE_MASK_clip_x_origin)
        g_string_append_printf (text, " clip_x_origin=%d", message->clip_x_origin);
    if (message->value_mask & XIG_GC_VALUE_MASK_clip_y_origin)
        g_string_append_printf (text, " clip_y_origin=%d", message->clip_y_origin);
    if (message->value_mask & XIG_GC_VALUE_MASK_clip_mask)
        g_string_append_printf (text, " clip_mask=0x%08X", message->clip_mask);
    if (message->value_mask & XIG_GC_VALUE_MASK_dash_offset)
        g_string_append_printf (text, " dash_offset=%d", message->dash_offset);
    if (message->value_mask & XIG_GC_VALUE_MASK_dashes)
        g_string_append_printf (text, " dashes=%d", message->dashes);
    if (message->value_mask & XIG_GC_VALUE_MASK_arc_mode)
        g_string_append_printf (text, " arc_mode=%d", message->arc_mode);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text->str);
    g_string_free (text, TRUE);
}

static void 
copy_gc_cb (XigCodec *codec, XigCopyGC *message, XigCodecLogger *logger)
{
    GString *text;

    text = g_string_new ("CopyGC");
    g_string_append_printf (text, " src_gc=0x%08X", message->src_gc);
    g_string_append_printf (text, " dst_gc=0x%08X", message->dst_gc);
    if (message->value_mask & XIG_GC_VALUE_MASK_function)
        g_string_append_printf (text, " function=%d", message->function);
    if (message->value_mask & XIG_GC_VALUE_MASK_plane_mask)
        g_string_append_printf (text, " plane_mask=0x%08X", message->plane_mask);
    if (message->value_mask & XIG_GC_VALUE_MASK_foreground)
        g_string_append_printf (text, " foreground=0x%08X", message->foreground);
    if (message->value_mask & XIG_GC_VALUE_MASK_background)
        g_string_append_printf (text, " background=0x%08X", message->background);
    if (message->value_mask & XIG_GC_VALUE_MASK_line_width)
        g_string_append_printf (text, " line_width=%d", message->line_width);
    if (message->value_mask & XIG_GC_VALUE_MASK_line_style)
        g_string_append_printf (text, " line_style=%d", message->line_style);
    if (message->value_mask & XIG_GC_VALUE_MASK_cap_style)
        g_string_append_printf (text, " cap_style=%d", message->cap_style);
    if (message->value_mask & XIG_GC_VALUE_MASK_join_style)
        g_string_append_printf (text, " join_style=%d", message->join_style);
    if (message->value_mask & XIG_GC_VALUE_MASK_fill_style)
        g_string_append_printf (text, " fill_style=%d", message->fill_style);
    if (message->value_mask & XIG_GC_VALUE_MASK_fill_rule)
        g_string_append_printf (text, " fill_rule=%d", message->fill_rule);
    if (message->value_mask & XIG_GC_VALUE_MASK_tile)
        g_string_append_printf (text, " tile=0x%08X", message->tile);
    if (message->value_mask & XIG_GC_VALUE_MASK_stipple)
        g_string_append_printf (text, " stipple=0x%08X", message->stipple);
    if (message->value_mask & XIG_GC_VALUE_MASK_tile_stipple_x_origin)
        g_string_append_printf (text, " tile_stipple_x_origin=%d", message->tile_stipple_x_origin);
    if (message->value_mask & XIG_GC_VALUE_MASK_tile_stipple_y_origin)
        g_string_append_printf (text, " tile_stipple_y_origin=%d", message->tile_stipple_y_origin);
    if (message->value_mask & XIG_GC_VALUE_MASK_font)
        g_string_append_printf (text, " font=0x%08X", message->font);
    if (message->value_mask & XIG_GC_VALUE_MASK_subwindow_mode)
        g_string_append_printf (text, " subwindow_mode=%d", message->subwindow_mode);
    if (message->value_mask & XIG_GC_VALUE_MASK_graphics_exposures ? "true" : "false")
        g_string_append_printf (text, " graphics_exposures=%s", message->graphics_exposures ? "true" : "false");
    if (message->value_mask & XIG_GC_VALUE_MASK_clip_x_origin)
        g_string_append_printf (text, " clip_x_origin=%d", message->clip_x_origin);
    if (message->value_mask & XIG_GC_VALUE_MASK_clip_y_origin)
        g_string_append_printf (text, " clip_y_origin=%d", message->clip_y_origin);
    if (message->value_mask & XIG_GC_VALUE_MASK_clip_mask)
        g_string_append_printf (text, " clip_mask=0x%08X", message->clip_mask);
    if (message->value_mask & XIG_GC_VALUE_MASK_dash_offset)
        g_string_append_printf (text, " dash_offset=%d", message->dash_offset);
    if (message->value_mask & XIG_GC_VALUE_MASK_dashes)
        g_string_append_printf (text, " dashes=%d", message->dashes);
    if (message->value_mask & XIG_GC_VALUE_MASK_arc_mode)
        g_string_append_printf (text, " arc_mode=%d", message->arc_mode);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text->str);
    g_string_free (text, TRUE);
}

static void 
set_dashes_cb (XigCodec *codec, XigSetDashes *message, XigCodecLogger *logger)
{
    GString *text;
    int i;

    text = g_string_new ("SetDashes");
    g_string_append_printf (text, " gc=0x%08X", message->gc);
    g_string_append_printf (text, " dash-offset=%d", message->dash_offset);
    g_string_append (text, " dashes=0x");
    for (i = 0; i < message->dashes_length; i++)
        g_string_append_printf (text, "%02X", message->dashes[i]);

    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text->str);
    g_string_free (text, TRUE);   
}

static void 
set_clip_rectangles_cb (XigCodec *codec, XigSetClipRectangles *message, XigCodecLogger *logger)
{
    GString *text;

    text = g_string_new ("SetClipRectangles");
    g_string_append_printf (text, " gc=0x%08X", message->gc);
    g_string_append_printf (text, " ordering=%d", message->ordering);
    g_string_append_printf (text, " clip-x-origin=%d", message->clip_y_origin);
    g_string_append_printf (text, " clip-y-origin=%d", message->clip_y_origin);
    g_string_append (text, " rectangles=");
    if (message->rectangles_length > 0)
    {
        int i;
        g_string_append_printf (text, "{x=%d y=%d width=%d height=%d}", message->rectangles[0].x, message->rectangles[0].y, message->rectangles[0].width, message->rectangles[0].height);
        for (i = 1; i < message->rectangles_length; i++)
            g_string_append_printf (text, ",{x=%d y=%d width=%d height=%d}", message->rectangles[i].x, message->rectangles[i].y, message->rectangles[i].width, message->rectangles[i].height);          
    }

    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text->str);
    g_string_free (text, TRUE);
}

static void 
free_gc_cb (XigCodec *codec, XigFreeGC *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("FreeGC gc=0x%08X", message->gc);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
clear_area_cb (XigCodec *codec, XigClearArea *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("ClearArea window=0x%08X exposures=%s x=%d y=%d width=%d height=%d", message->window, message->exposures ? "true" : "false", message->x, message->y, message->width, message->height);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
copy_area_cb (XigCodec *codec, XigCopyArea *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("CopyArea src-drawable=0x%08X dst-drawable=0x%08X gc=0x%08X src-x=%d src-y=%d dst-x=%d dst-y=%d width=%d height=%d", message->src_drawable, message->dst_drawable, message->gc, message->src_x, message->src_y, message->dst_x, message->dst_y, message->width, message->height);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
copy_plane_cb (XigCodec *codec, XigCopyPlane *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("CopyPlane src-drawable=0x%08X dst-drawable=0x%08X gc=0x%08X src-x=%d src-y=%d dst-x=%d dst-y=%d width=%d height=%d bit-plane=0x%08X", message->src_drawable, message->dst_drawable, message->gc, message->src_x, message->src_y, message->dst_x, message->dst_y, message->width, message->height, message->bit_plane);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
poly_point_cb (XigCodec *codec, XigPolyPoint *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("PolyPoint ...");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
poly_line_cb (XigCodec *codec, XigPolyLine *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("PolyLine ...");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
poly_segment_cb (XigCodec *codec, XigPolySegment *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("PolySegment ...");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
poly_rectangle_cb (XigCodec *codec, XigPolyRectangle *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("PolyRectangle ...");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
poly_arc_cb (XigCodec *codec, XigPolyArc *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("PolyArc ...");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
fill_poly_cb (XigCodec *codec, XigFillPoly *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("FillPoly ...");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
poly_fill_rectangle_cb (XigCodec *codec, XigPolyFillRectangle *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("PolyFillRectangle ...");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
poly_fill_arc_cb (XigCodec *codec, XigPolyFillArc *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("PolyFillArc ...");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
put_image_cb (XigCodec *codec, XigPutImage *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("PutImage format=%d drawable=0x%08X gc=0x%08X width=%d height=%d dst-x=%d dst-y=%d left-pad=%d depth=%d data-length=%zi", message->format, message->drawable, message->gc, message->width, message->height, message->dst_x, message->dst_y, message->left_pad, message->depth, message->data_length);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
get_image_cb (XigCodec *codec, XigGetImage *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("GetImage format=%d drawable=0x%08X x=%d y=%d width=%d height=%d plane-mask=0x%08X", message->format, message->drawable, message->x, message->y, message->width, message->height, message->plane_mask);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
get_image_reply_cb (XigCodec *codec, XigGetImageReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("GetImage depth=%d visual=0x%08X data-length=%zi", message->depth, message->visual, message->data_length);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
create_colormap_cb (XigCodec *codec, XigCreateColormap *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("CreateColormap mid=0x%08X window=0x%08X visual=0x%08X alloc=%d", message->mid, message->window, message->visual, message->alloc);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
poly_text8_cb (XigCodec *codec, XigPolyText8 *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("PolyText8 ...");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
poly_text16_cb (XigCodec *codec, XigPolyText16 *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("PolyText16 ...");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
image_text8_cb (XigCodec *codec, XigImageText8 *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("ImageText8 ...");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
image_text16_cb (XigCodec *codec, XigImageText16 *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("ImageText16 ...");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
free_colormap_cb (XigCodec *codec, XigFreeColormap *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("FreeColormap cmap=0x%08X", message->cmap);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
copy_colormap_and_free_cb (XigCodec *codec, XigCopyColormapAndFree *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("CopyColormapAndFree mid=0x%08X src-cmap=0x%08X", message->mid, message->src_cmap);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
install_colormap_cb (XigCodec *codec, XigInstallColormap *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("InstallColormap cmap=0x%08X", message->cmap);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
uninstall_colormap_cb (XigCodec *codec, XigUninstallColormap *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("UninstallColormap cmap=0x%08X", message->cmap);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
list_installed_colormaps_cb (XigCodec *codec, XigListInstalledColormaps *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("ListInstalledColormaps window=0x%08X", message->window);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
list_installed_colormaps_reply_cb (XigCodec *codec, XigListInstalledColormapsReply *message, XigCodecLogger *logger)
{
    GString *text;
    int i;

    text = g_string_new ("ListInstalledColormaps cmaps=");

    if (message->cmaps_length > 0)
    {
        g_string_append_printf (text, "%d", message->cmaps[0]);
        for (i = 1; message->cmaps_length; i++)
            g_string_append_printf (text, ",%d", message->cmaps[i]);
    }

    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text->str);
    g_string_free (text, TRUE);
}

static void 
alloc_color_cb (XigCodec *codec, XigAllocColor *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("AllocColor cmap=0x%08X red=0x%04X green=0x%04X blue=0x%04X", message->cmap, message->red, message->green, message->blue);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
alloc_color_reply_cb (XigCodec *codec, XigAllocColorReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("AllocColor red=0x%04X green=0x%04X blue=0x%04X pixel=0x%08X", message->red, message->green, message->blue, message->pixel);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void 
alloc_named_color_cb (XigCodec *codec, XigAllocNamedColor *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("AllocNamedColor cmap=0x%08X name=%s", message->cmap, message->name);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
alloc_named_color_reply_cb (XigCodec *codec, XigAllocNamedColorReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("AllocNamedColor pixel=0x%08X exact-red=0x%04X exact-green=0x%04X exact-blue=0x%04X visual-red=0x%04X visual-green=0x%04X visual-blue=0x%04X", message->pixel, message->exact_red, message->exact_green, message->exact_blue, message->visual_red, message->visual_green, message->visual_blue);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
alloc_color_cells_cb (XigCodec *codec, XigAllocColorCells *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("AllocColorCells cmap=0x%08X colors=%d planes=%d contiguous=%s", message->cmap, message->colors, message->planes, message->contiguous ? "true" : "false");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void 
alloc_color_cells_reply_cb (XigCodec *codec, XigAllocColorCellsReply *message, XigCodecLogger *logger)
{
    GString *text;

    text = g_string_new ("AllocColorCells");
    g_string_append (text, " pixels=");
    if (message->pixels_length > 0)
    {
        int i;
        g_string_append_printf (text, "0x%08X", message->pixels[0]);
        for (i = 1; i < message->pixels_length; i++)
            g_string_append_printf (text, "0x%08X", message->pixels[i]);
    }
    g_string_append (text, " masks=");
    if (message->masks_length > 0)
    {
        int i;
        g_string_append_printf (text, "0x%08X", message->masks[0]);
        for (i = 1; i < message->masks_length; i++)
            g_string_append_printf (text, "0x%08X", message->masks[i]);
    }

    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text->str);
    g_string_free (text, TRUE);
}

static void
alloc_color_planes_cb (XigCodec *codec, XigAllocColorPlanes *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("AllocColorPlanes cmap=0x%08X colors=%d reds=%d greens=%d blues=%d contiguous=%s", message->cmap, message->colors, message->reds, message->greens, message->blues, message->contiguous ? "true" : "false");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void 
alloc_color_planes_reply_cb (XigCodec *codec, XigAllocColorPlanesReply *message, XigCodecLogger *logger)
{
    GString *text;

    text = g_string_new ("AllocColorPlanes");
    g_string_append_printf (text, " red-mask=0x%08X", message->red_mask);
    g_string_append_printf (text, " green-mask=0x%08X", message->green_mask);
    g_string_append_printf (text, " blue-mask=0x%08X", message->blue_mask);
    g_string_append (text, " pixels=");
    if (message->pixels_length > 0)
    {
        int i;
        g_string_append_printf (text, "0x%08X", message->pixels[0]);
        for (i = 1; i < message->pixels_length; i++)
            g_string_append_printf (text, "0x%08X", message->pixels[i]);
    }

    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text->str);
    g_string_free (text, TRUE);
}

static void 
free_colors_cb (XigCodec *codec, XigFreeColors *message, XigCodecLogger *logger)
{
    GString *text;
    int i;

    text = g_string_new ("FreeColors");
    g_string_append_printf (text, " cmap=0x%08X", message->cmap);
    g_string_append_printf (text, " plane-mask=0x%08X", message->plane_mask);
    g_string_append_printf (text, " pixels=");
    if (message->pixels_length > 0)
    {
        g_string_append_printf (text, "0x%08X", message->pixels[0]);
        for (i = 1; message->pixels_length; i++)
            g_string_append_printf (text, ",0x%08X", message->pixels[i]);
    }

    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text->str);
    g_string_free (text, TRUE);
}

static void 
store_colors_cb (XigCodec *codec, XigStoreColors *message, XigCodecLogger *logger)
{
    GString *text;

    text = g_string_new ("StoreColors");
    g_string_append_printf (text, " cmap=0x%08X", message->cmap);
    g_string_append (text, " items=");
    if (message->items_length > 0)
    {
        int i;
        g_string_append_printf (text, "{pixel=0x%08X red=0x%04X green=0x%04X blue=0x%04X do-red=%s do-green=%s do-blue=%s}", message->items[0].pixel, message->items[0].red, message->items[0].green, message->items[0].blue, message->items[0].do_red ? "true" : "false", message->items[0].do_green ? "true" : "false", message->items[0].do_blue ? "true" : "false");
        for (i = 1; i < message->items_length; i++)
            g_string_append_printf (text, ",{pixel=0x%08X red=0x%04X green=0x%04X blue=0x%04X do-red=%s do-green=%s do-blue=%s}", message->items[i].pixel, message->items[i].red, message->items[i].green, message->items[i].blue, message->items[i].do_red ? "true" : "false", message->items[i].do_green ? "true" : "false", message->items[i].do_blue ? "true" : "false");
    }

    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text->str);
    g_string_free (text, TRUE);
}

static void 
store_named_color_cb (XigCodec *codec, XigStoreNamedColor *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("StoreNamedColor cmap=0x%08X pixel=0x%08X do-red=%s do-green=%s do-blue=%s name=%s", message->cmap, message->pixel, message->do_red ? "true" : "false", message->do_green ? "true" : "false", message->do_blue ? "true" : "false", message->name);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
query_colors_cb (XigCodec *codec, XigQueryColors *message, XigCodecLogger *logger)
{
    GString *text;
    int i;

    text = g_string_new ("QueryColors");
    g_string_append_printf (text, " cmap=0x%08X", message->cmap);
    g_string_append_printf (text, " pixels=");
    if (message->pixels_length > 0)
    {
        g_string_append_printf (text, "0x%08X", message->pixels[0]);
        for (i = 1; message->pixels_length; i++)
            g_string_append_printf (text, ",0x%08X", message->pixels[i]);
    }

    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text->str);
    g_string_free (text, TRUE);
}

static void
query_colors_reply_cb (XigCodec *codec, XigQueryColorsReply *message, XigCodecLogger *logger)
{
    GString *text;
    int i;

    text = g_string_new ("QueryColors");
    g_string_append_printf (text, " colors=");
    if (message->colors_length > 0)
    {
        g_string_append_printf (text, "{red=0x%04X green=0x%04X blue=0x%04X}", message->colors[0].red, message->colors[0].green, message->colors[0].blue);
        for (i = 1; message->colors_length; i++)
            g_string_append_printf (text, "{red=0x%04X green=0x%04X blue=0x%04X}", message->colors[i].red, message->colors[i].green, message->colors[i].blue);
    }

    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text->str);
    g_string_free (text, TRUE);
}

static void 
lookup_color_cb (XigCodec *codec, XigLookupColor *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("LookupColor cmap=0x%08X name=%s", message->cmap, message->name);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
lookup_color_reply_cb (XigCodec *codec, XigLookupColorReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("LookupColor exact-red=0x%04X exact-green=0x%04X exact-blue=0x%04X visual-red=0x%04X visual-green=0x%04X visual-blue=0x%04X", message->exact_red, message->exact_green, message->exact_blue, message->visual_red, message->visual_green, message->visual_blue);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void 
create_cursor_cb (XigCodec *codec, XigCreateCursor *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("CreateCursor cid=0x%08X", message->cid);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
create_glyph_cursor_cb (XigCodec *codec, XigCreateGlyphCursor *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("CreateGlyphCursor cid=0x%08X", message->cid);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
free_cursor_cb (XigCodec *codec, XigFreeCursor *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("FreeCursor cursor=0x%08X", message->cursor);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
recolor_cursor_cb (XigCodec *codec, XigRecolorCursor *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("RecolorCursor cursor=0x%08X", message->cursor);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
query_best_size_cb (XigCodec *codec, XigQueryBestSize *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("QueryBestSize width=%u height=%u", message->width, message->height);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
query_best_size_reply_cb (XigCodec *codec, XigQueryBestSizeReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("QueryBestSize width=%u height=%u", message->width, message->height);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void 
query_extension_cb (XigCodec *codec, XigQueryExtension *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("QueryExtension name=%s", message->name);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
query_extension_reply_cb (XigCodec *codec, XigQueryExtensionReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("QueryExtension present=%s major-opcode=%d first-event=%d first-error=%d", message->present ? "true" : "false", message->major_opcode, message->first_event, message->first_error);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void 
list_extensions_cb (XigCodec *codec, XigListExtensions *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("ListExtensions");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
list_extensions_reply_cb (XigCodec *codec, XigListExtensionsReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("ListExtensions");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void 
change_keyboard_mapping_cb (XigCodec *codec, XigChangeKeyboardMapping *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("ChangeKeyboardMapping first-keycode=%d keysyms-per-keycode=%d keysyms=...", message->first_keycode, message->keysyms_per_keycode);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
get_keyboard_mapping_cb (XigCodec *codec, XigGetKeyboardMapping *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("GetKeyboardMapping first-keycode=%d count=%d", message->first_keycode, message->count);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
get_keyboard_mapping_reply_cb (XigCodec *codec, XigGetKeyboardMappingReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("GetKeyboardMapping keysyms-per-keycode=%d keysyms=...", message->keysyms_per_keycode);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
change_keyboard_control_cb (XigCodec *codec, XigChangeKeyboardControl *message, XigCodecLogger *logger)
{
    GString *text;

    text = g_string_new ("ChangeKeyboardControl");
    if (message->value_mask & XIG_KEYBOARD_CONTROL_key_click_percent)
        g_string_append_printf (text, " key-click-percent=%d", message->key_click_percent);
    if (message->value_mask & XIG_KEYBOARD_CONTROL_bell_percent)
        g_string_append_printf (text, " bell-percent=%d", message->bell_percent);
    if (message->value_mask & XIG_KEYBOARD_CONTROL_bell_pitch)
        g_string_append_printf (text, " bell-pitch=%d", message->bell_pitch);
    if (message->value_mask & XIG_KEYBOARD_CONTROL_bell_duration)
        g_string_append_printf (text, " bell-duration=%d", message->bell_duration);
    if (message->value_mask & XIG_KEYBOARD_CONTROL_led)
        g_string_append_printf (text, " led=%d", message->led);
    if (message->value_mask & XIG_KEYBOARD_CONTROL_led_mode)
        g_string_append_printf (text, " led-mode=%d", message->led_mode);
    if (message->value_mask & XIG_KEYBOARD_CONTROL_key)
        g_string_append_printf (text, " key=%d", message->key);
    if (message->value_mask & XIG_KEYBOARD_CONTROL_auto_repeat_mode)
        g_string_append_printf (text, " auto-repeat-mode=%d", message->auto_repeat_mode);

    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text->str);
    g_string_free (text, TRUE);
}

static void
get_keyboard_control_cb (XigCodec *codec, XigGetKeyboardControl *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("GetKeyboardControl");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
get_keyboard_control_reply_cb (XigCodec *codec, XigGetKeyboardControlReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("GetKeyboardControl global-auto-repeat=%d led-mask=%d key-click-percent=%d bell-percent=%d bell-pitch=%d bell-duration=%d auto-repeats=0x%02X%02X%02X%02X%02X0x%02X%02X%02X%02X%02X0x%02X%02X%02X%02X%02X0x%02X%02X%02X%02X%02X0x%02X%02X%02X%02X%02X0x%02X%02X%02X%02X%02X%02X%02X",
                            message->global_auto_repeat,
                            message->led_mask,
                            message->key_click_percent,
                            message->bell_percent,
                            message->bell_pitch,
                            message->bell_duration,
                            message->auto_repeats[0],
                            message->auto_repeats[1],
                            message->auto_repeats[2],
                            message->auto_repeats[3],
                            message->auto_repeats[4],
                            message->auto_repeats[5],
                            message->auto_repeats[6],
                            message->auto_repeats[7],
                            message->auto_repeats[8],
                            message->auto_repeats[9],
                            message->auto_repeats[10],
                            message->auto_repeats[11],
                            message->auto_repeats[12],
                            message->auto_repeats[13],
                            message->auto_repeats[14],
                            message->auto_repeats[15],
                            message->auto_repeats[16],
                            message->auto_repeats[17],
                            message->auto_repeats[18],
                            message->auto_repeats[19],
                            message->auto_repeats[20],
                            message->auto_repeats[21],
                            message->auto_repeats[22],
                            message->auto_repeats[23],
                            message->auto_repeats[24],
                            message->auto_repeats[25],
                            message->auto_repeats[26],
                            message->auto_repeats[27],
                            message->auto_repeats[28],
                            message->auto_repeats[29],
                            message->auto_repeats[30],
                            message->auto_repeats[31]);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void 
bell_cb (XigCodec *codec, XigBell *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("Bell");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
change_pointer_control_cb (XigCodec *codec, XigChangePointerControl *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("ChangePointerControl acceleration-numerator=%d acceleration-denominator=%d threshold=%d do-acceleration=%s do-threshold=%s", message->acceleration_numerator, message->acceleration_denominator, message->threshold, message->do_acceleration ? "true" : "false", message->do_threshold ? "true" : "false");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
get_pointer_control_cb (XigCodec *codec, XigGetPointerControl *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("GetPointerControl");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
get_pointer_control_reply_cb (XigCodec *codec, XigGetPointerControlReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("GetPointerControl acceleration-numerator=%d acceleration-denominator=%d threshold=%d do-acceleration=%s do-threshold=%s", message->acceleration_numerator, message->acceleration_denominator, message->threshold, message->do_acceleration ? "true" : "false", message->do_threshold ? "true" : "false");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void 
set_screen_saver_cb (XigCodec *codec, XigSetScreenSaver *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("SetScreenSaver timeout=%d interval=%d prefer_blanking=%s allow_exposures=%s", message->timeout, message->interval, message->prefer_blanking ? "true" : "false", message->allow_exposures ? "true" : "false");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
get_screen_saver_cb (XigCodec *codec, XigGetScreenSaver *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("GetScreenSaver");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
get_screen_saver_reply_cb (XigCodec *codec, XigGetScreenSaverReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("GetScreenSaver timeout=%d interval=%d prefer_blanking=%s allow_exposures=%s", message->timeout, message->interval, message->prefer_blanking ? "true" : "false", message->allow_exposures ? "true" : "false");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
change_hosts_cb (XigCodec *codec, XigChangeHosts *message, XigCodecLogger *logger)
{
    GString *text;

    text = g_string_new ("ChangeHosts");
    g_string_append_printf (text, " mode=%d", message->mode);
    g_string_append_printf (text, " family=%d", message->family);
    g_string_append (text, " address=");
    if (message->address_length > 0)
    {
        int i;
        g_string_append (text, "0x");
        for (i = 0; i < message->address_length; i++)
            g_string_append_printf (text, "%02X", message->address[i]);
    }

    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text->str);
    g_string_free (text, TRUE);
}

static void
list_hosts_cb (XigCodec *codec, XigListHosts *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("ListHosts");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
list_hosts_reply_cb (XigCodec *codec, XigListHostsReply *message, XigCodecLogger *logger)
{
    GString *text;

    text = g_string_new ("ListHosts");
    g_string_append_printf (text, " mode=%d", message->mode);
    g_string_append (text, " hosts=");
    if (message->hosts_length > 0)
    {
        int i;
        g_string_append_printf (text, "{family=%d address=", message->hosts[0].family);
        if (message->hosts[0].address_length > 0)
        {
            g_string_append (text, "0x");
            for (i = 0; i < message->hosts[0].address_length; i++)
                g_string_append_printf (text, "%02X", message->hosts[0].address[i]);
        }
        g_string_append (text, "}");
        for (i = 1; i < message->hosts_length; i++)
        {
            g_string_append_printf (text, ",{family=%d address=", message->hosts[i].family);
            if (message->hosts[i].address_length > 0)
            {
                int j;
                g_string_append (text, "0x");
                for (j = 0; j < message->hosts[i].address_length; j++)
                    g_string_append_printf (text, "%02X", message->hosts[i].address[j]);
            }
            g_string_append (text, "}");
        }
    }

    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text->str);
    g_string_free (text, TRUE);
}

static void
set_access_control_cb (XigCodec *codec, XigSetAccessControl *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("SetAccessControl mode=%d", message->mode);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
set_close_down_mode_cb (XigCodec *codec, XigSetCloseDownMode *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("SetCloseDownMode mode=%d", message->mode);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
kill_client_cb (XigCodec *codec, XigKillClient *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("KillClient resource=%d", message->resource);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
rotate_properties_cb (XigCodec *codec, XigRotateProperties *message, XigCodecLogger *logger)
{
    GString *text;

    text = g_string_new ("RotateProperties");
    g_string_append_printf (text, " delta=%d", message->delta);
    g_string_append (text, " properties=");
    if (message->properties_length > 0)
    {
        int i;
        g_string_append_printf (text, "0x%08X", message->properties[0]);
        for (i = 1; i < message->properties_length; i++)
            g_string_append_printf (text, ",0x%08X", message->properties[i]);
    }

    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text->str);
    g_string_free (text, TRUE);
}

static void 
force_screen_saver_cb (XigCodec *codec, XigForceScreenSaver *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("ForceScreenSaver mode=%d", message->mode);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
set_pointer_mapping_cb (XigCodec *codec, XigSetPointerMapping *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("SetPointerMapping ...");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
set_pointer_mapping_reply_cb (XigCodec *codec, XigSetPointerMappingReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("SetPointerMapping status=%d", message->status);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
get_pointer_mapping_cb (XigCodec *codec, XigGetPointerMapping *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("GetPointerMapping");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
get_pointer_mapping_reply_cb (XigCodec *codec, XigGetPointerMappingReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("GetPointerMapping ...");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
set_modifier_mapping_cb (XigCodec *codec, XigSetModifierMapping *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("SetModifierMapping ...");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
set_modifier_mapping_reply_cb (XigCodec *codec, XigSetModifierMappingReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("SetModifierMapping status=%d", message->status);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
get_modifier_mapping_cb (XigCodec *codec, XigGetModifierMapping *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("GetModifierMapping");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
get_modifier_mapping_reply_cb (XigCodec *codec, XigGetModifierMappingReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("GetModifierMapping ...");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
no_operation_cb (XigCodec *codec, XigNoOperation *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("NoOperation");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
BIGREQ_enable_cb (XigCodec *codec, XigBIGREQEnable *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("BIGREQ-Enable");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
BIGREQ_enable_reply_cb (XigCodec *codec, XigBIGREQEnableReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("BIGREQ-Enable maximum-request-length=%u", message->maximum_request_length);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
SYNC_initialize_cb (XigCodec *codec, XigSYNCInitialize *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("SYNC-Initialize desired-major-version=%d desired-minor-version=%d", message->desired_major_version, message->desired_minor_version);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
SYNC_initialize_reply_cb (XigCodec *codec, XigSYNCInitializeReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("SYNC-Initialize major-version=%d minor-version=%d", message->major_version, message->minor_version);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
RANDR_query_version_cb (XigCodec *codec, XigRANDRQueryVersion *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("RANDR-QueryVersion major-version=%d minor-version=%d", message->major_version, message->minor_version);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
RANDR_query_version_reply_cb (XigCodec *codec, XigRANDRQueryVersionReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("RANDR-QueryVersion major-version=%d minor-version=%d", message->major_version, message->minor_version);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
RANDR_select_input_cb (XigCodec *codec, XigRANDRSelectInput *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("RANDR-SelectInput window=0x%08X enable=0x%04X", message->window, message->enable);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
RANDR_get_screen_resources_cb (XigCodec *codec, XigRANDRGetScreenResources *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("RANDR-GetScreenResources window=0x%08X", message->window);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
RANDR_get_screen_resources_reply_cb (XigCodec *codec, XigRANDRGetScreenResourcesReply *message, XigCodecLogger *logger)
{
    GString *text;  

    text = g_string_new ("RANDR-GetScreenResourcesReply");  
    g_string_append_printf (text, " timestamp=%d", message->timestamp);
    g_string_append_printf (text, " config-timestamp=%d", message->config_timestamp);
    g_string_append (text, " crtcs=");
    if (message->crtcs_length > 0)
    {
        int i;
        g_string_append_printf (text, "0x%08X", message->crtcs[0]);
        for (i = 1; i < message->crtcs_length; i++)
            g_string_append_printf (text, ",0x%08X", message->crtcs[i]);
    }
    g_string_append (text, " outputs=");
    if (message->outputs_length > 0)
    {
        int i;
        g_string_append_printf (text, "0x%08X", message->outputs[0]);
        for (i = 1; i < message->outputs_length; i++)
            g_string_append_printf (text, ",0x%08X", message->outputs[i]);
    }
    g_string_append (text, " modes=");
    if (message->modes_length > 0)
    {
        int i;
        g_string_append_printf (text, "{width-in-pixels=%d height-in-pixels=%d dot-clock=%d h-sync-start=%d h-sync-end=%d h-total=%d h-skew=%d v-sync-start=%d v-sync-end=%d v-total=%d name=%s mode-flags=0x%08X}", message->modes[0].width_in_pixels, message->modes[0].height_in_pixels, message->modes[0].dot_clock, message->modes[0].h_sync_start, message->modes[0].h_sync_end, message->modes[0].h_total, message->modes[0].h_skew, message->modes[0].v_sync_start, message->modes[0].v_sync_end, message->modes[0].v_total, message->modes[0].name, message->modes[0].mode_flags);
        for (i = 1; i < message->modes_length; i++)
            g_string_append_printf (text, ",{width-in-pixels=%d height-in-pixels=%d dot-clock=%d h-sync-start=%d h-sync-end=%d h-total=%d h-skew=%d v-sync-start=%d v-sync-end=%d v-total=%d name=%s mode-flags=0x%08X}", message->modes[i].width_in_pixels, message->modes[i].height_in_pixels, message->modes[i].dot_clock, message->modes[i].h_sync_start, message->modes[i].h_sync_end, message->modes[i].h_total, message->modes[i].h_skew, message->modes[i].v_sync_start, message->modes[i].v_sync_end, message->modes[i].v_total, message->modes[i].name, message->modes[i].mode_flags);
    }

    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text->str);
    g_string_free (text, TRUE);
}

static void
RANDR_get_screen_resources_current_cb (XigCodec *codec, XigRANDRGetScreenResourcesCurrent *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("RANDR-GetScreenResourcesCurrent window=0x%08X", message->window);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
RANDR_get_screen_resources_current_reply_cb (XigCodec *codec, XigRANDRGetScreenResourcesCurrentReply *message, XigCodecLogger *logger)
{
    GString *text;  

    text = g_string_new ("RANDR-GetScreenResourcesCurrentReply");
    g_string_append_printf (text, " timestamp=%d", message->timestamp);
    g_string_append_printf (text, " config-timestamp=%d", message->config_timestamp);
    g_string_append (text, " crtcs=");
    if (message->crtcs_length > 0)
    {
        int i;
        g_string_append_printf (text, "0x%08X", message->crtcs[0]);
        for (i = 1; i < message->crtcs_length; i++)
            g_string_append_printf (text, ",0x%08X", message->crtcs[i]);
    }
    g_string_append (text, " outputs=");
    if (message->outputs_length > 0)
    {
        int i;
        g_string_append_printf (text, "0x%08X", message->outputs[0]);
        for (i = 1; i < message->outputs_length; i++)
            g_string_append_printf (text, ",0x%08X", message->outputs[i]);
    }
    g_string_append (text, " modes=");
    if (message->modes_length > 0)
    {
        int i;
        g_string_append_printf (text, "{width-in-pixels=%d height-in-pixels=%d dot-clock=%d h-sync-start=%d h-sync-end=%d h-total=%d h-skew=%d v-sync-start=%d v-sync-end=%d v-total=%d name=%s mode-flags=0x%08X}", message->modes[0].width_in_pixels, message->modes[0].height_in_pixels, message->modes[0].dot_clock, message->modes[0].h_sync_start, message->modes[0].h_sync_end, message->modes[0].h_total, message->modes[0].h_skew, message->modes[0].v_sync_start, message->modes[0].v_sync_end, message->modes[0].v_total, message->modes[0].name, message->modes[0].mode_flags);
        for (i = 1; i < message->modes_length; i++)
            g_string_append_printf (text, ",{width-in-pixels=%d height-in-pixels=%d dot-clock=%d h-sync-start=%d h-sync-end=%d h-total=%d h-skew=%d v-sync-start=%d v-sync-end=%d v-total=%d name=%s mode-flags=0x%08X}", message->modes[i].width_in_pixels, message->modes[i].height_in_pixels, message->modes[i].dot_clock, message->modes[i].h_sync_start, message->modes[i].h_sync_end, message->modes[i].h_total, message->modes[i].h_skew, message->modes[i].v_sync_start, message->modes[i].v_sync_end, message->modes[i].v_total, message->modes[i].name, message->modes[i].mode_flags);
    }

    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text->str);
    g_string_free (text, TRUE);
}


static void
RANDR_set_output_primary_cb (XigCodec *codec, XigRANDRSetOutputPrimary *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("RANDR-SetOutputPrimary window=0x%08X output=0x%08X", message->window, message->output);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
RANDR_get_output_primary_cb (XigCodec *codec, XigRANDRGetOutputPrimary *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("RANDR-GetOutputPrimary window=0x%08X", message->window);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
RANDR_get_output_primary_reply_cb (XigCodec *codec, XigRANDRGetOutputPrimaryReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("RANDR-GetOutputPrimary output=0x%08X", message->output);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void 
XINERAMA_is_active_cb (XigCodec *codec, XigXINERAMAIsActive *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("XINERAMA-IsActive");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
XINERAMA_is_active_reply_cb (XigCodec *codec, XigXINERAMAIsActiveReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("XINERAMA-IsActive is-active=%s", message->is_active ? "true" : "false");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void 
XINERAMA_query_screens_cb (XigCodec *codec, XigXINERAMAQueryScreens *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("XINERAMA-QueryScreens");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
XINERAMA_query_screens_reply_cb (XigCodec *codec, XigXINERAMAQueryScreensReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("XINERAMA-QueryScreens");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
XKB_use_extension_cb (XigCodec *codec, XigXKBUseExtension *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("XKB-UseExtension wanted-major=%d wanted-minor=%d", message->wanted_major, message->wanted_minor);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
XKB_use_extension_reply_cb (XigCodec *codec, XigXKBUseExtensionReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("XKB-UseExtension supported=%s server-major=%d server-minor=%d", message->supported ? "true" : "false", message->server_major, message->server_minor);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
XKB_select_events_cb (XigCodec *codec, XigXKBSelectEvents *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("XKB-SelectEvents device-spec=%d affect-which=0x%04X clear=0x%04X select-all=0x%04X affect-map=0x%04X map=0x%04X details=...",
                            message->device_spec,
                            message->affect_which,
                            message->clear,
                            message->select_all,
                            message->affect_map,
                            message->map);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
XKB_get_state_cb (XigCodec *codec, XigXKBGetState *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("XKB-GetState device-spec=0x%04X", message->device_spec);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
XKB_get_state_reply_cb (XigCodec *codec, XigXKBGetStateReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("XKB-GetState device-id=%d mods=0x%02X base-mods=0x%02X latched-mods=0x%02X locked-mods=0x%02X group=%d locked-group=%d base-group=%d latched-group=%d compat-state=0x%02X grab-mods=0x%02X compat-grab-mods=0x%02X lookup-mods=0x%02X compat-lookup-mods=0x%02X ptr-btn-state=0x%04X",
                            message->device_id,
                            message->mods,
                            message->base_mods,
                            message->latched_mods,
                            message->locked_mods,
                            message->group,
                            message->locked_group,
                            message->base_group,
                            message->latched_group,
                            message->compat_state,
                            message->grab_mods,
                            message->compat_grab_mods,
                            message->lookup_mods,
                            message->compat_lookup_mods,
                            message->ptr_btn_state);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
XKB_set_controls_cb (XigCodec *codec, XigXKBSetControls *message, XigCodecLogger *logger)
{
    GString *text;
    int i;

    text = g_string_new ("XKB-SetControls");
    g_string_append_printf (text, " device-spec=%d", message->device_spec);
    g_string_append_printf (text, " affect-internal-real-mods=0x%02X", message->affect_internal_real_mods);
    g_string_append_printf (text, " internal-real-mods=0x%02X", message->internal_real_mods);
    g_string_append_printf (text, " affect-ignore-lock-real-mods=0x%02X", message->affect_ignore_lock_real_mods);
    g_string_append_printf (text, " ignore-lock-real-mods=0x%02X", message->ignore_lock_real_mods);
    g_string_append_printf (text, " affect-internal-virtual-mods=0x%04X", message->affect_internal_virtual_mods);
    g_string_append_printf (text, " internal-virtual-mods=0x%04X", message->internal_virtual_mods);
    g_string_append_printf (text, " affect-ignore-lock-virtual-mods=0x%04X", message->affect_ignore_lock_virtual_mods);
    g_string_append_printf (text, " ignore-lock-virtual-mods=0x%04X", message->ignore_lock_virtual_mods);
    g_string_append_printf (text, " mouse-keys-dflt-btn=%d", message->mouse_keys_dflt_btn);
    g_string_append_printf (text, " groups-wrap=%d", message->groups_wrap);
    g_string_append_printf (text, " access-x-options=0x%04X", message->access_x_options);
    g_string_append_printf (text, " affect-enabled-controls=0x%08X", message->affect_enabled_controls);
    g_string_append_printf (text, " enabled-controls=0x%08X", message->enabled_controls);
    g_string_append_printf (text, " change-controls=0x%08X", message->change_controls);
    g_string_append_printf (text, " repeat-delay=%d", message->repeat_delay);
    g_string_append_printf (text, " repeat-interval=%d", message->repeat_interval);
    g_string_append_printf (text, " slow-keys-delay=%d", message->slow_keys_delay);
    g_string_append_printf (text, " debounce-delay=%d", message->debounce_delay);
    g_string_append_printf (text, " mouse-keys-delay=%d", message->mouse_keys_delay);
    g_string_append_printf (text, " mouse-keys-interval=%d", message->mouse_keys_interval);
    g_string_append_printf (text, " mouse-keys-time-to-max=%d", message->mouse_keys_time_to_max);
    g_string_append_printf (text, " mouse-keys-max-speed=%d", message->mouse_keys_max_speed);
    g_string_append_printf (text, " mouse-keys-curve=%d", message->mouse_keys_curve);
    g_string_append_printf (text, " access-x-timeout=%d", message->access_x_timeout);
    g_string_append_printf (text, " access-x-timeout-mask=0x%08X", message->access_x_timeout_mask);
    g_string_append_printf (text, " access-x-timeout-values=%d", message->access_x_timeout_values);
    g_string_append_printf (text, " access-x-timeout-options-mask=0x%08X", message->access_x_timeout_options_mask);
    g_string_append_printf (text, " access-x-timeout-options-values=%d", message->access_x_timeout_options_values);
    g_string_append_printf (text, " per-key-repeat=%d", message->per_key_repeat[0]);
    for (i = 1; i < 32; i++)
        g_string_append_printf (text, ",%d", message->per_key_repeat[i]);

    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text->str);
    g_string_free (text, TRUE);
}

static void
XKB_get_map_cb (XigCodec *codec, XigXKBGetMap *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("XKB-GetMap device-spec=0x%04X full=%d partial=%d first-type=%d n-types=%d first-key-sym=%d n-key-syms=%d first-key-action=%d n-key-actions=%d first-key-behavior=%d n-key-behaviors=%d virtual-mods=%d first-key-explicit=%d n-key-explicit=%d first-mod-map-key=%d n-mod-map-keys=%d first-v-mod-map-key=%d n-v-mod-map-keys=%d",
                            message->device_spec,
                            message->full,
                            message->partial,
                            message->first_type,
                            message->n_types,
                            message->first_key_sym,
                            message->n_key_syms,
                            message->first_key_action,
                            message->n_key_actions,
                            message->first_key_behavior,
                            message->n_key_behaviors,
                            message->virtual_mods,
                            message->first_key_explicit,
                            message->n_key_explicit,
                            message->first_mod_map_key,
                            message->n_mod_map_keys,
                            message->first_v_mod_map_key,
                            message->n_v_mod_map_keys);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
XKB_get_map_reply_cb (XigCodec *codec, XigXKBGetMapReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("XKB-GetMap ...");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
XKB_get_names_cb (XigCodec *codec, XigXKBGetNames *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("XKB-GetNames device-spec=0x%04X which=0x%08X", message->device_spec, message->which);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
XKB_get_names_reply_cb (XigCodec *codec, XigXKBGetNamesReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("XKB-GetNames ...");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void
XKB_per_client_flags_cb (XigCodec *codec, XigXKBPerClientFlags *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("XKB-PerClientFlags ...");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void
XKB_per_client_flags_reply_cb (XigCodec *codec, XigXKBPerClientFlagsReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("XKB-PerClientFlags ...");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void 
RENDER_query_version_cb (XigCodec *codec, XigRENDERQueryVersion *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("RENDER-QueryVersion client-major-version=%d client-minor-version=%d", message->client_major_version, message->client_minor_version);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
RENDER_query_version_reply_cb (XigCodec *codec, XigRENDERQueryVersionReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("RENDER-QueryVersion major-version=%d minor-version=%d", message->major_version, message->minor_version);
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

static void 
RENDER_query_pict_formats_cb (XigCodec *codec, XigRENDERQueryPictFormats *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("RENDER-QueryPictFormats");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_request (logger, text);
    g_free (text);
}

static void 
RENDER_query_pict_formats_reply_cb (XigCodec *codec, XigRENDERQueryPictFormatsReply *message, XigCodecLogger *logger)
{
    gchar *text;
    text = g_strdup_printf ("RENDER-QueryPictFormats ...");
    XIG_CODEC_LOGGER_GET_CLASS (logger)->log_reply (logger, text);
    g_free (text);
}

XigCodecLogger *
xig_codec_logger_new (XigCodec *codec)
{
    XigCodecLogger *logger;
  
    logger = g_object_new (xig_codec_logger_get_type (), NULL);
    logger->priv->codec = g_object_ref (codec);

    g_signal_connect (codec, "connect", G_CALLBACK (connect_cb), logger);
    g_signal_connect (codec, "failed", G_CALLBACK (failed_cb), logger);
    g_signal_connect (codec, "authenticate", G_CALLBACK (authenticate_cb), logger);
    g_signal_connect (codec, "success", G_CALLBACK (success_cb), logger);
    g_signal_connect (codec, "error", G_CALLBACK (error_cb), logger);  
    g_signal_connect (codec, "create-notify", G_CALLBACK (create_notify_cb), logger);
    g_signal_connect (codec, "destroy-notify", G_CALLBACK (destroy_notify_cb), logger);
    g_signal_connect (codec, "unmap-notify", G_CALLBACK (unmap_notify_cb), logger);
    g_signal_connect (codec, "map-notify", G_CALLBACK (map_notify_cb), logger);
    g_signal_connect (codec, "map-request", G_CALLBACK (map_request_cb), logger);
    g_signal_connect (codec, "reparent-notify", G_CALLBACK (reparent_notify_cb), logger);
    g_signal_connect (codec, "configure-notify", G_CALLBACK (configure_notify_cb), logger);
    g_signal_connect (codec, "configure-request", G_CALLBACK (configure_request_cb), logger);
    g_signal_connect (codec, "circulate-notify", G_CALLBACK (circulate_notify_cb), logger);
    g_signal_connect (codec, "circulate-request", G_CALLBACK (circulate_request_cb), logger);
    g_signal_connect (codec, "property-notify", G_CALLBACK (property_notify_cb), logger);
    g_signal_connect (codec, "create-window", G_CALLBACK (create_window_cb), logger);
    g_signal_connect (codec, "change-window-attributes", G_CALLBACK (change_window_attributes_cb), logger);
    g_signal_connect (codec, "get-window-attributes", G_CALLBACK (get_window_attributes_cb), logger);
    g_signal_connect (codec, "get-window-attributes-reply", G_CALLBACK (get_window_attributes_reply_cb), logger);
    g_signal_connect (codec, "destroy-window", G_CALLBACK (destroy_window_cb), logger);
    g_signal_connect (codec, "destroy-subwindows", G_CALLBACK (destroy_subwindows_cb), logger);
    g_signal_connect (codec, "change-save-set", G_CALLBACK (change_save_set_cb), logger);
    g_signal_connect (codec, "reparent-window", G_CALLBACK (reparent_window_cb), logger);
    g_signal_connect (codec, "map-window", G_CALLBACK (map_window_cb), logger);
    g_signal_connect (codec, "map-subwindows", G_CALLBACK (map_subwindows_cb), logger);
    g_signal_connect (codec, "unmap-window", G_CALLBACK (unmap_window_cb), logger);
    g_signal_connect (codec, "unmap-subwindows", G_CALLBACK (unmap_subwindows_cb), logger);
    g_signal_connect (codec, "configure-window", G_CALLBACK (configure_window_cb), logger);
    g_signal_connect (codec, "circulate-window", G_CALLBACK (circulate_window_cb), logger);
    g_signal_connect (codec, "get-geometry", G_CALLBACK (get_geometry_cb), logger);
    g_signal_connect (codec, "get-geometry-reply", G_CALLBACK (get_geometry_reply_cb), logger);
    g_signal_connect (codec, "query-tree", G_CALLBACK (query_tree_cb), logger);
    g_signal_connect (codec, "query-tree-reply", G_CALLBACK (query_tree_reply_cb), logger);
    g_signal_connect (codec, "intern-atom", G_CALLBACK (intern_atom_cb), logger);
    g_signal_connect (codec, "intern-atom-reply", G_CALLBACK (intern_atom_reply_cb), logger);
    g_signal_connect (codec, "get-atom-name", G_CALLBACK (get_atom_name_cb), logger);
    g_signal_connect (codec, "get-atom-name-reply", G_CALLBACK (get_atom_name_reply_cb), logger);
    g_signal_connect (codec, "change-property", G_CALLBACK (change_property_cb), logger);
    g_signal_connect (codec, "delete-property", G_CALLBACK (delete_property_cb), logger);
    g_signal_connect (codec, "get-property", G_CALLBACK (get_property_cb), logger);
    g_signal_connect (codec, "get-property-reply", G_CALLBACK (get_property_reply_cb), logger);
    g_signal_connect (codec, "list-properties", G_CALLBACK (list_properties_cb), logger);
    g_signal_connect (codec, "list-properties-reply", G_CALLBACK (list_properties_reply_cb), logger);
    g_signal_connect (codec, "set-selection-owner", G_CALLBACK (set_selection_owner_cb), logger);
    g_signal_connect (codec, "get-selection-owner", G_CALLBACK (get_selection_owner_cb), logger);
    g_signal_connect (codec, "get-selection-owner-reply", G_CALLBACK (get_selection_owner_reply_cb), logger);
    g_signal_connect (codec, "convert-selection", G_CALLBACK (convert_selection_cb), logger);
    g_signal_connect (codec, "send-event", G_CALLBACK (send_event_cb), logger);
    g_signal_connect (codec, "grab-pointer", G_CALLBACK (grab_pointer_cb), logger);
    g_signal_connect (codec, "grab-pointer-reply", G_CALLBACK (grab_pointer_reply_cb), logger);
    g_signal_connect (codec, "ungrab-pointer", G_CALLBACK (ungrab_pointer_cb), logger);
    g_signal_connect (codec, "grab-button", G_CALLBACK (grab_button_cb), logger);
    g_signal_connect (codec, "ungrab-button", G_CALLBACK (ungrab_button_cb), logger);
    g_signal_connect (codec, "change-active-pointer-grab", G_CALLBACK (change_active_pointer_grab_cb), logger);
    g_signal_connect (codec, "grab-keyboard", G_CALLBACK (grab_keyboard_cb), logger);
    g_signal_connect (codec, "grab-keyboard-reply", G_CALLBACK (grab_keyboard_reply_cb), logger);
    g_signal_connect (codec, "ungrab-keyboard", G_CALLBACK (ungrab_keyboard_cb), logger);  
    g_signal_connect (codec, "grab-key", G_CALLBACK (grab_key_cb), logger);
    g_signal_connect (codec, "ungrab-key", G_CALLBACK (ungrab_key_cb), logger);
    g_signal_connect (codec, "allow-events", G_CALLBACK (allow_events_cb), logger);
    g_signal_connect (codec, "grab-server", G_CALLBACK (grab_server_cb), logger);
    g_signal_connect (codec, "ungrab-server", G_CALLBACK (ungrab_server_cb), logger);
    g_signal_connect (codec, "query-pointer", G_CALLBACK (query_pointer_cb), logger);
    g_signal_connect (codec, "query-pointer-reply", G_CALLBACK (query_pointer_reply_cb), logger);
    g_signal_connect (codec, "get-motion-events", G_CALLBACK (get_motion_events_cb), logger);
    g_signal_connect (codec, "get-motion-events-reply", G_CALLBACK (get_motion_events_reply_cb), logger);
    g_signal_connect (codec, "translate-coordinates", G_CALLBACK (translate_coordinates_cb), logger);
    g_signal_connect (codec, "translate-coordinates-reply", G_CALLBACK (translate_coordinates_reply_cb), logger);
    g_signal_connect (codec, "warp-pointer", G_CALLBACK (warp_pointer_cb), logger);
    g_signal_connect (codec, "set-input-focus", G_CALLBACK (set_input_focus_cb), logger);
    g_signal_connect (codec, "get-input-focus", G_CALLBACK (get_input_focus_cb), logger);
    g_signal_connect (codec, "get-input-focus-reply", G_CALLBACK (get_input_focus_reply_cb), logger);
    g_signal_connect (codec, "query-keymap", G_CALLBACK (query_keymap_cb), logger);
    g_signal_connect (codec, "query-keymap-reply", G_CALLBACK (query_keymap_reply_cb), logger);
    g_signal_connect (codec, "open-font", G_CALLBACK (open_font_cb), logger);
    g_signal_connect (codec, "close-font", G_CALLBACK (close_font_cb), logger);
    g_signal_connect (codec, "query-font", G_CALLBACK (query_font_cb), logger);
    g_signal_connect (codec, "query-font-reply", G_CALLBACK (query_font_reply_cb), logger);
    g_signal_connect (codec, "query-text-extents", G_CALLBACK (query_text_extents_cb), logger);
    g_signal_connect (codec, "query-text-extents-reply", G_CALLBACK (query_text_extents_reply_cb), logger);
    g_signal_connect (codec, "list-fonts", G_CALLBACK (list_fonts_cb), logger);
    g_signal_connect (codec, "list-fonts-reply", G_CALLBACK (list_fonts_reply_cb), logger);
    g_signal_connect (codec, "list-fonts-with-info", G_CALLBACK (list_fonts_with_info_cb), logger);
    g_signal_connect (codec, "list-fonts-with-info-reply", G_CALLBACK (list_fonts_with_info_reply_cb), logger);
    g_signal_connect (codec, "set-font-path", G_CALLBACK (set_font_path_cb), logger);
    g_signal_connect (codec, "get-font-path", G_CALLBACK (get_font_path_cb), logger);
    g_signal_connect (codec, "get-font-path-reply", G_CALLBACK (get_font_path_reply_cb), logger);
    g_signal_connect (codec, "create-pixmap", G_CALLBACK (create_pixmap_cb), logger);
    g_signal_connect (codec, "free-pixmap", G_CALLBACK (free_pixmap_cb), logger);
    g_signal_connect (codec, "create-gc", G_CALLBACK (create_gc_cb), logger);
    g_signal_connect (codec, "change-gc", G_CALLBACK (change_gc_cb), logger);
    g_signal_connect (codec, "copy-gc", G_CALLBACK (copy_gc_cb), logger);
    g_signal_connect (codec, "set-dashes", G_CALLBACK (set_dashes_cb), logger);
    g_signal_connect (codec, "set-clip-rectangles", G_CALLBACK (set_clip_rectangles_cb), logger);
    g_signal_connect (codec, "free-gc", G_CALLBACK (free_gc_cb), logger);
    g_signal_connect (codec, "clear-area", G_CALLBACK (clear_area_cb), logger);
    g_signal_connect (codec, "copy-area", G_CALLBACK (copy_area_cb), logger);
    g_signal_connect (codec, "copy-plane", G_CALLBACK (copy_plane_cb), logger);
    g_signal_connect (codec, "poly-point", G_CALLBACK (poly_point_cb), logger);
    g_signal_connect (codec, "poly-line", G_CALLBACK (poly_line_cb), logger);
    g_signal_connect (codec, "poly-segment", G_CALLBACK (poly_segment_cb), logger);
    g_signal_connect (codec, "poly-rectangle", G_CALLBACK (poly_rectangle_cb), logger);
    g_signal_connect (codec, "poly-arc", G_CALLBACK (poly_arc_cb), logger);
    g_signal_connect (codec, "fill-poly", G_CALLBACK (fill_poly_cb), logger);
    g_signal_connect (codec, "poly-fill-rectangle", G_CALLBACK (poly_fill_rectangle_cb), logger);
    g_signal_connect (codec, "poly-fill-arc", G_CALLBACK (poly_fill_arc_cb), logger);
    g_signal_connect (codec, "put-image", G_CALLBACK (put_image_cb), logger);
    g_signal_connect (codec, "get-image", G_CALLBACK (get_image_cb), logger);
    g_signal_connect (codec, "get-image-reply", G_CALLBACK (get_image_reply_cb), logger);
    g_signal_connect (codec, "poly-text8", G_CALLBACK (poly_text8_cb), logger);
    g_signal_connect (codec, "poly-text16", G_CALLBACK (poly_text16_cb), logger);
    g_signal_connect (codec, "image-text8", G_CALLBACK (image_text8_cb), logger);
    g_signal_connect (codec, "image-text16", G_CALLBACK (image_text16_cb), logger);
    g_signal_connect (codec, "create-colormap", G_CALLBACK (create_colormap_cb), logger);
    g_signal_connect (codec, "free-colormap", G_CALLBACK (free_colormap_cb), logger);
    g_signal_connect (codec, "copy-colormap-and-free", G_CALLBACK (copy_colormap_and_free_cb), logger);
    g_signal_connect (codec, "install-colormap", G_CALLBACK (install_colormap_cb), logger);
    g_signal_connect (codec, "uninstall-colormap", G_CALLBACK (uninstall_colormap_cb), logger);
    g_signal_connect (codec, "list-installed-colormaps", G_CALLBACK (list_installed_colormaps_cb), logger);
    g_signal_connect (codec, "list-installed-colormaps-reply", G_CALLBACK (list_installed_colormaps_reply_cb), logger);
    g_signal_connect (codec, "alloc-color", G_CALLBACK (alloc_color_cb), logger);
    g_signal_connect (codec, "alloc-color-reply", G_CALLBACK (alloc_color_reply_cb), logger);
    g_signal_connect (codec, "alloc-named-color", G_CALLBACK (alloc_named_color_cb), logger);
    g_signal_connect (codec, "alloc-named-color-reply", G_CALLBACK (alloc_named_color_reply_cb), logger);
    g_signal_connect (codec, "alloc-color-cells", G_CALLBACK (alloc_color_cells_cb), logger);
    g_signal_connect (codec, "alloc-color-cells-reply", G_CALLBACK (alloc_color_cells_reply_cb), logger);
    g_signal_connect (codec, "alloc-color-planes", G_CALLBACK (alloc_color_planes_cb), logger);
    g_signal_connect (codec, "alloc-color-planes-reply", G_CALLBACK (alloc_color_planes_reply_cb), logger);
    g_signal_connect (codec, "free-colors", G_CALLBACK (free_colors_cb), logger);
    g_signal_connect (codec, "store-colors", G_CALLBACK (store_colors_cb), logger);
    g_signal_connect (codec, "store-named-color", G_CALLBACK (store_named_color_cb), logger);
    g_signal_connect (codec, "query-colors", G_CALLBACK (query_colors_cb), logger);
    g_signal_connect (codec, "query-colors-reply", G_CALLBACK (query_colors_reply_cb), logger);
    g_signal_connect (codec, "lookup-color", G_CALLBACK (lookup_color_cb), logger);
    g_signal_connect (codec, "lookup-color-reply", G_CALLBACK (lookup_color_reply_cb), logger);
    g_signal_connect (codec, "create-cursor", G_CALLBACK (create_cursor_cb), logger);
    g_signal_connect (codec, "create-glyph-cursor", G_CALLBACK (create_glyph_cursor_cb), logger);
    g_signal_connect (codec, "free-cursor", G_CALLBACK (free_cursor_cb), logger);
    g_signal_connect (codec, "recolor-cursor", G_CALLBACK (recolor_cursor_cb), logger);
    g_signal_connect (codec, "query-best-size", G_CALLBACK (query_best_size_cb), logger);
    g_signal_connect (codec, "query-best-size-reply", G_CALLBACK (query_best_size_reply_cb), logger);
    g_signal_connect (codec, "query-extension", G_CALLBACK (query_extension_cb), logger);
    g_signal_connect (codec, "query-extension-reply", G_CALLBACK (query_extension_reply_cb), logger);
    g_signal_connect (codec, "list-extensions", G_CALLBACK (list_extensions_cb), logger);
    g_signal_connect (codec, "list-extensions-reply", G_CALLBACK (list_extensions_reply_cb), logger);
    g_signal_connect (codec, "change-keyboard-mapping", G_CALLBACK (change_keyboard_mapping_cb), logger);
    g_signal_connect (codec, "get-keyboard-mapping", G_CALLBACK (get_keyboard_mapping_cb), logger);
    g_signal_connect (codec, "get-keyboard-mapping-reply", G_CALLBACK (get_keyboard_mapping_reply_cb), logger);
    g_signal_connect (codec, "change-keyboard-control", G_CALLBACK (change_keyboard_control_cb), logger);
    g_signal_connect (codec, "get-keyboard-control", G_CALLBACK (get_keyboard_control_cb), logger);
    g_signal_connect (codec, "get-keyboard-control-reply", G_CALLBACK (get_keyboard_control_reply_cb), logger);
    g_signal_connect (codec, "bell", G_CALLBACK (bell_cb), logger);
    g_signal_connect (codec, "change-pointer-control", G_CALLBACK (change_pointer_control_cb), logger);
    g_signal_connect (codec, "get-pointer-control", G_CALLBACK (get_pointer_control_cb), logger);
    g_signal_connect (codec, "get-pointer-control-reply", G_CALLBACK (get_pointer_control_reply_cb), logger);
    g_signal_connect (codec, "set-screen-saver", G_CALLBACK (set_screen_saver_cb), logger);
    g_signal_connect (codec, "get-screen-saver", G_CALLBACK (get_screen_saver_cb), logger);
    g_signal_connect (codec, "get-screen-saver-reply", G_CALLBACK (get_screen_saver_reply_cb), logger);
    g_signal_connect (codec, "change-hosts", G_CALLBACK (change_hosts_cb), logger);
    g_signal_connect (codec, "list-hosts", G_CALLBACK (list_hosts_cb), logger);
    g_signal_connect (codec, "list-hosts-reply", G_CALLBACK (list_hosts_reply_cb), logger);
    g_signal_connect (codec, "set-access-control", G_CALLBACK (set_access_control_cb), logger);
    g_signal_connect (codec, "set-close-down-mode", G_CALLBACK (set_close_down_mode_cb), logger);
    g_signal_connect (codec, "kill-client", G_CALLBACK (kill_client_cb), logger);
    g_signal_connect (codec, "rotate-properties", G_CALLBACK (rotate_properties_cb), logger);
    g_signal_connect (codec, "force-screen-saver", G_CALLBACK (force_screen_saver_cb), logger);
    g_signal_connect (codec, "set-pointer-mapping", G_CALLBACK (set_pointer_mapping_cb), logger);
    g_signal_connect (codec, "set-pointer-mapping-reply", G_CALLBACK (set_pointer_mapping_reply_cb), logger);
    g_signal_connect (codec, "get-pointer-mapping", G_CALLBACK (get_pointer_mapping_cb), logger);
    g_signal_connect (codec, "get-pointer-mapping-reply", G_CALLBACK (get_pointer_mapping_reply_cb), logger);
    g_signal_connect (codec, "set-modifier-mapping", G_CALLBACK (set_modifier_mapping_cb), logger);
    g_signal_connect (codec, "set-modifier-mapping-reply", G_CALLBACK (set_modifier_mapping_reply_cb), logger);
    g_signal_connect (codec, "get-modifier-mapping", G_CALLBACK (get_modifier_mapping_cb), logger);
    g_signal_connect (codec, "get-modifier-mapping-reply", G_CALLBACK (get_modifier_mapping_reply_cb), logger);
    g_signal_connect (codec, "no-operation", G_CALLBACK (no_operation_cb), logger);
    g_signal_connect (codec, "BIGREQ-enable", G_CALLBACK (BIGREQ_enable_cb), logger);
    g_signal_connect (codec, "BIGREQ-enable-reply", G_CALLBACK (BIGREQ_enable_reply_cb), logger);
    g_signal_connect (codec, "SYNC-initialize", G_CALLBACK (SYNC_initialize_cb), logger);
    g_signal_connect (codec, "SYNC-initialize-reply", G_CALLBACK (SYNC_initialize_reply_cb), logger);
    g_signal_connect (codec, "RANDR-query-version", G_CALLBACK (RANDR_query_version_cb), logger);
    g_signal_connect (codec, "RANDR-query-version-reply", G_CALLBACK (RANDR_query_version_reply_cb), logger);
    g_signal_connect (codec, "RANDR-select-input", G_CALLBACK (RANDR_select_input_cb), logger);
    g_signal_connect (codec, "RANDR-get-screen-resources", G_CALLBACK (RANDR_get_screen_resources_cb), logger);
    g_signal_connect (codec, "RANDR-get-screen-resources-reply", G_CALLBACK (RANDR_get_screen_resources_reply_cb), logger);
    g_signal_connect (codec, "RANDR-get-screen-resources-current", G_CALLBACK (RANDR_get_screen_resources_current_cb), logger);
    g_signal_connect (codec, "RANDR-get-screen-resources-current-reply", G_CALLBACK (RANDR_get_screen_resources_current_reply_cb), logger);
    g_signal_connect (codec, "RANDR-set-output-primary", G_CALLBACK (RANDR_set_output_primary_cb), logger);
    g_signal_connect (codec, "RANDR-get-output-primary", G_CALLBACK (RANDR_get_output_primary_cb), logger);
    g_signal_connect (codec, "RANDR-get-output-primary-reply", G_CALLBACK (RANDR_get_output_primary_reply_cb), logger);
    g_signal_connect (codec, "XINERAMA-is-active", G_CALLBACK (XINERAMA_is_active_cb), logger);
    g_signal_connect (codec, "XINERAMA-is-active-reply", G_CALLBACK (XINERAMA_is_active_reply_cb), logger);
    g_signal_connect (codec, "XINERAMA-query-screens", G_CALLBACK (XINERAMA_query_screens_cb), logger);
    g_signal_connect (codec, "XINERAMA-query-screens-reply", G_CALLBACK (XINERAMA_query_screens_reply_cb), logger);
    g_signal_connect (codec, "XKB-use-extension", G_CALLBACK (XKB_use_extension_cb), logger);
    g_signal_connect (codec, "XKB-use-extension-reply", G_CALLBACK (XKB_use_extension_reply_cb), logger);
    g_signal_connect (codec, "XKB-select-events", G_CALLBACK (XKB_select_events_cb), logger);
    g_signal_connect (codec, "XKB-get-state", G_CALLBACK (XKB_get_state_cb), logger);
    g_signal_connect (codec, "XKB-get-state-reply", G_CALLBACK (XKB_get_state_reply_cb), logger);
    g_signal_connect (codec, "XKB-set-controls", G_CALLBACK (XKB_set_controls_cb), logger);
    g_signal_connect (codec, "XKB-get-map", G_CALLBACK (XKB_get_map_cb), logger);
    g_signal_connect (codec, "XKB-get-map-reply", G_CALLBACK (XKB_get_map_reply_cb), logger);
    g_signal_connect (codec, "XKB-get-names", G_CALLBACK (XKB_get_names_cb), logger);
    g_signal_connect (codec, "XKB-get-names-reply", G_CALLBACK (XKB_get_names_reply_cb), logger);
    g_signal_connect (codec, "XKB-per-client-flags", G_CALLBACK (XKB_per_client_flags_cb), logger);
    g_signal_connect (codec, "XKB-per-client-flags-reply", G_CALLBACK (XKB_per_client_flags_reply_cb), logger);
    g_signal_connect (codec, "RENDER-query-version", G_CALLBACK (RENDER_query_version_cb), logger);
    g_signal_connect (codec, "RENDER-query-version-reply", G_CALLBACK (RENDER_query_version_reply_cb), logger);
    g_signal_connect (codec, "RENDER-query-pict-formats", G_CALLBACK (RENDER_query_pict_formats_cb), logger);
    g_signal_connect (codec, "RENDER-query-pict-formats-reply", G_CALLBACK (RENDER_query_pict_formats_reply_cb), logger);

    return logger;
}

static void
xig_codec_logger_real_log_request (XigCodecLogger *logger, const gchar *text)
{
    g_printerr (">%s\n", text);
}

static void
xig_codec_logger_real_log_reply (XigCodecLogger *logger, const gchar *text)
{
    g_printerr ("<%s\n", text);
}

static void
xig_codec_logger_init (XigCodecLogger *logger)
{
    logger->priv = G_TYPE_INSTANCE_GET_PRIVATE (logger, xig_codec_logger_get_type (), XigCodecLoggerPrivate);
}

static void
xig_codec_logger_finalize (GObject *object)
{
    XigCodecLogger *logger = XIG_CODEC_LOGGER (object);
    g_object_unref (logger->priv->codec);

    G_OBJECT_CLASS (xig_codec_logger_parent_class)->finalize (object);
}

static void
xig_codec_logger_class_init (XigCodecLoggerClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    klass->log_request = xig_codec_logger_real_log_request;
    klass->log_reply = xig_codec_logger_real_log_reply;
    object_class->finalize = xig_codec_logger_finalize;

    g_type_class_add_private (klass, sizeof (XigCodecLoggerPrivate));
}
