/*
 * Copyright (C) 2011 Canonical Ltd.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation. See http://www.gnu.org/copyleft/lgpl.html the
 * full text of the license.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <gio/gio.h>
#include <gio/gunixsocketaddress.h>

#include "xig-remote-client-private.h"
#include "xig/xig-version.h"
#include "xig-server-private.h"
#include "xig-screen-private.h"
#include "xig-visual-private.h"
#include "xig-window-private.h"
#include "xig-drawable-private.h"
#include "xig-pixmap-private.h"
#include "xig-pixmap-format-private.h"
#include "xig-common.h"

G_DEFINE_TYPE (XigRemoteClient, xig_remote_client, xig_server_codec_get_type ());

struct XigRemoteClientPrivate
{
    /* Server this client is part of */
    XigServer *server;

    /* Socket to client */
    GSocket *socket;  
    GIOChannel *channel;

    /* Close down mode */
    guint8 close_down_mode;

    /* Sequence number of last request */
    guint16 sequence_number;

    /* Allocated resource IDs */
    guint32 resource_id_base;
    guint32 resource_id_mask;

    /* Event masks */
    GHashTable *window_events;
};

enum
{
    DISCONNECTED,
    LAST_SIGNAL
};
static guint signals[LAST_SIGNAL] = { 0 };

GInetAddress *
xig_remote_client_get_address (XigRemoteClient *client)
{
    GSocketAddress *socket_address;
    GError *error = NULL;

    socket_address = g_socket_get_remote_address (client->priv->socket, &error);
    if (error)
        g_warning ("Error getting remote socket address");
    g_clear_error (&error);
    if (!socket_address)
        return NULL;

    if (G_IS_INET_SOCKET_ADDRESS (socket_address))
        return g_inet_socket_address_get_address (G_INET_SOCKET_ADDRESS (socket_address));
    else
        return NULL;
}

static void
create_notify_cb (XigWindow *window, XigWindow *child, XigRemoteClient *client)
{
    XigCreateNotify message;

    message.sequence_number = client->priv->sequence_number;
    message.parent = xig_window_get_id (window);
    message.window = xig_window_get_id (child);
    message.x = xig_window_get_x (child);
    message.y = xig_window_get_y (child);
    message.width = xig_window_get_width (child);
    message.height = xig_window_get_height (child);
    message.border_width = xig_window_get_border_width (child);
    message.override_redirect = xig_window_get_override_redirect (child);

    xig_codec_feed_create_notify (XIG_CODEC (client), &message);
}

static void
send_destroy_notify (XigRemoteClient *client, XigWindow *event, XigWindow *window)
{
    XigDestroyNotify message;

    message.sequence_number = client->priv->sequence_number;
    message.event = xig_window_get_id (event);
    message.window = xig_window_get_id (window);

    xig_codec_feed_destroy_notify (XIG_CODEC (client), &message);
}

static void
destroy_notify_cb (XigWindow *window, XigRemoteClient *client)
{
    send_destroy_notify (client, window, window);
}

static void
child_destroy_notify_cb (XigWindow *window, XigWindow *child, XigRemoteClient *client)
{
    send_destroy_notify (client, child, window);
}

static void
send_unmap_notify (XigRemoteClient *client, XigWindow *event, XigWindow *window)
{
    XigUnmapNotify message;

    message.sequence_number = client->priv->sequence_number;
    message.event = xig_window_get_id (event);
    message.window = xig_window_get_id (window);
    message.from_configure = FALSE; // FIXME

    xig_codec_feed_unmap_notify (XIG_CODEC (client), &message);
}

static void
unmap_notify_cb (XigWindow *window, XigRemoteClient *client)
{
    send_unmap_notify (client, window, window);
}

static void
child_unmap_notify_cb (XigWindow *window, XigWindow *child, XigRemoteClient *client)
{
    send_unmap_notify (client, child, window);
}

static void
send_map_notify (XigRemoteClient *client, XigWindow *event, XigWindow *window)
{
    XigMapNotify message;

    message.sequence_number = client->priv->sequence_number;
    message.event = xig_window_get_id (event);
    message.window = xig_window_get_id (window);
    message.override_redirect = xig_window_get_override_redirect (event);

    xig_codec_feed_map_notify (XIG_CODEC (client), &message);
}

void
_xig_remote_client_map_request (XigRemoteClient *client, XigWindow *window)
{
    XigMapRequest message;

    message.sequence_number = client->priv->sequence_number;
    message.parent = xig_window_get_id (xig_window_get_parent (window));
    message.window = xig_window_get_id (window);

    xig_codec_feed_map_request (XIG_CODEC (client), &message);
}

static void
map_notify_cb (XigWindow *window, XigRemoteClient *client)
{
    send_map_notify (client, window, window);
}

static void
child_map_notify_cb (XigWindow *window, XigWindow *child, XigRemoteClient *client)
{
    send_map_notify (client, child, window);
}

static void
send_reparent_notify (XigRemoteClient *client, XigWindow *event, XigWindow *window)
{
    XigReparentNotify message;

    message.sequence_number = client->priv->sequence_number;
    message.event = xig_window_get_id (event);
    message.window = xig_window_get_id (window);
    message.parent = xig_window_get_id (xig_window_get_parent (event));
    message.x = xig_window_get_x (event);
    message.y = xig_window_get_y (event);
    message.override_redirect = xig_window_get_override_redirect (event);

    xig_codec_feed_reparent_notify (XIG_CODEC (client), &message);
}

static void
reparent_notify_cb (XigWindow *window, XigRemoteClient *client)
{
    send_reparent_notify (client, window, window);
}

static void
child_reparent_notify_cb (XigWindow *window, XigWindow *child, XigRemoteClient *client)
{
    send_reparent_notify (client, child, window);
}

static void
send_configure_notify (XigRemoteClient *client, XigWindow *event, XigWindow *window)
{
    XigWindow *below;
    XigConfigureNotify message;

    below = xig_window_get_window_below (event);

    message.sequence_number = client->priv->sequence_number;
    message.event = xig_window_get_id (event);
    message.window = xig_window_get_id (window);
    message.above_sibling = below ?  xig_window_get_id (below) : XIG_WINDOW_ID_NONE;
    message.x = xig_window_get_x (event);
    message.y = xig_window_get_y (event);
    message.width = xig_window_get_width (event);
    message.height = xig_window_get_height (event);
    message.border_width = xig_window_get_border_width (event);
    message.override_redirect = xig_window_get_override_redirect (event);

    xig_codec_feed_configure_notify (XIG_CODEC (client), &message);
}

void
_xig_remote_client_configure_request (XigRemoteClient *client, XigWindow *window, gint16 x, gint16 y, guint16 width, guint16 height, guint16 border_width, XigWindow *sibling, XigWindowStackMode stack_mode, XigConfigureWindowValueMask value_mask)
{
    XigConfigureRequest message;

    message.sequence_number = client->priv->sequence_number;
    message.stack_mode = stack_mode;
    message.parent = xig_window_get_id (xig_window_get_parent (window));
    message.window = xig_window_get_id (window);
    message.sibling = sibling ? xig_window_get_id (sibling) : XIG_WINDOW_ID_NONE;
    message.x = x;
    message.y = y;
    message.width = width;
    message.height = height;
    message.border_width = border_width;
    message.value_mask = value_mask;

    xig_codec_feed_configure_request (XIG_CODEC (client), &message);
}

static void
configure_notify_cb (XigWindow *window, XigRemoteClient *client)
{
    send_configure_notify (client, window, window);
}

static void
child_configure_notify_cb (XigWindow *window, XigWindow *child, XigRemoteClient *client)
{
    send_configure_notify (client, child, window);
}

static void
send_circulate_notify (XigRemoteClient *client, XigWindow *event, XigWindow *window)
{
    XigCirculateNotify message;

    message.sequence_number = client->priv->sequence_number;
    message.event = xig_window_get_id (event);
    message.window = xig_window_get_id (window);
    message.place = xig_window_get_is_on_bottom (event) ? 1 : 0;

    xig_codec_feed_circulate_notify (XIG_CODEC (client), &message);
}

void
_xig_remote_client_circulate_request (XigRemoteClient *client, XigWindow *window, XigCirculatePlace place)
{
    XigCirculateRequest message;

    message.sequence_number = client->priv->sequence_number;
    message.parent = xig_window_get_id (xig_window_get_parent (window));
    message.window = xig_window_get_id (window);
    message.place = place;

    xig_codec_feed_circulate_request (XIG_CODEC (client), &message);
}

static void
circulate_notify_cb (XigWindow *window, XigRemoteClient *client)
{
    send_circulate_notify (client, window, window);
}

static void
child_circulate_notify_cb (XigWindow *window, XigWindow *child, XigRemoteClient *client)
{
    send_circulate_notify (client, child, window);
}

static void
property_notify_cb (XigWindow *window, const gchar *name, XigRemoteClient *client)
{
    XigPropertyNotify message;

    message.sequence_number = client->priv->sequence_number;
    message.window = xig_window_get_id (window);
    message.atom = xig_server_get_atom_id (client->priv->server, name);
    message.time = 0; // FIXME
    message.state = xig_window_get_property (window, name) ? 0 : 1;

    xig_codec_feed_property_notify (XIG_CODEC (client), &message);
}

static void
set_window_events (XigRemoteClient *client, XigWindow *window, guint32 events)
{
    guint32 old_events, changed;

    old_events = xig_remote_client_get_window_events (client, window);
    g_hash_table_insert (client->priv->window_events, g_object_ref (window), GINT_TO_POINTER (events));

    changed = old_events ^ events;
    if ((changed & XIG_EVENT_StructureNotify) != 0)
    {
        if ((events & XIG_EVENT_StructureNotify) != 0)
        {
            g_signal_connect (window, "destroy-notify", G_CALLBACK (destroy_notify_cb), client);          
            g_signal_connect (window, "unmap-notify", G_CALLBACK (map_notify_cb), client);
            g_signal_connect (window, "map-notify", G_CALLBACK (map_notify_cb), client);
            g_signal_connect (window, "reparent-notify", G_CALLBACK (reparent_notify_cb), client);
            g_signal_connect (window, "configure-notify", G_CALLBACK (configure_notify_cb), client);
            g_signal_connect (window, "circulate-notify", G_CALLBACK (circulate_notify_cb), client);
        }
        else
        {
            g_signal_handlers_disconnect_by_func (window, destroy_notify_cb, client);          
            g_signal_handlers_disconnect_by_func (window, unmap_notify_cb, client);
            g_signal_handlers_disconnect_by_func (window, map_notify_cb, client);
            g_signal_handlers_disconnect_by_func (window, reparent_notify_cb, client);
            g_signal_handlers_disconnect_by_func (window, configure_notify_cb, client);
            g_signal_handlers_disconnect_by_func (window, circulate_notify_cb, client);
        }
    }
    if ((changed & XIG_EVENT_SubstructureNotify) != 0)
    {
        if ((events & XIG_EVENT_SubstructureNotify) != 0)
        {
            g_signal_connect (window, "create-notify", G_CALLBACK (create_notify_cb), client);
            g_signal_connect (window, "child-destroy-notify", G_CALLBACK (child_destroy_notify_cb), client);
            g_signal_connect (window, "child-unmap-notify", G_CALLBACK (child_unmap_notify_cb), client);
            g_signal_connect (window, "child-map-notify", G_CALLBACK (child_map_notify_cb), client);
            g_signal_connect (window, "child-reparent-notify", G_CALLBACK (child_reparent_notify_cb), client);
            g_signal_connect (window, "child-configure-notify", G_CALLBACK (child_configure_notify_cb), client);
            g_signal_connect (window, "child-circulate-notify", G_CALLBACK (child_circulate_notify_cb), client);
        }
        else
        {
            g_signal_handlers_disconnect_by_func (window, create_notify_cb, client);
            g_signal_handlers_disconnect_by_func (window, child_destroy_notify_cb, client);
            g_signal_handlers_disconnect_by_func (window, child_unmap_notify_cb, client);
            g_signal_handlers_disconnect_by_func (window, child_map_notify_cb, client);
            g_signal_handlers_disconnect_by_func (window, child_reparent_notify_cb, client);
            g_signal_handlers_disconnect_by_func (window, child_configure_notify_cb, client);
            g_signal_handlers_disconnect_by_func (window, child_circulate_notify_cb, client);
        }
    }
    if ((changed & XIG_EVENT_PropertyChange) != 0)
    {
        if ((events & XIG_EVENT_PropertyChange) != 0)
            g_signal_connect (window, "property-notify", G_CALLBACK (property_notify_cb), client);
        else
            g_signal_handlers_disconnect_by_func (window, property_notify_cb, client);
    }
}

guint32
xig_remote_client_get_window_events (XigRemoteClient *client, XigWindow *window)
{
    return GPOINTER_TO_INT (g_hash_table_lookup (client->priv->window_events, window));
}

void
xig_remote_client_disconnect (XigRemoteClient *client)
{
    g_io_channel_shutdown (client->priv->channel, TRUE, NULL);
}

static void
send_error (XigRemoteClient *client, guint8 type, guint8 major_opcode, guint16 minor_opcode, guint16 sequence_number, guint32 data)
{
    XigError message;

    message.type = type;
    message.major_opcode = major_opcode;
    message.minor_opcode = minor_opcode;
    message.sequence_number = sequence_number;
    message.data = data;

    xig_codec_feed_error (XIG_CODEC (client), &message);
}

static gboolean
socket_data_cb (GIOChannel *channel, GIOCondition condition, gpointer data)
{
    XigRemoteClient *client = data;
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gssize n_read;
  
    n_read = recv (g_io_channel_unix_get_fd (channel), buffer, MAXIMUM_REQUEST_LENGTH, 0);
    if (n_read < 0)
        g_warning ("Error reading from socket: %s", strerror (errno));
    else if (n_read == 0)
    {      
        g_signal_emit (client, signals[DISCONNECTED], 0);
        return FALSE;
    }
    else
        xig_codec_feed_data (XIG_CODEC (client), buffer, n_read);

    return TRUE;
}

static void 
xig_remote_client_write (XigCodec *codec, const guint8 *data, gsize data_length)
{
    XigRemoteClient *client = XIG_REMOTE_CLIENT (codec);
    send (g_io_channel_unix_get_fd (client->priv->channel), data, data_length, 0);
}

static void
xig_remote_client_connect (XigCodec *codec, XigConnect *message)
{
    XigRemoteClient *client = XIG_REMOTE_CLIENT (codec);
    XigServer *server = client->priv->server;
    XigSuccess reply;
    GList *link;
    int i;

    reply.protocol_major_version = XIG_PROTOCOL_MAJOR_VERSION;
    reply.protocol_minor_version = XIG_PROTOCOL_MINOR_VERSION;
    reply.release_number = XIG_RELEASE_NUMBER;
    reply.resource_id_base = client->priv->resource_id_base;
    reply.resource_id_mask = client->priv->resource_id_mask;
    reply.motion_buffer_size = xig_server_get_motion_buffer_size (server);
    reply.maximum_request_length = MAXIMUM_REQUEST_LENGTH;
    reply.image_byte_order = xig_server_get_image_byte_order (server);
    reply.bitmap_format_bit_order = xig_server_get_bitmap_format_bit_order (server);
    reply.bitmap_format_scanline_unit = 32; // FIXME
    reply.bitmap_format_scanline_pad = 32; // FIXME
    reply.min_keycode = xig_server_get_min_keycode (server);
    reply.max_keycode = xig_server_get_max_keycode (server);
    reply.vendor = (gchar *) xig_server_get_vendor (server);
    reply.pixmap_formats_length = g_list_length (xig_server_get_pixmap_formats (server));
    reply.pixmap_formats = calloc (reply.pixmap_formats_length, sizeof (XigFORMAT));
    for (link = xig_server_get_pixmap_formats (server), i = 0; link; link = link->next, i++)
    {
        XigPixmapFormat *format = link->data;
        reply.pixmap_formats[i].depth = xig_pixmap_format_get_depth (format);
        reply.pixmap_formats[i].bits_per_pixel = xig_pixmap_format_get_bits_per_pixel (format);
        reply.pixmap_formats[i].scanline_pad = xig_pixmap_format_get_scanline_pad (format);
    }
    reply.screens_length = g_list_length (xig_server_get_screens (server));
    reply.screens = calloc (reply.screens_length, sizeof (XigSCREEN));
    for (link = xig_server_get_screens (server), i = 0; link; link = link->next, i++)
    {
        XigSCREEN *s = &reply.screens[i];
        XigScreen *screen = link->data;
        XigWindow *root;
        XigVisual *root_visual;
        guint8 depths[255];
        GList *client_link;
        int j;

        root = xig_screen_get_root (screen);
        root_visual = xig_window_get_visual (root);

        s->current_input_masks = 0x00000000;
        for (client_link = xig_server_get_clients (server); client_link; client_link = client_link->next)
        {
            XigRemoteClient *c = client_link->data;
            s->current_input_masks |= xig_remote_client_get_window_events (c, root);
        }

        s->root = xig_window_get_id (root);
        s->default_colormap = 32; // FIXME
        s->white_pixel = xig_screen_get_white_pixel (screen);
        s->black_pixel = xig_screen_get_black_pixel (screen);
        s->width_in_pixels = xig_screen_get_width_in_pixels (screen);
        s->height_in_pixels = xig_screen_get_height_in_pixels (screen);
        s->width_in_millimeters = xig_screen_get_width_in_millimeters (screen);
        s->height_in_millimeters = xig_screen_get_height_in_millimeters (screen);
        s->min_installed_maps = 1; // FIXME
        s->max_installed_maps = 1; // FIXME
        s->root_visual = xig_visual_get_id (root_visual);
        s->backing_stores = 0; // FIXME
        s->save_unders = 0; // FIXME
        s->root_depth = xig_visual_get_depth (root_visual);

        /* Work out the number of supported depths */
        s->allowed_depths_length = 0;
        depths[0] = 0;
        while (TRUE)
        {
            GList *visual_link;
            int next_depth = 256;

            for (visual_link = xig_screen_get_visuals (screen); visual_link; visual_link = visual_link->next)
            {
                XigVisual *visual = visual_link->data;
                guint8 depth = xig_visual_get_depth (visual);
                if (depth > depths[s->allowed_depths_length] && depth < next_depth)
                    next_depth = depth;
            }

            if (next_depth == 256)
                break;

            depths[s->allowed_depths_length] = next_depth;
            s->allowed_depths_length++;
        }
        s->allowed_depths = calloc (s->allowed_depths_length, sizeof (XigDEPTH));

        for (j = 0; j < s->allowed_depths_length; j++)
        {
            XigDEPTH *d = &s->allowed_depths[j];
            GList *visual_link;
            guint16 n_visuals = 0;
            int k;

            d->depth = depths[j];
            for (visual_link = xig_screen_get_visuals (screen); visual_link; visual_link = visual_link->next)
            {
                XigVisual *visual = visual_link->data;
                if (xig_visual_get_depth (visual) == d->depth)
                    n_visuals++;
            }
            d->visuals_length = n_visuals;
            d->visuals = calloc (d->visuals_length, sizeof (XigVISUALTYPE));
            for (visual_link = xig_screen_get_visuals (screen), k = 0; visual_link; visual_link = visual_link->next)
            {
                XigVisual *visual = visual_link->data;

                if (xig_visual_get_depth (visual) != d->depth)
                    continue;

                d->visuals[k].visual_id = xig_visual_get_id (visual);
                d->visuals[k].class = xig_visual_get_class (visual);
                d->visuals[k].bits_per_rgb_value = xig_visual_get_bits_per_rgb_value (visual);
                d->visuals[k].colormap_entries = xig_visual_get_colormap_entries (visual);
                d->visuals[k].red_mask = xig_visual_get_red_mask (visual);
                d->visuals[k].green_mask = xig_visual_get_green_mask (visual);
                d->visuals[k].blue_mask = xig_visual_get_blue_mask (visual);
                k++;
            }
        }
    }

    xig_codec_feed_success (XIG_CODEC (codec), &reply);

    g_free (reply.pixmap_formats);
    for (i = 0; i < reply.screens_length; i++)
    {
        int j;
        for (j = 0; j < reply.screens_length; j++)
            g_free (reply.screens[i].allowed_depths[j].visuals);
        g_free (reply.screens[i].allowed_depths);
    }
    g_free (reply.screens);
}

static void 
xig_remote_client_create_window (XigCodec *codec, XigCreateWindow *message)
{
    XigRemoteClient *client = XIG_REMOTE_CLIENT (codec);
    GString *text;
    XigWindow *parent, *window;
    guint16 class;
    XigVisual *visual;
    XigPixmap *background_pixmap = NULL, *border_pixmap = NULL;

    text = g_string_new (">CreateWindow");
    g_string_append_printf (text, " wid=0x%08X parent=0x%08X x=%d y=%d width=%u height=%u",
                            message->wid, message->parent,
                            message->x, message->y,
                            message->width, message->height);
    if ((message->value_mask & XIG_WINDOW_VALUE_MASK_background_pixmap) != 0)
        g_string_append_printf (text, " background-pixmap=0x%08X", message->background_pixmap);
    if ((message->value_mask & XIG_WINDOW_VALUE_MASK_background_pixel) != 0)
        g_string_append_printf (text, " background-pixel=0x%08X", message->background_pixel);
    if ((message->value_mask & XIG_WINDOW_VALUE_MASK_border_pixmap) != 0)
        g_string_append_printf (text, " border-pixmap=0x%08X", message->border_pixmap);
    if ((message->value_mask & XIG_WINDOW_VALUE_MASK_border_pixel) != 0)
        g_string_append_printf (text, " border-pixel=0x%08X", message->border_pixel);
    if ((message->value_mask & XIG_WINDOW_VALUE_MASK_bit_gravity) != 0)
        g_string_append_printf (text, " bit-gravity=%d", message->bit_gravity);
    if ((message->value_mask & XIG_WINDOW_VALUE_MASK_win_gravity) != 0)
        g_string_append_printf (text, " win-gravity=%d", message->win_gravity);
    if ((message->value_mask & XIG_WINDOW_VALUE_MASK_backing_store) != 0)
        g_string_append_printf (text, " backing-store=%d", message->backing_store);
    if ((message->value_mask & XIG_WINDOW_VALUE_MASK_backing_planes) != 0)
        g_string_append_printf (text, " backing-planes=%d", message->backing_planes);
    if ((message->value_mask & XIG_WINDOW_VALUE_MASK_backing_pixel) != 0)
        g_string_append_printf (text, " backing-pixel=0x%08X", message->backing_pixel);
    if ((message->value_mask & XIG_WINDOW_VALUE_MASK_override_redirect) != 0)
        g_string_append_printf (text, " override-redirect=%s", message->override_redirect ? "true" : "false");
    if ((message->value_mask & XIG_WINDOW_VALUE_MASK_save_under) != 0)
        g_string_append_printf (text, " save-under=%s", message->save_under ? "true" : "false");
    if ((message->value_mask & XIG_WINDOW_VALUE_MASK_event_mask) != 0)
        g_string_append_printf (text, " event-mask=0x%08X", message->event_mask);
    if ((message->value_mask & XIG_WINDOW_VALUE_MASK_do_not_propagate_mask) != 0)
        g_string_append_printf (text, " do-not-propagate-mask=0x%08X", message->do_not_propagate_mask);
    if ((message->value_mask & XIG_WINDOW_VALUE_MASK_colormap) != 0)
        g_string_append_printf (text, " colormap=0x%08X", message->colormap);
    if ((message->value_mask & XIG_WINDOW_VALUE_MASK_cursor) != 0)
        g_string_append_printf (text, " cursor=0x%08X", message->cursor);
    g_string_free (text, TRUE);

    // FIXME: Check valid wid

    /* Check drawable doesn't already exist */
    if (xig_server_get_drawable (client->priv->server, message->wid))
    {
        send_error (client, IDChoice, XIG_REQUEST_CreateWindow, 0, message->sequence_number, message->wid);
        return;
    }

    parent = xig_server_get_window (client->priv->server, message->parent);
    if (!parent)
    {
        send_error (client, Window, XIG_REQUEST_CreateWindow, 0, message->sequence_number, message->parent);
        return;
    }

    if (message->width == 0)
    {
        send_error (client, Value, XIG_REQUEST_CreateWindow, 0, message->sequence_number, message->width);
        return;
    }
    if (message->height == 0)
    {
        send_error (client, Value, XIG_REQUEST_CreateWindow, 0, message->sequence_number, message->height);
        return;
    }

    class = message->class;
    if (class == XIG_WINDOW_CLASS_CopyFromParent)
        class = xig_window_get_class (parent);

    if (class == XIG_WINDOW_CLASS_InputOnly && message->border_width != 0)
    {
        send_error (client, Match, XIG_REQUEST_CreateWindow, 0, message->sequence_number, 0);
        return;
    }

    // FIXME
    //if (message->visual == 0)
    visual = xig_window_get_visual (parent);
    //else
    //    visual = ?_get_visual (message->visual);

    if ((message->value_mask & XIG_WINDOW_VALUE_MASK_background_pixmap) != 0)
    {
        background_pixmap = xig_server_get_pixmap (client->priv->server, message->background_pixmap);
        if (!background_pixmap && message->background_pixmap != XIG_PIXMAP_ID_NONE)
        {
            send_error (client, Pixmap, XIG_REQUEST_CreateWindow, 0, message->sequence_number, message->background_pixmap);
            return;
        }
    }
    if ((message->value_mask & XIG_WINDOW_VALUE_MASK_border_pixmap) != 0)
    {
        border_pixmap = xig_server_get_pixmap (client->priv->server, message->border_pixmap);
        if (!border_pixmap && message->border_pixmap != XIG_PIXMAP_ID_NONE)
        {
            send_error (client, Pixmap, XIG_REQUEST_CreateWindow, 0, message->sequence_number, message->border_pixmap);
            return;
        }
    }

    // FIXME: Check valid cursor

    // FIXME: colormap and cursor

    window = xig_window_add_child (parent,
                                   client,
                                   message->wid,
                                   class,
                                   message->x,
                                   message->y,
                                   message->width,
                                   message->height,
                                   message->border_width,
                                   visual,
                                   message->value_mask,
                                   background_pixmap,
                                   message->background_pixel,
                                   border_pixmap,
                                   message->border_pixel,
                                   message->bit_gravity,
                                   message->win_gravity,
                                   message->backing_store,
                                   message->backing_planes,
                                   message->backing_pixel,
                                   message->override_redirect,
                                   message->save_under);

    if ((message->value_mask & XIG_WINDOW_VALUE_MASK_event_mask) != 0)
        set_window_events (client, window, message->event_mask); // FIXME: Check if Redirect is set and it's already set elsewhere
}

static void 
xig_remote_client_change_window_attributes (XigCodec *codec, XigChangeWindowAttributes *message)
{
    XigRemoteClient *client = XIG_REMOTE_CLIENT (codec);
    XigWindow *window;
    XigPixmap *background_pixmap = NULL, *border_pixmap = NULL;

    window = xig_server_get_window (client->priv->server, message->window);
    if (!window)
    {
        send_error (client, Window, XIG_REQUEST_ChangeWindowAttributes, 0, message->sequence_number, message->window);
        return;
    }

    if (message->value_mask & XIG_WINDOW_VALUE_MASK_background_pixmap)
    {
        background_pixmap = xig_server_get_pixmap (client->priv->server, message->background_pixmap);
        if (!background_pixmap && message->background_pixmap != XIG_PIXMAP_ID_NONE)
        {
            send_error (client, Pixmap, XIG_REQUEST_ChangeWindowAttributes, 0, message->sequence_number, message->background_pixmap);
            return;
        }
    }
    if (message->value_mask & XIG_WINDOW_VALUE_MASK_border_pixmap)
    {
        border_pixmap = xig_server_get_pixmap (client->priv->server, message->border_pixmap);
        if (!border_pixmap && message->border_pixmap != XIG_PIXMAP_ID_NONE)
        {
            send_error (client, Pixmap, XIG_REQUEST_ChangeWindowAttributes, 0, message->sequence_number, message->border_pixmap);
            return;
        }     
    }

    // FIXME: Check valid cursor

    // FIXME: colormap and cursor

    xig_window_change_attributes (window,
                                  message->value_mask,
                                  background_pixmap,
                                  message->background_pixel,
                                  border_pixmap,
                                  message->border_pixel,
                                  message->bit_gravity,
                                  message->win_gravity,
                                  message->backing_store,
                                  message->backing_planes,
                                  message->backing_pixel,
                                  message->override_redirect,
                                  message->save_under);
    if ((message->value_mask & XIG_WINDOW_VALUE_MASK_event_mask) != 0)
        set_window_events (client, window, message->event_mask); // FIXME: Check if Redirect is set and it's already set elsewhere
}

static void 
xig_remote_client_get_window_attributes (XigCodec *codec, XigGetWindowAttributes *message)
{
    XigRemoteClient *client = XIG_REMOTE_CLIENT (codec);
    XigWindow *window;
    XigGetWindowAttributesReply reply;
    GList *link;

    window = xig_server_get_window (client->priv->server, message->window);
    if (!window)
    {
        send_error (client, Window, XIG_REQUEST_GetWindowAttributes, 0, message->sequence_number, message->window);
        return;
    }

    reply.sequence_number = message->sequence_number;
    reply.backing_store = xig_window_get_backing_store (window);
    reply.visual = xig_visual_get_id (xig_window_get_visual (window));
    reply.class = xig_window_get_class (window);
    reply.bit_gravity = xig_window_get_bit_gravity (window);
    reply.win_gravity = xig_window_get_win_gravity (window);
    reply.backing_planes = xig_window_get_backing_planes (window);
    reply.backing_pixel = xig_window_get_backing_pixel (window);
    reply.save_under = xig_window_get_save_under (window);
    reply.map_is_installed = TRUE; // FIXME
    reply.map_state = 2; // FIXME
    reply.override_redirect = xig_window_get_override_redirect (window);
    reply.colormap = 32; // FIXME
    reply.all_event_masks = 0x00000000;
    reply.your_event_mask = xig_remote_client_get_window_events (client, window);
    reply.do_not_propagate_mask = 0x0000; // FIXME

    for (link = xig_server_get_clients (client->priv->server); link; link = link->next)
    {
        XigRemoteClient *c = link->data;
        reply.all_event_masks |= xig_remote_client_get_window_events (c, window);
    }

    xig_codec_feed_get_window_attributes_reply (XIG_CODEC (codec), &reply);
}

static void 
xig_remote_client_destroy_window (XigCodec *codec, XigDestroyWindow *message)
{
    XigRemoteClient *client = XIG_REMOTE_CLIENT (codec);
    XigWindow *window;

    window = xig_server_get_window (client->priv->server, message->window);
    if (!window)
    {
        send_error (client, Window, XIG_REQUEST_DestroyWindow, 0, message->sequence_number, message->window);
        return;
    }

    xig_window_destroy (window);
}

static void 
xig_remote_client_destroy_subwindows (XigCodec *codec, XigDestroySubwindows *message)
{
    XigRemoteClient *client = XIG_REMOTE_CLIENT (codec);
    XigWindow *window;

    window = xig_server_get_window (client->priv->server, message->window);
    if (!window)
    {
        send_error (client, Window, XIG_REQUEST_DestroySubwindows, 0, message->sequence_number, message->window);
        return;
    }

    xig_window_destroy_subwindows (window);
}

static void 
xig_remote_client_change_save_set (XigCodec *codec, XigChangeSaveSet *message)
{
}

static void 
xig_remote_client_reparent_window (XigCodec *codec, XigReparentWindow *message)
{
    XigRemoteClient *client = XIG_REMOTE_CLIENT (codec);
    XigWindow *window, *parent;

    if (message->window == message->parent)
    {
        send_error (client, Match, XIG_REQUEST_ReparentWindow, 0, message->sequence_number, message->window);
        return;
    }

    window = xig_server_get_window (client->priv->server, message->window);
    if (!window)
    {
        send_error (client, Window, XIG_REQUEST_ReparentWindow, 0, message->sequence_number, message->window);
        return;
    }
    parent = xig_server_get_window (client->priv->server, message->parent);
    if (!parent)
    {
        send_error (client, Window, XIG_REQUEST_ReparentWindow, 0, message->sequence_number, message->parent);
        return;
    }

    xig_window_reparent (window, parent, message->x, message->y);
}

static void
xig_remote_client_map_window (XigCodec *codec, XigMapWindow *message)
{
    XigRemoteClient *client = XIG_REMOTE_CLIENT (codec);
    XigWindow *window;

    window = xig_server_get_window (client->priv->server, message->window);
    if (!window)
    {
        send_error (client, Window, XIG_REQUEST_MapWindow, 0, message->sequence_number, message->window);
        return;
    }

    xig_window_map (window, client);
}

static void 
xig_remote_client_map_subwindows (XigCodec *codec, XigMapSubwindows *message)
{
    XigRemoteClient *client = XIG_REMOTE_CLIENT (codec);
    XigWindow *window;
    GList *link;

    window = xig_server_get_window (client->priv->server, message->window);
    if (!window)
    {
        send_error (client, Window, XIG_REQUEST_MapSubwindows, 0, message->sequence_number, message->window);
        return;
    }

    for (link = g_list_last (xig_window_get_children (window)); link; link = link->prev)
    {
        XigWindow *w = link->data;
        xig_window_map (w, client);
    }
}

static void 
xig_remote_client_unmap_window (XigCodec *codec, XigUnmapWindow *message)
{
    XigRemoteClient *client = XIG_REMOTE_CLIENT (codec);
    XigWindow *window;

    window = xig_server_get_window (client->priv->server, message->window);
    if (!window)
    {
        send_error (client, Window, XIG_REQUEST_UnmapWindow, 0, message->sequence_number, message->window);
        return;
    }

    xig_window_unmap (window);
}

static void 
xig_remote_client_unmap_subwindows (XigCodec *codec, XigUnmapSubwindows *message)
{
    XigRemoteClient *client = XIG_REMOTE_CLIENT (codec);
    XigWindow *window;
    GList *link;

    window = xig_server_get_window (client->priv->server, message->window);
    if (!window)
    {
        send_error (client, Window, XIG_REQUEST_UnmapSubwindows, 0, message->sequence_number, message->window);
        return;
    }

    for (link = xig_window_get_children (window); link; link = link->next)
    {
        XigWindow *w = link->data;
        xig_window_unmap (w);
    }
}

static void 
xig_remote_client_configure_window (XigCodec *codec, XigConfigureWindow *message)
{
    XigRemoteClient *client = XIG_REMOTE_CLIENT (codec);
    XigWindow *window, *sibling = NULL;
    gint16 x, y;
    guint16 width, height, border_width;

    window = xig_server_get_window (client->priv->server, message->window);
    if (!window)
    {
        send_error (client, Window, XIG_REQUEST_ConfigureWindow, 0, message->sequence_number, message->window);
        return;
    }

    /* Can't configure root window */
    if (xig_window_is_root (window))
        return;

    if ((message->value_mask & XIG_CONFIGURE_WINDOW_VALUE_MASK_x) != 0)
        x = message->x;
    else
        x = xig_window_get_x (window);
    if ((message->value_mask & XIG_CONFIGURE_WINDOW_VALUE_MASK_y) != 0)
        y = message->y;
    else
        y = xig_window_get_y (window);
    if ((message->value_mask & XIG_CONFIGURE_WINDOW_VALUE_MASK_width) != 0)
    {
        if (message->width == 0)
        {
            send_error (client, Value, XIG_REQUEST_ConfigureWindow, 0, message->sequence_number, message->width);
            return;
        }
        width = message->width;
    }
    else
        width = xig_window_get_width (window);
    if ((message->value_mask & XIG_CONFIGURE_WINDOW_VALUE_MASK_height) != 0)
    {
        if (message->height == 0)
        {
            send_error (client, Value, XIG_REQUEST_ConfigureWindow, 0, message->sequence_number, message->height);
            return;
        }
        height = message->height;
    }
    else
        height = xig_window_get_height (window);
    if ((message->value_mask & XIG_CONFIGURE_WINDOW_VALUE_MASK_border_width) != 0)
    {
        if (xig_window_get_class (window) == XIG_WINDOW_CLASS_InputOnly && message->border_width != 0)
        {
            send_error (client, Match, XIG_REQUEST_ConfigureWindow, 0, message->sequence_number, 0);
            return;
        }
        border_width = message->border_width;
    }
    else
        border_width = xig_window_get_border_width (window);
    if ((message->value_mask & XIG_CONFIGURE_WINDOW_VALUE_MASK_sibling) != 0)
    {
        sibling = xig_server_get_window (client->priv->server, message->sibling);
        if (!sibling && message->sibling != XIG_WINDOW_ID_NONE)
        {
            send_error (client, Window, XIG_REQUEST_ConfigureWindow, 0, message->sequence_number, message->window);
            return;
        }
        if (sibling && (!xig_window_has_sibling (window, sibling)))
        {
            send_error (client, Match, XIG_REQUEST_ConfigureWindow, 0, message->sequence_number, message->sibling);
            return;
        }
        if (sibling && (message->value_mask & XIG_CONFIGURE_WINDOW_VALUE_MASK_stack_mode) == 0)
        {
            send_error (client, Match, XIG_REQUEST_ConfigureWindow, 0, message->sequence_number, message->sibling);
            return;
        }
    }

    xig_window_configure (window, client, x, y, width, height, border_width, sibling, message->stack_mode, message->value_mask);
}

static void 
xig_remote_client_circulate_window (XigCodec *codec, XigCirculateWindow *message)
{
    XigRemoteClient *client = XIG_REMOTE_CLIENT (codec);
    XigWindow *window;
 
    window = xig_server_get_window (client->priv->server, message->window);
    if (!window)
    {
        send_error (client, Window, XIG_REQUEST_CirculateWindow, 0, message->sequence_number, message->window);
        return;
    }

    switch (message->direction)
    {
    case XIG_CIRCULATE_DIRECTION_RAISE_LOWEST:
    case XIG_CIRCULATE_DIRECTION_LOWER_HIGHEST:
        xig_window_circulate (window, client, message->direction);
        break;
    default:
        send_error (client, Value, XIG_REQUEST_CirculateWindow, 0, message->sequence_number, message->direction);
        return;
    }
}

static void 
xig_remote_client_get_geometry (XigCodec *codec, XigGetGeometry *message)
{
    XigRemoteClient *client = XIG_REMOTE_CLIENT (codec);
    XigWindow *window;
    XigGetGeometryReply reply;

    // FIXME: Also handle pixmaps
    window = xig_server_get_window (client->priv->server, message->drawable);
    if (!window)
    {
        send_error (client, Window, XIG_REQUEST_GetGeometry, 0, message->sequence_number, message->drawable);
        return;
    }

    reply.sequence_number = message->sequence_number;
    reply.depth = xig_visual_get_depth (xig_window_get_visual (window));
    reply.root = xig_window_get_id (xig_window_get_root (window));
    reply.x = xig_window_get_x (window);
    reply.y = xig_window_get_y (window);
    reply.width = xig_window_get_width (window);
    reply.height = xig_window_get_height (window);
    reply.border_width = xig_window_get_border_width (window);

    xig_codec_feed_get_geometry_reply (XIG_CODEC (codec), &reply);
}

static void 
xig_remote_client_query_tree (XigCodec *codec, XigQueryTree *message)
{
    XigRemoteClient *client = XIG_REMOTE_CLIENT (codec);
    XigWindow *window;
    XigQueryTreeReply reply;
    GList *children, *link;
    int i;

    window = xig_server_get_window (client->priv->server, message->window);
    if (!window)
    {
        send_error (client, Window, XIG_REQUEST_QueryTree, 0, message->sequence_number, message->window);
        return;
    }

    reply.sequence_number = message->sequence_number;
    reply.root = xig_window_get_id (xig_window_get_root (window));
    reply.parent = xig_window_get_parent (window) ? xig_window_get_id (xig_window_get_parent (window)) : XIG_WINDOW_ID_NONE;
    children = xig_window_get_children (window);
    reply.children_length = g_list_length (children);
    reply.children = calloc (reply.children_length, sizeof (guint32));
    for (link = children, i = 0; link; link = link->next, i++)
    {
        XigWindow *child = link->data;
        reply.children[i] = xig_window_get_id (child);
    }

    xig_codec_feed_query_tree_reply (XIG_CODEC (codec), &reply);

    g_free (reply.children);
}

static void 
xig_remote_client_intern_atom (XigCodec *codec, XigInternAtom *message)
{
    XigRemoteClient *client = XIG_REMOTE_CLIENT (codec);  
    XigInternAtomReply reply;

    reply.sequence_number = message->sequence_number;
    reply.atom = xig_server_intern_atom (client->priv->server, message->name, message->only_if_exists);

    xig_codec_feed_intern_atom_reply (XIG_CODEC (codec), &reply);
}

static void 
xig_remote_client_get_atom_name (XigCodec *codec, XigGetAtomName *message)
{
    XigRemoteClient *client = XIG_REMOTE_CLIENT (codec);  
    XigGetAtomNameReply reply;

    reply.sequence_number = message->sequence_number;
    reply.name = (gchar *) xig_server_get_atom_name (client->priv->server, message->atom);
    if (!reply.name)
    {
        send_error (client, Atom, XIG_REQUEST_GetAtomName, 0, message->sequence_number, message->atom);
        return;
    }

    xig_codec_feed_get_atom_name_reply (XIG_CODEC (codec), &reply);
}

static void 
xig_remote_client_change_property (XigCodec *codec, XigChangeProperty *message)
{
    XigRemoteClient *client = XIG_REMOTE_CLIENT (codec);  
    XigWindow *window;
    const gchar *name, *type;

    window = xig_server_get_window (client->priv->server, message->window);
    if (!window)
    {
        send_error (client, Window, XIG_REQUEST_ChangeProperty, 0, message->sequence_number, message->window);
        return;
    }
    name = xig_server_get_atom_name (client->priv->server, message->property);
    if (!name)
    {
        send_error (client, Atom, XIG_REQUEST_ChangeProperty, 0, message->sequence_number, message->property);
        return;
    }
    type = xig_server_get_atom_name (client->priv->server, message->type);
    if (!type)
    {
        send_error (client, Atom, XIG_REQUEST_ChangeProperty, 0, message->sequence_number, message->type);
        return;
    }

    /* Check format is valid */
    switch (message->data_format)
    {
    case 8:
    case 16:
    case 32:
        break;
    default:
        send_error (client, Value, XIG_REQUEST_ChangeProperty, 0, message->sequence_number, message->data_format);
        return;
    }

    switch (message->mode)
    {
    case XIG_CHANGE_PROPERTY_MODE_REPLACE:
        xig_window_replace_property (window, name, type, message->data_format, message->data, message->data_length * (message->data_format / 8));
        break;
    case XIG_CHANGE_PROPERTY_MODE_APPEND:
        // FIXME: Check property exists and type and format matches
        xig_window_append_property (window, name, message->data, message->data_length);
        break;
    case XIG_CHANGE_PROPERTY_MODE_PREPEND:
        // FIXME: Check property exists and type and format matches
        xig_window_prepend_property (window, name, message->data, message->data_length);
        break;
    default:
        send_error (client, Value, XIG_REQUEST_ChangeProperty, 0, message->sequence_number, message->mode);
        return;
    }
}

static void 
xig_remote_client_delete_property (XigCodec *codec, XigDeleteProperty *message)
{
    XigRemoteClient *client = XIG_REMOTE_CLIENT (codec);  
    XigWindow *window;
    const gchar *name;

    window = xig_server_get_window (client->priv->server, message->window);
    if (!window)
    {
        send_error (client, Window, XIG_REQUEST_DeleteProperty, 0, message->sequence_number, message->window);
        return;
    }
    name = xig_server_get_atom_name (client->priv->server, message->property);
    if (!name)
    {
        send_error (client, Atom, XIG_REQUEST_DeleteProperty, 0, message->sequence_number, message->property);
        return;
    }

    xig_window_delete_property (window, name);
}

static void 
xig_remote_client_get_property (XigCodec *codec, XigGetProperty *message)
{
    XigRemoteClient *client = XIG_REMOTE_CLIENT (codec);  
    XigWindow *window;
    const gchar *name, *type_name = NULL;
    XigWindowProperty *property;
    XigGetPropertyReply reply;

    window = xig_server_get_window (client->priv->server, message->window);
    if (!window)
    {
        send_error (client, Window, XIG_REQUEST_GetProperty, 0, message->sequence_number, message->window);
        return;
    }
    name = xig_server_get_atom_name (client->priv->server, message->property);
    if (!name)
    {
        send_error (client, Atom, XIG_REQUEST_GetProperty, 0, message->sequence_number, message->property);
        return;
    }
    if (message->type != 0)
    {
        type_name = xig_server_get_atom_name (client->priv->server, message->type);
        if (!type_name)
        {
            send_error (client, Atom, XIG_REQUEST_GetProperty, 0, message->sequence_number, message->type);
            return;
        }
    }

    reply.sequence_number = message->sequence_number;
    reply.type = 0;
    reply.bytes_after = 0;
    reply.value_format = 0;
    reply.value = NULL;
    reply.value_length = 0;

    property = xig_window_get_property (window, name);
    if (property)
    {
        /* If the type is wrong then return the correct type and size */
        if (type_name && strcmp (property->type, type_name) != 0)
        {
            reply.type = xig_server_get_atom_id (client->priv->server, property->type);
            reply.value_format = property->format;
            reply.bytes_after = property->data_length;
        }
        else
        {
            gsize offset;

            reply.type = xig_server_get_atom_id (client->priv->server, property->type);
            reply.value_format = property->format;

            /* Get the subset of the data requested */
            offset = MIN (property->data_length, 4 * message->long_offset);
            if (offset > property->data_length)
            {
                send_error (client, Value, XIG_REQUEST_GetProperty, 0, message->sequence_number, message->long_offset);
                return;
            }

            reply.value_length = MIN (property->data_length - offset, 4 * message->long_length);
            reply.value = property->data + offset;
            reply.bytes_after = property->data_length - (offset + reply.value_length);
        }
    }

    xig_codec_feed_get_property_reply (XIG_CODEC (codec), &reply);

    if (message->delete && reply.bytes_after == 0)
        xig_window_delete_property (window, name);
}

static void 
xig_remote_client_list_properties (XigCodec *codec, XigListProperties *message)
{
    XigRemoteClient *client = XIG_REMOTE_CLIENT (codec);  
    XigWindow *window;
    XigListPropertiesReply reply;
    GHashTable *properties;
    GHashTableIter iter;
    int i = 0;

    window = xig_server_get_window (client->priv->server, message->window);
    if (!window)
    {
        send_error (client, Window, XIG_REQUEST_ListProperties, 0, message->sequence_number, message->window);
        return;
    }
    properties = xig_window_get_properties (window);

    reply.sequence_number = message->sequence_number;
    reply.atoms_length = g_hash_table_size (properties);
    reply.atoms = calloc (reply.atoms_length, sizeof (guint32));

    g_hash_table_iter_init (&iter, properties);
    while (TRUE)
    {
        const gchar *name;
        XigWindowProperty *property;

        if (!g_hash_table_iter_next (&iter, (gpointer) &name, (gpointer) &property))
            break;
      
        reply.atoms[i] = xig_server_get_atom_id (client->priv->server, name);
        i++;
    }

    xig_codec_feed_list_properties_reply (XIG_CODEC (codec), &reply);
                          
    g_free (reply.atoms);
}

static void
xig_remote_client_set_selection_owner (XigCodec *codec, XigSetSelectionOwner *message)
{
    XigRemoteClient *client = XIG_REMOTE_CLIENT (codec);  
    XigWindow *window = NULL;
    const gchar *name;

    window = xig_server_get_window (client->priv->server, message->owner);
    if (!window && message->owner != XIG_WINDOW_ID_NONE)
    {
        send_error (client, Window, XIG_REQUEST_SetSelectionOwner, 0, message->sequence_number, message->owner);
        return;
    }
    name = xig_server_get_atom_name (client->priv->server, message->selection);
    if (!name)
    {
        send_error (client, Atom, XIG_REQUEST_SetSelectionOwner, 0, message->sequence_number, message->selection);
        return;
    }

    _xig_server_set_selection_owner (client->priv->server, name, window);
}

static void
xig_remote_client_get_selection_owner (XigCodec *codec, XigGetSelectionOwner *message)
{
    XigRemoteClient *client = XIG_REMOTE_CLIENT (codec);  
    const gchar *name;
    XigWindow *owner;
    XigGetSelectionOwnerReply reply;

    name = xig_server_get_atom_name (client->priv->server, message->selection);
    if (!name)
    {
        send_error (client, Atom, XIG_REQUEST_GetSelectionOwner, 0, message->sequence_number, message->selection);
        return;
    }

    owner = xig_server_get_selection_owner (client->priv->server, name);

    reply.sequence_number = message->sequence_number;
    reply.owner = owner ? xig_window_get_id (owner) : XIG_WINDOW_ID_NONE;

    xig_codec_feed_get_selection_owner_reply (XIG_CODEC (codec), &reply);
}

static void
xig_remote_client_convert_selection (XigCodec *codec, XigConvertSelection *message)
{
}

static void
xig_remote_client_send_event (XigCodec *codec, XigSendEvent *message)
{
    // FIXME:
}

static void
xig_remote_client_grab_pointer (XigCodec *codec, XigGrabPointer *message)
{
    XigGrabPointerReply reply;
 
    reply.sequence_number = message->sequence_number;
    reply.status = 0; // FIXME

    xig_codec_feed_grab_pointer_reply (XIG_CODEC (codec), &reply);
}

static void
xig_remote_client_ungrab_pointer (XigCodec *codec, XigUngrabPointer *message)
{
}

static void
xig_remote_client_grab_button (XigCodec *codec, XigGrabButton *message)
{
}

static void
xig_remote_client_ungrab_button (XigCodec *codec, XigUngrabButton *message)
{
}

static void
xig_remote_client_change_active_pointer_grab (XigCodec *codec, XigChangeActivePointerGrab *message)
{
}

static void
xig_remote_client_grab_keyboard (XigCodec *codec, XigGrabKeyboard *message)
{
    XigGrabKeyboardReply reply;

    reply.sequence_number = message->sequence_number;
    reply.status = 0; // FIXME

    xig_codec_feed_grab_keyboard_reply (XIG_CODEC (codec), &reply);
}

static void
xig_remote_client_ungrab_keyboard (XigCodec *codec, XigUngrabKeyboard *message)
{
}
  
static void
xig_remote_client_grab_key (XigCodec *codec, XigGrabKey *message)
{
}

static void
xig_remote_client_ungrab_key (XigCodec *codec, XigUngrabKey *message)
{
}

static void
xig_remote_client_allow_events (XigCodec *codec, XigAllowEvents *message)
{
}

static void
xig_remote_client_grab_server (XigCodec *codec, XigGrabServer *message)
{
}

static void
xig_remote_client_ungrab_server (XigCodec *codec, XigUngrabServer *message)
{
}

static void
xig_remote_client_query_pointer (XigCodec *codec, XigQueryPointer *message)
{
    XigRemoteClient *client = XIG_REMOTE_CLIENT (codec);
    XigWindow *window;
    XigQueryPointerReply reply;

    window = xig_server_get_window (client->priv->server, message->window);
    if (!window)
    {
        send_error (client, Window, XIG_REQUEST_QueryPointer, 0, message->sequence_number, message->window);
        return;
    }

    reply.sequence_number = message->sequence_number;
    reply.same_screen = TRUE; // FIXME
    reply.root = xig_window_get_id (xig_window_get_root (window));
    reply.child = 0; // FIXME
    reply.root_x = 0; // FIXME
    reply.root_y = 0; // FIXME
    reply.win_x = 0; // FIXME
    reply.win_y = 0; // FIXME
    reply.mask = 0x0000; // FIXME

    xig_codec_feed_query_pointer_reply (XIG_CODEC (codec), &reply);
}

static void
xig_remote_client_get_motion_events (XigCodec *codec, XigGetMotionEvents *message)
{
    XigRemoteClient *client = XIG_REMOTE_CLIENT (codec);
    XigWindow *window;
    XigGetMotionEventsReply reply;

    window = xig_server_get_window (client->priv->server, message->window);
    if (!window)
    {
        send_error (client, Window, XIG_REQUEST_GetMotionEvents, 0, message->sequence_number, message->window);
        return;
    }

    reply.sequence_number = message->sequence_number;
    reply.events_length = 0; // FIXME
    reply.events = NULL; // FIXME

    xig_codec_feed_get_motion_events_reply (XIG_CODEC (codec), &reply);

    g_free (reply.events);
}

static void
xig_remote_client_translate_coordinates (XigCodec *codec, XigTranslateCoordinates *message)
{
    XigRemoteClient *client = XIG_REMOTE_CLIENT (codec);
    XigWindow *src_window, *dst_window;
    XigTranslateCoordinatesReply reply;

    src_window = xig_server_get_window (client->priv->server, message->src_window);
    if (!src_window)
    {
        send_error (client, Window, XIG_REQUEST_TranslateCoordinates, 0, message->sequence_number, message->src_window);
        return;
    }
    dst_window = xig_server_get_window (client->priv->server, message->dst_window);
    if (!dst_window)
    {
        send_error (client, Window, XIG_REQUEST_TranslateCoordinates, 0, message->sequence_number, message->dst_window);
        return;
    }

    reply.sequence_number = message->sequence_number;
    reply.same_screen = TRUE; // FIXME
    reply.child = 0; // FIXME
    reply.dst_x = 0; // FIXME
    reply.dst_y = 0; // FIXME

    xig_codec_feed_translate_coordinates_reply (XIG_CODEC (codec), &reply);
}

static void
xig_remote_client_warp_pointer (XigCodec *codec, XigWarpPointer *message)
{
}

static void
xig_remote_client_set_input_focus (XigCodec *codec, XigSetInputFocus *message)
{
    XigRemoteClient *client = XIG_REMOTE_CLIENT (codec);
    XigWindow *window;

    window = xig_server_get_window (client->priv->server, message->focus);
    if (!window)
    {
        send_error (client, Window, 42, 0, message->sequence_number, message->focus);
        return;
    }

    xig_server_set_input_focus (client->priv->server, window, message->revert_to);
}

static void 
xig_remote_client_get_input_focus (XigCodec *codec, XigGetInputFocus *message)
{
    XigRemoteClient *client = XIG_REMOTE_CLIENT (codec);  
    XigWindow *focus;
    XigGetInputFocusReply reply;

    reply.sequence_number = message->sequence_number;
    focus = xig_server_get_input_focus (client->priv->server, &reply.revert_to);
    reply.focus = focus ? xig_window_get_id (focus) : XIG_WINDOW_ID_NONE;

    xig_codec_feed_get_input_focus_reply (XIG_CODEC (codec), &reply);
}

static void 
xig_remote_client_query_keymap (XigCodec *codec, XigQueryKeymap *message)
{
    XigQueryKeymapReply reply;
    int i;

    reply.sequence_number = message->sequence_number;
    for (i = 0; i < 32; i++)
        reply.keys[i] = 0x00; // FIXME

    xig_codec_feed_query_keymap_reply (XIG_CODEC (codec), &reply);
}

static void 
xig_remote_client_open_font (XigCodec *codec, XigOpenFont *message)
{
}

static void 
xig_remote_client_close_font (XigCodec *codec, XigCloseFont *message)
{
}

static void 
xig_remote_client_query_font (XigCodec *codec, XigQueryFont *message)
{
    XigQueryFontReply reply;
  
    reply.sequence_number = message->sequence_number;
    reply.min_bounds.left_side_bearing = 0; // FIXME
    reply.min_bounds.right_side_bearing = 0; // FIXME
    reply.min_bounds.character_width = 0; // FIXME
    reply.min_bounds.ascent = 0; // FIXME
    reply.min_bounds.descent = 0; // FIXME
    reply.min_bounds.attributes = 0x0000; // FIXME
    reply.max_bounds.left_side_bearing = 0; // FIXME
    reply.max_bounds.right_side_bearing = 0; // FIXME
    reply.max_bounds.character_width = 0; // FIXME
    reply.max_bounds.ascent = 0; // FIXME
    reply.max_bounds.descent = 0; // FIXME
    reply.max_bounds.attributes = 0x0000; // FIXME
    reply.min_char_or_byte2 = 0; // FIXME
    reply.max_char_or_byte2 = 0; // FIXME
    reply.default_char = 0; // FIXME
    reply.draw_direction = 0; // FIXME
    reply.min_byte1 = 0; // FIXME
    reply.max_byte1 = 0; // FIXME
    reply.all_chars_exist = FALSE; // FIXME
    reply.font_ascent = 0; // FIXME
    reply.font_descent = 0; // FIXME
    reply.properties_length = 0; // FIXME
    reply.properties = NULL; // FIXME
    reply.char_infos_length = 0; // FIXME
    reply.char_infos = NULL; // FIXME

    xig_codec_feed_query_font_reply (XIG_CODEC (codec), &reply);
  
    g_free (reply.properties);
    g_free (reply.char_infos);
}

static void 
xig_remote_client_query_text_extents (XigCodec *codec, XigQueryTextExtents *message)
{
    XigQueryTextExtentsReply reply;

    reply.sequence_number = message->sequence_number;
    reply.draw_direction = 0; // FIXME
    reply.font_ascent = 0; // FIXME
    reply.font_descent = 0; // FIXME
    reply.overall_ascent = 0; // FIXME
    reply.overall_descent = 0; // FIXME
    reply.overall_width = 0; // FIXME
    reply.overall_left = 0; // FIXME
    reply.overall_right = 0; // FIXME

    xig_codec_feed_query_text_extents_reply (XIG_CODEC (codec), &reply);
}

static void 
xig_remote_client_list_fonts (XigCodec *codec, XigListFonts *message)
{
    XigListFontsReply reply;

    reply.sequence_number = message->sequence_number;
    reply.names = g_strsplit ("", " ", -1); // FIXME

    xig_codec_feed_list_fonts_reply (XIG_CODEC (codec), &reply);

    g_strfreev (reply.names);
}

static void
xig_remote_client_list_fonts_with_info (XigCodec *codec, XigListFontsWithInfo *message)
{
    // FIXME emit reply
}

static void
xig_remote_client_set_font_path (XigCodec *codec, XigSetFontPath *message)
{
}

static void
xig_remote_client_get_font_path (XigCodec *codec, XigGetFontPath *message)
{
    XigGetFontPathReply reply;

    reply.sequence_number = message->sequence_number;
    reply.path = g_strsplit ("", " ", -1); // FIXME

    xig_codec_feed_get_font_path_reply (XIG_CODEC (codec), &reply);

    g_strfreev (reply.path);
}

static void
xig_remote_client_create_pixmap (XigCodec *codec, XigCreatePixmap *message)
{
    XigRemoteClient *client = XIG_REMOTE_CLIENT (codec);
    XigPixmap *pixmap;
    XigDrawable *drawable;

    /* Check drawable doesn't already exist */
    if (xig_server_get_drawable (client->priv->server, message->pid))
    {
        send_error (client, IDChoice, XIG_REQUEST_CreatePixmap, 0, message->sequence_number, message->pid);
        return;
    }
    drawable = xig_server_get_drawable (client->priv->server, message->drawable);
    if (!drawable)
    {
        send_error (client, Drawable, XIG_REQUEST_CreatePixmap, 0, message->sequence_number, message->drawable);
        return;
    }

    pixmap = _xig_pixmap_new (message->pid, drawable, message->width, message->height);
    _xig_server_add_pixmap (client->priv->server, pixmap);
}

static void 
xig_remote_client_free_pixmap (XigCodec *codec, XigFreePixmap *message)
{
    XigRemoteClient *client = XIG_REMOTE_CLIENT (codec);
    XigPixmap *pixmap;

    pixmap = xig_server_get_pixmap (client->priv->server, message->pixmap);
    if (!pixmap)
    {
        send_error (client, Pixmap, XIG_REQUEST_FreePixmap, 0, message->sequence_number, message->pixmap);
        return;
    }

    _xig_server_remove_pixmap (client->priv->server, pixmap);
}

static void 
xig_remote_client_create_gc (XigCodec *codec, XigCreateGC *message)
{
}

static void 
xig_remote_client_change_gc (XigCodec *codec, XigChangeGC *message)
{
}

static void 
xig_remote_client_copy_gc (XigCodec *codec, XigCopyGC *message)
{
}

static void 
xig_remote_client_set_dashes (XigCodec *codec, XigSetDashes *message)
{
}

static void 
xig_remote_client_set_clip_rectangles (XigCodec *codec, XigSetClipRectangles *message)
{
}

static void 
xig_remote_client_free_gc (XigCodec *codec, XigFreeGC *message)
{
}

static void 
xig_remote_client_clear_area (XigCodec *codec, XigClearArea *message)
{
}

static void 
xig_remote_client_copy_area (XigCodec *codec, XigCopyArea *message)
{
}

static void 
xig_remote_client_copy_plane (XigCodec *codec, XigCopyPlane *message)
{
}

static void
xig_remote_client_poly_point (XigCodec *codec, XigPolyPoint *message)
{
}

static void
xig_remote_client_poly_line (XigCodec *codec, XigPolyLine *message)
{
}

static void
xig_remote_client_poly_segment (XigCodec *codec, XigPolySegment *message)
{
}

static void
xig_remote_client_poly_rectangle (XigCodec *codec, XigPolyRectangle *message)
{
}

static void
xig_remote_client_poly_arc (XigCodec *codec, XigPolyArc *message)
{
}

static void
xig_remote_client_fill_poly (XigCodec *codec, XigFillPoly *message)
{
}

static void
xig_remote_client_poly_fill_rectangle (XigCodec *codec, XigPolyFillRectangle *message)
{
}

static void
xig_remote_client_poly_fill_arc (XigCodec *codec, XigPolyFillArc *message)
{
}

static void 
xig_remote_client_put_image (XigCodec *codec, XigPutImage *message)
{
}

static void
xig_remote_client_get_image (XigCodec *codec, XigGetImage *message)
{
    XigGetImageReply reply;

    reply.sequence_number = message->sequence_number;
    reply.depth = 0; // FIXME
    reply.visual = 0; // FIXME
    reply.data = NULL; // FIXME
    reply.data_length = 0; // FIXME

    xig_codec_feed_get_image_reply (XIG_CODEC (codec), &reply);
}

static void
xig_remote_client_poly_text8 (XigCodec *codec, XigPolyText8 *message)
{
}

static void
xig_remote_client_poly_text16 (XigCodec *codec, XigPolyText16 *message)
{
}

static void
xig_remote_client_image_text8 (XigCodec *codec, XigImageText8 *message)
{
}

static void
xig_remote_client_image_text16 (XigCodec *codec, XigImageText16 *message)
{
}
 
static void
xig_remote_client_create_colormap (XigCodec *codec, XigCreateColormap *message)
{
}

static void
xig_remote_client_free_colormap (XigCodec *codec, XigFreeColormap *message)
{
}

static void
xig_remote_client_copy_colormap_and_free (XigCodec *codec, XigCopyColormapAndFree *message)
{
}

static void
xig_remote_client_install_colormap (XigCodec *codec, XigInstallColormap *message)
{
}

static void
xig_remote_client_uninstall_colormap (XigCodec *codec, XigUninstallColormap *message)
{
}

static void
xig_remote_client_list_installed_colormaps (XigCodec *codec, XigListInstalledColormaps *message)
{
    XigListInstalledColormapsReply reply;

    reply.sequence_number = message->sequence_number;
    reply.cmaps_length = 0;
    reply.cmaps = NULL;

    xig_codec_feed_list_installed_colormaps_reply (XIG_CODEC (codec), &reply);

    g_free (reply.cmaps);
}

static void 
xig_remote_client_alloc_color (XigCodec *codec, XigAllocColor *message)
{
    XigAllocColorReply reply;

    reply.sequence_number = message->sequence_number;
    reply.red = message->red;
    reply.green = message->green;
    reply.blue = message->blue;
    reply.pixel = (message->red >> 8) << 24 | (message->green >> 8) << 8 | (message->blue >> 8);

    xig_codec_feed_alloc_color_reply (XIG_CODEC (codec), &reply);
}

static void 
xig_remote_client_alloc_named_color (XigCodec *codec, XigAllocNamedColor *message)
{
    XigAllocNamedColorReply reply;

    reply.sequence_number = message->sequence_number;
    reply.exact_red = 0; // FIXME
    reply.exact_green = 0; // FIXME
    reply.exact_blue = 0; // FIXME
    reply.visual_red = 0; // FIXME
    reply.visual_green = 0; // FIXME
    reply.visual_blue = 0; // FIXME
    reply.pixel = 0; // FIXME

    xig_codec_feed_alloc_named_color_reply (XIG_CODEC (codec), &reply);
}

static void
xig_remote_client_alloc_color_cells (XigCodec *codec, XigAllocColorCells *message)
{
    XigAllocColorCellsReply reply;

    reply.sequence_number = message->sequence_number;
    reply.pixels_length = 0; // FIXME
    reply.pixels = NULL; // FIXME
    reply.masks_length = 0; // FIXME
    reply.masks = NULL; // FIXME

    xig_codec_feed_alloc_color_cells_reply (XIG_CODEC (codec), &reply);
  
    g_free (reply.pixels);
    g_free (reply.masks);
}

static void
xig_remote_client_alloc_color_planes (XigCodec *codec, XigAllocColorPlanes *message)
{
    XigAllocColorPlanesReply reply;

    reply.sequence_number = message->sequence_number;
    reply.pixels_length = 0; // FIXME
    reply.pixels = NULL; // FIXME
    reply.red_mask = 0x00000000; // FIXME
    reply.green_mask = 0x00000000; // FIXME
    reply.blue_mask = 0x00000000; // FIXME

    xig_codec_feed_alloc_color_planes_reply (XIG_CODEC (codec), &reply);
  
    g_free (reply.pixels);
}

static void 
xig_remote_client_free_colors (XigCodec *codec, XigFreeColors *message)
{
}

static void 
xig_remote_client_store_colors (XigCodec *codec, XigStoreColors *message)
{
}

static void 
xig_remote_client_store_named_color (XigCodec *codec, XigStoreNamedColor *message)
{
}

static void 
xig_remote_client_query_colors (XigCodec *codec, XigQueryColors *message)
{
    XigQueryColorsReply reply;

    reply.sequence_number = message->sequence_number;
    reply.colors_length = 0; // FIXME
    reply.colors = NULL; // FIXME

    xig_codec_feed_query_colors_reply (XIG_CODEC (codec), &reply);
  
    g_free (reply.colors);
}

static void 
xig_remote_client_lookup_color (XigCodec *codec, XigLookupColor *message)
{
    XigLookupColorReply reply;

    reply.sequence_number = message->sequence_number;
    reply.exact_red = 0; // FIXME
    reply.exact_green = 0; // FIXME
    reply.exact_blue = 0; // FIXME
    reply.visual_red = 0; // FIXME
    reply.visual_green = 0; // FIXME
    reply.visual_blue = 0; // FIXME

    xig_codec_feed_lookup_color_reply (XIG_CODEC (codec), &reply);
}

static void 
xig_remote_client_create_cursor (XigCodec *codec, XigCreateCursor *message)
{
}

static void 
xig_remote_client_create_glyph_cursor (XigCodec *codec, XigCreateGlyphCursor *message)
{
}

static void 
xig_remote_client_free_cursor (XigCodec *codec, XigFreeCursor *message)
{
}

static void 
xig_remote_client_recolor_cursor (XigCodec *codec, XigRecolorCursor *message)
{
}

static void 
xig_remote_client_query_best_size (XigCodec *codec, XigQueryBestSize *message)
{
    XigQueryBestSizeReply reply;

    reply.sequence_number = message->sequence_number;
    reply.width = message->width;
    reply.height = message->height;

    xig_codec_feed_query_best_size_reply (XIG_CODEC (codec), &reply);
}

static void 
xig_remote_client_query_extension (XigCodec *codec, XigQueryExtension *message)
{
    XigQueryExtensionReply reply;

    reply.sequence_number = message->sequence_number;
    reply.major_opcode = 0;
    reply.first_event = 0;
    reply.first_error = 0;

    if (strcmp (message->name, "BIG-REQUESTS") == 0)
        reply.major_opcode = XIG_REQUEST_BIGREQ;
    else if (strcmp (message->name, "SYNC") == 0)
        reply.major_opcode = XIG_REQUEST_SYNC;
    else if (strcmp (message->name, "RANDR") == 0)
        reply.major_opcode = XIG_REQUEST_RANDR;
    else if (strcmp (message->name, "XINERAMA") == 0)
        reply.major_opcode = XIG_REQUEST_XINERAMA;
    else if (strcmp (message->name, "XKEYBOARD") == 0)
        reply.major_opcode = XIG_REQUEST_XKEYBOARD;
    //else if (strcmp (message->name, "RENDER") == 0)
    //    reply.major_opcode = XIG_REQUEST_RENDER;
    //else if (strcmp (message->name, "SHAPE") == 0)
    //    reply.major_opcode = ?;
    //else if (strcmp (message->name, "DAMAGE") == 0)
    //    reply.major_opcode = ?;
    //else if (strcmp (message->name, "Composite") == 0)
    //    reply.major_opcode = ?;
    //else if (strcmp (message->name, "GLX") == 0)
    //    reply.major_opcode = ?;
    //else if (strcmp (message->name, "Generic Event Extension") == 0)
    //    reply.major_opcode = ?;
    //else if (strcmp (message->name, "XInputExtension") == 0)
    //    reply.major_opcode = ?;
    //else if (strcmp (message->name, "XFIXES") == 0)
    //    reply.major_opcode = ?;
    else
        ;//g_warning ("Unknown extension %s", message->name);

    reply.present = reply.major_opcode != 0;

    xig_codec_feed_query_extension_reply (XIG_CODEC (codec), &reply);
}

static void 
xig_remote_client_list_extensions (XigCodec *codec, XigListExtensions *message)
{
    XigListExtensionsReply reply;

    reply.sequence_number = message->sequence_number;
    reply.names = g_strsplit ("", " ", -1);

    xig_codec_feed_list_extensions_reply (XIG_CODEC (codec), &reply);

    g_strfreev (reply.names);
}

static void 
xig_remote_client_change_keyboard_mapping (XigCodec *codec, XigChangeKeyboardMapping *message)
{
}

static void 
xig_remote_client_get_keyboard_mapping (XigCodec *codec, XigGetKeyboardMapping *message)
{
    XigGetKeyboardMappingReply reply;
    int i, j;
 
    // FIXME: Throw Value error if first-keycode and count are outside allowed range

    reply.sequence_number = message->sequence_number;
    reply.keysyms_per_keycode = 1; // FIXME
    reply.keysyms_length = message->count * reply.keysyms_per_keycode;
    reply.keysyms = calloc (reply.keysyms_length, sizeof (guint32));
    for (i = 0; i < message->count; i++)
        for (j = 0; j < reply.keysyms_per_keycode; j++)
            reply.keysyms[i * reply.keysyms_per_keycode + j] = 0;

    xig_codec_feed_get_keyboard_mapping_reply (XIG_CODEC (codec), &reply);

    g_free (reply.keysyms);
}

static void
xig_remote_client_change_keyboard_control (XigCodec *codec, XigChangeKeyboardControl *message)
{
}

static void
xig_remote_client_get_keyboard_control (XigCodec *codec, XigGetKeyboardControl *message)
{
    XigGetKeyboardControlReply reply;
    int i;

    reply.sequence_number = message->sequence_number;
    reply.global_auto_repeat = 0; // FIXME
    reply.led_mask = 0x00000000; // FIXME
    reply.key_click_percent = 0; // FIXME
    reply.bell_percent = 0; // FIXME
    reply.bell_pitch = 0; // FIXME
    reply.bell_duration = 0; // FIXME
    for (i = 0; i < 32; i++)
        reply.auto_repeats[i] = 0;

    xig_codec_feed_get_keyboard_control_reply (XIG_CODEC (codec), &reply);
}


static void 
xig_remote_client_bell (XigCodec *codec, XigBell *message)
{
}

static void 
xig_remote_client_change_pointer_control (XigCodec *codec, XigChangePointerControl *message)
{
}

static void 
xig_remote_client_get_pointer_control (XigCodec *codec, XigGetPointerControl *message)
{
    XigGetPointerControlReply reply;
  
    reply.sequence_number = message->sequence_number;

    xig_codec_feed_get_pointer_control_reply (XIG_CODEC (codec), &reply);
}

static void 
xig_remote_client_set_screen_saver (XigCodec *codec, XigSetScreenSaver *message)
{
}

static void 
xig_remote_client_get_screen_saver (XigCodec *codec, XigGetScreenSaver *message)
{
    XigGetScreenSaverReply reply;

    reply.sequence_number = message->sequence_number;
    reply.timeout = 0; // FIXME
    reply.interval = 0; // FIXME
    reply.prefer_blanking = FALSE; // FIXME
    reply.allow_exposures = FALSE; // FIXME

    xig_codec_feed_get_screen_saver_reply (XIG_CODEC (codec), &reply);
}

static void
xig_remote_client_change_hosts (XigCodec *codec, XigChangeHosts *message)
{
}

static void
xig_remote_client_list_hosts (XigCodec *codec, XigListHosts *message)
{
    XigListHostsReply reply;
  
    reply.sequence_number = message->sequence_number;
    reply.mode = 0; // FIXME
    reply.hosts_length = 0;
    reply.hosts = NULL;

    xig_codec_feed_list_hosts_reply (XIG_CODEC (codec), &reply);

    g_free (reply.hosts);
}

static void
xig_remote_client_set_access_control (XigCodec *codec, XigSetAccessControl *message)
{
}

static void 
xig_remote_client_set_close_down_mode (XigCodec *codec, XigSetCloseDownMode *message)
{
    XigRemoteClient *client = XIG_REMOTE_CLIENT (codec);
    client->priv->close_down_mode = message->mode;
}

static void 
xig_remote_client_kill_client (XigCodec *codec, XigKillClient *message)
{
}

static void 
xig_remote_client_rotate_properties (XigCodec *codec, XigRotateProperties *message)
{
}

static void 
xig_remote_client_force_screen_saver (XigCodec *codec, XigForceScreenSaver *message)
{
}

static void
xig_remote_client_set_pointer_mapping (XigCodec *codec, XigSetPointerMapping *message)
{
    XigSetPointerMappingReply reply;

    reply.sequence_number = message->sequence_number;
    reply.status = 0; // FIXME

    xig_codec_feed_set_pointer_mapping_reply (XIG_CODEC (codec), &reply);
}

static void
xig_remote_client_get_pointer_mapping (XigCodec *codec, XigGetPointerMapping *message)
{
    XigGetPointerMappingReply reply;

    reply.sequence_number = message->sequence_number;

    xig_codec_feed_get_pointer_mapping_reply (XIG_CODEC (codec), &reply);
}

static void
xig_remote_client_set_modifier_mapping (XigCodec *codec, XigSetModifierMapping *message)
{
    XigSetModifierMappingReply reply;

    reply.sequence_number = message->sequence_number;
    reply.status = 0; // FIXME

    xig_codec_feed_set_modifier_mapping_reply (XIG_CODEC (codec), &reply);
}

static void
xig_remote_client_get_modifier_mapping (XigCodec *codec, XigGetModifierMapping *message)
{
    XigGetModifierMappingReply reply;
    int i, j;

    reply.sequence_number = message->sequence_number;
    reply.keycodes_per_modifier = 1; // FIXME
    reply.keycodes_length = 8 * reply.keycodes_per_modifier;
    reply.keycodes = g_malloc (sizeof (guint32) * reply.keycodes_length);
    for (i = 0; i < 8; i++)
        for (j = 0; j < reply.keycodes_per_modifier; j++)
            reply.keycodes[i * reply.keycodes_per_modifier + j] = 0;

    xig_codec_feed_get_modifier_mapping_reply (XIG_CODEC (codec), &reply);

    g_free (reply.keycodes);
}

static void
xig_remote_client_no_operation (XigCodec *codec, XigNoOperation *message)
{
}

static void 
xig_remote_client_BIGREQ_enable (XigCodec *codec, XigBIGREQEnable *message)
{
    XigBIGREQEnableReply reply;

    xig_server_codec_set_BIGREQ_enabled (XIG_SERVER_CODEC (codec), TRUE);

    reply.sequence_number = message->sequence_number;
    reply.maximum_request_length = MAXIMUM_REQUEST_LENGTH;

    xig_codec_feed_BIGREQ_enable_reply (XIG_CODEC (codec), &reply);
}

static void
xig_remote_client_SYNC_initialize (XigCodec *codec, XigSYNCInitialize *message)
{
    XigSYNCInitializeReply reply;

    reply.sequence_number = message->sequence_number;
    reply.major_version = message->desired_major_version;
    reply.minor_version = message->desired_minor_version;
  
    xig_codec_feed_SYNC_initialize_reply (XIG_CODEC (codec), &reply);
}

static void
xig_remote_client_RANDR_query_version (XigCodec *codec, XigRANDRQueryVersion *message)
{
    XigRANDRQueryVersionReply reply;

    reply.sequence_number = message->sequence_number;
    reply.major_version = message->major_version;
    reply.minor_version = message->minor_version;

    xig_codec_feed_RANDR_query_version_reply (XIG_CODEC (codec), &reply);
}

static void
xig_remote_client_RANDR_select_input (XigCodec *codec, XigRANDRSelectInput *message)
{
}

static void
xig_remote_client_RANDR_get_screen_resources (XigCodec *codec, XigRANDRGetScreenResources *message)
{
    XigRANDRGetScreenResourcesReply reply;

    reply.sequence_number = message->sequence_number;
    reply.timestamp = 0;
    reply.config_timestamp = 0;
    reply.crtcs = NULL;
    reply.crtcs_length = 0;
    reply.outputs = NULL;
    reply.outputs_length = 0;
    reply.modes = NULL;
    reply.modes_length = 0;

    xig_codec_feed_RANDR_get_screen_resources_reply (XIG_CODEC (codec), &reply);
}

static void
xig_remote_client_RANDR_set_output_primary (XigCodec *codec, XigRANDRSetOutputPrimary *message)
{
}

static void
xig_remote_client_RANDR_get_output_primary (XigCodec *codec, XigRANDRGetOutputPrimary *message)
{
    XigRANDRGetOutputPrimaryReply reply;

    reply.sequence_number = message->sequence_number;
    reply.output = 0; // FIXME

    xig_codec_feed_RANDR_get_output_primary_reply (XIG_CODEC (codec), &reply);
}

static void
xig_remote_client_RANDR_get_screen_resources_current (XigCodec *codec, XigRANDRGetScreenResourcesCurrent *message)
{
    XigRANDRGetScreenResourcesCurrentReply reply;

    reply.sequence_number = message->sequence_number;
    reply.timestamp = 0;
    reply.config_timestamp = 0;
    reply.crtcs = NULL;
    reply.crtcs_length = 0;
    reply.outputs = NULL;
    reply.outputs_length = 0;
    reply.modes = NULL;
    reply.modes_length = 0;

    xig_codec_feed_RANDR_get_screen_resources_current_reply (XIG_CODEC (codec), &reply);
}

static void 
xig_remote_client_XINERAMA_is_active (XigCodec *codec, XigXINERAMAIsActive *message)
{
    XigXINERAMAIsActiveReply reply;

    reply.sequence_number = message->sequence_number;
    reply.is_active = FALSE;

    xig_codec_feed_XINERAMA_is_active_reply (XIG_CODEC (codec), &reply);
}

static void 
xig_remote_client_XINERAMA_query_screens (XigCodec *codec, XigXINERAMAQueryScreens *message)
{
    XigRemoteClient *client = XIG_REMOTE_CLIENT (codec);
    XigXINERAMAQueryScreensReply reply;
    XigScreen *screen;

    screen = xig_server_get_screen (client->priv->server, 0);

    reply.sequence_number = message->sequence_number;
    reply.info_length = 1; // FIXME
    reply.info = calloc (reply.info_length, sizeof (XigXINERAMAScreenInfo));
    reply.info[0].x = 0;
    reply.info[0].y = 0;
    reply.info[0].width = xig_screen_get_width_in_pixels (screen);
    reply.info[0].height = xig_screen_get_height_in_pixels (screen);

    xig_codec_feed_XINERAMA_query_screens_reply (XIG_CODEC (codec), &reply);
  
    g_free (reply.info);
}

static void
xig_remote_client_XKB_use_extension (XigCodec *codec, XigXKBUseExtension *message)
{
    XigXKBUseExtensionReply reply;

    reply.sequence_number = message->sequence_number;
    reply.supported = TRUE; // FIXME
    reply.server_major = message->wanted_major; // FIXME
    reply.server_minor = message->wanted_minor; // FIXME

    xig_codec_feed_XKB_use_extension_reply (XIG_CODEC (codec), &reply);
}

static void
xig_remote_client_XKB_select_events (XigCodec *codec, XigXKBSelectEvents *message)
{
}

static void
xig_remote_client_XKB_get_state (XigCodec *codec, XigXKBGetState *message)
{
    XigXKBGetStateReply reply;

    reply.sequence_number = message->sequence_number;
    reply.device_id = 0; // FIXME
    reply.mods = 0; // FIXME
    reply.base_mods = 0; // FIXME
    reply.latched_mods = 0; // FIXME
    reply.locked_mods = 0; // FIXME
    reply.group = 0; // FIXME
    reply.locked_group = 0; // FIXME
    reply.base_group = 0; // FIXME
    reply.latched_group = 0; // FIXME
    reply.compat_state = 0; // FIXME
    reply.grab_mods = 0; // FIXME
    reply.compat_grab_mods = 0; // FIXME
    reply.lookup_mods = 0; // FIXME
    reply.compat_lookup_mods = 0; // FIXME
    reply.ptr_btn_state = 0; // FIXME

    xig_codec_feed_XKB_get_state_reply (XIG_CODEC (codec), &reply);
}

static void
xig_remote_client_XKB_set_controls (XigCodec *codec, XigXKBSetControls *message)
{
}

static void
xig_remote_client_XKB_get_map (XigCodec *codec, XigXKBGetMap *message)
{
    XigXKBGetMapReply reply;
 
    reply.sequence_number = message->sequence_number;
    reply.device_id = 0; // FIXME

    xig_codec_feed_XKB_get_map_reply (XIG_CODEC (codec), &reply);
}

static void
xig_remote_client_XKB_get_names (XigCodec *codec, XigXKBGetNames *message)
{
    XigXKBGetNamesReply reply;

    reply.sequence_number = message->sequence_number;
    reply.device_id = 0;

    xig_codec_feed_XKB_get_names_reply (XIG_CODEC (codec), &reply);
}

static void
xig_remote_client_XKB_per_client_flags (XigCodec *codec, XigXKBPerClientFlags *message)
{
    XigXKBPerClientFlagsReply reply;

    reply.sequence_number = message->sequence_number;
    reply.device_id = 0; // FIXME
    reply.supported = 0x00000000; // FIXME
    reply.value = 0x00000000; // FIXME
    reply.auto_ctrls = 0x00000000; // FIXME
    reply.auto_ctrls_values = 0x00000000; // FIXME

    xig_codec_feed_XKB_per_client_flags_reply (XIG_CODEC (codec), &reply);
}

static void 
xig_remote_client_RENDER_query_version (XigCodec *codec, XigRENDERQueryVersion *message)
{
    XigRENDERQueryVersionReply reply;
  
    reply.sequence_number = message->sequence_number;
    reply.major_version = 1;
    reply.minor_version = 0;

    xig_codec_feed_RENDER_query_version_reply (XIG_CODEC (codec), &reply);
}

static void 
xig_remote_client_RENDER_query_pict_formats (XigCodec *codec, XigRENDERQueryPictFormats *message)
{
    XigRENDERQueryPictFormatsReply reply;
  
    reply.sequence_number = message->sequence_number;
    // FIXME

    xig_codec_feed_RENDER_query_pict_formats_reply (XIG_CODEC (codec), &reply);
}

XigRemoteClient *
_xig_remote_client_new (XigServer *server, GSocket *socket)
{
    XigRemoteClient *client;

    client = g_object_new (xig_remote_client_get_type (), NULL);  
    client->priv->server = server;
    client->priv->socket = g_object_ref (socket);
    client->priv->channel = g_io_channel_unix_new (g_socket_get_fd (socket));
    g_io_add_watch (client->priv->channel, G_IO_IN, socket_data_cb, client);

    return client;
}

XigServer *
xig_remote_client_get_server (XigRemoteClient *client)
{
    return client->priv->server;
}

static void
xig_remote_client_init (XigRemoteClient *client)
{
    client->priv = G_TYPE_INSTANCE_GET_PRIVATE (client, xig_remote_client_get_type (), XigRemoteClientPrivate);
    client->priv->sequence_number = 0;
    client->priv->window_events = g_hash_table_new_full (g_direct_hash, g_direct_equal, g_object_unref, NULL);
    // FIXME
    client->priv->resource_id_base = 0x00a00000;
    client->priv->resource_id_mask = 0x001fffff;
}

static void
xig_remote_client_finalize (GObject *object)
{
    XigRemoteClient *client = XIG_REMOTE_CLIENT (object);

    g_object_unref (client->priv->socket);
    g_io_channel_unref (client->priv->channel);
    g_hash_table_unref (client->priv->window_events);

    G_OBJECT_CLASS (xig_remote_client_parent_class)->finalize (object);
}

static void
xig_remote_client_class_init (XigRemoteClientClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);
    XigCodecClass *codec_class = XIG_CODEC_CLASS (klass);

    object_class->finalize = xig_remote_client_finalize;

    g_type_class_add_private (klass, sizeof (XigRemoteClientPrivate));

    codec_class->write = xig_remote_client_write;
    codec_class->connect = xig_remote_client_connect;
    codec_class->create_window = xig_remote_client_create_window;
    codec_class->change_window_attributes = xig_remote_client_change_window_attributes;
    codec_class->get_window_attributes = xig_remote_client_get_window_attributes;
    codec_class->destroy_window = xig_remote_client_destroy_window;
    codec_class->destroy_subwindows = xig_remote_client_destroy_subwindows;
    codec_class->change_save_set = xig_remote_client_change_save_set;
    codec_class->reparent_window = xig_remote_client_reparent_window;
    codec_class->map_window = xig_remote_client_map_window;
    codec_class->map_subwindows = xig_remote_client_map_subwindows;
    codec_class->unmap_window = xig_remote_client_unmap_window;
    codec_class->unmap_subwindows = xig_remote_client_unmap_subwindows;
    codec_class->configure_window = xig_remote_client_configure_window;
    codec_class->circulate_window = xig_remote_client_circulate_window;
    codec_class->get_geometry = xig_remote_client_get_geometry;
    codec_class->query_tree = xig_remote_client_query_tree;
    codec_class->intern_atom = xig_remote_client_intern_atom;
    codec_class->get_atom_name = xig_remote_client_get_atom_name;
    codec_class->change_property = xig_remote_client_change_property;
    codec_class->delete_property = xig_remote_client_delete_property;
    codec_class->get_property = xig_remote_client_get_property;
    codec_class->list_properties = xig_remote_client_list_properties;
    codec_class->set_selection_owner = xig_remote_client_set_selection_owner;
    codec_class->get_selection_owner = xig_remote_client_get_selection_owner;
    codec_class->convert_selection = xig_remote_client_convert_selection;
    codec_class->send_event = xig_remote_client_send_event;
    codec_class->grab_pointer = xig_remote_client_grab_pointer;
    codec_class->ungrab_pointer = xig_remote_client_ungrab_pointer;
    codec_class->grab_button = xig_remote_client_grab_button;
    codec_class->ungrab_button = xig_remote_client_ungrab_button;
    codec_class->change_active_pointer_grab = xig_remote_client_change_active_pointer_grab; 
    codec_class->grab_keyboard = xig_remote_client_grab_keyboard;
    codec_class->ungrab_keyboard = xig_remote_client_ungrab_keyboard;
    codec_class->grab_key = xig_remote_client_grab_key;
    codec_class->ungrab_key = xig_remote_client_ungrab_key;
    codec_class->allow_events = xig_remote_client_allow_events;
    codec_class->grab_server = xig_remote_client_grab_server;
    codec_class->ungrab_server = xig_remote_client_ungrab_server;
    codec_class->query_pointer = xig_remote_client_query_pointer;
    codec_class->get_motion_events = xig_remote_client_get_motion_events;
    codec_class->translate_coordinates = xig_remote_client_translate_coordinates;
    codec_class->warp_pointer = xig_remote_client_warp_pointer;
    codec_class->set_input_focus = xig_remote_client_set_input_focus;
    codec_class->get_input_focus = xig_remote_client_get_input_focus;
    codec_class->query_keymap = xig_remote_client_query_keymap;
    codec_class->open_font = xig_remote_client_open_font;
    codec_class->close_font = xig_remote_client_close_font;
    codec_class->query_font = xig_remote_client_query_font;
    codec_class->query_text_extents = xig_remote_client_query_text_extents;
    codec_class->list_fonts = xig_remote_client_list_fonts;
    codec_class->list_fonts_with_info = xig_remote_client_list_fonts_with_info;
    codec_class->set_font_path = xig_remote_client_set_font_path;
    codec_class->get_font_path = xig_remote_client_get_font_path;
    codec_class->create_pixmap = xig_remote_client_create_pixmap;
    codec_class->free_pixmap = xig_remote_client_free_pixmap;
    codec_class->create_gc = xig_remote_client_create_gc;
    codec_class->change_gc = xig_remote_client_change_gc;
    codec_class->copy_gc = xig_remote_client_copy_gc;
    codec_class->set_dashes = xig_remote_client_set_dashes;
    codec_class->set_clip_rectangles = xig_remote_client_set_clip_rectangles;
    codec_class->free_gc = xig_remote_client_free_gc;
    codec_class->clear_area = xig_remote_client_clear_area;
    codec_class->copy_area = xig_remote_client_copy_area;
    codec_class->copy_plane = xig_remote_client_copy_plane;
    codec_class->poly_point = xig_remote_client_poly_point;
    codec_class->poly_line = xig_remote_client_poly_line;
    codec_class->poly_segment = xig_remote_client_poly_segment;
    codec_class->poly_rectangle = xig_remote_client_poly_rectangle;
    codec_class->poly_arc = xig_remote_client_poly_arc;
    codec_class->fill_poly = xig_remote_client_fill_poly;
    codec_class->poly_fill_rectangle = xig_remote_client_poly_fill_rectangle;
    codec_class->poly_fill_arc = xig_remote_client_poly_fill_arc;
    codec_class->put_image = xig_remote_client_put_image;
    codec_class->get_image = xig_remote_client_get_image;
    codec_class->poly_text8 = xig_remote_client_poly_text8;
    codec_class->poly_text16 = xig_remote_client_poly_text16;
    codec_class->image_text8 = xig_remote_client_image_text8;
    codec_class->image_text16 = xig_remote_client_image_text16;
    codec_class->create_colormap = xig_remote_client_create_colormap;
    codec_class->free_colormap = xig_remote_client_free_colormap;
    codec_class->copy_colormap_and_free = xig_remote_client_copy_colormap_and_free;
    codec_class->install_colormap = xig_remote_client_install_colormap;
    codec_class->uninstall_colormap = xig_remote_client_uninstall_colormap;
    codec_class->list_installed_colormaps = xig_remote_client_list_installed_colormaps;
    codec_class->alloc_color = xig_remote_client_alloc_color;
    codec_class->alloc_named_color = xig_remote_client_alloc_named_color;
    codec_class->alloc_color_cells = xig_remote_client_alloc_color_cells;
    codec_class->alloc_color_planes = xig_remote_client_alloc_color_planes;
    codec_class->free_colors = xig_remote_client_free_colors;
    codec_class->store_colors = xig_remote_client_store_colors;
    codec_class->store_named_color = xig_remote_client_store_named_color;
    codec_class->query_colors = xig_remote_client_query_colors;
    codec_class->lookup_color = xig_remote_client_lookup_color;
    codec_class->create_cursor = xig_remote_client_create_cursor;
    codec_class->create_glyph_cursor = xig_remote_client_create_glyph_cursor;
    codec_class->free_cursor = xig_remote_client_free_cursor;
    codec_class->recolor_cursor = xig_remote_client_recolor_cursor;
    codec_class->query_best_size = xig_remote_client_query_best_size;
    codec_class->query_extension = xig_remote_client_query_extension;
    codec_class->list_extensions = xig_remote_client_list_extensions;
    codec_class->change_keyboard_mapping = xig_remote_client_change_keyboard_mapping;
    codec_class->get_keyboard_mapping = xig_remote_client_get_keyboard_mapping;
    codec_class->change_keyboard_control = xig_remote_client_change_keyboard_control;
    codec_class->get_keyboard_control = xig_remote_client_get_keyboard_control;
    codec_class->bell = xig_remote_client_bell;
    codec_class->change_pointer_control = xig_remote_client_change_pointer_control;
    codec_class->get_pointer_control = xig_remote_client_get_pointer_control;
    codec_class->set_screen_saver = xig_remote_client_set_screen_saver;
    codec_class->get_screen_saver = xig_remote_client_get_screen_saver;
    codec_class->change_hosts = xig_remote_client_change_hosts;
    codec_class->list_hosts = xig_remote_client_list_hosts;
    codec_class->set_access_control = xig_remote_client_set_access_control;
    codec_class->set_close_down_mode = xig_remote_client_set_close_down_mode;
    codec_class->kill_client = xig_remote_client_kill_client;
    codec_class->rotate_properties = xig_remote_client_rotate_properties;
    codec_class->force_screen_saver = xig_remote_client_force_screen_saver;
    codec_class->set_pointer_mapping = xig_remote_client_set_pointer_mapping;
    codec_class->get_pointer_mapping = xig_remote_client_get_pointer_mapping;
    codec_class->set_modifier_mapping = xig_remote_client_set_modifier_mapping;
    codec_class->get_modifier_mapping = xig_remote_client_get_modifier_mapping;
    codec_class->no_operation = xig_remote_client_no_operation;
    codec_class->BIGREQ_enable = xig_remote_client_BIGREQ_enable;
    codec_class->SYNC_initialize = xig_remote_client_SYNC_initialize;
    codec_class->RANDR_query_version = xig_remote_client_RANDR_query_version;
    codec_class->RANDR_select_input = xig_remote_client_RANDR_select_input;
    codec_class->RANDR_get_screen_resources = xig_remote_client_RANDR_get_screen_resources;
    codec_class->RANDR_set_output_primary = xig_remote_client_RANDR_set_output_primary;
    codec_class->RANDR_get_output_primary = xig_remote_client_RANDR_get_output_primary;
    codec_class->RANDR_get_screen_resources_current = xig_remote_client_RANDR_get_screen_resources_current;
    codec_class->XINERAMA_is_active = xig_remote_client_XINERAMA_is_active;
    codec_class->XINERAMA_query_screens = xig_remote_client_XINERAMA_query_screens;
    codec_class->XKB_use_extension = xig_remote_client_XKB_use_extension;
    codec_class->XKB_select_events = xig_remote_client_XKB_select_events;
    codec_class->XKB_get_state = xig_remote_client_XKB_get_state;
    codec_class->XKB_set_controls = xig_remote_client_XKB_set_controls;
    codec_class->XKB_get_map = xig_remote_client_XKB_get_map;
    codec_class->XKB_get_names = xig_remote_client_XKB_get_names;
    codec_class->XKB_per_client_flags = xig_remote_client_XKB_per_client_flags;
    codec_class->RENDER_query_version = xig_remote_client_RENDER_query_version;
    codec_class->RENDER_query_pict_formats = xig_remote_client_RENDER_query_pict_formats;

    signals[DISCONNECTED] =
        g_signal_new ("disconnected",
                      G_TYPE_FROM_CLASS (klass),
                      G_SIGNAL_RUN_LAST,
                      G_STRUCT_OFFSET (XigRemoteClientClass, disconnected),
                      NULL, NULL,
                      g_cclosure_marshal_VOID__VOID,
                      G_TYPE_NONE, 0);
}
