/*
 * Copyright (C) 2011 Canonical Ltd.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation. See http://www.gnu.org/copyleft/lgpl.html the
 * full text of the license.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <gio/gio.h>
#include <gio/gunixsocketaddress.h>

#include "xig-server-private.h"
#include "xig-screen-private.h"
#include "xig-window-private.h"
#include "xig-drawable-private.h"
#include "xig-pixmap-format-private.h"
#include "xig-remote-client-private.h"

G_DEFINE_TYPE (XigServer, xig_server, G_TYPE_OBJECT);

struct XigServerPrivate
{
    /* Name of server vendor */
    gchar *vendor;

    /* The display number for this server */
    guint display_number;

    /* FIXME: Something about pointer motion */
    guint32 motion_buffer_size;

    /* Byte and bit order of images */
    guint8 image_byte_order;
    guint8 bitmap_format_bit_order;

    /* Allowed keycodes */
    guint8 min_keycode;
    guint8 max_keycode;

    /* Pixmap formats supported */
    GList *pixmap_formats;

    /* Screens on this server */
    GList *screens;

    /* Unix socket */
    gboolean listen_unix;
    gchar *socket_path;
    GSocket *unix_socket;
    GIOChannel *unix_channel;

    /* TCP/IP socket */
    gboolean listen_tcp;
    gint tcp_port;
    GSocket *tcp_socket;
    GIOChannel *tcp_channel;

    /* Clients connected to this server */
    GList *clients;

    /* Drawables */
    GHashTable *drawables;
    guint32 last_drawable_id;

    /* Window with focus */
    XigWindow *focus_window;
    guint8 focus_revert_to;

    /* Last atom ID allocated */
    guint32 last_atom_id;

    /* Atom table */
    GHashTable *atoms_by_name;
    GHashTable *atoms_by_id;

    /* Selection owners */
    GHashTable *selection_owners;
};

#define MAXIMUM_REQUEST_LENGTH 65535

enum {
    CLIENT_CONNECTED,
    CLIENT_DISCONNECTED,
    LAST_SIGNAL
};
static guint signals[LAST_SIGNAL] = { 0 };

XigServer *
xig_server_new (const gchar *vendor, guint display_number)
{
    XigServer *server = g_object_new (xig_server_get_type (), NULL);
    g_free (server->priv->vendor);
    server->priv->vendor = g_strdup (vendor);
    server->priv->display_number = display_number;
    server->priv->tcp_port = 6000 + display_number;
    return server;
}

const gchar *
xig_server_get_vendor (XigServer *server)
{
    return server->priv->vendor;
}

guint
xig_server_get_display_number (XigServer *server)
{
    return server->priv->display_number;
}

XigScreen *
xig_server_add_screen (XigServer *server, guint32 white_pixel, guint32 black_pixel, guint16 width_in_pixels, guint16 height_in_pixels, guint16 width_in_millimeters, guint16 height_in_millimeters)
{
    XigScreen *screen;

    screen = _xig_screen_new (server, white_pixel, black_pixel, width_in_pixels, height_in_pixels, width_in_millimeters, height_in_millimeters);
    server->priv->screens = g_list_append (server->priv->screens, screen);

    return screen;
}

XigPixmapFormat *
xig_server_add_pixmap_format (XigServer *server, guint8 depth, guint8 bits_per_pixel, guint8 scanline_pad)
{
    XigPixmapFormat *format;
  
    format = _xig_pixmap_format_new (depth, bits_per_pixel, scanline_pad);
    server->priv->pixmap_formats = g_list_append (server->priv->pixmap_formats, format);

    return format;
}

void
xig_server_set_listen_unix (XigServer *server, gboolean listen_unix)
{
    server->priv->listen_unix = listen_unix;
}

void
xig_server_set_listen_tcp (XigServer *server, gboolean listen_tcp)
{
    server->priv->listen_tcp = listen_tcp;
}

static void
xig_remote_client_disconnected_cb (XigRemoteClient *client, XigServer *server)
{
    g_signal_handlers_disconnect_matched (client, G_SIGNAL_MATCH_DATA, 0, 0, NULL, NULL, server);
    server->priv->clients = g_list_remove (server->priv->clients, client);
    g_signal_emit (server, signals[CLIENT_DISCONNECTED], 0, client);
    g_object_unref (client);
}

static gboolean
socket_connect_cb (GIOChannel *channel, GIOCondition condition, gpointer data)
{
    XigServer *server = data;
    GSocket *data_socket;
    XigRemoteClient *client;
    GError *error = NULL;

    if (channel == server->priv->unix_channel)
        data_socket = g_socket_accept (server->priv->unix_socket, NULL, &error);
    else
        data_socket = g_socket_accept (server->priv->tcp_socket, NULL, &error);
    if (error)
        g_warning ("Error accepting connection: %s", strerror (errno));
    g_clear_error (&error);
    if (!data_socket)
        return FALSE;

    client = _xig_remote_client_new (server, data_socket);
    g_signal_connect (client, "disconnected", G_CALLBACK (xig_remote_client_disconnected_cb), server);
    server->priv->clients = g_list_append (server->priv->clients, client);
    g_signal_emit (server, signals[CLIENT_CONNECTED], 0, client);

    g_object_unref (data_socket);

    return TRUE;
}

void
xig_server_set_motion_buffer_size (XigServer *server, guint32 size)
{
    server->priv->motion_buffer_size = size;
}

guint32
xig_server_get_motion_buffer_size (XigServer *server)
{
    return server->priv->motion_buffer_size;
}

guint8
xig_server_get_min_keycode (XigServer *server)
{
    return server->priv->min_keycode;
}

guint8
xig_server_get_max_keycode (XigServer *server)
{
    return server->priv->max_keycode;
}

XigServerImageByteOrder
xig_server_get_image_byte_order (XigServer *server)
{
    return server->priv->image_byte_order;
}

XigServerBitmapFormatBitOrder
xig_server_get_bitmap_format_bit_order (XigServer *server)
{
    return server->priv->bitmap_format_bit_order;
}

GList *
xig_server_get_pixmap_formats (XigServer *server)
{
    return server->priv->pixmap_formats;
}

GList *
xig_server_get_screens (XigServer *server)
{
    return server->priv->screens;
}

XigScreen *
xig_server_get_screen (XigServer *server, int number)
{
    return g_list_nth_data (server->priv->screens, number);
}

XigWindow *
xig_server_get_root (XigServer *server, int screen)
{
    XigScreen *s = g_list_nth_data (server->priv->screens, screen);
    if (!s)
        return NULL;
    return xig_screen_get_root (s);
}

guint32
_xig_server_get_drawable_id (XigServer *server)
{
    return ++server->priv->last_drawable_id;
}

void
_xig_server_add_window (XigServer *server, XigWindow *window)
{
    g_hash_table_insert (server->priv->drawables, GINT_TO_POINTER (xig_drawable_get_id (XIG_DRAWABLE (window))), g_object_ref (window));
}

void
_xig_server_set_selection_owner (XigServer *server, const gchar *name, XigWindow *window)
{
    g_hash_table_insert (server->priv->selection_owners, g_strdup (name), g_object_ref (window));  
}

XigWindow *
xig_server_get_selection_owner (XigServer *server, const gchar *name)
{
    return g_hash_table_lookup (server->priv->selection_owners, name);
}

void
_xig_server_add_pixmap (XigServer *server, XigPixmap *pixmap)
{
    g_hash_table_insert (server->priv->drawables, GINT_TO_POINTER (xig_drawable_get_id (XIG_DRAWABLE (pixmap))), g_object_ref (pixmap));
}

void
_xig_server_remove_pixmap (XigServer *server, XigPixmap *pixmap)
{
    g_hash_table_remove (server->priv->drawables, GINT_TO_POINTER (xig_drawable_get_id (XIG_DRAWABLE (pixmap))));
}

XigDrawable *
xig_server_get_drawable (XigServer *server, guint32 drawable)
{
    return g_hash_table_lookup (server->priv->drawables, GINT_TO_POINTER (drawable));
}

XigWindow *
xig_server_get_window (XigServer *server, guint32 wid)
{
    XigDrawable *drawable;

    if (wid == XIG_WINDOW_ID_NONE)
        return NULL;

    if (wid == XIG_WINDOW_ID_POINTER_ROOT)
    {
        XigScreen *screen = server->priv->screens->data;
        return xig_screen_get_root (screen);
    }

    drawable = xig_server_get_drawable (server, wid);
    if (!drawable || !XIG_IS_WINDOW (drawable))
        return NULL;

    return XIG_WINDOW (drawable);
}

XigRemoteClient *
xig_server_get_resize_redirect_client (XigServer *server, XigWindow *window)
{
    GList *link;

    for (link = server->priv->clients; link; link = link->next)
    {
        XigRemoteClient *client = link->data;
        if ((xig_remote_client_get_window_events (client, window) & XIG_EVENT_ResizeRedirect) != 0)
            return client;
    }

    return NULL;
}

XigRemoteClient *
xig_server_get_substructure_redirect_client (XigServer *server, XigWindow *window)
{
    GList *link;

    for (link = server->priv->clients; link; link = link->next)
    {
        XigRemoteClient *client = link->data;
        if ((xig_remote_client_get_window_events (client, window) & XIG_EVENT_SubstructureRedirect) != 0)
            return client;
    }

    return NULL;
}

void
xig_server_set_input_focus (XigServer *server, XigWindow *window, guint8 revert_to)
{
    server->priv->focus_window = g_object_ref (window);
    // FIXME: Catch delete event
    server->priv->focus_revert_to = revert_to;
}

XigWindow *
xig_server_get_input_focus (XigServer *server, guint8 *revert_to)
{
    if (revert_to)
        *revert_to = server->priv->focus_revert_to;
    return server->priv->focus_window;
}

XigPixmap *
xig_server_get_pixmap (XigServer *server, guint32 pid)
{
    XigDrawable *drawable;

    if (pid == XIG_PIXMAP_ID_NONE)
        return NULL;

    drawable = xig_server_get_drawable (server, pid);
    if (!drawable || !XIG_IS_PIXMAP (drawable))
        return NULL;

    return XIG_PIXMAP (drawable);
}

GList *
xig_server_get_clients (XigServer *server)
{
    return server->priv->clients;
}

guint32
xig_server_intern_atom (XigServer *server, const gchar *name, gboolean only_if_exists)
{
    guint32 id;

    id = xig_server_get_atom_id (server, name);
    if (id == 0 && only_if_exists)
        return 0;

    server->priv->last_atom_id++;
    id = server->priv->last_atom_id;
    g_hash_table_insert (server->priv->atoms_by_name, g_strdup (name), GINT_TO_POINTER (id));
    g_hash_table_insert (server->priv->atoms_by_id, GINT_TO_POINTER (id), g_strdup (name));

    return id;
}

const gchar *
xig_server_get_atom_name (XigServer *server, guint32 id)
{
    return g_hash_table_lookup (server->priv->atoms_by_id, GINT_TO_POINTER (id));
}

const guint32
xig_server_get_atom_id (XigServer *server, const gchar *name)
{
    return GPOINTER_TO_INT (g_hash_table_lookup (server->priv->atoms_by_name, name));
}

gboolean
xig_server_start (XigServer *server, GError **error)
{
    if (server->priv->listen_unix)
    {
        server->priv->socket_path = g_strdup_printf ("/tmp/.X11-unix/X%d", server->priv->display_number);

        server->priv->unix_socket = g_socket_new (G_SOCKET_FAMILY_UNIX, G_SOCKET_TYPE_STREAM, G_SOCKET_PROTOCOL_DEFAULT, error);
        if (!server->priv->unix_socket)
            return FALSE;
        if (!g_socket_bind (server->priv->unix_socket, g_unix_socket_address_new (server->priv->socket_path), TRUE, error))
            return FALSE;
        if (!g_socket_listen (server->priv->unix_socket, error))
            return FALSE;
        server->priv->unix_channel = g_io_channel_unix_new (g_socket_get_fd (server->priv->unix_socket));
        g_io_add_watch (server->priv->unix_channel, G_IO_IN, socket_connect_cb, server);
    }

    if (server->priv->listen_tcp)
    {
        server->priv->tcp_socket = g_socket_new (G_SOCKET_FAMILY_IPV4, G_SOCKET_TYPE_STREAM, G_SOCKET_PROTOCOL_TCP, error);
        if (!server->priv->tcp_socket)
            return FALSE;
        if (!g_socket_bind (server->priv->tcp_socket, g_inet_socket_address_new (g_inet_address_new_any (G_SOCKET_FAMILY_IPV4), server->priv->tcp_port), TRUE, error))
            return FALSE;
        if (!g_socket_listen (server->priv->tcp_socket, error))
            return FALSE;
        server->priv->tcp_channel = g_io_channel_unix_new (g_socket_get_fd (server->priv->tcp_socket));
        g_io_add_watch (server->priv->tcp_channel, G_IO_IN, socket_connect_cb, server);
    }

    return TRUE;
}

void
xig_server_stop (XigServer *server)
{
    GList *link;

    /* Shutdown the Unix connection */
    if (server->priv->socket_path)
        unlink (server->priv->socket_path);
    if (server->priv->unix_channel)
        g_io_channel_unref (server->priv->unix_channel);
    server->priv->unix_channel = NULL;
    if (server->priv->unix_socket)
    {
        g_socket_shutdown (server->priv->unix_socket, TRUE, TRUE, NULL);
        g_object_unref (server->priv->unix_socket);
    }
    server->priv->unix_socket = NULL;

    /* Shutdown the TCP/IP connection */
    if (server->priv->tcp_channel)
        g_io_channel_unref (server->priv->tcp_channel);
    server->priv->tcp_channel = NULL;
    if (server->priv->tcp_socket)
    {
        g_socket_shutdown (server->priv->tcp_socket, TRUE, TRUE, NULL);
        g_object_unref (server->priv->tcp_socket);
    }
  
    /* Shutdown the clients */
    for (link = server->priv->clients; link; link = link->next)
    {
        XigRemoteClient *client = link->data;
        xig_remote_client_disconnect (client);
    }
}

static void
xig_server_init (XigServer *server)
{
    server->priv = G_TYPE_INSTANCE_GET_PRIVATE (server, xig_server_get_type (), XigServerPrivate);
    server->priv->vendor = g_strdup ("");
    server->priv->image_byte_order = XIG_SERVER_IMAGE_BYTE_ORDER_LSBFirst;
    server->priv->bitmap_format_bit_order = XIG_SERVER_IMAGE_BYTE_ORDER_LeastSignificant;
    server->priv->min_keycode = 8;
    server->priv->min_keycode = 255;
    server->priv->screens = NULL;
    server->priv->listen_unix = TRUE;
    server->priv->listen_tcp = TRUE;
    server->priv->last_drawable_id = 2;
    server->priv->drawables = g_hash_table_new_full (g_direct_hash, g_direct_equal, NULL, g_object_unref);
    server->priv->last_atom_id = 0;
    server->priv->atoms_by_id = g_hash_table_new_full (g_direct_hash, g_direct_equal, NULL, g_free);
    server->priv->atoms_by_name = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
    server->priv->selection_owners = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_object_unref);

    xig_server_intern_atom (server, "PRIMARY", FALSE);
    xig_server_intern_atom (server, "SECONDARY", FALSE);
    xig_server_intern_atom (server, "ARC", FALSE);
    xig_server_intern_atom (server, "ATOM", FALSE);
    xig_server_intern_atom (server, "BITMAP", FALSE);
    xig_server_intern_atom (server, "CARDINAL", FALSE);
    xig_server_intern_atom (server, "COLORMAP", FALSE);
    xig_server_intern_atom (server, "CURSOR", FALSE);
    xig_server_intern_atom (server, "CUT_BUFFER0", FALSE);
    xig_server_intern_atom (server, "CUT_BUFFER1", FALSE);
    xig_server_intern_atom (server, "CUT_BUFFER2", FALSE);
    xig_server_intern_atom (server, "CUT_BUFFER3", FALSE);
    xig_server_intern_atom (server, "CUT_BUFFER4", FALSE);
    xig_server_intern_atom (server, "CUT_BUFFER5", FALSE);
    xig_server_intern_atom (server, "CUT_BUFFER6", FALSE);
    xig_server_intern_atom (server, "CUT_BUFFER7", FALSE);
    xig_server_intern_atom (server, "DRAWABLE", FALSE);
    xig_server_intern_atom (server, "FONT", FALSE);
    xig_server_intern_atom (server, "INTEGER", FALSE);
    xig_server_intern_atom (server, "PIXMAP", FALSE);
    xig_server_intern_atom (server, "POINT", FALSE);
    xig_server_intern_atom (server, "RECTANGLE", FALSE);
    xig_server_intern_atom (server, "RESOURCE_MANAGER", FALSE);
    xig_server_intern_atom (server, "RGB_COLOR_MAP", FALSE);
    xig_server_intern_atom (server, "RGB_BEST_MAP", FALSE);
    xig_server_intern_atom (server, "RGB_BLUE_MAP", FALSE);
    xig_server_intern_atom (server, "RGB_DEFAULT_MAP", FALSE);
    xig_server_intern_atom (server, "RGB_GRAY_MAP", FALSE);
    xig_server_intern_atom (server, "RGB_GREEN_MAP", FALSE);
    xig_server_intern_atom (server, "RGB_RED_MAP", FALSE);
    xig_server_intern_atom (server, "STRING", FALSE);
    xig_server_intern_atom (server, "VISUALID", FALSE);
    xig_server_intern_atom (server, "WINDOW", FALSE);
    xig_server_intern_atom (server, "WM_COMMAND", FALSE);
    xig_server_intern_atom (server, "WM_HINTS", FALSE);
    xig_server_intern_atom (server, "WM_CLIENT_MACHINE", FALSE);
    xig_server_intern_atom (server, "WM_ICON_NAME", FALSE);
    xig_server_intern_atom (server, "WM_ICON_SIZE", FALSE);
    xig_server_intern_atom (server, "WM_NAME", FALSE);
    xig_server_intern_atom (server, "WM_NORMAL_HINTS", FALSE);
    xig_server_intern_atom (server, "WM_SIZE_HINTS", FALSE);
    xig_server_intern_atom (server, "WM_ZOOM_HINTS", FALSE);
    xig_server_intern_atom (server, "MIN_SPACE", FALSE);
    xig_server_intern_atom (server, "NORM_SPACE", FALSE);
    xig_server_intern_atom (server, "MAX_SPACE", FALSE);
    xig_server_intern_atom (server, "END_SPACE", FALSE);
    xig_server_intern_atom (server, "SUPERSCRIPT_X", FALSE);
    xig_server_intern_atom (server, "SUPERSCRIPT_Y", FALSE);
    xig_server_intern_atom (server, "SUBSCRIPT_X", FALSE);
    xig_server_intern_atom (server, "SUBSCRIPT_Y", FALSE);
    xig_server_intern_atom (server, "UNDERLINE_POSITION", FALSE);
    xig_server_intern_atom (server, "UNDERLINE_THICKNESS", FALSE);
    xig_server_intern_atom (server, "STRIKEOUT_ASCENT", FALSE);
    xig_server_intern_atom (server, "STRIKEOUT_DESCENT", FALSE);
    xig_server_intern_atom (server, "ITALIC_ANGLE", FALSE);
    xig_server_intern_atom (server, "X_HEIGHT", FALSE);
    xig_server_intern_atom (server, "QUAD_WIDTH", FALSE);
    xig_server_intern_atom (server, "WEIGHT", FALSE);
    xig_server_intern_atom (server, "POINT_SIZE", FALSE);
    xig_server_intern_atom (server, "RESOLUTION", FALSE);
    xig_server_intern_atom (server, "COPYRIGHT", FALSE);
    xig_server_intern_atom (server, "NOTICE", FALSE);
    xig_server_intern_atom (server, "FONT_NAME", FALSE);
    xig_server_intern_atom (server, "FAMILY_NAME", FALSE);
    xig_server_intern_atom (server, "FULL_NAME", FALSE);
    xig_server_intern_atom (server, "CAP_HEIGHT", FALSE);
    xig_server_intern_atom (server, "WM_CLASS", FALSE);
    xig_server_intern_atom (server, "WM_TRANSIENT_FOR", FALSE);
}

static void
xig_server_finalize (GObject *object)
{
    XigServer *server = (XigServer *) object;
    g_free (server->priv->vendor);
    g_list_free_full (server->priv->pixmap_formats, g_object_unref);
    g_list_free_full (server->priv->screens, g_object_unref);
  
    /* Remove Unix socket */
    if (server->priv->socket_path)
        unlink (server->priv->socket_path);
    g_free (server->priv->socket_path);
    if (server->priv->unix_socket)
        g_object_unref (server->priv->unix_socket);
    if (server->priv->unix_channel)
        g_io_channel_unref (server->priv->unix_channel);

    /* Remove TCP/IP socket */
    if (server->priv->tcp_socket)
        g_object_unref (server->priv->tcp_socket);
    if (server->priv->tcp_channel)
        g_io_channel_unref (server->priv->tcp_channel);

    g_list_free_full (server->priv->clients, g_object_unref);

    g_hash_table_unref (server->priv->drawables);

    if (server->priv->focus_window)
        g_object_unref (server->priv->focus_window);

    g_hash_table_unref (server->priv->atoms_by_id);
    g_hash_table_unref (server->priv->atoms_by_name);

    g_hash_table_unref (server->priv->selection_owners);
}

static void
xig_server_class_init (XigServerClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);
    object_class->finalize = xig_server_finalize;
    g_type_class_add_private (klass, sizeof (XigServerPrivate));
    signals[CLIENT_CONNECTED] =
        g_signal_new ("client-connected",
                      G_TYPE_FROM_CLASS (klass),
                      G_SIGNAL_RUN_LAST,
                      G_STRUCT_OFFSET (XigServerClass, client_connected),
                      NULL, NULL,
                      g_cclosure_marshal_VOID__OBJECT,
                      G_TYPE_NONE, 1, xig_remote_client_get_type ());
    signals[CLIENT_DISCONNECTED] =
        g_signal_new ("client-disconnected",
                      G_TYPE_FROM_CLASS (klass),
                      G_SIGNAL_RUN_LAST,
                      G_STRUCT_OFFSET (XigServerClass, client_disconnected),
                      NULL, NULL,
                      g_cclosure_marshal_VOID__OBJECT,
                      G_TYPE_NONE, 1, xig_remote_client_get_type ());
}
