/*
 * Copyright (C) 2011 Canonical Ltd.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation. See http://www.gnu.org/copyleft/lgpl.html the
 * full text of the license.
 */

#include <string.h>

#include "xig-window-private.h"
#include "xig-drawable-private.h"
#include "xig-remote-client-private.h"
#include "xig-server-private.h"

G_DEFINE_TYPE (XigWindow, xig_window, xig_drawable_get_type ());

struct XigWindowPrivate
{
    XigServer *server;
    XigRemoteClient *client;

    XigWindow *parent;

    XigWindowClassType class;

    gint16 x;
    gint16 y;
    guint16 width;
    guint16 height;
    guint16 border_width;

    XigVisual *visual;

    XigPixmap *background_pixmap;
    guint32 background_pixel;

    XigPixmap *border_pixmap;
    guint32 border_pixel;
  
    XigBitGravity bit_gravity;
    XigWinGravity win_gravity;
  
    guint8 backing_store;
    guint32 backing_planes;
    guint32 backing_pixel;

    gboolean override_redirect;
  
    gboolean save_under;

    GHashTable *properties;

    GList *children;

    gboolean mapped;
};

enum {
    CREATE_NOTIFY,
    DESTROY_NOTIFY,
    CHILD_DESTROY_NOTIFY,
    UNMAP_NOTIFY,
    CHILD_UNMAP_NOTIFY,
    MAP_NOTIFY,
    CHILD_MAP_NOTIFY,
    REPARENT_NOTIFY,
    CHILD_REPARENT_NOTIFY,
    CONFIGURE_NOTIFY,
    CHILD_CONFIGURE_NOTIFY,
    CIRCULATE_NOTIFY,
    CHILD_CIRCULATE_NOTIFY,
    PROPERTY_NOTIFY,
    LAST_SIGNAL
};
static guint signals[LAST_SIGNAL] = { 0 };

XigWindow *
_xig_window_new (XigServer *server, XigRemoteClient *client, guint32 wid, XigWindow *parent, XigWindowClassType class, gint16 x, gint16 y, guint16 width, guint16 height, guint16 border_width, XigVisual *visual)
{
    XigWindow *window = g_object_new (xig_window_get_type (), NULL);

    _xig_drawable_set_id (XIG_DRAWABLE (window), wid);

    if (parent)
        parent->priv->children = g_list_append (parent->priv->children, window);

    window->priv->server = server;
    window->priv->client = client;
    window->priv->parent = parent;
    window->priv->class = class;
    window->priv->x = x;
    window->priv->y = y;
    window->priv->width = width;
    window->priv->height = height;
    window->priv->border_width = border_width;
    window->priv->visual = g_object_ref (visual);

    if (parent)
        g_signal_emit (parent, signals[CREATE_NOTIFY], 0, window);

    return window;
}

XigServer *
xig_window_get_server (XigWindow *window)
{
    return window->priv->server;
}

XigRemoteClient *
xig_window_get_client (XigWindow *window)
{
    return window->priv->client;
}

guint32
xig_window_get_id (XigWindow *window)
{
    return xig_drawable_get_id (XIG_DRAWABLE (window));
}

XigWindowClassType
xig_window_get_class (XigWindow *window)
{
    return window->priv->class;
}

gint16
xig_window_get_x (XigWindow *window)
{
    return window->priv->x;
}

gint16
xig_window_get_y (XigWindow *window)
{
    return window->priv->y;
}

guint16
xig_window_get_width (XigWindow *window)
{
    return window->priv->width;
}

guint16
xig_window_get_height (XigWindow *window)
{
    return window->priv->height;
}

guint16
xig_window_get_border_width (XigWindow *window)
{
    return window->priv->border_width;
}

XigVisual *
xig_window_get_visual (XigWindow *window)
{
    return window->priv->visual;
}

XigPixmap *
xig_window_get_background_pixmap (XigWindow *window)
{
    return window->priv->background_pixmap;
}

guint32
xig_window_get_background_pixel (XigWindow *window)
{
    return window->priv->background_pixel;
}

XigPixmap *
xig_window_get_border_pixmap (XigWindow *window)
{
    return window->priv->border_pixmap;
}

guint32
xig_window_get_border_pixel (XigWindow *window)
{
    return window->priv->border_pixel;
}

XigBitGravity
xig_window_get_bit_gravity (XigWindow *window)
{
    return window->priv->bit_gravity;
}

XigWinGravity
xig_window_get_win_gravity (XigWindow *window)
{
    return window->priv->win_gravity;
}

guint8
xig_window_get_backing_store (XigWindow *window)
{
    return window->priv->backing_store;
}

guint32
xig_window_get_backing_planes (XigWindow *window)
{
    return window->priv->backing_planes;
}

guint32
xig_window_get_backing_pixel (XigWindow *window)
{
    return window->priv->backing_pixel;
}

gboolean
xig_window_get_override_redirect (XigWindow *window)
{
    return window->priv->override_redirect;  
}

gboolean
xig_window_get_save_under (XigWindow *window)
{
    return window->priv->save_under;  
}

GHashTable *
xig_window_get_properties (XigWindow *window)
{
    return window->priv->properties;
}

XigWindowProperty *
xig_window_get_property (XigWindow *window, const gchar *name)
{
    return g_hash_table_lookup (window->priv->properties, name);
}

GList *
xig_window_get_children (XigWindow *window)
{
    return window->priv->children;
}

gboolean
xig_window_is_root (XigWindow *window)
{
    return window->priv->parent == NULL;
}

XigWindow *
xig_window_get_root (XigWindow *window)
{
    while (window->priv->parent)
        window = window->priv->parent;
    return window;
}

XigWindow *
xig_window_get_parent (XigWindow *window)
{
    return window->priv->parent;
}

XigWindow *
xig_window_get_window_below (XigWindow *window)
{
    XigWindow *parent = window->priv->parent;
    GList *link;
  
    link = g_list_find (parent->priv->children, window);
    if (link && link->prev)
        return link->prev->data;
    else
        return NULL;
}

XigWindow *
xig_window_get_window_above (XigWindow *window)
{
    XigWindow *parent = window->priv->parent;
    GList *link;
  
    link = g_list_find (parent->priv->children, window);
    if (link && link->next)
        return link->next->data;
    else
        return NULL;
}

gboolean
xig_window_has_child (XigWindow *window, XigWindow *child)
{
    return g_list_find (window->priv->children, child) != NULL;
}

gboolean
xig_window_get_is_on_bottom (XigWindow *window)
{
    GList *link;

    link = g_list_find (window->priv->parent->priv->children, window);
    g_return_val_if_fail (link != NULL, FALSE);

    return link->prev == NULL;
}

gboolean
xig_window_get_is_on_top (XigWindow *window)
{
    GList *link;

    link = g_list_find (window->priv->parent->priv->children, window);
    g_return_val_if_fail (link != NULL, FALSE);

    return link->next == NULL;
}

gboolean
xig_window_has_sibling (XigWindow *window, XigWindow *sibling)
{
    return xig_window_has_child (window->priv->parent, sibling);
}

XigWindow *
xig_window_add_child (XigWindow *window, XigRemoteClient *client, guint32 wid, XigWindowClassType class, gint16 x, gint16 y, guint16 width, guint16 height, guint16 border_width, XigVisual *visual, XigCreateWindowValueMask value_mask, XigPixmap *background_pixmap, guint32 background_pixel, XigPixmap *border_pixmap, guint32 border_pixel, XigBitGravity bit_gravity, XigWinGravity win_gravity, guint8 backing_store, guint32 backing_planes, guint32 backing_pixel, gboolean override_redirect, gboolean save_under)
{
    XigWindow *child;

    child = _xig_window_new (window->priv->server, client, wid, window, class, x, y, width, height, border_width, visual);
    xig_window_change_attributes (child, value_mask, background_pixmap, background_pixel, border_pixmap, border_pixel, bit_gravity, win_gravity, backing_store, backing_planes, backing_pixel, override_redirect, save_under);
    _xig_server_add_window (xig_window_get_server (window), child);

    return child;
}

void
xig_window_change_attributes (XigWindow *window, XigCreateWindowValueMask value_mask, XigPixmap *background_pixmap, guint32 background_pixel, XigPixmap *border_pixmap, guint32 border_pixel, XigBitGravity bit_gravity, XigWinGravity win_gravity, guint8 backing_store, guint32 backing_planes, guint32 backing_pixel, gboolean override_redirect, gboolean save_under)
{
    if (value_mask & XIG_WINDOW_VALUE_MASK_background_pixmap)
    {
        if (window->priv->background_pixmap)
            g_object_unref (window->priv->background_pixmap);
        window->priv->background_pixmap = NULL;
        if (background_pixmap)
            window->priv->background_pixmap = g_object_ref (background_pixmap);
    }
    if (value_mask & XIG_WINDOW_VALUE_MASK_background_pixel)
        window->priv->background_pixel = background_pixel;
    if (value_mask & XIG_WINDOW_VALUE_MASK_border_pixmap)
    {
        if (window->priv->border_pixmap)
            g_object_unref (window->priv->border_pixmap);
        window->priv->border_pixmap = NULL;
        if (border_pixmap)
            window->priv->border_pixmap = g_object_ref (border_pixmap);
    }
    if (value_mask & XIG_WINDOW_VALUE_MASK_border_pixel)
        window->priv->border_pixel = border_pixel;
    if (value_mask & XIG_WINDOW_VALUE_MASK_bit_gravity)
        window->priv->bit_gravity = bit_gravity;
    if (value_mask & XIG_WINDOW_VALUE_MASK_win_gravity)
        window->priv->win_gravity = win_gravity;
    if (value_mask & XIG_WINDOW_VALUE_MASK_backing_store)
        window->priv->backing_store = backing_store;
    if (value_mask & XIG_WINDOW_VALUE_MASK_backing_planes)
        window->priv->backing_planes = backing_planes;
    if (value_mask & XIG_WINDOW_VALUE_MASK_backing_pixel)
        window->priv->backing_pixel = backing_pixel;
    if (value_mask & XIG_WINDOW_VALUE_MASK_override_redirect)
        window->priv->override_redirect = override_redirect;
    if (value_mask & XIG_WINDOW_VALUE_MASK_save_under)
        window->priv->save_under = save_under;
}

void
xig_window_reparent (XigWindow *window, XigWindow *parent, gint16 x, gint16 y)
{
    XigWindow *old_parent;
    gboolean was_mapped;

    if (window->priv->mapped)
    {
        xig_window_unmap (window);
        was_mapped = TRUE;
    }

    /* Move across */
    old_parent = window->priv->parent;
    old_parent->priv->children = g_list_remove (old_parent->priv->children, window);
    parent->priv->children = g_list_append (parent->priv->children, window);
    window->priv->x = x;
    window->priv->y = y;

    g_signal_emit (window, signals[REPARENT_NOTIFY], 0);
    g_signal_emit (old_parent, signals[CHILD_REPARENT_NOTIFY], 0, window);
    g_signal_emit (parent, signals[CHILD_REPARENT_NOTIFY], 0, window);

    if (was_mapped)
        xig_window_map (window, NULL);
}

void
xig_window_map (XigWindow *window, XigRemoteClient *client)
{
    XigRemoteClient *redirect_client;

    redirect_client = xig_server_get_substructure_redirect_client (xig_window_get_server (window), window);
    if (!xig_window_get_override_redirect (window) && redirect_client != NULL && redirect_client != client)
    {
        _xig_remote_client_map_request (redirect_client, window);
        return;
    }
  
    if (window->priv->mapped)
        return;

    window->priv->mapped = TRUE;
    g_signal_emit (window, signals[MAP_NOTIFY], 0);
    g_signal_emit (window->priv->parent, signals[CHILD_MAP_NOTIFY], 0, window);
}

void
xig_window_unmap (XigWindow *window)
{
    if (!window->priv->mapped)
        return;

    window->priv->mapped = FALSE;
    g_signal_emit (window, signals[UNMAP_NOTIFY], 0);
    g_signal_emit (window->priv->parent, signals[CHILD_UNMAP_NOTIFY], 0, window);
}

gboolean
xig_window_get_mapped (XigWindow *window)
{
    return window->priv->mapped;
}

void
xig_window_configure (XigWindow *window, XigRemoteClient *client, gint16 x, gint16 y, guint16 width, guint16 height, guint16 border_width, XigWindow *sibling, XigWindowStackMode stack_mode, XigConfigureWindowValueMask value_mask)
{
    XigRemoteClient *redirect_client;
    gboolean changed = FALSE;

    redirect_client = xig_server_get_substructure_redirect_client (xig_window_get_server (window), window);
    if (!xig_window_get_override_redirect (window) && redirect_client != NULL && redirect_client != client)
    {
        _xig_remote_client_configure_request (redirect_client, window, x, y, width, height, border_width, sibling, stack_mode, value_mask);
        return;
    }

    if ((value_mask & XIG_CONFIGURE_WINDOW_VALUE_MASK_x) && window->priv->x != x)
        changed = TRUE;
    if ((value_mask & XIG_CONFIGURE_WINDOW_VALUE_MASK_y) && window->priv->y != y)
        changed = TRUE;
    if ((value_mask & XIG_CONFIGURE_WINDOW_VALUE_MASK_width) && window->priv->width != width)
        changed = TRUE;
    if ((value_mask & XIG_CONFIGURE_WINDOW_VALUE_MASK_height) && window->priv->height != height)
        changed = TRUE;
    if ((value_mask & XIG_CONFIGURE_WINDOW_VALUE_MASK_border_width) && window->priv->border_width != border_width)
        changed = TRUE;

    /* Ignore if already in this state */
    // FIXME: Check if stacking will change
    if (!changed)
        return;

    if (value_mask & XIG_CONFIGURE_WINDOW_VALUE_MASK_x)
        window->priv->x = x;
    if (value_mask & XIG_CONFIGURE_WINDOW_VALUE_MASK_y)
        window->priv->y = y;
    if (value_mask & XIG_CONFIGURE_WINDOW_VALUE_MASK_width)
        window->priv->width = width;
    if (value_mask & XIG_CONFIGURE_WINDOW_VALUE_MASK_height)
        window->priv->height = height;
    if (value_mask & XIG_CONFIGURE_WINDOW_VALUE_MASK_border_width)
        window->priv->border_width = border_width;
    if (value_mask & XIG_CONFIGURE_WINDOW_VALUE_MASK_stack_mode)
    {
        XigWindow *parent;
        GList *sibling_link;

        parent = window->priv->parent;
        parent->priv->children = g_list_remove (parent->priv->children, window);
        sibling_link = g_list_find (parent->priv->children, sibling);
        switch (stack_mode)
        {
        case XIG_WINDOW_STACK_MODE_ABOVE:
            if (sibling)
                parent->priv->children = g_list_insert_before (parent->priv->children, sibling_link->next, window);
            else
                parent->priv->children = g_list_append (parent->priv->children, window);
            break;
        case XIG_WINDOW_STACK_MODE_BELOW:
            if (sibling)
                parent->priv->children = g_list_insert_before (parent->priv->children, sibling_link, window);
            else
                parent->priv->children = g_list_prepend (parent->priv->children, window);
            break;
        default:
            g_warning ("Unsupported stack mode %d", stack_mode);
            break;
        }
    }

    g_signal_emit (window, signals[CONFIGURE_NOTIFY], 0);
    g_signal_emit (window->priv->parent, signals[CHILD_CONFIGURE_NOTIFY], 0, window);
}

void
xig_window_circulate (XigWindow *window, XigRemoteClient *client, XigCirculateDirection direction)
{
    XigRemoteClient *redirect_client;
    XigWindow *child;

    redirect_client = xig_server_get_substructure_redirect_client (xig_window_get_server (window), window);
    if (!xig_window_get_override_redirect (window) && redirect_client != NULL && redirect_client != client)
    {
        XigCirculatePlace place;
        if (direction == XIG_CIRCULATE_DIRECTION_RAISE_LOWEST)
            place = XIG_CIRCULATE_PLACE_BOTTOM;
        else
            place = XIG_CIRCULATE_PLACE_TOP;
        _xig_remote_client_circulate_request (redirect_client, window, place);
        return; 
    }
  
    /* Need at least two windows to cycle */
    if (!window->priv->children || !window->priv->children->next)
        return;

    switch (direction)
    {
    /* Move from the start to the end */
    case XIG_CIRCULATE_DIRECTION_RAISE_LOWEST:
        child = window->priv->children->data;
        window->priv->children = g_list_remove (window->priv->children, child);
        window->priv->children = g_list_append (window->priv->children, child);
        break;

    /* Move from the end to the start */
    case XIG_CIRCULATE_DIRECTION_LOWER_HIGHEST:
        child = g_list_last (window->priv->children)->data;
        window->priv->children = g_list_remove (window->priv->children, child);
        window->priv->children = g_list_prepend (window->priv->children, child);
        break;

    default:
        return;
    }

    g_signal_emit (child, signals[CIRCULATE_NOTIFY], 0);
    g_signal_emit (window, signals[CHILD_CIRCULATE_NOTIFY], 0, child);
}

static void
property_free (XigWindowProperty *property)
{
    g_free (property->name);
    g_free (property->type);
    g_free (property->data);  
    g_free (property);
}

XigWindowProperty *
xig_window_replace_property (XigWindow *window, const gchar *name, const gchar *type, guint8 format, guint8 *data, gsize data_length)
{
    XigWindowProperty *property;

    g_hash_table_remove (window->priv->properties, name);
    property = g_malloc0 (sizeof (XigWindowProperty));
    g_hash_table_insert (window->priv->properties, g_strdup (name), property);

    g_free (property->name);
    property->name = g_strdup (name);
    g_free (property->type);
    property->type = g_strdup (type);
    property->format = format;
    g_free (property->data);
    property->data = g_malloc (data_length);
    memcpy (property->data, data, data_length);
    property->data_length = data_length;

    g_signal_emit (window, signals[PROPERTY_NOTIFY], 0, name);

    return property;
}

XigWindowProperty *
xig_window_append_property (XigWindow *window, const gchar *name, guint8 *data, gsize data_length)
{
    XigWindowProperty *property;

    property = g_hash_table_lookup (window->priv->properties, name);
    g_return_val_if_fail (property != NULL, NULL);

    property->data = g_realloc (property->data, property->data_length + data_length);
    memcpy (property->data + property->data_length, data, data_length);
    property->data_length += data_length;
  
    return property;
}

XigWindowProperty *
xig_window_prepend_property (XigWindow *window, const gchar *name, guint8 *data, gsize data_length)
{
    XigWindowProperty *property;
    int i;

    property = g_hash_table_lookup (window->priv->properties, name);
    g_return_val_if_fail (property != NULL, NULL);

    property->data = g_realloc (property->data, data_length + property->data_length);
    for (i = property->data_length + data_length - 1; i >= data_length; i--)
        property->data[i] = property->data[i - data_length];
    memcpy (property->data, data, data_length);
    property->data_length += data_length;
  
    return property;
}

void
xig_window_delete_property (XigWindow *window, const gchar *name)
{
    g_hash_table_remove (window->priv->properties, name);
    g_signal_emit (window, signals[PROPERTY_NOTIFY], 0, name);
}

static void
destroy_recursive (XigWindow *window)
{
    GList *link;

    /* Destroy all children first */
    for (link = window->priv->children; link; link = link->next)
    {
        XigWindow *w = link->data;
        destroy_recursive (w);
    }

    g_signal_emit (window, signals[DESTROY_NOTIFY], 0);
    g_signal_emit (window->priv->parent, signals[CHILD_DESTROY_NOTIFY], 0, window);

    g_object_unref (window);
}

void
xig_window_destroy (XigWindow *window)
{
    window->priv->parent->priv->children = g_list_remove (window->priv->parent->priv->children, window);
    destroy_recursive (window);
}

void
xig_window_destroy_subwindows (XigWindow *window)
{
    GList *link;
    for (link = window->priv->children; link; link = link->next)
    {
        XigWindow *w = link->data;
        destroy_recursive (w);
    }
    g_list_free (window->priv->children);
    window->priv->children = NULL;
}

static void
xig_window_init (XigWindow *window)
{
    window->priv = G_TYPE_INSTANCE_GET_PRIVATE (window, xig_window_get_type (), XigWindowPrivate);
    window->priv->properties = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, (GDestroyNotify) property_free);
  
    window->priv->background_pixmap = NULL;
    window->priv->border_pixmap = NULL; // FIXME: Copy from parent
    window->priv->bit_gravity = XIG_BIT_GRAVITY_Forget;
    window->priv->win_gravity = XIG_WIN_GRAVITY_NorthWest;
    //window->priv->backing_store = NOT_USEFUL;
    window->priv->backing_planes = 0xFFFFFFFF;
    window->priv->backing_pixel = 0x00000000;
    window->priv->override_redirect = FALSE;
    window->priv->save_under = FALSE;
}

static void
xig_window_finalize (GObject *object)
{
    XigWindow *window = (XigWindow *) object;
    if (window->priv->visual)
        g_object_unref (window->priv->visual);
    if (window->priv->background_pixmap)
        g_object_unref (window->priv->background_pixmap);
    if (window->priv->border_pixmap)
        g_object_unref (window->priv->border_pixmap);
    g_list_free_full (window->priv->children, g_object_unref);
    g_hash_table_unref (window->priv->properties);
}

static void
xig_window_class_init (XigWindowClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);
    object_class->finalize = xig_window_finalize;
    g_type_class_add_private (klass, sizeof (XigWindowPrivate));

    signals[CREATE_NOTIFY] =
        g_signal_new ("create-notify",
                      G_TYPE_FROM_CLASS (klass),
                      G_SIGNAL_RUN_LAST,
                      G_STRUCT_OFFSET (XigWindowClass, create_notify),
                      NULL, NULL,
                      g_cclosure_marshal_VOID__OBJECT,
                      G_TYPE_NONE, 1, xig_window_get_type ());
    signals[DESTROY_NOTIFY] =
        g_signal_new ("destroy-notify",
                      G_TYPE_FROM_CLASS (klass),
                      G_SIGNAL_RUN_LAST,
                      G_STRUCT_OFFSET (XigWindowClass, destroy_notify),
                      NULL, NULL,
                      g_cclosure_marshal_VOID__VOID,
                      G_TYPE_NONE, 0);
    signals[CHILD_DESTROY_NOTIFY] =
        g_signal_new ("child-destroy-notify",
                      G_TYPE_FROM_CLASS (klass),
                      G_SIGNAL_RUN_LAST,
                      G_STRUCT_OFFSET (XigWindowClass, child_destroy_notify),
                      NULL, NULL,
                      g_cclosure_marshal_VOID__OBJECT,
                      G_TYPE_NONE, 1, xig_window_get_type ());
    signals[UNMAP_NOTIFY] =
        g_signal_new ("unmap-notify",
                      G_TYPE_FROM_CLASS (klass),
                      G_SIGNAL_RUN_LAST,
                      G_STRUCT_OFFSET (XigWindowClass, unmap_notify),
                      NULL, NULL,
                      g_cclosure_marshal_VOID__VOID,
                      G_TYPE_NONE, 0);
    signals[CHILD_UNMAP_NOTIFY] =
        g_signal_new ("child-unmap-notify",
                      G_TYPE_FROM_CLASS (klass),
                      G_SIGNAL_RUN_LAST,
                      G_STRUCT_OFFSET (XigWindowClass, child_unmap_notify),
                      NULL, NULL,
                      g_cclosure_marshal_VOID__OBJECT,
                      G_TYPE_NONE, 1, xig_window_get_type ());
    signals[MAP_NOTIFY] =
        g_signal_new ("map-notify",
                      G_TYPE_FROM_CLASS (klass),
                      G_SIGNAL_RUN_LAST,
                      G_STRUCT_OFFSET (XigWindowClass, map_notify),
                      NULL, NULL,
                      g_cclosure_marshal_VOID__VOID,
                      G_TYPE_NONE, 0);
    signals[CHILD_MAP_NOTIFY] =
        g_signal_new ("child-map-notify",
                      G_TYPE_FROM_CLASS (klass),
                      G_SIGNAL_RUN_LAST,
                      G_STRUCT_OFFSET (XigWindowClass, child_map_notify),
                      NULL, NULL,
                      g_cclosure_marshal_VOID__OBJECT,
                      G_TYPE_NONE, 1, xig_window_get_type ());
    signals[REPARENT_NOTIFY] =
        g_signal_new ("reparent-notify",
                      G_TYPE_FROM_CLASS (klass),
                      G_SIGNAL_RUN_LAST,
                      G_STRUCT_OFFSET (XigWindowClass, reparent_notify),
                      NULL, NULL,
                      g_cclosure_marshal_VOID__VOID,
                      G_TYPE_NONE, 0);
    signals[CHILD_REPARENT_NOTIFY] =
        g_signal_new ("child-reparent-notify",
                      G_TYPE_FROM_CLASS (klass),
                      G_SIGNAL_RUN_LAST,
                      G_STRUCT_OFFSET (XigWindowClass, child_reparent_notify),
                      NULL, NULL,
                      g_cclosure_marshal_VOID__OBJECT,
                      G_TYPE_NONE, 1, xig_window_get_type ());
    signals[CONFIGURE_NOTIFY] =
        g_signal_new ("configure-notify",
                      G_TYPE_FROM_CLASS (klass),
                      G_SIGNAL_RUN_LAST,
                      G_STRUCT_OFFSET (XigWindowClass, configure_notify),
                      NULL, NULL,
                      g_cclosure_marshal_VOID__VOID,
                      G_TYPE_NONE, 0);
    signals[CHILD_CONFIGURE_NOTIFY] =
        g_signal_new ("child-configure-notify",
                      G_TYPE_FROM_CLASS (klass),
                      G_SIGNAL_RUN_LAST,
                      G_STRUCT_OFFSET (XigWindowClass, child_configure_notify),
                      NULL, NULL,
                      g_cclosure_marshal_VOID__OBJECT,
                      G_TYPE_NONE, 1, xig_window_get_type ());
    signals[CIRCULATE_NOTIFY] =
        g_signal_new ("circulate-notify",
                      G_TYPE_FROM_CLASS (klass),
                      G_SIGNAL_RUN_LAST,
                      G_STRUCT_OFFSET (XigWindowClass, circulate_notify),
                      NULL, NULL,
                      g_cclosure_marshal_VOID__VOID,
                      G_TYPE_NONE, 0);
    signals[CHILD_CIRCULATE_NOTIFY] =
        g_signal_new ("child-circulate-notify",
                      G_TYPE_FROM_CLASS (klass),
                      G_SIGNAL_RUN_LAST,
                      G_STRUCT_OFFSET (XigWindowClass, child_circulate_notify),
                      NULL, NULL,
                      g_cclosure_marshal_VOID__OBJECT,
                      G_TYPE_NONE, 1, xig_window_get_type ());
    signals[PROPERTY_NOTIFY] =
        g_signal_new ("property-notify",
                      G_TYPE_FROM_CLASS (klass),
                      G_SIGNAL_RUN_LAST,
                      G_STRUCT_OFFSET (XigWindowClass, property_notify),
                      NULL, NULL,
                      g_cclosure_marshal_VOID__STRING,
                      G_TYPE_NONE, 1, G_TYPE_STRING);
}
