/*
 * Copyright (C) 2011 Canonical Ltd.
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation. See http://www.gnu.org/copyleft/lgpl.html the
 * full text of the license.
 */

#include <string.h>

#include "xig/xig-server-codec.h"
#include "xig-common.h" // FIXME: Rename to xig-codec.h

G_DEFINE_TYPE (XigServerCodec, xig_server_codec, xig_codec_get_type ());

struct XigServerCodecPrivate
{
    /* TRUE if client has connected */
    gboolean connected;

    /* Byte order client is using */
    ByteOrder byte_order;

    /* TRUE if client can send big requests */
    gboolean BIGREQ_enabled;

    /* Sequence number of last request */
    guint16 sequence_number;
};

XigServerCodec *
xig_server_codec_new (void)
{
    return g_object_new (xig_server_codec_get_type (), NULL);
}

void
xig_server_codec_set_BIGREQ_enabled (XigServerCodec *codec, gboolean BIGREQ_enabled)
{
    g_return_if_fail (codec != NULL);
    codec->priv->BIGREQ_enabled = BIGREQ_enabled;
}

static void
decode_POINT (XigServerCodec *codec, const guint8 *buffer, gsize buffer_length, gsize *offset, XigPOINT *value)
{
    value->x = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    value->y = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
}

static void
decode_RECTANGLE (XigServerCodec *codec, const guint8 *buffer, gsize buffer_length, gsize *offset, XigRECTANGLE *value)
{
    value->x = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    value->y = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    value->width = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    value->height = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
}

static void
decode_ARC (XigServerCodec *codec, const guint8 *buffer, gsize buffer_length, gsize *offset, XigARC *value)
{
    value->x = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    value->y = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    value->width = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    value->height = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    value->angle1 = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    value->angle2 = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
}

static void
decode_connection_request (XigServerCodec *codec, const guint8 *buffer, gsize buffer_length)
{
    guint8 byte_order;
    gsize offset = 0;
    guint16 n;
    XigConnect message;

    byte_order = read_card8 (buffer, buffer_length, &offset);
    if (!(byte_order == 'B' || byte_order == 'l'))
    {
        g_warning ("Invalid byte order");
        return;
    }

    message.byte_order = byte_order == 'B' ? BYTE_ORDER_MSB : BYTE_ORDER_LSB;
    read_padding (1, &offset);
    message.protocol_major_version = read_card16 (buffer, buffer_length, message.byte_order, &offset);
    message.protocol_minor_version = read_card16 (buffer, buffer_length, message.byte_order, &offset);
    n = read_card16 (buffer, buffer_length, message.byte_order, &offset);
    message.authorization_protocol_data_length = read_card16 (buffer, buffer_length, message.byte_order, &offset);
    read_padding (2, &offset);
    message.authorization_protocol_name = read_padded_string (buffer, buffer_length, n, &offset);
    message.authorization_protocol_data = read_string8 (buffer, buffer_length, message.authorization_protocol_data_length, &offset);
    read_padding (pad (message.authorization_protocol_data_length), &offset);

    /* Store information about the client */
    codec->priv->byte_order = message.byte_order;
    codec->priv->connected = TRUE;

    xig_codec_feed_connect (XIG_CODEC (codec), &message);

    g_free (message.authorization_protocol_name);
    g_free (message.authorization_protocol_data);
}

static void
decode_create_window (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigCreateWindow message;

    message.sequence_number = codec->priv->sequence_number;
    message.depth = data;
    message.wid = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.parent = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.x = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.y = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.width = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.height = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.border_width = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.class = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.visual = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.value_mask = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_WINDOW_VALUE_MASK_background_pixmap) != 0)
        message.background_pixmap = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_WINDOW_VALUE_MASK_background_pixel) != 0)
        message.background_pixel = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_WINDOW_VALUE_MASK_border_pixmap) != 0)
        message.border_pixmap = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_WINDOW_VALUE_MASK_border_pixel) != 0)
        message.border_pixel = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_WINDOW_VALUE_MASK_bit_gravity) != 0)
    {
        message.bit_gravity = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_WINDOW_VALUE_MASK_win_gravity) != 0)
    {
        message.win_gravity = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_WINDOW_VALUE_MASK_backing_store) != 0)
    {
        message.backing_store = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_WINDOW_VALUE_MASK_backing_planes) != 0)
        message.backing_planes = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_WINDOW_VALUE_MASK_backing_pixel) != 0)
        message.backing_pixel = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_WINDOW_VALUE_MASK_override_redirect) != 0)
    {
        message.override_redirect = read_card8 (buffer, buffer_length, offset) != 0;
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_WINDOW_VALUE_MASK_save_under) != 0)
    {
        message.save_under = read_card8 (buffer, buffer_length, offset) != 0;
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_WINDOW_VALUE_MASK_event_mask) != 0)
        message.event_mask = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_WINDOW_VALUE_MASK_do_not_propagate_mask) != 0)
        message.do_not_propagate_mask = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_WINDOW_VALUE_MASK_colormap) != 0)
        message.colormap = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_WINDOW_VALUE_MASK_cursor) != 0)
        message.cursor = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_create_window (XIG_CODEC (codec), &message);
}

static void
decode_change_window_attributes (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigChangeWindowAttributes message;

    message.sequence_number = codec->priv->sequence_number;
    message.window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.value_mask = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_WINDOW_VALUE_MASK_background_pixmap) != 0)
        message.background_pixmap = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_WINDOW_VALUE_MASK_background_pixel) != 0)
        message.background_pixel = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_WINDOW_VALUE_MASK_border_pixmap) != 0)
        message.border_pixmap = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_WINDOW_VALUE_MASK_border_pixel) != 0)
        message.border_pixel = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_WINDOW_VALUE_MASK_bit_gravity) != 0)
    {
        message.bit_gravity = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_WINDOW_VALUE_MASK_win_gravity) != 0)
    {
        message.win_gravity = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_WINDOW_VALUE_MASK_backing_store) != 0)
    {
        message.backing_store = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_WINDOW_VALUE_MASK_backing_planes) != 0)
        message.backing_planes = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_WINDOW_VALUE_MASK_backing_pixel) != 0)
        message.backing_pixel = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_WINDOW_VALUE_MASK_override_redirect) != 0)
    {
        message.override_redirect = read_card8 (buffer, buffer_length, offset) != 0;
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_WINDOW_VALUE_MASK_save_under) != 0)
    {
        message.save_under = read_card8 (buffer, buffer_length, offset) != 0;
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_WINDOW_VALUE_MASK_event_mask) != 0)
        message.event_mask = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_WINDOW_VALUE_MASK_do_not_propagate_mask) != 0)
        message.do_not_propagate_mask = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_WINDOW_VALUE_MASK_colormap) != 0)
        message.colormap = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_WINDOW_VALUE_MASK_cursor) != 0)
        message.cursor = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_change_window_attributes (XIG_CODEC (codec), &message);
}

static void
decode_get_window_attributes (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGetWindowAttributes message;

    message.sequence_number = codec->priv->sequence_number;
    message.window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_get_window_attributes (XIG_CODEC (codec), &message);
}

static void
decode_destroy_window (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigDestroyWindow message;

    message.sequence_number = codec->priv->sequence_number;
    message.window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_destroy_window (XIG_CODEC (codec), &message);
}

static void
decode_destroy_subwindows (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigDestroySubwindows message;

    message.sequence_number = codec->priv->sequence_number;
    message.window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_destroy_subwindows (XIG_CODEC (codec), &message);
}

static void
decode_change_save_set (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigChangeSaveSet message;

    message.sequence_number = codec->priv->sequence_number;
    message.mode = data;
    message.window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_change_save_set (XIG_CODEC (codec), &message);
}

static void
decode_reparent_window (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigReparentWindow message;

    message.sequence_number = codec->priv->sequence_number;
    message.window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.parent = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.x = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.y = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_reparent_window (XIG_CODEC (codec), &message);
}

static void
decode_map_window (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigMapWindow message;

    message.sequence_number = codec->priv->sequence_number;
    message.window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_map_window (XIG_CODEC (codec), &message);
}

static void
decode_map_subwindows (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigMapSubwindows message;

    message.sequence_number = codec->priv->sequence_number;
    message.window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_map_subwindows (XIG_CODEC (codec), &message);
}

static void
decode_unmap_window (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigUnmapWindow message;

    message.sequence_number = codec->priv->sequence_number;
    message.window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_unmap_window (XIG_CODEC (codec), &message);
}

static void
decode_unmap_subwindows (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigUnmapSubwindows message;

    message.sequence_number = codec->priv->sequence_number;
    message.window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_unmap_subwindows (XIG_CODEC (codec), &message);
}

static void
decode_configure_window (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigConfigureWindow message;

    message.sequence_number = codec->priv->sequence_number;
    message.window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.value_mask = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (2, offset);
    if ((message.value_mask & XIG_CONFIGURE_WINDOW_VALUE_MASK_x) != 0)
    {
        message.x = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
        read_padding (2, offset);
    }
    if ((message.value_mask & XIG_CONFIGURE_WINDOW_VALUE_MASK_y) != 0)
    {
        message.y = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
        read_padding (2, offset);
    }
    if ((message.value_mask & XIG_CONFIGURE_WINDOW_VALUE_MASK_width) != 0)
    {
        message.width = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
        read_padding (2, offset);
    }
    if ((message.value_mask & XIG_CONFIGURE_WINDOW_VALUE_MASK_height) != 0)
    {
        message.height = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
        read_padding (2, offset);
    }
    if ((message.value_mask & XIG_CONFIGURE_WINDOW_VALUE_MASK_border_width) != 0)
    {
        message.border_width = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
        read_padding (2, offset);
    }
    if ((message.value_mask & XIG_CONFIGURE_WINDOW_VALUE_MASK_sibling) != 0)
        message.sibling = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_CONFIGURE_WINDOW_VALUE_MASK_stack_mode) != 0)
    {
        message.stack_mode = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }

    xig_codec_feed_configure_window (XIG_CODEC (codec), &message);
}

static void
decode_circulate_window (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigCirculateWindow message;

    message.sequence_number = codec->priv->sequence_number;
    message.direction = data;
    message.window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_circulate_window (XIG_CODEC (codec), &message);
}

static void
decode_get_geometry (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGetGeometry message;

    message.sequence_number = codec->priv->sequence_number;
    message.drawable = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_get_geometry (XIG_CODEC (codec), &message);
}

static void
decode_query_tree (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigQueryTree message;

    message.sequence_number = codec->priv->sequence_number;
    message.window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_query_tree (XIG_CODEC (codec), &message);
}

static void
decode_intern_atom (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigInternAtom message;
    gsize name_length;

    message.sequence_number = codec->priv->sequence_number;
    message.only_if_exists = data != 0;
    name_length = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (2, offset);
    message.name = read_padded_string (buffer, buffer_length, name_length, offset);

    xig_codec_feed_intern_atom (XIG_CODEC (codec), &message);

    g_free (message.name);
}

static void
decode_get_atom_name (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGetAtomName message;

    message.sequence_number = codec->priv->sequence_number;
    message.atom = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_get_atom_name (XIG_CODEC (codec), &message);
}

static void
decode_change_property (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigChangeProperty message;
    guint8 *data8;
    guint16 *data16;
    guint32 *data32;
    int i;

    message.sequence_number = codec->priv->sequence_number;
    message.mode = data;
    message.window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.property = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.type = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.data_format = read_card8 (buffer, buffer_length, offset);
    read_padding (3, offset);
    message.data_length = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    switch (message.data_format)
    {
    default:
    case 8:
        message.data = data8 = calloc (message.data_length, sizeof (guint8));
        for (i = 0; i < message.data_length; i++)
            data8[i] = read_card8 (buffer, buffer_length, offset);
        break;
    case 16:
        message.data = data16 = calloc (message.data_length, sizeof (guint16));
        for (i = 0; i < message.data_length; i++)
            data16[i] = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
        break;
    case 32:
        message.data = data32 = calloc (message.data_length, sizeof (guint32));
        for (i = 0; i < message.data_length; i++)
            data32[i] = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
        break;
    }
    read_padding (pad (message.data_length * message.data_format / 8), offset);

    xig_codec_feed_change_property (XIG_CODEC (codec), &message);

    g_free (message.data);
}

static void
decode_delete_property (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigDeleteProperty message;

    message.sequence_number = codec->priv->sequence_number;
    message.window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.property = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_delete_property (XIG_CODEC (codec), &message);
}

static void
decode_get_property (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGetProperty message;

    message.sequence_number = codec->priv->sequence_number;
    message.delete = data != 0;
    message.window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.property = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.type = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.long_offset = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.long_length = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_get_property (XIG_CODEC (codec), &message);
}

static void
decode_list_properties (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigListProperties message;

    message.sequence_number = codec->priv->sequence_number;
    message.window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_list_properties (XIG_CODEC (codec), &message);
}

static void
decode_set_selection_owner (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigSetSelectionOwner message;

    message.sequence_number = codec->priv->sequence_number;
    message.owner = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.selection = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.time = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_set_selection_owner (XIG_CODEC (codec), &message);
}

static void
decode_get_selection_owner (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGetSelectionOwner message;

    message.sequence_number = codec->priv->sequence_number;
    message.selection = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_get_selection_owner (XIG_CODEC (codec), &message);
}

static void
decode_convert_selection (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigConvertSelection message;

    message.sequence_number = codec->priv->sequence_number;
    message.requestor = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.selection = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.target = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.property = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.time = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_convert_selection (XIG_CODEC (codec), &message);
}

static void
decode_send_event (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigSendEvent message;

    message.sequence_number = codec->priv->sequence_number;
    message.propagate = data != 0;
    message.destination = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.event_mask = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    // FIXME message.event = ;

    xig_codec_feed_send_event (XIG_CODEC (codec), &message);
}

static void
decode_grab_pointer (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGrabPointer message;

    message.sequence_number = codec->priv->sequence_number;
    message.owner_events = data != 0;
    message.grab_window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.event_mask = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.pointer_mode = read_card8 (buffer, buffer_length, offset);
    message.keyboard_mode = read_card8 (buffer, buffer_length, offset);
    message.confine_to = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.cursor = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.time = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_grab_pointer (XIG_CODEC (codec), &message);
}

static void
decode_ungrab_pointer (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigUngrabPointer message;

    message.sequence_number = codec->priv->sequence_number;
    message.time = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_ungrab_pointer (XIG_CODEC (codec), &message);
}

static void
decode_grab_button (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGrabButton message;

    message.sequence_number = codec->priv->sequence_number;
    message.owner_events = data != 0;
    message.grab_window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.event_mask = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.pointer_mode = read_card8 (buffer, buffer_length, offset);
    message.keyboard_mode = read_card8 (buffer, buffer_length, offset);
    message.confine_to = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.cursor = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.button = read_card8 (buffer, buffer_length, offset);
    read_padding (1, offset);
    message.modifiers = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_grab_button (XIG_CODEC (codec), &message);
}

static void
decode_ungrab_button (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigUngrabButton message;

    message.sequence_number = codec->priv->sequence_number;
    message.button = data;
    message.grab_window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.modifiers = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (2, offset);

    xig_codec_feed_ungrab_button (XIG_CODEC (codec), &message);
}

static void
decode_change_active_pointer_grab (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigChangeActivePointerGrab message;

    message.sequence_number = codec->priv->sequence_number;
    message.cursor = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.time = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.event_mask = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (2, offset);

    xig_codec_feed_change_active_pointer_grab (XIG_CODEC (codec), &message);
}

static void
decode_grab_keyboard (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGrabKeyboard message;

    message.sequence_number = codec->priv->sequence_number;
    message.owner_events = data != 0;
    message.grab_window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.time = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.pointer_mode = read_card8 (buffer, buffer_length, offset);
    message.keyboard_mode = read_card8 (buffer, buffer_length, offset);
    read_padding (2, offset);

    xig_codec_feed_grab_keyboard (XIG_CODEC (codec), &message);
}

static void
decode_ungrab_keyboard (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigUngrabKeyboard message;

    message.sequence_number = codec->priv->sequence_number;
    message.time = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_ungrab_keyboard (XIG_CODEC (codec), &message);
}

static void
decode_grab_key (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGrabKey message;

    message.sequence_number = codec->priv->sequence_number;
    message.owner_events = data != 0;
    message.grab_window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.modifiers = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.key = read_card8 (buffer, buffer_length, offset);
    message.pointer_mode = read_card8 (buffer, buffer_length, offset);
    message.keyboard_mode = read_card8 (buffer, buffer_length, offset);
    read_padding (3, offset);

    xig_codec_feed_grab_key (XIG_CODEC (codec), &message);
}

static void
decode_ungrab_key (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigUngrabKey message;

    message.sequence_number = codec->priv->sequence_number;
    message.key = data;
    message.grab_window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.modifiers = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (2, offset);

    xig_codec_feed_ungrab_key (XIG_CODEC (codec), &message);
}

static void
decode_allow_events (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigAllowEvents message;

    message.sequence_number = codec->priv->sequence_number;
    message.mode = data;
    message.time = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_allow_events (XIG_CODEC (codec), &message);
}

static void
decode_grab_server (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGrabServer message;

    message.sequence_number = codec->priv->sequence_number;

    xig_codec_feed_grab_server (XIG_CODEC (codec), &message);
}

static void
decode_ungrab_server (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigUngrabServer message;

    message.sequence_number = codec->priv->sequence_number;

    xig_codec_feed_ungrab_server (XIG_CODEC (codec), &message);
}

static void
decode_query_pointer (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigQueryPointer message;

    message.sequence_number = codec->priv->sequence_number;
    message.window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_query_pointer (XIG_CODEC (codec), &message);
}

static void
decode_get_motion_events (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGetMotionEvents message;

    message.sequence_number = codec->priv->sequence_number;
    message.window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.start = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.stop = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_get_motion_events (XIG_CODEC (codec), &message);
}

static void
decode_translate_coordinates (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigTranslateCoordinates message;

    message.sequence_number = codec->priv->sequence_number;
    message.src_window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.dst_window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.src_x = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.src_y = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_translate_coordinates (XIG_CODEC (codec), &message);
}

static void
decode_warp_pointer (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigWarpPointer message;

    message.sequence_number = codec->priv->sequence_number;
    message.src_window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.dst_window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.src_x = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.src_y = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.src_width = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.src_height = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.dst_x = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.dst_y = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_warp_pointer (XIG_CODEC (codec), &message);
}

static void
decode_set_input_focus (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigSetInputFocus message;

    message.sequence_number = codec->priv->sequence_number;
    message.revert_to = data;
    message.focus = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.time = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_set_input_focus (XIG_CODEC (codec), &message);
}

static void
decode_get_input_focus (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGetInputFocus message;

    message.sequence_number = codec->priv->sequence_number;

    xig_codec_feed_get_input_focus (XIG_CODEC (codec), &message);
}

static void
decode_query_keymap (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigQueryKeymap message;

    message.sequence_number = codec->priv->sequence_number;

    xig_codec_feed_query_keymap (XIG_CODEC (codec), &message);
}

static void
decode_open_font (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigOpenFont message;
    gsize name_length;

    message.sequence_number = codec->priv->sequence_number;
    message.fid = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    name_length = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (2, offset);
    message.name = read_padded_string (buffer, buffer_length, name_length, offset);

    xig_codec_feed_open_font (XIG_CODEC (codec), &message);

    g_free (message.name);
}

static void
decode_close_font (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigCloseFont message;

    message.sequence_number = codec->priv->sequence_number;
    message.font = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_close_font (XIG_CODEC (codec), &message);
}

static void
decode_query_font (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigQueryFont message;

    message.sequence_number = codec->priv->sequence_number;
    message.font = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_query_font (XIG_CODEC (codec), &message);
}

static void
decode_query_text_extents (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigQueryTextExtents message;
    int i;

    message.sequence_number = codec->priv->sequence_number;
    message.font = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.string_length = (buffer_length - 4) / 2;
    if (data != 0)
        message.string_length--;
    message.string = calloc (message.string_length + 1, sizeof (guint16));
    for (i = 0; i < message.string_length; i++)
        message.string[i] = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.string[i] = '\0';

    xig_codec_feed_query_text_extents (XIG_CODEC (codec), &message);

    g_free (message.string);
}

static void
decode_list_fonts (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigListFonts message;
    gsize pattern_length;

    message.sequence_number = codec->priv->sequence_number;
    message.max_names = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    pattern_length = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.pattern = read_padded_string (buffer, buffer_length, pattern_length, offset);

    xig_codec_feed_list_fonts (XIG_CODEC (codec), &message);

    g_free (message.pattern);
}

static void
decode_list_fonts_with_info (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigListFontsWithInfo message;
    gsize pattern_length;

    message.sequence_number = codec->priv->sequence_number;
    message.max_names = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    pattern_length = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.pattern = read_padded_string (buffer, buffer_length, pattern_length, offset);

    xig_codec_feed_list_fonts_with_info (XIG_CODEC (codec), &message);

    g_free (message.pattern);
}

static void
decode_set_font_path (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigSetFontPath message;
    gsize path_length;
    int i;

    message.sequence_number = codec->priv->sequence_number;
    path_length = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (2, offset);
    message.path = calloc (path_length + 1, sizeof (gchar *));
    for (i = 0; i < path_length; i++)
        message.path[i] = read_string (buffer, buffer_length, read_card8 (buffer, buffer_length, offset), offset);
    message.path[i] = NULL;

    xig_codec_feed_set_font_path (XIG_CODEC (codec), &message);

    g_free (message.path);
}

static void
decode_get_font_path (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGetFontPath message;

    message.sequence_number = codec->priv->sequence_number;

    xig_codec_feed_get_font_path (XIG_CODEC (codec), &message);
}

static void
decode_create_pixmap (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigCreatePixmap message;

    message.sequence_number = codec->priv->sequence_number;
    message.depth = data;
    message.pid = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.drawable = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.width = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.height = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_create_pixmap (XIG_CODEC (codec), &message);
}

static void
decode_free_pixmap (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigFreePixmap message;

    message.sequence_number = codec->priv->sequence_number;
    message.pixmap = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_free_pixmap (XIG_CODEC (codec), &message);
}

static void
decode_create_gc (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigCreateGC message;

    message.sequence_number = codec->priv->sequence_number;
    message.cid = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.drawable = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.value_mask = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_GC_VALUE_MASK_function) != 0)
    {
        message.function = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_plane_mask) != 0)
        message.plane_mask = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_GC_VALUE_MASK_foreground) != 0)
        message.foreground = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_GC_VALUE_MASK_background) != 0)
        message.background = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_GC_VALUE_MASK_line_width) != 0)
    {
        message.line_width = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
        read_padding (2, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_line_style) != 0)
    {
        message.line_style = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_cap_style) != 0)
    {
        message.cap_style = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_join_style) != 0)
    {
        message.join_style = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_fill_style) != 0)
    {
        message.fill_style = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_fill_rule) != 0)
    {
        message.fill_rule = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_tile) != 0)
        message.tile = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_GC_VALUE_MASK_stipple) != 0)
        message.stipple = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_GC_VALUE_MASK_tile_stipple_x_origin) != 0)
    {
        message.tile_stipple_x_origin = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
        read_padding (2, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_tile_stipple_y_origin) != 0)
    {
        message.tile_stipple_y_origin = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
        read_padding (2, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_font) != 0)
        message.font = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_GC_VALUE_MASK_subwindow_mode) != 0)
    {
        message.subwindow_mode = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_graphics_exposures) != 0)
    {
        message.graphics_exposures = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_clip_x_origin) != 0)
    {
        message.clip_x_origin = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
        read_padding (2, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_clip_y_origin) != 0)
    {
        message.clip_y_origin = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
        read_padding (2, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_clip_mask) != 0)
        message.clip_mask = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_GC_VALUE_MASK_dash_offset) != 0)
    {
        message.dash_offset = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
        read_padding (2, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_dashes) != 0)
    {
        message.dashes = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_arc_mode) != 0)
    {
        message.arc_mode = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }

    xig_codec_feed_create_gc (XIG_CODEC (codec), &message);
}

static void
decode_change_gc (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigChangeGC message;

    message.sequence_number = codec->priv->sequence_number;
    message.gc = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.value_mask = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_GC_VALUE_MASK_function) != 0)
    {
        message.function = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_plane_mask) != 0)
        message.plane_mask = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_GC_VALUE_MASK_foreground) != 0)
        message.foreground = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_GC_VALUE_MASK_background) != 0)
        message.background = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_GC_VALUE_MASK_line_width) != 0)
    {
        message.line_width = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
        read_padding (2, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_line_style) != 0)
    {
        message.line_style = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_cap_style) != 0)
    {
        message.cap_style = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_join_style) != 0)
    {
        message.join_style = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_fill_style) != 0)
    {
        message.fill_style = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_fill_rule) != 0)
    {
        message.fill_rule = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_tile) != 0)
        message.tile = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_GC_VALUE_MASK_stipple) != 0)
        message.stipple = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_GC_VALUE_MASK_tile_stipple_x_origin) != 0)
    {
        message.tile_stipple_x_origin = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
        read_padding (2, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_tile_stipple_y_origin) != 0)
    {
        message.tile_stipple_y_origin = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
        read_padding (2, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_font) != 0)
        message.font = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_GC_VALUE_MASK_subwindow_mode) != 0)
    {
        message.subwindow_mode = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_graphics_exposures) != 0)
    {
        message.graphics_exposures = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_clip_x_origin) != 0)
    {
        message.clip_x_origin = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
        read_padding (2, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_clip_y_origin) != 0)
    {
        message.clip_y_origin = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
        read_padding (2, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_clip_mask) != 0)
        message.clip_mask = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_GC_VALUE_MASK_dash_offset) != 0)
    {
        message.dash_offset = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
        read_padding (2, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_dashes) != 0)
    {
        message.dashes = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_arc_mode) != 0)
    {
        message.arc_mode = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }

    xig_codec_feed_change_gc (XIG_CODEC (codec), &message);
}

static void
decode_copy_gc (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigCopyGC message;

    message.sequence_number = codec->priv->sequence_number;
    message.src_gc = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.dst_gc = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.value_mask = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_GC_VALUE_MASK_function) != 0)
    {
        message.function = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_plane_mask) != 0)
        message.plane_mask = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_GC_VALUE_MASK_foreground) != 0)
        message.foreground = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_GC_VALUE_MASK_background) != 0)
        message.background = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_GC_VALUE_MASK_line_width) != 0)
    {
        message.line_width = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
        read_padding (2, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_line_style) != 0)
    {
        message.line_style = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_cap_style) != 0)
    {
        message.cap_style = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_join_style) != 0)
    {
        message.join_style = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_fill_style) != 0)
    {
        message.fill_style = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_fill_rule) != 0)
    {
        message.fill_rule = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_tile) != 0)
        message.tile = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_GC_VALUE_MASK_stipple) != 0)
        message.stipple = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_GC_VALUE_MASK_tile_stipple_x_origin) != 0)
    {
        message.tile_stipple_x_origin = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
        read_padding (2, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_tile_stipple_y_origin) != 0)
    {
        message.tile_stipple_y_origin = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
        read_padding (2, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_font) != 0)
        message.font = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_GC_VALUE_MASK_subwindow_mode) != 0)
    {
        message.subwindow_mode = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_graphics_exposures) != 0)
    {
        message.graphics_exposures = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_clip_x_origin) != 0)
    {
        message.clip_x_origin = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
        read_padding (2, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_clip_y_origin) != 0)
    {
        message.clip_y_origin = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
        read_padding (2, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_clip_mask) != 0)
        message.clip_mask = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    if ((message.value_mask & XIG_GC_VALUE_MASK_dash_offset) != 0)
    {
        message.dash_offset = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
        read_padding (2, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_dashes) != 0)
    {
        message.dashes = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }
    if ((message.value_mask & XIG_GC_VALUE_MASK_arc_mode) != 0)
    {
        message.arc_mode = read_card8 (buffer, buffer_length, offset);
        read_padding (3, offset);
    }

    xig_codec_feed_copy_gc (XIG_CODEC (codec), &message);
}

static void
decode_set_dashes (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigSetDashes message;
    int i;

    message.sequence_number = codec->priv->sequence_number;
    message.gc = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.dash_offset = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.dashes_length = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.dashes = calloc (message.dashes_length, sizeof (guint8));
    for (i = 0; i < message.dashes_length; i++)
        message.dashes[i] = read_card8 (buffer, buffer_length, offset);

    xig_codec_feed_set_dashes (XIG_CODEC (codec), &message);

    g_free (message.dashes);
}

static void
decode_set_clip_rectangles (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigSetClipRectangles message;
    int i;

    message.sequence_number = codec->priv->sequence_number;
    message.ordering = data;
    message.gc = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.clip_x_origin = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.clip_y_origin = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.rectangles_length = (buffer_length - 12) / 2;
    message.rectangles = calloc (message.rectangles_length, sizeof (XigRECTANGLE));
    for (i = 0; i < message.rectangles_length; i++)
        decode_RECTANGLE (codec, buffer, buffer_length, offset, &message.rectangles[i]);

    xig_codec_feed_set_clip_rectangles (XIG_CODEC (codec), &message);

    g_free (message.rectangles);
}

static void
decode_free_gc (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigFreeGC message;

    message.sequence_number = codec->priv->sequence_number;
    message.gc = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_free_gc (XIG_CODEC (codec), &message);
}

static void
decode_clear_area (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigClearArea message;

    message.sequence_number = codec->priv->sequence_number;
    message.exposures = data != 0;
    message.window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.x = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.y = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.width = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.height = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_clear_area (XIG_CODEC (codec), &message);
}

static void
decode_copy_area (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigCopyArea message;

    message.sequence_number = codec->priv->sequence_number;
    message.src_drawable = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.dst_drawable = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.gc = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.src_x = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.src_y = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.dst_x = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.dst_y = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.width = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.height = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_copy_area (XIG_CODEC (codec), &message);
}

static void
decode_copy_plane (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigCopyPlane message;

    message.sequence_number = codec->priv->sequence_number;
    message.src_drawable = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.dst_drawable = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.gc = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.src_x = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.src_y = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.dst_x = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.dst_y = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.width = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.height = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.bit_plane = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_copy_plane (XIG_CODEC (codec), &message);
}

static void
decode_poly_point (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigPolyPoint message;
    int i;

    message.sequence_number = codec->priv->sequence_number;
    message.coordinate_mode = data;
    message.drawable = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.gc = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.points_length = (buffer_length - 12) / 4;
    message.points = calloc (message.points_length, sizeof (XigPOINT));
    for (i = 0; i < message.points_length; i++)
        decode_POINT (codec, buffer, buffer_length, offset, &message.points[i]);

    xig_codec_feed_poly_point (XIG_CODEC (codec), &message);

    g_free (message.points);
}

static void
decode_poly_line (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigPolyLine message;
    int i;

    message.sequence_number = codec->priv->sequence_number;
    message.coordinate_mode = data;
    message.drawable = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.gc = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.points_length = (buffer_length - *offset) / 4;
    message.points = calloc (message.points_length, sizeof (XigPOINT));
    for (i = 0; i < message.points_length; i++)
        decode_POINT (codec, buffer, buffer_length, offset, &message.points[i]);

    xig_codec_feed_poly_line (XIG_CODEC (codec), &message);

    g_free (message.points);
}

static void
decode_SEGMENT (XigServerCodec *codec, const guint8 *buffer, gsize buffer_length, gsize *offset, XigSEGMENT *value)
{
    value->x1 = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    value->y1 = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    value->x2 = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    value->y2 = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
}

static void
decode_poly_segment (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigPolySegment message;
    int i;

    message.sequence_number = codec->priv->sequence_number;
    message.drawable = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.gc = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.segments_length = (buffer_length - *offset) / 8;
    message.segments = calloc (message.segments_length, sizeof (XigSEGMENT));
    for (i = 0; i < message.segments_length; i++)
        decode_SEGMENT (codec, buffer, buffer_length, offset, &message.segments[i]);

    xig_codec_feed_poly_segment (XIG_CODEC (codec), &message);

    g_free (message.segments);
}

static void
decode_poly_rectangle (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigPolyRectangle message;
    int i;

    message.sequence_number = codec->priv->sequence_number;
    message.drawable = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.gc = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.rectangles_length = (buffer_length - 12) / 8;
    message.rectangles = calloc (message.rectangles_length, sizeof (XigRECTANGLE));
    for (i = 0; i < message.rectangles_length; i++)
        decode_RECTANGLE (codec, buffer, buffer_length, offset, &message.rectangles[i]);

    xig_codec_feed_poly_rectangle (XIG_CODEC (codec), &message);

    g_free (message.rectangles);
}

static void
decode_poly_arc (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigPolyArc message;
    int i;

    message.sequence_number = codec->priv->sequence_number;
    message.drawable = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.gc = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.arcs_length = (buffer_length - 12) / 8;
    message.arcs = calloc (message.arcs_length, sizeof (XigARC));
    for (i = 0; i < message.arcs_length; i++)
        decode_ARC (codec, buffer, buffer_length, offset, &message.arcs[i]);

    xig_codec_feed_poly_arc (XIG_CODEC (codec), &message);

    g_free (message.arcs);
}

static void
decode_fill_poly (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigFillPoly message;
    int i;

    message.sequence_number = codec->priv->sequence_number;
    message.drawable = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.gc = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.shape = read_card8 (buffer, buffer_length, offset);
    message.coordinate_mode = read_card8 (buffer, buffer_length, offset);
    read_padding (2, offset);
    message.points_length = (buffer_length - 16) / 4;
    message.points = calloc (message.points_length, sizeof (XigPOINT));
    for (i = 0; i < message.points_length; i++)
        decode_POINT (codec, buffer, buffer_length, offset, &message.points[i]);

    xig_codec_feed_fill_poly (XIG_CODEC (codec), &message);

    g_free (message.points);
}

static void
decode_poly_fill_rectangle (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigPolyFillRectangle message;
    int i;

    message.sequence_number = codec->priv->sequence_number;
    message.drawable = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.gc = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.rectangles_length = (buffer_length - 12) / 8;
    message.rectangles = calloc (message.rectangles_length, sizeof (XigRECTANGLE));
    for (i = 0; i < message.rectangles_length; i++)
        decode_RECTANGLE (codec, buffer, buffer_length, offset, &message.rectangles[i]);

    xig_codec_feed_poly_fill_rectangle (XIG_CODEC (codec), &message);

    g_free (message.rectangles);
}

static void
decode_poly_fill_arc (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigPolyFillArc message;
    int i;

    message.sequence_number = codec->priv->sequence_number;
    message.drawable = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.gc = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.arcs_length = (buffer_length - 12) / 8;
    message.arcs = calloc (message.arcs_length, sizeof (XigARC));
    for (i = 0; i < message.arcs_length; i++)
        decode_ARC (codec, buffer, buffer_length, offset, &message.arcs[i]);

    xig_codec_feed_poly_fill_arc (XIG_CODEC (codec), &message);

    g_free (message.arcs);
}

static void
decode_put_image (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigPutImage message;
    int i;

    message.sequence_number = codec->priv->sequence_number;
    message.format = data;
    message.drawable = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.gc = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.width = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.height = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.dst_x = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.dst_y = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.left_pad = read_card8 (buffer, buffer_length, offset);
    message.depth = read_card8 (buffer, buffer_length, offset);
    read_padding (2, offset);
    message.data = calloc (message.data_length, sizeof (guint8));
    for (i = 0; i < message.data_length; i++)
        message.data[i] = read_card8 (buffer, buffer_length, offset);

    xig_codec_feed_put_image (XIG_CODEC (codec), &message);

    g_free (message.data);
}

static void
decode_get_image (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGetImage message;

    message.sequence_number = codec->priv->sequence_number;
    message.format = data;
    message.drawable = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.x = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.y = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.width = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.height = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.plane_mask = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_get_image (XIG_CODEC (codec), &message);
}

static void
decode_poly_text8 (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigPolyText8 message;
    int i;

    message.sequence_number = codec->priv->sequence_number;
    message.drawable = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.gc = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.x = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.y = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.items = calloc (message.items_length, sizeof (guint8));
    for (i = 0; i < message.items_length; i++)
        message.items[i] = read_card8 (buffer, buffer_length, offset);

    xig_codec_feed_poly_text8 (XIG_CODEC (codec), &message);

    g_free (message.items);
}

static void
decode_poly_text16 (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigPolyText16 message;
    int i;

    message.sequence_number = codec->priv->sequence_number;
    message.drawable = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.gc = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.x = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.y = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.items = calloc (message.items_length, sizeof (guint8));
    for (i = 0; i < message.items_length; i++)
        message.items[i] = read_card8 (buffer, buffer_length, offset);

    xig_codec_feed_poly_text16 (XIG_CODEC (codec), &message);

    g_free (message.items);
}

static void
decode_image_text8 (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigImageText8 message;
    gsize string_length;

    message.sequence_number = codec->priv->sequence_number;
    string_length = data;
    message.drawable = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.gc = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.x = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.y = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.string = read_padded_string (buffer, buffer_length, string_length, offset);

    xig_codec_feed_image_text8 (XIG_CODEC (codec), &message);

    g_free (message.string);
}

static void
decode_image_text16 (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigImageText16 message;
    int i;
    gsize string_length;

    message.sequence_number = codec->priv->sequence_number;
    string_length = data;
    message.drawable = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.gc = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.x = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.y = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.string = calloc (string_length, sizeof (guint16));
    for (i = 0; i < string_length; i++)
        message.string[i] = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.string[i] = '\0';

    xig_codec_feed_image_text16 (XIG_CODEC (codec), &message);

    g_free (message.string);
}

static void
decode_create_colormap (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigCreateColormap message;

    message.sequence_number = codec->priv->sequence_number;
    message.alloc = data;
    message.mid = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.visual = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_create_colormap (XIG_CODEC (codec), &message);
}

static void
decode_free_colormap (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigFreeColormap message;

    message.sequence_number = codec->priv->sequence_number;
    message.cmap = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_free_colormap (XIG_CODEC (codec), &message);
}

static void
decode_copy_colormap_and_free (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigCopyColormapAndFree message;

    message.sequence_number = codec->priv->sequence_number;
    message.mid = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.src_cmap = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_copy_colormap_and_free (XIG_CODEC (codec), &message);
}

static void
decode_install_colormap (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigInstallColormap message;

    message.sequence_number = codec->priv->sequence_number;
    message.cmap = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_install_colormap (XIG_CODEC (codec), &message);
}

static void
decode_uninstall_colormap (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigUninstallColormap message;

    message.sequence_number = codec->priv->sequence_number;
    message.cmap = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_uninstall_colormap (XIG_CODEC (codec), &message);
}

static void
decode_list_installed_colormaps (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigListInstalledColormaps message;

    message.sequence_number = codec->priv->sequence_number;
    message.window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_list_installed_colormaps (XIG_CODEC (codec), &message);
}

static void
decode_alloc_color (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigAllocColor message;

    message.sequence_number = codec->priv->sequence_number;
    message.cmap = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.red = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.green = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.blue = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (2, offset);

    xig_codec_feed_alloc_color (XIG_CODEC (codec), &message);
}

static void
decode_alloc_named_color (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigAllocNamedColor message;
    gsize name_length;

    message.sequence_number = codec->priv->sequence_number;
    message.cmap = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    name_length = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (2, offset);
    message.name = read_padded_string (buffer, buffer_length, name_length, offset);

    xig_codec_feed_alloc_named_color (XIG_CODEC (codec), &message);

    g_free (message.name);
}

static void
decode_alloc_color_cells (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigAllocColorCells message;

    message.sequence_number = codec->priv->sequence_number;
    message.contiguous = data != 0;
    message.cmap = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.colors = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.planes = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_alloc_color_cells (XIG_CODEC (codec), &message);
}

static void
decode_alloc_color_planes (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigAllocColorPlanes message;

    message.sequence_number = codec->priv->sequence_number;
    message.contiguous = data != 0;
    message.cmap = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.colors = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.reds = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.greens = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.blues = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_alloc_color_planes (XIG_CODEC (codec), &message);
}

static void
decode_free_colors (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigFreeColors message;
    int i;

    message.sequence_number = codec->priv->sequence_number;
    message.cmap = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.plane_mask = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.pixels_length = (buffer_length - 12) / 4;
    message.pixels = calloc (message.pixels_length, sizeof (guint32));
    for (i = 0; i < message.pixels_length; i++)
        message.pixels[i] = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_free_colors (XIG_CODEC (codec), &message);

    g_free (message.pixels);
}

static void
decode_COLORITEM (XigServerCodec *codec, const guint8 *buffer, gsize buffer_length, gsize *offset, XigCOLORITEM *value)
{
    guint8 do_mask;

    value->pixel = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    value->red = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    value->green = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    value->blue = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    do_mask = read_card8 (buffer, buffer_length, offset);
    value->do_red = do_mask & 0x01;
    value->do_green = do_mask & 0x02;
    value->do_blue = do_mask & 0x04;
    read_padding (1, offset);
}

static void
decode_store_colors (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigStoreColors message;
    int i;

    message.sequence_number = codec->priv->sequence_number;
    message.cmap = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.items_length = (buffer_length - 8) / 12;
    message.items = calloc (message.items_length, sizeof (XigCOLORITEM));
    for (i = 0; i < message.items_length; i++)
        decode_COLORITEM (codec, buffer, buffer_length, offset, &message.items[i]);

    xig_codec_feed_store_colors (XIG_CODEC (codec), &message);

    g_free (message.items);
}

static void
decode_store_named_color (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigStoreNamedColor message;
    gsize name_length;

    message.sequence_number = codec->priv->sequence_number;
    message.cmap = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.pixel = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.do_red = data & 0x01;
    message.do_green = data & 0x02;
    message.do_blue = data & 0x04;
    name_length = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (2, offset);
    message.name = read_padded_string (buffer, buffer_length, name_length, offset);

    xig_codec_feed_store_named_color (XIG_CODEC (codec), &message);

    g_free (message.name);
}

static void
decode_query_colors (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigQueryColors message;
    int i;

    message.sequence_number = codec->priv->sequence_number;
    message.cmap = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.pixels_length = (buffer_length - 8) / 4;
    message.pixels = calloc (message.pixels_length, sizeof (guint32));
    for (i = 0; i < message.pixels_length; i++)
        message.pixels[i] = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_query_colors (XIG_CODEC (codec), &message);

    g_free (message.pixels);
}

static void
decode_lookup_color (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigLookupColor message;
    gsize name_length;

    message.sequence_number = codec->priv->sequence_number;
    message.cmap = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    name_length = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (2, offset);
    message.name = read_padded_string (buffer, buffer_length, name_length, offset);

    xig_codec_feed_lookup_color (XIG_CODEC (codec), &message);

    g_free (message.name);
}

static void
decode_create_cursor (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigCreateCursor message;

    message.sequence_number = codec->priv->sequence_number;
    message.cid = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.source = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.mask = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.fore_red = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.fore_green = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.fore_blue = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.back_red = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.back_green = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.back_blue = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.x = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.y = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_create_cursor (XIG_CODEC (codec), &message);
}

static void
decode_create_glyph_cursor (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigCreateGlyphCursor message;

    message.sequence_number = codec->priv->sequence_number;
    message.cid = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.source_font = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.mask_font = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.source_char = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.mask_char = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.fore_red = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.fore_green = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.fore_blue = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.back_red = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.back_green = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.back_blue = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_create_glyph_cursor (XIG_CODEC (codec), &message);
}

static void
decode_free_cursor (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigFreeCursor message;

    message.sequence_number = codec->priv->sequence_number;
    message.cursor = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_free_cursor (XIG_CODEC (codec), &message);
}

static void
decode_recolor_cursor (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigRecolorCursor message;

    message.sequence_number = codec->priv->sequence_number;
    message.cursor = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.fore_red = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.fore_green = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.fore_blue = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.back_red = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.back_green = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.back_blue = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_recolor_cursor (XIG_CODEC (codec), &message);
}

static void
decode_query_best_size (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigQueryBestSize message;

    message.sequence_number = codec->priv->sequence_number;
    message.class = data;
    message.drawable = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.width = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.height = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_query_best_size (XIG_CODEC (codec), &message);
}

static void
decode_query_extension (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigQueryExtension message;
    gsize name_length;

    message.sequence_number = codec->priv->sequence_number;
    name_length = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (2, offset);
    message.name = read_padded_string (buffer, buffer_length, name_length, offset);

    xig_codec_feed_query_extension (XIG_CODEC (codec), &message);

    g_free (message.name);
}

static void
decode_list_extensions (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigListExtensions message;

    message.sequence_number = codec->priv->sequence_number;

    xig_codec_feed_list_extensions (XIG_CODEC (codec), &message);
}

static void
decode_change_keyboard_mapping (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigChangeKeyboardMapping message;
    int i;
    guint8 keycode_count;

    message.sequence_number = codec->priv->sequence_number;
    keycode_count = data;
    message.first_keycode = read_card8 (buffer, buffer_length, offset);
    message.keysyms_per_keycode = read_card8 (buffer, buffer_length, offset);
    read_padding (2, offset);
    message.keysyms_length = keycode_count * message.keysyms_per_keycode;
    message.keysyms = calloc (message.keysyms_length, sizeof (guint32));
    for (i = 0; i < message.keysyms_length; i++)
        message.keysyms[i] = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_change_keyboard_mapping (XIG_CODEC (codec), &message);

    g_free (message.keysyms);
}

static void
decode_get_keyboard_mapping (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGetKeyboardMapping message;

    message.sequence_number = codec->priv->sequence_number;
    message.first_keycode = read_card8 (buffer, buffer_length, offset);
    message.count = read_card8 (buffer, buffer_length, offset);
    read_padding (2, offset);

    xig_codec_feed_get_keyboard_mapping (XIG_CODEC (codec), &message);
}

static void
decode_change_keyboard_control (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigChangeKeyboardControl message;

    message.sequence_number = codec->priv->sequence_number;
    message.value_mask = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    if (message.value_mask & XIG_KEYBOARD_CONTROL_key_click_percent)
        message.key_click_percent = read_card8 (buffer, buffer_length, offset);
    if (message.value_mask & XIG_KEYBOARD_CONTROL_bell_percent)
        message.bell_percent = read_card8 (buffer, buffer_length, offset);
    if (message.value_mask & XIG_KEYBOARD_CONTROL_bell_pitch)
        message.bell_pitch = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    if (message.value_mask & XIG_KEYBOARD_CONTROL_bell_duration)
        message.bell_duration = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    if (message.value_mask & XIG_KEYBOARD_CONTROL_led)
        message.led = read_card8 (buffer, buffer_length, offset);
    if (message.value_mask & XIG_KEYBOARD_CONTROL_led_mode)
        message.led_mode = read_card8 (buffer, buffer_length, offset);
    if (message.value_mask & XIG_KEYBOARD_CONTROL_key)
        message.key = read_card8 (buffer, buffer_length, offset);
    if (message.value_mask & XIG_KEYBOARD_CONTROL_auto_repeat_mode)
        message.auto_repeat_mode = read_card8 (buffer, buffer_length, offset);

    xig_codec_feed_change_keyboard_control (XIG_CODEC (codec), &message);
}

static void
decode_get_keyboard_control (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGetKeyboardControl message;

    message.sequence_number = codec->priv->sequence_number;

    xig_codec_feed_get_keyboard_control (XIG_CODEC (codec), &message);
}

static void
decode_bell (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigBell message;

    message.sequence_number = codec->priv->sequence_number;
    message.percent = data;

    xig_codec_feed_bell (XIG_CODEC (codec), &message);
}

static void
decode_change_pointer_control (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigChangePointerControl message;

    message.sequence_number = codec->priv->sequence_number;
    message.acceleration_numerator = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.acceleration_denominator = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.threshold = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.do_acceleration = read_card8 (buffer, buffer_length, offset) != 0;
    message.do_threshold = read_card8 (buffer, buffer_length, offset) != 0;

    xig_codec_feed_change_pointer_control (XIG_CODEC (codec), &message);
}

static void
decode_get_pointer_control (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGetPointerControl message;

    message.sequence_number = codec->priv->sequence_number;

    xig_codec_feed_get_pointer_control (XIG_CODEC (codec), &message);
}

static void
decode_set_screen_saver (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigSetScreenSaver message;

    message.sequence_number = codec->priv->sequence_number;
    message.timeout = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.interval = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.prefer_blanking = read_card8 (buffer, buffer_length, offset);
    message.allow_exposures = read_card8 (buffer, buffer_length, offset);
    read_padding (2, offset);

    xig_codec_feed_set_screen_saver (XIG_CODEC (codec), &message);
}

static void
decode_get_screen_saver (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGetScreenSaver message;

    message.sequence_number = codec->priv->sequence_number;

    xig_codec_feed_get_screen_saver (XIG_CODEC (codec), &message);
}

static void
decode_change_hosts (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigChangeHosts message;
    int i;

    message.sequence_number = codec->priv->sequence_number;
    message.mode = data;
    message.family = read_card8 (buffer, buffer_length, offset);
    read_padding (1, offset);
    message.address_length = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.address = calloc (message.address_length, sizeof (guint8));
    for (i = 0; i < message.address_length; i++)
        message.address[i] = read_card8 (buffer, buffer_length, offset);

    xig_codec_feed_change_hosts (XIG_CODEC (codec), &message);

    g_free (message.address);
}

static void
decode_list_hosts (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigListHosts message;

    message.sequence_number = codec->priv->sequence_number;

    xig_codec_feed_list_hosts (XIG_CODEC (codec), &message);
}

static void
decode_set_access_control (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigSetAccessControl message;

    message.sequence_number = codec->priv->sequence_number;
    message.mode = data;

    xig_codec_feed_set_access_control (XIG_CODEC (codec), &message);
}

static void
decode_set_close_down_mode (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigSetCloseDownMode message;

    message.sequence_number = codec->priv->sequence_number;
    message.mode = data;

    xig_codec_feed_set_close_down_mode (XIG_CODEC (codec), &message);
}

static void
decode_kill_client (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigKillClient message;

    message.sequence_number = codec->priv->sequence_number;
    message.resource = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_kill_client (XIG_CODEC (codec), &message);
}

static void
decode_rotate_properties (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigRotateProperties message;
    int i;

    message.sequence_number = codec->priv->sequence_number;
    message.window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.properties_length = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.delta = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.properties = calloc (message.properties_length, sizeof (guint32));
    for (i = 0; i < message.properties_length; i++)
        message.properties[i] = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_rotate_properties (XIG_CODEC (codec), &message);

    g_free (message.properties);
}

static void
decode_force_screen_saver (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigForceScreenSaver message;

    message.sequence_number = codec->priv->sequence_number;
    message.mode = data;

    xig_codec_feed_force_screen_saver (XIG_CODEC (codec), &message);
}

static void
decode_set_pointer_mapping (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigSetPointerMapping message;
    int i;

    message.sequence_number = codec->priv->sequence_number;
    message.map_length = data;
    message.map = calloc (message.map_length, sizeof (guint8));
    for (i = 0; i < message.map_length; i++)
        message.map[i] = read_card8 (buffer, buffer_length, offset);

    xig_codec_feed_set_pointer_mapping (XIG_CODEC (codec), &message);

    g_free (message.map);
}

static void
decode_get_pointer_mapping (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGetPointerMapping message;

    message.sequence_number = codec->priv->sequence_number;

    xig_codec_feed_get_pointer_mapping (XIG_CODEC (codec), &message);
}

static void
decode_set_modifier_mapping (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigSetModifierMapping message;
    guint8 keycodes_per_modifier;
    int i;

    message.sequence_number = codec->priv->sequence_number;
    keycodes_per_modifier = data;
    message.keycodes_length = keycodes_per_modifier * 8;
    message.keycodes = calloc (sizeof (guint8), message.keycodes_length);
    for (i = 0; i < message.keycodes_length; i++)
        message.keycodes[i] = read_card8 (buffer, buffer_length, offset);

    xig_codec_feed_set_modifier_mapping (XIG_CODEC (codec), &message);

    g_free (message.keycodes);
}

static void
decode_get_modifier_mapping (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigGetModifierMapping message;

    message.sequence_number = codec->priv->sequence_number;

    xig_codec_feed_get_modifier_mapping (XIG_CODEC (codec), &message);
}

static void
decode_no_operation (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigNoOperation message;

    message.sequence_number = codec->priv->sequence_number;

    xig_codec_feed_no_operation (XIG_CODEC (codec), &message);
}

static void
decode_BIGREQ (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigBIGREQEnable message;

    message.sequence_number = codec->priv->sequence_number;

    xig_codec_feed_BIGREQ_enable (XIG_CODEC (codec), &message);
}

static void
decode_SYNC_initialize (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigSYNCInitialize message;

    message.sequence_number = codec->priv->sequence_number;
    message.desired_major_version = data;
    message.desired_minor_version = read_card8 (buffer, buffer_length, offset);
    read_padding (2, offset);

    xig_codec_feed_SYNC_initialize (XIG_CODEC (codec), &message);
}

static void
send_implementation_error (XigServerCodec *codec, guint16 major_opcode, guint16 minor_opcode)
{
    XigError message;

    message.sequence_number = codec->priv->sequence_number;
    message.type = Implementation;
    message.data = 0;
    message.major_opcode = major_opcode;
    message.minor_opcode = minor_opcode;

    xig_codec_feed_error (XIG_CODEC (codec), &message);
}

static void
decode_SYNC (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    switch (data)
    {
    case 0:
        decode_SYNC_initialize (codec, data, buffer, buffer_length, offset);
        break;
    default:
        send_implementation_error (codec, XIG_REQUEST_SYNC, data);
        break;
    }
}

static void
decode_RANDR_query_version (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigRANDRQueryVersion message;

    message.sequence_number = codec->priv->sequence_number;
    message.major_version = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.minor_version = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_RANDR_query_version (XIG_CODEC (codec), &message);
}

static void
decode_RANDR_select_input (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigRANDRSelectInput message;

    message.sequence_number = codec->priv->sequence_number;
    message.window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.enable = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (2, offset);

    xig_codec_feed_RANDR_select_input (XIG_CODEC (codec), &message);
}

static void
decode_RANDR_get_screen_resources (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigRANDRGetScreenResources message;

    message.sequence_number = codec->priv->sequence_number;
    message.window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_RANDR_get_screen_resources (XIG_CODEC (codec), &message);
}

static void
decode_RANDR_get_screen_resources_current (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigRANDRGetScreenResourcesCurrent message;

    message.sequence_number = codec->priv->sequence_number;
    message.window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_RANDR_get_screen_resources_current (XIG_CODEC (codec), &message);
}

static void
decode_RANDR_set_output_primary (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigRANDRSetOutputPrimary message;

    message.sequence_number = codec->priv->sequence_number;
    message.window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.output = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_RANDR_set_output_primary (XIG_CODEC (codec), &message);
}

static void
decode_RANDR_get_output_primary (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigRANDRGetOutputPrimary message;

    message.sequence_number = codec->priv->sequence_number;
    message.window = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_RANDR_get_output_primary (XIG_CODEC (codec), &message);
}

static void
decode_RANDR (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    switch (data)
    {
    case 0:
        decode_RANDR_query_version (codec, data, buffer, buffer_length, offset);
        break;
    case 4:
        decode_RANDR_select_input (codec, data, buffer, buffer_length, offset);
        break;
    case 8:
        decode_RANDR_get_screen_resources (codec, data, buffer, buffer_length, offset);
        break;
    case 25:
        decode_RANDR_get_screen_resources_current (codec, data, buffer, buffer_length, offset);
        break;
    case 30:
        decode_RANDR_set_output_primary (codec, data, buffer, buffer_length, offset);
        break;
    case 31:
        decode_RANDR_get_output_primary (codec, data, buffer, buffer_length, offset);
        break;
    default:
        send_implementation_error (codec, XIG_REQUEST_RANDR, data);
        break;
    }
}

static void
decode_XINERAMA_is_active (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigXINERAMAIsActive message;

    message.sequence_number = codec->priv->sequence_number;

    xig_codec_feed_XINERAMA_is_active (XIG_CODEC (codec), &message);
}

static void
decode_XINERAMA_query_screens (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigXINERAMAQueryScreens message;

    message.sequence_number = codec->priv->sequence_number;

    xig_codec_feed_XINERAMA_query_screens (XIG_CODEC (codec), &message);
}

static void
decode_XINERAMA (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    switch (data)
    {
    case 4:
        decode_XINERAMA_is_active (codec, data, buffer, buffer_length, offset);
        break;
    case 5:
        decode_XINERAMA_query_screens (codec, data, buffer, buffer_length, offset);
        break;
    default:
        send_implementation_error (codec, XIG_REQUEST_XINERAMA, data);
        break;
    }
}

static void
decode_XKB_use_extension (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigXKBUseExtension message;

    message.sequence_number = codec->priv->sequence_number;
    message.wanted_major = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.wanted_minor = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_XKB_use_extension (XIG_CODEC (codec), &message);
}

static void
decode_XKB_select_events (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigXKBSelectEvents message;

    message.sequence_number = codec->priv->sequence_number;
    message.device_spec = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.affect_which = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.clear = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.select_all = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.affect_map = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.map = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    // FIXME: details

    xig_codec_feed_XKB_select_events (XIG_CODEC (codec), &message);
}

static void
decode_XKB_get_state (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigXKBGetState message;

    message.sequence_number = codec->priv->sequence_number;
    message.device_spec = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (2, offset);

    xig_codec_feed_XKB_get_state (XIG_CODEC (codec), &message);
}

static void
decode_XKB_set_controls (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigXKBSetControls message;
    int i;

    message.sequence_number = codec->priv->sequence_number;
    message.device_spec = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.affect_internal_real_mods = read_card8 (buffer, buffer_length, offset);
    message.internal_real_mods = read_card8 (buffer, buffer_length, offset);
    message.affect_ignore_lock_real_mods = read_card8 (buffer, buffer_length, offset);
    message.ignore_lock_real_mods = read_card8 (buffer, buffer_length, offset);
    message.affect_internal_virtual_mods = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.internal_virtual_mods = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.affect_ignore_lock_virtual_mods = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.ignore_lock_virtual_mods = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.mouse_keys_dflt_btn = read_card8 (buffer, buffer_length, offset);
    message.groups_wrap = read_card8 (buffer, buffer_length, offset);
    message.access_x_options = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (2, offset);
    message.affect_enabled_controls = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.enabled_controls = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.change_controls = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.repeat_delay = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.repeat_interval = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.slow_keys_delay = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.debounce_delay = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.mouse_keys_delay = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.mouse_keys_interval = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.mouse_keys_time_to_max = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.mouse_keys_max_speed = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.mouse_keys_curve = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.access_x_timeout = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.access_x_timeout_mask = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.access_x_timeout_values = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.access_x_timeout_options_mask = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.access_x_timeout_options_values = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    for (i = 0; i < 32; i++)
        message.per_key_repeat[i] = read_card8 (buffer, buffer_length, offset);

    xig_codec_feed_XKB_set_controls (XIG_CODEC (codec), &message);

    g_free (message.per_key_repeat);
}

static void
decode_XKB_get_map (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigXKBGetMap message;

    message.sequence_number = codec->priv->sequence_number;
    message.device_spec = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.full = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.partial = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.first_type = read_card8 (buffer, buffer_length, offset);
    message.n_types = read_card8 (buffer, buffer_length, offset);
    message.first_key_sym = read_card8 (buffer, buffer_length, offset);
    message.n_key_syms = read_card8 (buffer, buffer_length, offset);
    message.first_key_action = read_card8 (buffer, buffer_length, offset);
    message.n_key_actions = read_card8 (buffer, buffer_length, offset);
    message.first_key_behavior = read_card8 (buffer, buffer_length, offset);
    message.n_key_behaviors = read_card8 (buffer, buffer_length, offset);
    message.virtual_mods = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.first_key_explicit = read_card8 (buffer, buffer_length, offset);
    message.n_key_explicit = read_card8 (buffer, buffer_length, offset);
    message.first_mod_map_key = read_card8 (buffer, buffer_length, offset);
    message.n_mod_map_keys = read_card8 (buffer, buffer_length, offset);
    message.first_v_mod_map_key = read_card8 (buffer, buffer_length, offset);
    message.n_v_mod_map_keys = read_card8 (buffer, buffer_length, offset);
    read_padding (2, offset);

    xig_codec_feed_XKB_get_map (XIG_CODEC (codec), &message);
}

static void
decode_XKB_get_names (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigXKBGetNames message;

    message.sequence_number = codec->priv->sequence_number;
    message.device_spec = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (2, offset);
    message.which = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_XKB_get_names (XIG_CODEC (codec), &message);
}

static void
decode_XKB_per_client_flags (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigXKBPerClientFlags message;

    message.sequence_number = codec->priv->sequence_number;
    message.device_spec = read_card16 (buffer, buffer_length, codec->priv->byte_order, offset);
    read_padding (2, offset);
    message.change = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.value = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.ctrls_to_change = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.auto_ctrls = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.auto_ctrls_values = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_XKB_per_client_flags (XIG_CODEC (codec), &message);
}

static void
decode_XKB (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    switch (data)
    {
    case 0:
        decode_XKB_use_extension (codec, data, buffer, buffer_length, offset);
        break;
    case 1:
        decode_XKB_select_events (codec, data, buffer, buffer_length, offset);
        break;
    /*case 3:
        decode_XKB_bell (codec, data, buffer, buffer_length, offset);
        break;*/
    case 4:
        decode_XKB_get_state (codec, data, buffer, buffer_length, offset);
        break;
    /*case 5:
        decode_XKB_latch_lock_state (codec, data, buffer, buffer_length, offset);
        break;
    case 6:
        decode_XKB_get_controls (codec, data, buffer, buffer_length, offset);
        break;*/
    case 7:
        decode_XKB_set_controls (codec, data, buffer, buffer_length, offset);
        break;
    case 8:
        decode_XKB_get_map (codec, data, buffer, buffer_length, offset);
        break;
    /*case 9:
        decode_XKB_set_map (codec, data, buffer, buffer_length, offset);
        break;
    case 10:
        decode_XKB_get_compat_map (codec, data, buffer, buffer_length, offset);
        break;
    case 11:
        decode_XKB_set_compat_map (codec, data, buffer, buffer_length, offset);
        break;
    case 12:
        decode_XKB_get_indicator_state (codec, data, buffer, buffer_length, offset);
        break;
    case 13:
        decode_XKB_get_indicator_map (codec, data, buffer, buffer_length, offset);
        break;
    case 14:
        decode_XKB_set_indicator_map (codec, data, buffer, buffer_length, offset);
        break;
    case 15:
        decode_XKB_get_named_indicator (codec, data, buffer, buffer_length, offset);
        break;
    case 16:
        decode_XKB_set_named_indicator (codec, data, buffer, buffer_length, offset);
        break;*/
    case 17:
        decode_XKB_get_names (codec, data, buffer, buffer_length, offset);
        break;
    /*case 18:
        decode_XKB_set_names (codec, data, buffer, buffer_length, offset);
        break;
    case 19:
        decode_XKB_get_geometry (codec, data, buffer, buffer_length, offset);
        break;
    case 20:
        decode_XKB_set_geometry (codec, data, buffer, buffer_length, offset);
        break;*/
    case 21:
        decode_XKB_per_client_flags (codec, data, buffer, buffer_length, offset);
        break;
    /*case 22:
        decode_XKB_list_components (codec, data, buffer, buffer_length, offset);
        break;
    case 23:
        decode_XKB_get_kbd_by_name (codec, data, buffer, buffer_length, offset);
        break;
    case 24:
        decode_XKB_get_device_info (codec, data, buffer, buffer_length, offset);
        break;
    case 25:
        decode_XKB_set_device_info (codec, data, buffer, buffer_length, offset);
        break;
    case 101:
        decode_XKB_set_debugging_flags (codec, data, buffer, buffer_length, offset);
        break;*/
    default:
        send_implementation_error (codec, XIG_REQUEST_XKEYBOARD, data);
        return;
    }
}

static void
decode_RENDER_query_version (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigRENDERQueryVersion message;

    message.sequence_number = codec->priv->sequence_number;
    message.client_major_version = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);
    message.client_minor_version = read_card32 (buffer, buffer_length, codec->priv->byte_order, offset);

    xig_codec_feed_RENDER_query_version (XIG_CODEC (codec), &message);
}

static void
decode_RENDER_query_pict_formats (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    XigRENDERQueryPictFormats message;

    message.sequence_number = codec->priv->sequence_number;

    xig_codec_feed_RENDER_query_pict_formats (XIG_CODEC (codec), &message);
}

static void
decode_RENDER (XigServerCodec *codec, guint8 data, const guint8 *buffer, gsize buffer_length, gsize *offset)
{
    switch (data)
    {
    case 0:
        decode_RENDER_query_version (codec, data, buffer, buffer_length, offset);
        break;
    case 1:
        decode_RENDER_query_pict_formats (codec, data, buffer, buffer_length, offset);
        break;
    /*case 2:
        decode_RENDER_query_pict_index_values (codec, data, buffer, buffer_length, offset);
        break;
    case 3:
        decode_RENDER_query_filters (codec, data, buffer, buffer_length, offset);
        break;
    case 4:
        decode_RENDER_create_picture (codec, data, buffer, buffer_length, offset);
        break;
    case 5:
        decode_RENDER_change_picture (codec, data, buffer, buffer_length, offset);
        break;
    case 6:
        decode_RENDER_set_picture_clip_rectangles (codec, data, buffer, buffer_length, offset);
        break;
    case 7:
        decode_RENDER_set_picture_transform (codec, data, buffer, buffer_length, offset);
        break;
    case 8:
        decode_RENDER_set_picture_filter (codec, data, buffer, buffer_length, offset);
        break;
    case 9:
        decode_RENDER_free_picture (codec, data, buffer, buffer_length, offset);
        break;
    case 10:
        decode_RENDER_composite (codec, data, buffer, buffer_length, offset);
        break;
    case 11:
        decode_RENDER_fill_rectangles (codec, data, buffer, buffer_length, offset);
        break;
    case 12:
        decode_RENDER_trapezoids (codec, data, buffer, buffer_length, offset);
        break;
    case 13:
        decode_RENDER_triangles (codec, data, buffer, buffer_length, offset);
        break;
    case 14:
        decode_RENDER_tri_strip (codec, data, buffer, buffer_length, offset);
        break;
    case 15:
        decode_RENDER_tri_fan (codec, data, buffer, buffer_length, offset);
        break;
    case 16:
        decode_RENDER_create_glyph_set (codec, data, buffer, buffer_length, offset);
        break;
    case 17:
        decode_RENDER_reference_glyph_set (codec, data, buffer, buffer_length, offset);
        break;
    case 18:
        decode_RENDER_free_glyph_set (codec, data, buffer, buffer_length, offset);
        break;
    case 19:
        decode_RENDER_add_glyphs (codec, data, buffer, buffer_length, offset);
        break;
    case 20:
        decode_RENDER_free_glyphs (codec, data, buffer, buffer_length, offset);
        break;
    case 21:
        decode_RENDER_composite_glyphs8 (codec, data, buffer, buffer_length, offset);
        break;
    case 22:
        decode_RENDER_composite_glyphs16 (codec, data, buffer, buffer_length, offset);
        break;
    case 23:
        decode_RENDER_composite_glyphs32 (codec, data, buffer, buffer_length, offset);
        break;
    case 24:
        decode_RENDER_create_cursor (codec, data, buffer, buffer_length, offset);
        break;
    case 25:
        decode_RENDER_create_anim_cursor (codec, data, buffer, buffer_length, offset);
        break;
    case 26:
        decode_RENDER_add_traps (codec, data, buffer, buffer_length, offset);
        break;
    case 27:
        decode_RENDER_create_solid_fill (codec, data, buffer, buffer_length, offset);
        break;
    case 28:
        decode_RENDER_create_linear_gradient (codec, data, buffer, buffer_length, offset);
        break;
    case 29:
        decode_RENDER_create_radial_gradient (codec, data, buffer, buffer_length, offset);
        break;
    case 30:
        decode_RENDER_create_conical_gradient (codec, data, buffer, buffer_length, offset);
        break;*/
    default:
        send_implementation_error (codec, XIG_REQUEST_RENDER, data);
        return;
    }
}

static void
decode_request (XigServerCodec *codec, const guint8 *buffer, gsize buffer_length)
{
    int opcode;
    gsize offset = 0;

    while (offset < buffer_length)
    {
        gsize start_offset;
        guint8 data;
        guint16 length, remaining;

        start_offset = offset;
        codec->priv->sequence_number++;
        opcode = read_card8 (buffer, buffer_length, &offset);
        data = read_card8 (buffer, buffer_length, &offset);
        length = read_card16 (buffer, buffer_length, codec->priv->byte_order, &offset) * 4;
        if (length == 0 && codec->priv->BIGREQ_enabled)
            length = read_card32 (buffer, buffer_length, codec->priv->byte_order, &offset) * 4;
        remaining = start_offset + length;

        switch (opcode)
        {
        case XIG_REQUEST_CreateWindow:
            decode_create_window (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_ChangeWindowAttributes:
            decode_change_window_attributes (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_GetWindowAttributes:
            decode_get_window_attributes (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_DestroyWindow:
            decode_destroy_window (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_DestroySubwindows:
            decode_destroy_subwindows (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_ChangeSaveSet:
            decode_change_save_set (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_ReparentWindow:
            decode_reparent_window (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_MapWindow:
            decode_map_window (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_MapSubwindows:
            decode_map_subwindows (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_UnmapWindow:
            decode_unmap_window (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_UnmapSubwindows:
            decode_unmap_subwindows (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_ConfigureWindow:
            decode_configure_window (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_CirculateWindow:
            decode_circulate_window (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_GetGeometry:
            decode_get_geometry (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_QueryTree:
            decode_query_tree (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_InternAtom:
            decode_intern_atom (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_GetAtomName:
            decode_get_atom_name (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_ChangeProperty:
            decode_change_property (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_DeleteProperty:
            decode_delete_property (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_GetProperty:
            decode_get_property (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_ListProperties:
            decode_list_properties (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_SetSelectionOwner:
            decode_set_selection_owner (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_GetSelectionOwner:
            decode_get_selection_owner (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_ConvertSelection:
            decode_convert_selection (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_SendEvent:
            decode_send_event (codec, data, buffer, remaining, &offset);          
            break;
        case XIG_REQUEST_GrabPointer:
            decode_grab_pointer (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_UngrabPointer:
            decode_ungrab_pointer (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_GrabButton:
            decode_grab_button (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_UngrabButton:
            decode_ungrab_button (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_ChangeActivePointerGrab:
            decode_change_active_pointer_grab (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_GrabKeyboard:
            decode_grab_keyboard (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_UngrabKeyboard:
            decode_ungrab_keyboard (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_GrabKey:
            decode_grab_key (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_UngrabKey:
            decode_ungrab_key (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_AllowEvents:
            decode_allow_events (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_GrabServer:
            decode_grab_server (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_UngrabServer:
            decode_ungrab_server (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_QueryPointer:
            decode_query_pointer (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_GetMotionEvents:
            decode_get_motion_events (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_TranslateCoordinates:
            decode_translate_coordinates (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_WarpPointer:
            decode_warp_pointer (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_SetInputFocus:
            decode_set_input_focus (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_GetInputFocus:
            decode_get_input_focus (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_QueryKeymap:
            decode_query_keymap (codec, data, buffer, remaining, &offset);          
            break;
        case XIG_REQUEST_OpenFont:
            decode_open_font (codec, data, buffer, remaining, &offset);          
            break;
        case XIG_REQUEST_CloseFont:
            decode_close_font (codec, data, buffer, remaining, &offset);          
            break;
        case XIG_REQUEST_QueryFont:
            decode_query_font (codec, data, buffer, remaining, &offset);          
            break;
        case XIG_REQUEST_QueryTextExtents:
            decode_query_text_extents (codec, data, buffer, remaining, &offset);          
            break;
        case XIG_REQUEST_ListFonts:
            decode_list_fonts (codec, data, buffer, remaining, &offset);          
            break;
        case XIG_REQUEST_ListFontsWithInfo:
            decode_list_fonts_with_info (codec, data, buffer, remaining, &offset);          
            break;
        case XIG_REQUEST_SetFontPath:
            decode_set_font_path (codec, data, buffer, remaining, &offset);          
            break;
        case XIG_REQUEST_GetFontPath:
            decode_get_font_path (codec, data, buffer, remaining, &offset);          
            break;
        case XIG_REQUEST_CreatePixmap:
            decode_create_pixmap (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_FreePixmap:
            decode_free_pixmap (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_CreateGC:
            decode_create_gc (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_ChangeGC:
            decode_change_gc (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_CopyGC:
            decode_copy_gc (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_SetDashes:
            decode_set_dashes (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_SetClipRectangles:
            decode_set_clip_rectangles (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_FreeGC:
            decode_free_gc (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_ClearArea:
            decode_clear_area (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_CopyArea:
            decode_copy_area (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_CopyPlane:
            decode_copy_plane (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_PolyPoint:
            decode_poly_point (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_PolyLine:
            decode_poly_line (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_PolySegment:
            decode_poly_segment (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_PolyRectangle:
            decode_poly_rectangle (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_PolyArc:
            decode_poly_arc (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_FillPoly:
            decode_fill_poly (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_PolyFillRectangle:
            decode_poly_fill_rectangle (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_PolyFillArc:
            decode_poly_fill_arc (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_PutImage:
            decode_put_image (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_GetImage:
            decode_get_image (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_PolyText8:
            decode_poly_text8 (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_PolyText16:
            decode_poly_text16 (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_ImageText8:
            decode_image_text8 (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_ImageText16:
            decode_image_text16 (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_CreateColormap:
            decode_create_colormap (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_FreeColormap:
            decode_free_colormap (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_CopyColormapAndFree:
            decode_copy_colormap_and_free (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_InstallColormap:
            decode_install_colormap (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_UninstallColormap:
            decode_uninstall_colormap (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_ListInstalledColormaps:
            decode_list_installed_colormaps (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_AllocColor:
            decode_alloc_color (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_AllocNamedColor:
            decode_alloc_named_color (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_AllocColorCells:
            decode_alloc_color_cells (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_AllocColorPlanes:
            decode_alloc_color_planes (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_FreeColors:
            decode_free_colors (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_StoreColors:
            decode_store_colors (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_StoreNamedColor:
            decode_store_named_color (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_QueryColors:
            decode_query_colors (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_LookupColor:
            decode_lookup_color (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_CreateCursor:
            decode_create_cursor (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_CreateGlyphCursor:
            decode_create_glyph_cursor (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_FreeCursor:
            decode_free_cursor (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_RecolorCursor:
            decode_recolor_cursor (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_QueryBestSize:
            decode_query_best_size (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_QueryExtension:
            decode_query_extension (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_ListExtensions:
            decode_list_extensions (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_ChangeKeyboardMapping:
            decode_change_keyboard_mapping (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_GetKeyboardMapping:
            decode_get_keyboard_mapping (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_ChangeKeyboardControl:
            decode_change_keyboard_control (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_GetKeyboardControl:
            decode_get_keyboard_control (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_Bell:
            decode_bell (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_ChangePointerControl:
            decode_change_pointer_control (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_GetPointerControl:
            decode_get_pointer_control (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_SetScreenSaver:
            decode_set_screen_saver (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_GetScreenSaver:
            decode_get_screen_saver (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_ChangeHosts:
            decode_change_hosts (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_ListHosts:
            decode_list_hosts (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_SetAccessControl:
            decode_set_access_control (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_SetCloseDownMode:
            decode_set_close_down_mode (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_KillClient:
            decode_kill_client (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_RotateProperties:
            decode_rotate_properties (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_ForceScreenSaver:
            decode_force_screen_saver (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_SetPointerMapping:
            decode_set_pointer_mapping (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_GetPointerMapping:
            decode_get_pointer_mapping (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_SetModifierMapping:
            decode_set_modifier_mapping (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_GetModifierMapping:
            decode_get_modifier_mapping (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_NoOperation:
            decode_no_operation (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_BIGREQ:
            decode_BIGREQ (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_SYNC:
            decode_SYNC (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_RANDR:
            decode_RANDR (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_XINERAMA:
            decode_XINERAMA (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_XKEYBOARD:
            decode_XKB (codec, data, buffer, remaining, &offset);
            break;
        case XIG_REQUEST_RENDER:
            decode_RENDER (codec, data, buffer, remaining, &offset);
            break;
        default:
            send_implementation_error (codec, opcode, 0);
            break;
        }

        offset = start_offset + length;
    }
}

static void
xig_server_codec_real_feed_data (XigCodec *codec, const guint8 *data, gsize data_length)
{
    XigServerCodec *server_codec = XIG_SERVER_CODEC (codec);

    if (server_codec->priv->connected)
        decode_request (server_codec, data, data_length);
    else
        decode_connection_request (server_codec, data, data_length);
}

static void
xig_server_codec_real_failed (XigCodec *codec, XigFailed *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0, length_offset;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, Failed, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, strlen (message->reason), &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->protocol_major_version, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->protocol_minor_version, &n_written);
    length_offset = n_written;
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padded_string (buffer, MAXIMUM_REQUEST_LENGTH, message->reason, &n_written);

    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, (n_written - length_offset) / 4, &length_offset);
    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_authenticate (XigCodec *codec, XigAuthenticate *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0, length_offset;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, Authenticate, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 5, &n_written);
    length_offset = n_written;
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padded_string (buffer, MAXIMUM_REQUEST_LENGTH, message->reason, &n_written);

    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, (n_written - length_offset) / 4, &length_offset);
    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_success (XigCodec *codec, XigSuccess *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0, length_offset;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, Success, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->protocol_major_version, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->protocol_minor_version, &n_written);
    length_offset = n_written;
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->release_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->resource_id_base, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->resource_id_mask, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->motion_buffer_size, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, strlen (message->vendor), &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->maximum_request_length, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->screens_length, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->pixmap_formats_length, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->image_byte_order, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->bitmap_format_bit_order, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->bitmap_format_scanline_unit, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->bitmap_format_scanline_pad, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->min_keycode, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->max_keycode, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 4, &n_written);
    write_padded_string (buffer, MAXIMUM_REQUEST_LENGTH, message->vendor, &n_written);

    for (i = 0; i < message->pixmap_formats_length; i++)
    {
        write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->pixmap_formats[i].depth, &n_written);
        write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->pixmap_formats[i].bits_per_pixel, &n_written);
        write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->pixmap_formats[i].scanline_pad, &n_written);
        write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 5, &n_written);
    }

    for (i = 0; i < message->screens_length; i++)
    {
        XigSCREEN *screen = &message->screens[i];
        int j;

        write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, screen->root, &n_written);
        write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, screen->default_colormap, &n_written);
        write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, screen->white_pixel, &n_written);
        write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, screen->black_pixel, &n_written);
        write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, screen->current_input_masks, &n_written);
        write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, screen->width_in_pixels, &n_written);
        write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, screen->height_in_pixels, &n_written);
        write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, screen->width_in_millimeters, &n_written);
        write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, screen->height_in_millimeters, &n_written);
        write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, screen->min_installed_maps, &n_written);
        write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, screen->max_installed_maps, &n_written);
        write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, screen->root_visual, &n_written);
        write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, screen->backing_stores, &n_written);
        write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, screen->save_unders ? 1 : 0, &n_written);
        write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, screen->root_visual, &n_written);
        write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, screen->allowed_depths_length, &n_written);

        for (j = 0; j < screen->allowed_depths_length; j++)
        {
            XigDEPTH *depth = &screen->allowed_depths[j];
            int k;

            write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, depth->depth, &n_written);
            write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
            write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, depth->visuals_length, &n_written);
            write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 4, &n_written);

            for (k = 0; k < depth->visuals_length; k++)
            {
                XigVISUALTYPE *visual = &depth->visuals[k];

                write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, visual->visual_id, &n_written);
                write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, visual->class, &n_written);
                write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, visual->bits_per_rgb_value, &n_written);
                write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, visual->colormap_entries, &n_written);
                write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, visual->red_mask, &n_written);
                write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, visual->green_mask, &n_written);
                write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, visual->blue_mask, &n_written);
                write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 4, &n_written);
            }
        }
    }

    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, (n_written - length_offset) / 4, &length_offset);
    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_error (XigCodec *codec, XigError *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Error, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->type, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->data, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->minor_opcode, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->major_opcode, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 21, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_create_notify (XigCodec *codec, XigCreateNotify *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_CreateNotify, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->parent, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->window, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->x, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->y, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->width, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->height, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->border_width, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->override_redirect ? 1 : 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 9, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_destroy_notify (XigCodec *codec, XigDestroyNotify *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_DestroyNotify, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->event, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->window, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 20, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_unmap_notify (XigCodec *codec, XigUnmapNotify *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_UnmapNotify, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->event, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->window, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->from_configure ? 1 : 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 19, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_map_notify (XigCodec *codec, XigMapNotify *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_MapNotify, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->event, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->window, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->override_redirect ? 1 : 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 19, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_map_request (XigCodec *codec, XigMapRequest *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_MapRequest, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->parent, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->window, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 20, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_reparent_notify (XigCodec *codec, XigReparentNotify *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_ReparentNotify, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->event, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->window, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->parent, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->x, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->y, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->override_redirect ? 1 : 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 11, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_configure_notify (XigCodec *codec, XigConfigureNotify *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_ConfigureNotify, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->event, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->window, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->above_sibling, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->x, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->y, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->width, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->height, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->border_width, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->override_redirect ? 1 : 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 5, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_configure_request (XigCodec *codec, XigConfigureRequest *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_ConfigureRequest, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->stack_mode, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->parent, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->window, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sibling, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->x, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->y, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->width, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->height, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->border_width, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->value_mask, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 4, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_circulate_notify (XigCodec *codec, XigCirculateNotify *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_CirculateNotify, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->event, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->window, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 4, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->place, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 15, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_circulate_request (XigCodec *codec, XigCirculateRequest *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_MapRequest, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->parent, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->window, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 4, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->place, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 15, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_property_notify (XigCodec *codec, XigPropertyNotify *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_PropertyNotify, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->window, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->atom, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->time, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->state, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 15, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_get_window_attributes_reply (XigCodec *codec, XigGetWindowAttributesReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->backing_store, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 3, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->visual, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->class, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->bit_gravity, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->win_gravity, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->backing_planes, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->backing_pixel, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->save_under ? 1 : 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->map_is_installed ? 1 : 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->map_state, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->override_redirect ? 1 : 0, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->colormap, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->all_event_masks, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->your_event_mask, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->do_not_propagate_mask, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 2, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_get_geometry_reply (XigCodec *codec, XigGetGeometryReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->depth, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->root, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->x, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->y, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->width, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->height, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->border_width, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 10, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_query_tree_reply (XigCodec *codec, XigQueryTreeReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0, length_offset, payload_offset;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    length_offset = n_written;
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->root, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->parent, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->children_length, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 14, &n_written);
    payload_offset = n_written;
    for (i = 0; i < message->children_length; i++)
        write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->children[i], &n_written);

    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, (n_written - payload_offset) / 4, &length_offset);
    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_intern_atom_reply (XigCodec *codec, XigInternAtomReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->atom, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 20, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_get_atom_name_reply (XigCodec *codec, XigGetAtomNameReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0, length_offset, payload_offset;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    length_offset = n_written;
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, strlen (message->name), &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 22, &n_written);
    payload_offset = n_written;
    write_padded_string (buffer, MAXIMUM_REQUEST_LENGTH, message->name, &n_written);

    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, (n_written - payload_offset) / 4, &length_offset);
    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_get_property_reply (XigCodec *codec, XigGetPropertyReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0, length_offset, payload_offset;
    guint32 length = 0;
    guint8 *value8;
    guint16 *value16;
    guint32 *value32;
    int i;

    if (message->value_format != 0)
        length = message->value_length / (message->value_format / 8);

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->value_format, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    length_offset = n_written;
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->type, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->bytes_after, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, length, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 12, &n_written);
    payload_offset = n_written;
    switch (message->value_format)
    {
    default:
    case 8:
        value8 = message->value;
        for (i = 0; i < message->value_length; i++)
            write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, value8[i], &n_written);
        break;
    case 16:
        value16 = message->value;
        for (i = 0; i < message->value_length; i++)
            write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, value16[i], &n_written);
        break;
    case 32:
        value32 = message->value;
        for (i = 0; i < message->value_length; i++)
            write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, value32[i], &n_written);
        break;
    }
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, pad (n_written), &n_written);

    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, (n_written - payload_offset) / 4, &length_offset);
    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_list_properties_reply (XigCodec *codec, XigListPropertiesReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0, length_offset, payload_offset;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    length_offset = n_written;
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->atoms_length, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 22, &n_written);
    payload_offset = n_written;
    for (i = 0; i < message->atoms_length; i++)
        write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->atoms[i], &n_written);

    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, (n_written - payload_offset) / 4, &length_offset);
    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_get_selection_owner_reply (XigCodec *codec, XigGetSelectionOwnerReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->owner, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 20, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_grab_pointer_reply (XigCodec *codec, XigGrabPointerReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->status, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 24, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_grab_keyboard_reply (XigCodec *codec, XigGrabKeyboardReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->status, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 24, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_query_pointer_reply (XigCodec *codec, XigQueryPointerReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->same_screen ? 1 : 0, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->root, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->child, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->root_x, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->root_y, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->win_x, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->win_y, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->mask, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 6, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_get_motion_events_reply (XigCodec *codec, XigGetMotionEventsReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0, length_offset, payload_offset;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    length_offset = n_written;
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->events_length, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 20, &n_written);
    payload_offset = n_written;
    for (i = 0; i < message->events_length; i++)
    {
        write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->events[i].time, &n_written);
        write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->events[i].x, &n_written);
        write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->events[i].y, &n_written);
    }

    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, (n_written - payload_offset) / 4, &length_offset);
    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_translate_coordinates_reply (XigCodec *codec, XigTranslateCoordinatesReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->same_screen ? 1 : 0, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->child, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->dst_x, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->dst_y, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 16, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_get_input_focus_reply (XigCodec *codec, XigGetInputFocusReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->revert_to, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->focus, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 20, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_query_keymap_reply (XigCodec *codec, XigQueryKeymapReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 2, &n_written);
    for (i = 0; i < 32; i++)
        write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->keys[i], &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_query_font_reply (XigCodec *codec, XigQueryFontReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0, length_offset, payload_offset;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    length_offset = n_written;
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->min_bounds.left_side_bearing, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->min_bounds.right_side_bearing, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->min_bounds.character_width, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->min_bounds.ascent, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->min_bounds.descent, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->min_bounds.attributes, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 4, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->max_bounds.left_side_bearing, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->max_bounds.right_side_bearing, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->max_bounds.character_width, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->max_bounds.ascent, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->max_bounds.descent, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->max_bounds.attributes, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 4, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->min_char_or_byte2, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->max_char_or_byte2, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->default_char, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->properties_length, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->draw_direction, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->min_byte1, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->max_byte1, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->all_chars_exist ? 1 : 0, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->font_ascent, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->font_descent, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->char_infos_length, &n_written);
    for (i = 0; i < message->properties_length; i++)
    {
        write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->properties[i].name, &n_written);
        write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->properties[i].value, &n_written);
    }
    for (i = 0; i < message->char_infos_length; i++)
    {
        write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->char_infos[i].left_side_bearing, &n_written);
        write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->char_infos[i].right_side_bearing, &n_written);
        write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->char_infos[i].character_width, &n_written);
        write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->char_infos[i].ascent, &n_written);
        write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->char_infos[i].descent, &n_written);
        write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->char_infos[i].attributes, &n_written);
    }
    payload_offset = n_written;

    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, (n_written - payload_offset) / 4, &length_offset);
    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_query_text_extents_reply (XigCodec *codec, XigQueryTextExtentsReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->draw_direction, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->font_ascent, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->font_descent, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->overall_ascent, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->overall_descent, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->overall_width, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->overall_left, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->overall_right, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 4, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_list_fonts_reply (XigCodec *codec, XigListFontsReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0, length_offset, payload_offset;
    gchar **i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    length_offset = n_written;
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, g_strv_length (message->names), &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 22, &n_written);
    payload_offset = n_written;
    for (i = message->names; *i; i++)
    {
        write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, strlen (*i), &n_written);
        write_string (buffer, MAXIMUM_REQUEST_LENGTH, *i, &n_written);
    }
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, pad (n_written), &n_written);

    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, (n_written - payload_offset) / 4, &length_offset);
    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_list_fonts_with_info_reply (XigCodec *codec, XigListFontsWithInfoReply *message)
{
    // FIXME
}

static void
xig_server_codec_real_get_font_path_reply (XigCodec *codec, XigGetFontPathReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0, length_offset, payload_offset;
    gchar **s;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    length_offset = n_written;
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, g_strv_length (message->path), &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 22, &n_written);
    payload_offset = n_written;
    for (s = message->path; *s; s++)
    {
        write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, strlen (*s), &n_written);
        write_string (buffer, MAXIMUM_REQUEST_LENGTH, *s, &n_written);
    }
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, pad (n_written), &n_written);

    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, (n_written - payload_offset) / 4, &length_offset);
    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_get_image_reply (XigCodec *codec, XigGetImageReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0, length_offset, payload_offset;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    length_offset = n_written;
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->visual, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 20, &n_written);
    payload_offset = n_written;
    for (i = 0; i < message->data_length; i++)
        write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->data[i], &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, pad (n_written), &n_written);

    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, (n_written - payload_offset) / 4, &length_offset);
    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_list_installed_colormaps_reply (XigCodec *codec, XigListInstalledColormapsReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0, length_offset, payload_offset;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    length_offset = n_written;
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->cmaps_length, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 22, &n_written);
    payload_offset = n_written;
    for (i = 0; i < message->cmaps_length; i++)
        write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->cmaps[i], &n_written);

    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, (n_written - payload_offset) / 4, &length_offset);
    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_alloc_color_reply (XigCodec *codec, XigAllocColorReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->red, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->green, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->blue, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 2, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->pixel, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 12, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_alloc_named_color_reply (XigCodec *codec, XigAllocNamedColorReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->pixel, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->exact_red, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->exact_green, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->exact_blue, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->visual_red, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->visual_green, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->visual_blue, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 8, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_alloc_color_cells_reply (XigCodec *codec, XigAllocColorCellsReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0, length_offset, payload_offset;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    length_offset = n_written;
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->pixels_length, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->masks_length, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 20, &n_written);
    payload_offset = n_written;
    for (i = 0; i < message->pixels_length; i++)
        write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->pixels[i], &n_written);
    for (i = 0; i < message->masks_length; i++)
        write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->masks[i], &n_written);

    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, (n_written - payload_offset) / 4, &length_offset);
    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_alloc_color_planes_reply (XigCodec *codec, XigAllocColorPlanesReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0, length_offset, payload_offset;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    length_offset = n_written;
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->pixels_length, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 2, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->red_mask, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->green_mask, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->blue_mask, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 8, &n_written);
    payload_offset = n_written;
    for (i = 0; i < message->pixels_length; i++)
        write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->pixels[i], &n_written);

    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, (n_written - payload_offset) / 4, &length_offset);
    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_query_colors_reply (XigCodec *codec, XigQueryColorsReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0, length_offset, payload_offset;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    length_offset = n_written;
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->colors_length, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 22, &n_written);
    payload_offset = n_written;
    for (i = 0; i < message->colors_length; i++)
    {
        write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->colors[i].red, &n_written);
        write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->colors[i].green, &n_written);
        write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->colors[i].blue, &n_written);
        write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 2, &n_written);
    }

    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, (n_written - payload_offset) / 4, &length_offset);
    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_lookup_color_reply (XigCodec *codec, XigLookupColorReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->exact_red, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->exact_green, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->exact_blue, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->visual_red, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->visual_green, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->visual_blue, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 12, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_query_best_size_reply (XigCodec *codec, XigQueryBestSizeReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->width, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->height, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 20, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_query_extension_reply (XigCodec *codec, XigQueryExtensionReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->present ? 1 : 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->major_opcode, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->first_event, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->first_error, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 20, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_list_extensions_reply (XigCodec *codec, XigListExtensionsReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0, length_offset, payload_offset;
    gchar **s;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, g_strv_length (message->names), &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    length_offset = n_written;
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 24, &n_written);
    payload_offset = n_written;
    for (s = message->names; *s; s++)
    {
        write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, strlen (*s), &n_written);
        write_string (buffer, MAXIMUM_REQUEST_LENGTH, *s, &n_written);
    }

    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, (n_written - payload_offset) / 4, &length_offset);
    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_get_keyboard_mapping_reply (XigCodec *codec, XigGetKeyboardMappingReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0, length_offset, payload_offset;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->keysyms_per_keycode, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    length_offset = n_written;
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 24, &n_written);
    payload_offset = n_written;
    for (i = 0; i < message->keysyms_length; i++)
        write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->keysyms[i], &n_written);

    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, (n_written - payload_offset) / 4, &length_offset);
    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_get_keyboard_control_reply (XigCodec *codec, XigGetKeyboardControlReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->global_auto_repeat, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 5, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->led_mask, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->key_click_percent, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->bell_percent, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->bell_pitch, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->bell_duration, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 2, &n_written);
    for (i = 0; i < 32; i++);
        write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->auto_repeats[i], &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_get_pointer_control_reply (XigCodec *codec, XigGetPointerControlReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);;
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->acceleration_numerator, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->acceleration_denominator, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->threshold, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 18, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_get_screen_saver_reply (XigCodec *codec, XigGetScreenSaverReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);;
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->timeout, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->interval, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->prefer_blanking, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->allow_exposures, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 18, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_list_hosts_reply (XigCodec *codec, XigListHostsReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0, length_offset, payload_offset;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->mode, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    length_offset = n_written;
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->hosts_length, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 22, &n_written);
    payload_offset = n_written;
    for (i = 0; i < message->hosts_length; i++)
    {
        write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->hosts[i].family, &n_written);
        write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
        write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->hosts[i].address_length, &n_written);
        write_string8 (buffer, MAXIMUM_REQUEST_LENGTH, message->hosts[i].address, message->hosts[i].address_length, &n_written);
        write_padding (buffer, MAXIMUM_REQUEST_LENGTH, pad (n_written), &n_written);
    }

    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, (n_written - payload_offset) / 4, &length_offset);
    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_set_pointer_mapping_reply (XigCodec *codec, XigSetPointerMappingReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->status, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 24, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_get_pointer_mapping_reply (XigCodec *codec, XigGetPointerMappingReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0, length_offset, payload_offset;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->map_length, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    length_offset = n_written;
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 24, &n_written);
    payload_offset = n_written;
    for (i = 0; i < message->map_length; i++)
        write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->map[i], &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, pad (n_written), &n_written);

    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, (n_written - payload_offset) / 4, &length_offset);
    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_set_modifier_mapping_reply (XigCodec *codec, XigSetModifierMappingReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->status, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 24, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_get_modifier_mapping_reply (XigCodec *codec, XigGetModifierMappingReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0, length_offset, payload_offset;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->keycodes_per_modifier, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    length_offset = n_written;
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 24, &n_written);
    payload_offset = n_written;
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, 0, &n_written);

    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, (n_written - payload_offset) / 4, &length_offset);
    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_BIGREQ_enable_reply (XigCodec *codec, XigBIGREQEnableReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->maximum_request_length, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 20, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_SYNC_initialize_reply (XigCodec *codec, XigSYNCInitializeReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->major_version, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->minor_version, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 22, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_RANDR_query_version_reply (XigCodec *codec, XigRANDRQueryVersionReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->major_version, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->minor_version, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 16, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
write_RANDR_MODEINFO (XigCodec *codec, guint8 *buffer, gsize buffer_length, XigRANDRMODEINFO *value, gsize *n_written)
{
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, value->id, n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, value->width_in_pixels, n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, value->height_in_pixels, n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, value->dot_clock, n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, value->h_sync_start, n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, value->h_sync_end, n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, value->h_total, n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, value->h_skew, n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, value->v_sync_start, n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, value->v_sync_end, n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, value->v_total, n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, strlen (value->name), n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, value->mode_flags, n_written);
}

static void
xig_server_codec_real_RANDR_get_screen_resources_reply (XigCodec *codec, XigRANDRGetScreenResourcesReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0, length_offset, payload_offset;
    guint16 names_length;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    length_offset = n_written;
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->timestamp, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->config_timestamp, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->crtcs_length, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->outputs_length, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->modes_length, &n_written);
    names_length = 0;
    for (i = 0; i < message->modes_length; i++)
        names_length += strlen (message->modes[i].name);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, names_length, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 8, &n_written);
    payload_offset = n_written;
    for (i = 0; i < message->crtcs_length; i++)
        write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->crtcs[i], &n_written);
    for (i = 0; i < message->outputs_length; i++)
        write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->outputs[i], &n_written);
    for (i = 0; i < message->modes_length; i++)
        write_RANDR_MODEINFO (codec, buffer, MAXIMUM_REQUEST_LENGTH, &message->modes[i], &n_written);
    for (i = 0; i < message->modes_length; i++)
        write_string (buffer, MAXIMUM_REQUEST_LENGTH, message->modes[i].name, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, pad (n_written), &n_written);

    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, (n_written - payload_offset) / 4, &length_offset);
    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_RANDR_get_screen_resources_current_reply (XigCodec *codec, XigRANDRGetScreenResourcesCurrentReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0, length_offset, payload_offset;
    guint16 names_length;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    length_offset = n_written;
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->timestamp, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->config_timestamp, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->crtcs_length, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->outputs_length, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->modes_length, &n_written);
    names_length = 0;
    for (i = 0; i < message->modes_length; i++)
        names_length += strlen (message->modes[i].name);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, names_length, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 8, &n_written);
    payload_offset = n_written;
    for (i = 0; i < message->crtcs_length; i++)
        write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->crtcs[i], &n_written);
    for (i = 0; i < message->outputs_length; i++)
        write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->outputs[i], &n_written);
    for (i = 0; i < message->modes_length; i++)
        write_RANDR_MODEINFO (codec, buffer, MAXIMUM_REQUEST_LENGTH, &message->modes[i], &n_written);
    for (i = 0; i < message->modes_length; i++)
        write_string (buffer, MAXIMUM_REQUEST_LENGTH, message->modes[i].name, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, pad (n_written), &n_written);

    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, (n_written - payload_offset) / 4, &length_offset);
    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_RANDR_get_output_primary_reply (XigCodec *codec, XigRANDRGetOutputPrimaryReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->output, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 20, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_XINERAMA_is_active_reply (XigCodec *codec, XigXINERAMAIsActiveReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->is_active ? 1 : 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 20, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void 
xig_server_codec_real_XINERAMA_query_screens_reply (XigCodec *codec, XigXINERAMAQueryScreensReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0, length_offset, payload_offset;
    int i;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    length_offset = n_written;
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->info_length, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 20, &n_written);
    payload_offset = n_written;
    for (i = 0; i < message->info_length; i++)
    {
        XigXINERAMAScreenInfo *info = &message->info[i];
        write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, info->x, &n_written);
        write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, info->y, &n_written);
        write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, info->width, &n_written);
        write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, info->height, &n_written);
    }

    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, (n_written - payload_offset) / 4, &length_offset);  
    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_XKB_use_extension_reply (XigCodec *codec, XigXKBUseExtensionReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->supported ? 1 : 0, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->server_major, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->server_minor, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 20, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_XKB_get_state_reply (XigCodec *codec, XigXKBGetStateReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->device_id, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->mods, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->base_mods, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->latched_mods, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->locked_mods, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->group, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->locked_group, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->base_group, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->latched_group, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->compat_state, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->grab_mods, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->compat_grab_mods, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->lookup_mods, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->compat_lookup_mods, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->ptr_btn_state, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 6, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_XKB_get_map_reply (XigCodec *codec, XigXKBGetMapReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->device_id, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    // FIXME
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 24, &n_written);
    /*write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 2, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, 0, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, 0, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, 0, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, 0, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, 0, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);*/

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_XKB_get_names_reply (XigCodec *codec, XigXKBGetNamesReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->device_id, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    // FIXME
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 24, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);
}

static void
xig_server_codec_real_XKB_per_client_flags_reply (XigCodec *codec, XigXKBPerClientFlagsReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, message->device_id, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->supported, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->value, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->auto_ctrls, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->auto_ctrls_values, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 8, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);        
}

static void
xig_server_codec_real_RENDER_query_version_reply (XigCodec *codec, XigRENDERQueryVersionReply *message)
{
    guint8 buffer[MAXIMUM_REQUEST_LENGTH];
    gsize n_written = 0;

    write_card8 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_RESPONSE_Reply, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 1, &n_written);
    write_card16 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->sequence_number, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, 0, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->major_version, &n_written);
    write_card32 (buffer, MAXIMUM_REQUEST_LENGTH, XIG_SERVER_CODEC (codec)->priv->byte_order, message->minor_version, &n_written);
    write_padding (buffer, MAXIMUM_REQUEST_LENGTH, 16, &n_written);

    XIG_CODEC_GET_CLASS (codec)->write (codec, buffer, n_written);        
}

static void
xig_server_codec_real_RENDER_query_pict_formats_reply (XigCodec *codec, XigRENDERQueryPictFormatsReply *message)
{
    // FIXME
}

static void
xig_server_codec_init (XigServerCodec *codec)
{
    codec->priv = G_TYPE_INSTANCE_GET_PRIVATE (codec, xig_server_codec_get_type (), XigServerCodecPrivate);
    codec->priv->sequence_number = 0;
}

static void
xig_server_codec_class_init (XigServerCodecClass *klass)
{
    XigCodecClass *codec_class = XIG_CODEC_CLASS (klass);

    g_type_class_add_private (klass, sizeof (XigServerCodecPrivate));

    codec_class->feed_data = xig_server_codec_real_feed_data;
    codec_class->failed = xig_server_codec_real_failed;
    codec_class->authenticate = xig_server_codec_real_authenticate;
    codec_class->success = xig_server_codec_real_success;
    codec_class->error = xig_server_codec_real_error;
    codec_class->create_notify = xig_server_codec_real_create_notify;
    codec_class->destroy_notify = xig_server_codec_real_destroy_notify;
    codec_class->unmap_notify = xig_server_codec_real_unmap_notify;
    codec_class->map_notify = xig_server_codec_real_map_notify;
    codec_class->map_request = xig_server_codec_real_map_request;
    codec_class->reparent_notify = xig_server_codec_real_reparent_notify;
    codec_class->configure_notify = xig_server_codec_real_configure_notify;
    codec_class->configure_request = xig_server_codec_real_configure_request;
    codec_class->circulate_notify = xig_server_codec_real_circulate_notify;
    codec_class->circulate_request = xig_server_codec_real_circulate_request;
    codec_class->property_notify = xig_server_codec_real_property_notify;
    codec_class->get_window_attributes_reply = xig_server_codec_real_get_window_attributes_reply;
    codec_class->get_geometry_reply = xig_server_codec_real_get_geometry_reply;
    codec_class->query_tree_reply = xig_server_codec_real_query_tree_reply;
    codec_class->intern_atom_reply = xig_server_codec_real_intern_atom_reply;
    codec_class->get_atom_name_reply = xig_server_codec_real_get_atom_name_reply;
    codec_class->get_property_reply = xig_server_codec_real_get_property_reply;
    codec_class->list_properties_reply = xig_server_codec_real_list_properties_reply;
    codec_class->get_selection_owner_reply = xig_server_codec_real_get_selection_owner_reply;
    codec_class->grab_pointer_reply = xig_server_codec_real_grab_pointer_reply;
    codec_class->grab_keyboard_reply = xig_server_codec_real_grab_keyboard_reply;
    codec_class->query_pointer_reply = xig_server_codec_real_query_pointer_reply;
    codec_class->get_motion_events_reply = xig_server_codec_real_get_motion_events_reply;
    codec_class->translate_coordinates_reply = xig_server_codec_real_translate_coordinates_reply;
    codec_class->get_input_focus_reply = xig_server_codec_real_get_input_focus_reply;
    codec_class->query_keymap_reply = xig_server_codec_real_query_keymap_reply;
    codec_class->query_font_reply = xig_server_codec_real_query_font_reply;
    codec_class->query_text_extents_reply = xig_server_codec_real_query_text_extents_reply;
    codec_class->list_fonts_reply = xig_server_codec_real_list_fonts_reply;
    codec_class->list_fonts_with_info_reply = xig_server_codec_real_list_fonts_with_info_reply;
    codec_class->get_font_path_reply = xig_server_codec_real_get_font_path_reply;
    codec_class->get_image_reply = xig_server_codec_real_get_image_reply;
    codec_class->list_installed_colormaps_reply = xig_server_codec_real_list_installed_colormaps_reply;
    codec_class->alloc_color_reply = xig_server_codec_real_alloc_color_reply;
    codec_class->alloc_named_color_reply = xig_server_codec_real_alloc_named_color_reply;
    codec_class->alloc_color_cells_reply = xig_server_codec_real_alloc_color_cells_reply;
    codec_class->alloc_color_planes_reply = xig_server_codec_real_alloc_color_planes_reply;
    codec_class->query_colors_reply = xig_server_codec_real_query_colors_reply;
    codec_class->lookup_color_reply = xig_server_codec_real_lookup_color_reply;
    codec_class->query_best_size_reply = xig_server_codec_real_query_best_size_reply;
    codec_class->query_extension_reply = xig_server_codec_real_query_extension_reply;
    codec_class->list_extensions_reply = xig_server_codec_real_list_extensions_reply;
    codec_class->get_keyboard_mapping_reply = xig_server_codec_real_get_keyboard_mapping_reply;
    codec_class->get_keyboard_control_reply = xig_server_codec_real_get_keyboard_control_reply;
    codec_class->get_pointer_control_reply = xig_server_codec_real_get_pointer_control_reply;
    codec_class->get_screen_saver_reply = xig_server_codec_real_get_screen_saver_reply;
    codec_class->list_hosts_reply = xig_server_codec_real_list_hosts_reply;
    codec_class->set_pointer_mapping_reply = xig_server_codec_real_set_pointer_mapping_reply;
    codec_class->get_pointer_mapping_reply = xig_server_codec_real_get_pointer_mapping_reply;
    codec_class->set_modifier_mapping_reply = xig_server_codec_real_set_modifier_mapping_reply;
    codec_class->get_modifier_mapping_reply = xig_server_codec_real_get_modifier_mapping_reply;
    codec_class->BIGREQ_enable_reply = xig_server_codec_real_BIGREQ_enable_reply;
    codec_class->SYNC_initialize_reply = xig_server_codec_real_SYNC_initialize_reply;
    codec_class->RANDR_query_version_reply = xig_server_codec_real_RANDR_query_version_reply;
    codec_class->RANDR_get_screen_resources_reply = xig_server_codec_real_RANDR_get_screen_resources_reply;
    codec_class->RANDR_get_screen_resources_current_reply = xig_server_codec_real_RANDR_get_screen_resources_current_reply;
    codec_class->RANDR_get_output_primary_reply = xig_server_codec_real_RANDR_get_output_primary_reply;
    codec_class->XINERAMA_is_active_reply = xig_server_codec_real_XINERAMA_is_active_reply;
    codec_class->XINERAMA_query_screens_reply = xig_server_codec_real_XINERAMA_query_screens_reply;
    codec_class->XKB_use_extension_reply = xig_server_codec_real_XKB_use_extension_reply;
    codec_class->XKB_get_state_reply = xig_server_codec_real_XKB_get_state_reply;
    codec_class->XKB_get_map_reply = xig_server_codec_real_XKB_get_map_reply;
    codec_class->XKB_get_names_reply = xig_server_codec_real_XKB_get_names_reply;
    codec_class->XKB_per_client_flags_reply = xig_server_codec_real_XKB_per_client_flags_reply;
    codec_class->RENDER_query_version_reply = xig_server_codec_real_RENDER_query_version_reply;
    codec_class->RENDER_query_pict_formats_reply = xig_server_codec_real_RENDER_query_pict_formats_reply;
}
